#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Broker",
    ]

import logging
import socket

from checkbox.lib.param import getInteger

from checkbox.io.socktcp import SockTCP
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )

from checkbox.connector.operation import (
    OperationCode,
    get_operation,
    put_operation,
    )
from checkbox.connector.relay import (
    RelayProtocol,
    Relay,
    )


BROKER_LOCAL_PORT = 65430
BROKER_PUBLIC_PORT = 65432

DEFAULT_MAX_FD = 1024
DEFAULT_MAX_RELAYERS = 50


max_fd = DEFAULT_MAX_RELAYERS


class BrokerClient:

    __slots__ = (
        "protocol",
        "cip",
        "cport",
        "active",
        "ip",
        "port",
        "listener",
        "relay",
        )

    def __init__(
        self, protocol, cip, cport, active, ip, port, listener, relay):
        self.protocol = protocol
        self.cip = cip
        self.cport = cport
        self.active = active
        self.ip = ip
        self.port = port
        self.listener = listener
        self.relay = relay


class Broker:

    __slots__ = (
        "_address",
        "_local",
        "_public",
        "_clients",
        "_relays",
        "_relay_count",
        "_relay_max",
        )

    def __init__(self, address):
        self._address = address
        self._local = self.listen(address, BROKER_LOCAL_PORT)
        logging.info(
            "Local socket listening at %s:%d", address, BROKER_LOCAL_PORT)

        self._public = self.listen(address, BROKER_PUBLIC_PORT)
        logging.info(
            "Public socket listening at %s:%d", address, BROKER_PUBLIC_PORT)

        self._clients = {}

        self._relays = set()
        self._relay_count = 0
        self._relay_max = getInteger("MAX_RELAYERS", DEFAULT_MAX_RELAYERS)

    def listen(self, address, port):
        sock = SockTCP()
        if not sock.setOptions(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1):
            raise Exception("Failed to set socket options")

        if not sock.bind(address, port):
            raise Exception("Failed to bind socket to %d" % port)

        if not sock.listen():
            raise Exception("Failed to listen on socket")

        return sock

    def createRelay(self):
        relay = Relay(self._address)
        self._relays.add(relay)
        return relay

    def deleteRelay(self, relay):
        self._relays.remove(relay)

        # TODO: Delete clients the relay is servicing
        # self._clients = [client for client in self._clients
        #     if client.relay != relay]

    def deleteClient(self, client):
        if client.relay:
            client.relay.deleteClient()

        # delete from _clients

    def handleClientRequest(self):
        logging.debug("handleClientRequest")

        client = self._local.accept()
        if client is None:
            logging.info("Failed to acept new connection on local socket")
            return

        operation = get_operation(client)
        if operation is None:
            logging.info("Failed to read request on local socket")
            client.close()
            return

        if operation["code"] == OperationCode.NEW_SOCK:
            for relay in self._relays:
                if relay.has_more_clients:
                    break
            else:
                if self._relay_count > self._relay_max:
                    logging.info("No more relays available")
                    client.close()
                    return

                relay = self.createRelay()

            put_operation(
                client, OperationCode.RELAY_SERVER,
                ip=relay.mip, port=relay.mport)

        elif operation["code"] == OperationCode.DATA_QUERY:
            count = sum([relay.client_max - relay.client_count
                for relay in self._relays])
            count += (
                (self._relay_max, self._relay_count) *
                self._relays[0].client_max)
            put_operation(
                client, OperationCode.DATA_RESPONSE,
                cip=operation["cip"], cport=count)

        else:
            logging.info(
                "Invalid operation code received on local socket: %s",
                operation["code"])

        client.close()

    def handleConnectionRequest(self):
        logging.debug("handleConnectionRequest")

        # Accept the connection
        client = self._public.accept()
        if client is None:
            logging.info("Failed to acept new connection on public socket")
            return

        # Receive a packet
        operation = get_operation(client)
        if operation is None:
            logging.info("Failed to read request on public socket")
            client.close()
            return

        # Find the listening client registered with dip:dport
        if operation["code"] == OperationCode.CONNECT:
            protocol = RelayProtocol.TCP
        elif operation["code"] == OperationCode.INIT_COMM:
            protocol = RelayProtocol.UDP
        else:
            logging.info(
                "Invalid operation code received on local socket: %s",
                operation["code"])
            client.close()
            return

        # TODO: check that index exists
        receiver = self._clients.get((operation["dip"], operation["dport"]))
        if (not receiver or
            (receiver.protocol != protocol and not receiver.listener)):
            put_operation(client, OperationCode.NAK, 0)
            client.close()
            return

        # If the client can accept a connection, or
        # connector is also a client of this broker and on the same network,
        # then let the connector make a direct connection.
        if (receiver.active or
            (operation["cip"] == self.address
                and operation["ip"] == receiver.ip)):
            if operation["code"] == OperationCode.CONNECT:
                code = OperationCode.ACTIVE_CONNECT
            else:
                code = OperationCode.ACTIVE_INIT

            put_operation(client, code, ip=receiver.ip, port=receiver.port)
            client.close()
            return

        # If the listener is active and the connector is passive,
        # then let the listener connect to the connector.
        elif not operation["active"]:
            if operation["code"] == OperationCode.CONNECT:
                if not put_operation(client, OperationCode.PASSIVE_CONNECT):
                    logging.info(
                        "Failed to put PASSIVE_CONNECT after CONNECT")
                    client.close()
                    return

                operation = get_operation(client)
                if operation is None:
                    logging.info("Failed to get after PASSIVE_CONNECT")
                    client.close()
                    return

                if operation["code"] != OperationCode.ACK:
                    logging.info(
                        "Invalid operation received after PASSIVE_CONNECT")
                    client.close()
                    return

                # Connection to the remote client can be closed
                client.close()

                if not put_operation(
                    receiver.relay.outbound, OperationCode.CONTACT_FWD,
                    cip=receiver.cip, cport=receiver.cport,
                    ip=operation["ip"], port=operation["port"]):
                    logging.info("Failed to put CONTACT_FWD after CONNECT")
                    self.deleteRelay(receiver.relay)
                    return

            else:  # INIT_COMM
                if not put_operation(client, OperationCode.PASSIVE_INIT):
                    logging.info("Failed to put PASSIVE_INIT after INIT_COMM")
                    client.close()
                    return

                # Connection to the remote client can be closed
                client.close()

                if not put_operation(
                    receiver.relay.outbound, OperationCode.TRIGGER_FWD,
                    cip=operation["cip"], cport=operation["cport"],
                    ip=operation["ip"], port=operation["port"],
                    dip=receiver.cip, dport=receiver.cport):
                    logging.info("Failed to put CONTACT_FWD after CONNECT")
                    self.deleteRelay(receiver.relay)
                    return

        else:  # Both parties are active
            if operation["code"] == OperationCode.CONNECT:
                code = OperationCode.ACTIVE_CONNECT
            else:
                code = OperationCode.ACTIVE_INIT

            put_operation(client, code, ip=receiver.cip, port=receiver.cport)
            client.close()
            return

    def handleRelayServerEvent(self, relay):
        logging.debug("handleRelayServerEvent")

        operation = get_operation(relay.inbound)
        if operation is None:
            logging.info("Failed to get operation for relay")
            self.deleteRelay(relay)
            return

        if (operation["code"] == OperationCode.TCP_CLIENT or
            operation["code"] == OperationCode.UDP_CLIENT):
            # Create the client
            if operation["code"] == OperationCode.TCP_BIND:
                protocol, listener = (RelayProtocol.TCP, False)
            else:
                protocol, listener = (RelayProtocol.UDP, True)

            client = BrokerClient(
                protocol, operation["cip"], operation["cport"], True,
                operation["ip"], operation["port"], listener, relay)

            # Register the client
            relay.addClient()
            self._clients[(operation["cip"], operation["cport"])] = client
            logging.info(
                "Listening at %s:%s on behalf of %s:%s",
                client.cip, client.cport, client.ip, client.port)

        elif operation["code"] == OperationCode.LISTEN:
            # Find the client and tag it as a listener
            client = self._clients[(operation["cip"], operation["cport"])]
            client.listener = True

        elif (operation["code"] == OperationCode.TCP_DEL or
            operation["code"] == OperationCode.UDP_DEL):
            # Find the client and delete it
            client = self._clients[(operation["cip"], operation["cport"])]
            logging.info(
                "Closing %s:%s on behalf of %s:%s",
                client.cip, client.cport, client.ip, client.port)
            self.deleteClient(client)

        else:
            logging.info("Unexpected operation received")

    def run(self):
        while True:
            selector = Selector()
            selector.add_fd(self._local.fileno(), SelectorIO.READ)
            selector.add_fd(self._public.fileno(), SelectorIO.READ)
            for relay in self._relays:
                selector.add_fd(relay.inbound.fileno(), SelectorIO.READ)

            selector.execute()
            if selector.has_signaled:
                continue
            elif selector.has_failed:
                logging.info(
                    "Failed to select descriptors: %s", selector.get_errstr())
                return

            if selector.check_fd(self._local.fileno(), SelectorIO.READ):
                self.handleClientRequest()
            if selector.check_fd(self._public.fileno(), SelectorIO.READ):
                self.handleConnectionRequest()

            for relay in self._relays:
                if selector.check_fd(
                    relay.inbound.fileno(), SelectorIO.READ):
                    self.handleRelayServerEvent(relay)
