#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "get_broker",
    "get_relay_server",
    "create_manager",
    "delete_manager",
    ]

import logging

from checkbox.lib import param
from checkbox.lib.enum import Enum

from checkbox.io.socktcp import SockTCP

from checkbox.connector.broker import BROKER_LOCAL_PORT
from checkbox.connector.operation import (
    OperationCode,
    get_operation,
    put_operation,
    )


ManagerStatus = Enum(
    "NO_MNG",
    "NEED_RECOVER",
    )


def get_broker():
    broker = param.get("BROKER_INAGENT", None)
    if not broker:
        return

    # TODO: validate broker as IP address
    # TODO: optimize to only use broker when necessary
    # TODO: return address and port
    return broker


def get_relay_server():
    # Get broker address
    broker = get_broker()
    if not broker:
        return None

    # Connect to the broker
    sock = SockTCP()
    if not sock.connect(broker, BROKER_LOCAL_PORT):
        logging.error("Failed to connect to the broker")
        sock.close()
        return None

    # Send NEW_SOCK
    if not put_operation(sock, OperationCode.NEW_SOCK):
        logging.error("Failed to send request to the broker for a socket")
        sock.close()
        return None

    # Receive RELAY_SERVER
    operation = get_operation(sock)
    if operation is None:
        logging.error("Failed to receive confirmation from broker")
        sock.close()
        return None

    if operation["code"] != OperationCode.RELAY_SERVER:
        logging.error("Broker returned invalid response to request")
        sock.close()
        return None

    sock.close()
    logging.debug(
        "Received relay server %s:%s", operation["ip"], operation["port"])

    return (operation["ip"], operation["port"])


def create_manager():
    address = get_relay_server()
    if address is None:
        logging.error("Failed to arrange a relay server through the broker")
        return None

    manager = SockTCP()
    if not manager.connect(*address):
        logging.error("Failed to connect to the new relay server")
        manager.close()
        return None

    return manager


def delete_manager(manager):
    manager.close()
    return ManagerStatus.NO_MNG
