#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "OperationCode",
    "OperationError",
    "get_operation",
    "put_operation",
    ]

import logging
import socket
import struct

from checkbox.lib.enum import (
    Enum,
    enum_value_to_name_string,
    )


OperationCode = Enum(
    "ACK",
    "ACTIVE_CONNECT",
    "ACTIVE_INIT",
    "BIND_OK",
    "CONNECT",
    "CONTACT",
    "CONTACT_FWD",
    "DATA_QUERY",
    "DATA_RESPONSE",
    "DUMMY",
    "DUMMY_ACK",
    "HEARTBEAT",
    "INIT_COMM",
    "LISTEN",
    "NAK",
    "NEW_SOCK",
    "PASSIVE_CONNECT",
    "PASSIVE_INIT",
    "RELAY_SERVER",
    "TCP_BIND",
    "TCP_CLIENT",
    "TCP_DEL",
    "TRIGGER",
    "TRIGGER_FWD",
    "UDP_BIND",
    "UDP_CLIENT",
    "UDP_DEL",
    )

OperationMessage = Enum(
    "ADDRESS",
    "CANONICAL_CLIENT",
    "CANONICAL",
    "CANONICAL_DESTINATION",
    "CANONICAL_PHYSICAL",
    "CODE",
    "DUMMY",
    "ERROR",
    "OFFICIAL_LOCAL",
    "QUERY",
    "QUERY_RESPONSE",
    )

MESSAGE_TO_KEYS = {
    OperationMessage.ADDRESS: (
        "code", "ip", "port",),
    OperationMessage.CANONICAL_CLIENT: (
        "code", "cip", "cport", "ip", "port", "dip", "dport",),
    OperationMessage.CANONICAL: (
        "code", "cip", "cport",),
    OperationMessage.CANONICAL_DESTINATION: (
        "code", "cip", "cport", "active", "ip", "port", "dip", "dport",),
    OperationMessage.CANONICAL_PHYSICAL: (
        "code", "cip", "cport", "ip", "port",),
    OperationMessage.CODE: (
        "code",),
    OperationMessage.DUMMY: (
        "code", "cip", "cport", "dip", "dport",),
    OperationMessage.ERROR: (
        "code", "errcode",),
    OperationMessage.OFFICIAL_LOCAL: (
        "code", "cip", "cport", "active", "ip", "port",),
    OperationMessage.QUERY: (
        "code", "cip",),
    OperationMessage.QUERY_RESPONSE: (
        "code", "ip", "cip",),
    }

CODE_TO_MESSAGE = {
    OperationCode.ACK: OperationMessage.CODE,
    OperationCode.ACTIVE_CONNECT: OperationMessage.ADDRESS,
    OperationCode.ACTIVE_INIT: OperationMessage.ADDRESS,
    OperationCode.BIND_OK: OperationMessage.CANONICAL,
    OperationCode.CONNECT: OperationMessage.CANONICAL_DESTINATION,
    OperationCode.CONTACT_FWD: OperationMessage.CANONICAL_PHYSICAL,
    OperationCode.CONTACT: OperationMessage.ADDRESS,
    OperationCode.DATA_QUERY: OperationMessage.QUERY,
    OperationCode.DATA_RESPONSE: OperationMessage.QUERY_RESPONSE,
    OperationCode.DUMMY_ACK: OperationMessage.DUMMY,
    OperationCode.DUMMY: OperationMessage.DUMMY,
    OperationCode.HEARTBEAT: OperationMessage.CODE,
    OperationCode.INIT_COMM: OperationMessage.CANONICAL_DESTINATION,
    OperationCode.LISTEN: OperationMessage.CANONICAL,
    OperationCode.NAK: OperationMessage.ERROR,
    OperationCode.NEW_SOCK: OperationMessage.CODE,
    OperationCode.PASSIVE_CONNECT: OperationMessage.CODE,
    OperationCode.PASSIVE_INIT: OperationMessage.CODE,
    OperationCode.RELAY_SERVER: OperationMessage.ADDRESS,
    OperationCode.TCP_BIND: OperationMessage.OFFICIAL_LOCAL,
    OperationCode.TCP_CLIENT: OperationMessage.OFFICIAL_LOCAL,
    OperationCode.TCP_DEL: OperationMessage.CANONICAL,
    OperationCode.TRIGGER_FWD: OperationMessage.CANONICAL_CLIENT,
    OperationCode.TRIGGER: OperationMessage.CANONICAL_PHYSICAL,
    OperationCode.UDP_BIND: OperationMessage.OFFICIAL_LOCAL,
    OperationCode.UDP_CLIENT: OperationMessage.OFFICIAL_LOCAL,
    OperationCode.UDP_DEL: OperationMessage.CANONICAL,
    }

KEY_TO_FORMAT = {
    "active": "?",
    "cip": "4s",
    "code": "b",
    "cport": "H",
    "dip": "4s",
    "dport": "H",
    "errcode": "b",
    "ip": "4s",
    "port": "H",
    }

CODE_TO_KEYS = tuple(
    MESSAGE_TO_KEYS[message]
    for code, message in sorted(CODE_TO_MESSAGE.iteritems()))

CODE_TO_FORMAT_KEYS = tuple(
    ("!" + "".join([KEY_TO_FORMAT[key] for key in keys]), keys,)
    for code, keys in enumerate(CODE_TO_KEYS))


class OperationError(Exception):

    pass


def put_operation(stream, code, **kwargs):
    """Put an operation from the stream.

    :param stream: Stream instance.
    :param code: OperationCode value.
    :param kwargs: Keyword arguments for the given code.
    :raises OperationError: if the keyword arguments are invalid.

    :return: True if the put was successful, False otherwise.
    """
    for key, value in kwargs.iteritems():
        if key.endswith("ip"):
            try:
                kwargs[key] = socket.inet_aton(value)
            except (TypeError, socket.error):
                raise OperationError(
                    "Invalid unpacked IP address: %s" % value)

    try:
        fmt, keys = CODE_TO_FORMAT_KEYS[code]
    except IndexError:
        raise OperationError("Invalid operation code: %s" % code)

    kwargs["code"] = code
    try:
        args = [kwargs[key] for key in keys]
    except KeyError, error:
        raise OperationError("Missing operation argument: %s" % error.args)

    try:
        bytes = struct.pack(fmt, *args)
    except (TypeError, struct.error):
        for key, value in kwargs.iteritems():
            if value is None:
                raise OperationError(
                    "Invalid operation value for %s: %s" % (key, value))
        else:
            raise

    logging.debug(
        "Putting %s operation on %s",
        enum_value_to_name_string(OperationCode, code),
        stream.get_description())
    return stream.putBytes(bytes) == len(bytes) and stream.putEnd()


def get_operation(stream):
    """Get an operation from the stream.

    :param stream: Stream instance.
    :raises OpeartionError: if the bytes on the stream are invalid.

    :return: Dictionary mapping operation to values, or None on timeout.
    """
    code = stream.peek()
    if code is None:
        return None

    code = struct.unpack("!b", code)[0]
    try:
        fmt, keys = CODE_TO_FORMAT_KEYS[code]
    except IndexError:
        raise OperationError("Invalid operation code: %s" % code)

    bytes = stream.getBytes(struct.calcsize(fmt))
    if not bytes:
        return None

    if not stream.getEnd():
        return None

    values = struct.unpack(fmt, bytes)
    kwargs = dict(zip(keys, values))
    for key, value in kwargs.iteritems():
        if key.endswith("ip"):
            try:
                kwargs[key] = socket.inet_ntoa(value)
            except socket.error:
                raise OperationError("Invalid packed IP address: %s" % value)

    logging.debug(
        "Got %s operation from %s",
        enum_value_to_name_string(OperationCode, code),
        stream.get_description())
    return kwargs
