#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Relay",
    ]

import os
import errno
import logging
import socket

from collections import deque
from signal import SIGTERM
from time import time

from checkbox.lib.enum import Enum
from checkbox.lib.param import getInteger

from checkbox.io.pipe import Pipe
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.sock import SockPort
from checkbox.io.socktcp import SockTCP
from checkbox.io.sockudp import SockUDP

from checkbox.connector.operation import (
    OperationCode,
    get_operation,
    put_operation,
    )


# TODO: change these to 0, 1, 2
READER_FD = 3
WRITER_FD = 4
MANAGER_FD = 5

DEFAULT_FLUSH_INTERVAL = 90
DEFAULT_MAX_CLIENTS = 10

MAX_STAY_BLOCKED = 50


RelayProtocol = Enum(
    ("NOT_DEFINED", 200),
    ("CLIENT", 201),
    ("MANAGER", 202),
    ("TCP", 203),
    ("UDP", 204),
    )

RelayStatus = Enum(
    ("NONE", 123),
    ("BOUND", 124),
    ("BLOCKED", 125),
    ("LISTENING", 126),
    ("HALF_CONNECTED", 127),
    ("CONNECTED", 128),
    )


class Relay:

    __slots__ = (
        "mip",
        "mport",
        "inbound",
        "outbound",
        "_pid",
        "client_count",
        "client_max",
        )

    def __init__(self, address):
        # Create pipes
        inbound_fds = os.pipe()
        outbound_fds = os.pipe()

        # Create listening socket for exchange with clients
        sock = SockTCP()
        if not sock.bind(address, port=SockPort.INBOUND):
            raise Exception("Failed to bind relay")
        if not sock.listen():
            raise Exception("Failed to listen relay")
        self.mip, self.mport = sock.get_host_and_port()

        # Fork
        pid = os.fork()
        if pid == 0:
            os.close(inbound_fds[0])
            os.close(outbound_fds[1])
            os.dup2(outbound_fds[0], READER_FD)
            os.dup2(inbound_fds[1], WRITER_FD)

            from checkbox.connector.broker import max_fd
            fileno = sock._sock.fileno()
            if hasattr(os, "closerange"):
                os.closerange(WRITER_FD + 1, fileno)
                os.closerange(fileno + 1, max_fd)
            else:
                for fd in xrange(WRITER_FD + 1, max_fd):
                    if fd != fileno:
                        try:
                            os.close(fd)
                        except OSError:
                            pass

            sock._sock = socket.fromfd(
                fileno, sock.get_family(), sock.get_type())
            assert(sock._sock.fileno() == MANAGER_FD)

            # Exec relay server
            server = RelayServer(sock)
            server.run()

        # Parent
        os.close(inbound_fds[1])
        os.close(outbound_fds[0])
        self.inbound = Pipe()
        self.inbound.listen(inbound_fds[0])
        self.outbound = Pipe()
        self.outbound.connect(outbound_fds[1])
        self._pid = pid
        self.client_count = 0
        self.client_max = getInteger("MAX_CLIENTS", DEFAULT_MAX_CLIENTS)
        sock.close()

    def __del__(self):
        os.close(self.inbound.fileno())
        os.close(self.outbound.fileno())
        try:
            os.kill(self._pid, SIGTERM)
        except OSError, error:
            if error.errno != errno.ESRCH:
                raise Exception(
                    "Failed to kill relay (pid: %d), errno %d %s"
                    % (self._pid, error.errno, error.errstr))

    @property
    def has_more_clients(self):
        return self.client_count < self.client_max

    def addClient(self):
        self.client_count += 1

    def deleteClient(self):
        self.client_count -= 1


class RelayInfo:

    def __init__(self, protocol, status, client):
        self.protocol = protocol
        self.status = status
        self.blocked_at = 0
        self.client = client
        self.sock = None

        # UDP
        self.udp_messages = deque()

        # TCP
        self.tcp_bytes = 0
        self.tcp_buffer = []
        self.skipped = -1


class RelayClient:

    def __init__(self, protocol, owner, manager, cip, cport, ip, port):
        self.protocol = protocol
        self.owner = owner
        self.manager = manager
        self.cip = cip
        self.cport = cport
        self.ip = ip
        self.port = port


class RelayServer:

    def __init__(self, manager):
        self._manager = manager
        self._cip = manager.get_host_and_port()[0]

        self._reader = Pipe()
        self._reader.listen(READER_FD)

        self._writer = Pipe()
        self._writer.connect(WRITER_FD)

        self._flush_time = time()

        self._clients = {}
        self._infos = {}

    def _flush(self):
        now = time()
        for sock, info in self._infos.items():
            if info.protocol == RelayProtocol.TCP:
                if (info.status == RelayStatus.LISTENING and
                    now - info.blocked_at > MAX_STAY_BLOCKED):
                    # Close the spouse
                    spouse = info.sock
                    del self._infos[spouse]
                    spouse.close()

                    # Close the sock
                    del self._infos[sock]
                    sock.close()

            elif info.protocol == RelayProtocol.UDP:
                info.skipped += 1
                if info.skipped > 1:
                    del self._infos[sock]
                    sock.close()

    def _deleteClient(self, client):
        if client.protocol == RelayProtocol.TCP:
            # Send TCP_DEL to the broker
            if not put_operation(
                self._writer, OperationCode.TCP_DEL,
                cip=client.cip, cport=client.cport):
                raise Exception("Failed to delete client")

            # Close CLIENT and MANAGER
            del self._infos[client.owner]
            client.owner.close()
            del self._infos[client.manager]
            client.manager.close()

            for sock, info in self._infos.iteritems():
                if (info.protocol == RelayProtocol.TCP and
                    info.client is not None and
                    info.client == client):
                    # Unlink the client
                    info.client = None

        else:
            raise NotImplementedError("UDP not supported")

        index = (client.cip, client.cport)
        del self._clients[index]

    def _handleBrokerEvent(self):
        operation = get_operation(self._reader)
        if operation is None:
            raise Exception("Failed to receive operation from broker")

        if operation["code"] == OperationCode.CONTACT_FWD:
            index = (operation["cip"], operation["cport"])
            for client in self._clients[index]:
                if client.protocol == RelayProtocol.TCP:
                    if not put_operation(
                        client.manager, OperationCode.CONTACT,
                        ip=operation["ip"], port=operation["port"]):
                        logging.info(
                            "CONTACT_FWD(%s:%s) failed to forward CONTACT"
                            % index)
                        self._deleteClient(client)
                    break

        elif operation["code"] == OperationCode.TRIGGER_FWD:
            index = (operation["dip"], operation["dport"])
            for client in self._clients[index]:
                if client.protocol == RelayProtocol.UDP:
                    if not put_operation(
                        client.manager, OperationCode.TRIGGER,
                        cip=operation["cip"], cport=operation["cport"],
                        ip=operation["ip"], port=operation["port"]):
                        logging.info(
                            "TRIGGER_FWD(%s:%s) failed to forward TRIGGER"
                            % index)
                        self._deleteClient(client)
                    break

        else:
            logging.info("Unexpected operation code received")

    def _handleNewClient(self):
        # Accept a connection
        manager = SockTCP()
        if not self._manager.acceptSock(manager):
            logging.info("Failed to accept from manager socket")
            return

        # Receive command from the client
        operation = get_operation(manager)
        if operation is None:
            logging.info("Failed to receive operation from manager")
            manager.close()
            return

        # Create a proxy socket of the appropriate type
        if operation["code"] == OperationCode.TCP_BIND:
            proxy = SockTCP()
        elif operation["code"] == OperationCode.UDP_BIND:
            proxy = SockUDP()
        else:
            logging.info("Neither TCP_BIND nor UDP_BIND received")
            manager.close()
            return

        # Bind the proxy socket to cport
        if not proxy.bind(self._cip, operation["cport"]):
            put_operation(manager, OperationCode.NAK, errcode=0)
            proxy.close()
            manager.close()
            return

        # Get cip and cport to reply to the client
        cip, cport = proxy.get_host_and_port()
        if cip is None:
            put_operation(manager, OperationCode.NAK, errcode=0)
            proxy.close()
            manager.close()
            return

        ip = operation["ip"]
        port = operation["port"]
        if not port:
            port = cport

        # Create a client
        if operation["code"] == OperationCode.TCP_BIND:
            protocol = RelayProtocol.TCP
        else:
            protocol = RelayProtocol.UDP

        client = RelayClient(protocol, proxy, manager, cip, cport, ip, port)

        # Answer BIND_OK
        if not put_operation(
            manager, OperationCode.BIND_OK, cip=cip, cport=cport):
            proxy.close()
            manager.close()
            return

        # Send *_CLIENT to the broker
        if operation["code"] == OperationCode.TCP_BIND:
            code = OperationCode.TCP_CLIENT
        else:
            code = OperationCode.UDP_CLIENT

        if not put_operation(
            self._writer, code, cip=cip, cport=cport, ip=ip, port=port,
            active=operation["active"]):
            raise Exception("Failed to send *_CLIENT to broker")

        # Register the client
        if operation["code"] == OperationCode.TCP_BIND:
            proxy_status = RelayStatus.BOUND
        else:
            proxy_status = RelayStatus.LISTENING

        self._infos[proxy] = RelayInfo(
            RelayProtocol.CLIENT, proxy_status, client)
        self._infos[manager] = RelayInfo(
            RelayProtocol.MANAGER, RelayStatus.CONNECTED, client)

        index = (cip, cport)
        clients = self._clients.setdefault(index, [])
        clients.append(client)

        logging.info("Added client")

    def _handleClientRequest(self, manager):
        operation = get_operation(manager)
        if operation is None:
            logging.info(
                "Failed to receive operation from the manager, "
                "deleting the client")
            self._deleteClient(self._infos[manager].client)
            return

        if operation["code"] == OperationCode.HEARTBEAT:
            pass  # Do nothing.
        elif operation["code"] == OperationCode.LISTEN:
            self._handleListen(manager, operation["cip"], operation["cport"])
        else:
            logging.info("Received unexpected operation from manager socket.")

    def _handleListen(self, manager, cip, cport):
        # Make the client passive and tag is as LISTENING
        info = self._infos[manager]
        owner = info.client.owner
        if not owner.listen():
            if not put_operation(manager, OperationCode.NAK, errcode=0):
                logging.info("Failed to answer NAK, deleting the client")
                self._deleteClient(info.client)

            return

        # Send LISTEN to the broker
        if not put_operation(
            self._writer, OperationCode.LISTEN, cip=cip, cport=cport):
            raise Exception("Failed to send LISTEN to broker")

        # Answer ACK to the client
        if not put_operation(manager, OperationCode.ACK):
            logging.info("Failed to answer AK, deleting the client")
            self._deleteClient(info.client)

        info.status = RelayStatus.LISTENING

    def _handlePacketToClient(self, fd):
        raise NotImplementedError

    def _handlePacketFromClient(self, fd):
        raise NotImplementedError

    def _startAddTCPRelay(self, fd):
        raise NotImplementedError

    def _finishPassiveAdd(self, fd):
        raise NotImplementedError

    def _finishActiveAdd(self, fd):
        raise NotImplementedError

    def _pushTCPData(self, fd):
        raise NotImplementedError

    def _relayTCPSock(self, sock):
        raise NotImplementedError

    def run(self):
        while True:
            current_time = time()
            if current_time - self._flush_time > DEFAULT_FLUSH_INTERVAL:
                self._flush()
                self._flush_time = current_time

            selector = Selector()
            selector.add_fd(READER_FD, SelectorIO.READ)
            selector.add_fd(MANAGER_FD, SelectorIO.READ)
            for sock, info in self._infos.iteritems():
                # Manager
                if ((info.protocol == RelayProtocol.MANAGER) or
                    # Listening client
                    (info.protocol == RelayProtocol.CLIENT and
                        info.status == RelayStatus.LISTENING) or
                    # Half-connected passive TCP connection
                    (info.protocol == RelayProtocol.TCP and
                        info.status == RelayStatus.LISTENING) or
                    # Connected TCP without pending messages
                    (info.protocol == RelayProtocol.TCP and
                        info.status == RelayStatus.CONNECTED and
                        not self._infos[info.sock].tcp_buffer) or
                    # Connected or blocked UDP
                    (info.protocol == RelayProtocol.UDP)):
                    selector.add_fd(sock.fileno(), SelectorIO.READ)

                # Connected TCP with pending messages
                if ((info.protocol == RelayProtocol.TCP and
                        info.status == RelayStatus.CONNECTED and
                        self._infos[info.sock].tcp_buffer) or
                    # Half-connected active TCP connection
                    (info.protocol == RelayProtocol.TCP and
                        info.status == RelayStatus.BLOCKED)):
                    selector.add_fd(sock.fileno(), SelectorIO.WRITE)

            selector.setTimeout(DEFAULT_FLUSH_INTERVAL)
            selector.execute()
            if selector.has_failed:
                logging.error(
                    "Failed to select descriptors: %s", selector.get_errstr())
                return
            elif not selector.has_ready:
                continue

            if selector.check_fd(READER_FD, SelectorIO.READ):
                self._handleBrokerEvent()
            if selector.check_fd(MANAGER_FD, SelectorIO.READ):
                self._handleNewClient()

            for sock, info in self._infos.items():
                if selector.check_fd(sock.fileno(), SelectorIO.READ):
                    if info.protocol == RelayProtocol.MANAGER:
                        self._handleClientRequest(sock)

                    elif info.protocol == RelayProtocol.CLIENT:
                        if info.client.protocol == RelayProtocol.TCP:
                            self._startAddTCPRelay(sock)
                        else:
                            self._handlePacketToClient(sock)

                    elif info.protocol == RelayProtocol.TCP:
                        if info.status == RelayStatus.LISTENING:
                            self._finishPassiveAdd(sock)
                        elif info.status == RelayStatus.CONNECTED:
                            self._relayTCPSock(sock)

                    elif info.protocol == RelayProtocol.UDP:
                        self._handlePacketFromClient(sock)

                    else:
                        logging.error(
                            "Invalid fd[%d] is ready", sock.fileno())

                if selector.check_fd(sock.fileno(), SelectorIO.WRITE):
                    if info.status == RelayStatus.BLOCKED:
                        self._finishActiveAdd(sock)
                    else:
                        self._pushTCPData(sock)
