#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Sock",
    ]

import os
import errno
import fcntl
import logging
import signal
import socket

from collections import deque
from time import time

from checkbox.lib.enum import Enum
from checkbox.lib.hostname import ipaddress

from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.sock import (
    SockAddress,
    SockPort,
    )
from checkbox.io.socktcp import SockTCP

from checkbox.connector.broker import BROKER_PUBLIC_PORT
from checkbox.connector.manager import (
    ManagerStatus,
    create_manager,
    delete_manager,
    get_broker,
    )
from checkbox.connector.operation import (
    OperationCode,
    get_operation,
    put_operation,
    )
from checkbox.connector.util import (
    do_accept,
    do_connect,
    )


CONN_TIMEOUT = 120
MAX_COLLISIONS = 20  # This is very scientific!

HEARTBEAT_RATE = 40
KEEP_EYE_RATE = 15


SocketType = Enum(
    "NOT_DEFINED",
    "TCP",
    "UDP",
    )

SocketStatus = Enum(
    "NONE",
    "CREATED",
    "BOUND",
    "LISTENING",
    "ACTIVE_CONNECTING",
    "PASSIVE_CONNECTING",
    "CONNECTED",
    "CONN_FAILED",
    )


# Rename this to something more Sock specific
class BrokerError(Exception):

    pass


class SockConnection:

    __slots__ = (
        "sock",
        "ip",
        "port",
        )

    def __init__(self, sock, ip, port):
        self.sock = sock
        self.ip = ip
        self.port = port


class Sock:

    __slots__ = (
        "_active",
        "_manager",
        "_status",
        "_type",
        "_cip",
        "_cport",
        "_ip",
        "_port",
        "_queue",
        "_sock",
        )

    _initialized = False
    _last_heart_beat = 0

    def __init__(
        self, family=socket.AF_INET, type=socket.SOCK_STREAM,
        proto=socket.IPPROTO_IP, _sock=None):
        if not Sock._initialized:
            # Let the process ignore SIGPIPE
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            Sock._initialized = True

        if _sock is None:
            _sock = socket.socket(family, type, proto, _sock)
        self._sock = _sock
        self._cip = None
        self._cport = 0
        self._ip = None
        self._port = 0
        self._queue = deque()

        # Check if the socket is in the supported family and type
        if (family != socket.AF_INET or
            (type != socket.SOCK_STREAM and type != socket.SOCK_DGRAM)):
            logging.info("Unsupported family and type, using direct socket.")
            self._type = SocketType.NOT_DEFINED

        else:
            self._status = SocketStatus.CREATED
            self._manager = ManagerStatus.NO_MNG

            if type == socket.SOCK_STREAM:
                self._type = SocketType.TCP
            else:
                raise NotImplementedError

            if get_broker():
                self._active = True
            else:
                self._active = False

    def fileno(self):
        return self._sock.fileno()

    def getsockname(self):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported getsockname, using direct socket.")
            return self._sock.getsockname()

        return (self._cip, self._cport,)

    def getsockopt(self, *args, **kwargs):
        return self._sock.getsockopt(*args, **kwargs)

    def setsockopt(self, *args, **kwargs):
        return self._sock.setsockopt(*args, **kwargs)

    def accept(self):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported accept, using direct socket.")
            return self._sock.accept()

        # Do sanity checking
        if self._status != SocketStatus.LISTENING:
            raise socket.error(errno.EINVAL, "Socket not LISTENING")

        # Check if the socket is nonblocking
        flags = fcntl.fcntl(self._sock.fileno(), fcntl.F_GETFL)
        while True:
            # Return the connection from the queue
            if self._queue:
                sock, address = self._queue.popleft()
                # TODO: fix this
                #sock._type = SocketType.NOT_DEFINED
                return (sock, address,)

            # If socket is nonblocking, return EAGAIN
            if flags & os.O_NONBLOCK:
                raise socket.error(
                    errno.EAGAIN, "Failed to accept on nonblocking socket")

            # Wait until the connection queue becomes nonempty
            self.select(SelectorIO.READ)

    def _bindInit(self, address):
        if not self._ip:
            if address[0] != socket.INADDR_ANY:
                self._ip = address[0]
            else:
                self._ip = ipaddress()
                if not self._ip:
                    raise socket.error(
                        errno.EINVAL, "Unable to determine local IP address")

    def _bindRegister(self, port):
        # Create a manager
        manager = self._manager = create_manager()
        if not manager:
            raise socket.error("Failed to connect to the broker")

        # Send BIND to relay server
        if self._type == SocketType.TCP:
            operation = OperationCode.TCP_BIND
        else:
            operation = OperationCode.UDP_BIND
        if not put_operation(
            manager, operation, cip=SockAddress.ANY, cport=port,
            active=self._active, ip=self._ip, port=SockPort.ANY):
            self._manager = delete_manager(manager)
            raise socket.error(
                errno.EPIPE, "Failed to request bind via broker")

        # Receive the result
        operation = get_operation(manager)
        if operation is None:
            self._manager = delete_manager(manager)
            raise socket.error(
                errno.ENOTCONN, "Failed to receive response from broker")

        if operation["code"] == OperationCode.BIND_OK:
            self._cip = operation["cip"]
            self._cport = operation["cport"]
            self._status = SocketStatus.BOUND
        elif operation["code"] == OperationCode.NAK:
            self._manager = delete_manager(manager)
            raise socket.error(
                operation["errcode"], "%s sent; NAK received", operation)
        else:
            self._manager = delete_manager(manager)
            raise socket.error("Received unexpected message")

    def _bindUnregister(self):
        if self._manager:
            self._manager = delete_manager(self._manager)

        self._cip = 0
        self._cport = 0
        self._status = SocketStatus.CREATED

    def _bindSave(self, is_local=False):
        # Find the port
        ip, port = self.getsockname()
        self._port = port

        # If there's nothing at the broker, record the same address
        if not self._active or is_local:
            self._cip = self._ip
            self._cport = self._port
            self._status = SocketStatus.BOUND

    def bind(self, address):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported bind, using direct socket.")
            self._sock.bind(address)

        logging.debug("Binding to broker %s:%s" % tuple(address))

        # Check if the socket has already been bound
        if self._status != SocketStatus.CREATED:
            raise socket.error(
                errno.EINVAL,
                "Socket not CREATED, it may have been bound elsewhere")

        if self._port:
            raise socket.error(
                errno.EINVAL,
                "Socket has not been bound but has a port associated with it")

        # Initialize the local IP address
        self._bindInit(address)

        # Try to use the same port number on both sides
        tweaked_address = list(address)
        collision_count = 0

        while True:
            if self._active:
                self._bindRegister(address[1])
                registered = True
                tweaked_address[1] = self._cport
            else:
                registered = False

            try:
                self._sock.bind(tuple(tweaked_address))
            except socket.error, error:
                if registered:
                    self._bindUnregister()
                    if error.errno == errno.EADDRINUSE and not address[1]:
                        logging.debug(
                            "Collision attempting to allocate same port")
                        collision_count += 1
                        if collision_count < MAX_COLLISIONS:
                            continue

                logging.error("Failed binding the local socket")
                raise
            else:
                break

        try:
            self._bindSave()
        except socket.error:
            if registered:
                self._bindUnregister()
            raise

    def close(self):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported close, using direct socket.")
            self._sock.close()

        logging.debug("Closing socket")

        # Close the manager
        if self._manager != ManagerStatus.NO_MNG:
            self._manager = delete_manager(self._manager)

        self._type = SocketType.NOT_DEFINED
        self._sock.close()

    def _connectDirect(self, address):
            if self._type == SocketType.UDP:
                raise NotImplementedError("UDP not supported")
            else:
                try:
                    do_connect(self._sock, address)
                except socket.error, error:
                    if error.errno == errno.EINPROGRESS:
                        self._status = SocketStatus.ACTIVE_CONNECTING
                    else:
                        self._status = SocketStatus.CONN_FAILED

                    raise

            self._status = SocketStatus.CONNECTED

    def _connectTCP(self, address):
        # Send connect request to the destination's broker
        temp_address = list(address)
        temp_address[1] = BROKER_PUBLIC_PORT
        temp_sock = SockTCP()
        if not temp_sock.connect(*temp_address):
            raise BrokerError(
                "Failed to connect to %s:%s" % tuple(temp_address))

        if not put_operation(
            temp_sock, OperationCode.CONNECT, cip=self._cip,
            cport=self._cport, active=self._active, ip=self._ip,
            port=self._port, dip=address[0], dport=address[1]):
            raise socket.error(
                errno.EPIPE, "Failed to send CONNECT to broker")

        # Receive response
        operation = get_operation(temp_sock)
        if operation is None:
            raise socket.error(
                errno.ENOTCONN,
                "Failed to receive response to CONNECT from broker")

        if operation["code"] == OperationCode.PASSIVE_CONNECT:
            try:
                self._sock.listen(5)
            except socket.error:
                logging.info(
                    "CONNECT sent; PASSIVE_CONNECT received; listen failed")

            if not put_operation(temp_sock, OperationCode.ACK):
                self._status = SocketStatus.CONN_FAILED
                raise socket.error(
                    errno.EPIPE, "Failed to send ACK for PASSIVE_CONNECT")

            temp_sock.close()
            self._status = SocketStatus.PASSIVE_CONNECTING
            try:
                new_sock = do_accept(self._sock, address, CONN_TIMEOUT)
            except socket.error, error:
                if error.errno == errno.EAGAIN:
                    raise socket.error(errno.EINPROGRESS, "Waiting on accept")
                else:
                    logging.info(
                        "CONNECT sent; PASSIVE_CONNECT received; "
                        "accept failed")
                    raise

            # Copy socket options
            flags = fcntl.fcntl(self.fileno(), fcntl.F_GETFL)
            fcntl.fcntl(new_sock.fileno(), fcntl.F_SETFL, flags)
            os.dup2(new_sock.fileno(), self.fileno())
            new_sock.close()
            self._status = SocketStatus.CONNECTED
            logging.info("Passively connected to %s:%s" % tuple(address))

        elif operation["code"] == OperationCode.ACTIVE_CONNECT:
            temp_sock.close()
            try:
                do_connect(
                    self._sock, (operation["ip"], operation["port"]),
                    CONN_TIMEOUT)
            except socket.error, error:
                if error.errno == errno.EINPROGRESS:
                    self._status = SocketStatus.ACTIVE_CONNECTING
                else:
                    self._status = SocketStatus.CONN_FAILED
                raise

            self._status = SocketStatus.CONNECTED

        elif operation["code"] == OperationCode.NAK:
            temp_sock.close()
            raise BrokerError("CONNECT sent; NAK received")

        else:
            temp_sock.close()
            raise BrokerError("Received unexpected message")

    def connect(self, address):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported connect, using direct socket.")
            return self._sock.connect(address)

        # Check address family
        if not isinstance(address, tuple) and len(address) != 2:
            logging.info("Unsupported family, using direct socket.")
            try:
                self._sock.connect(address)
            except socket.error, error:
                if error.errno == errno.EINPROGRESS:
                    self._status = SocketStatus.ACTIVE_CONNECTING
                else:
                    raise

        logging.debug("Connecting to broker %s:%s" % tuple(address))

        if self._type == SocketType.UDP:
            raise NotImplementedError("UDP not supported")

        # If TCP and registered to the RelayServer, unregister the socket
        elif self._type == SocketType.TCP:
            if self._manager == ManagerStatus.NEED_RECOVER:
                self._manager = ManagerStatus.NO_MNG
            elif self._manager != ManagerStatus.NO_MNG:
                self._manager.close()

            self._cport = 0

        # TODO: optimze by checking the routing table

        # Try a broker connection procedure
        if self._status == SocketStatus.CREATED:
            self._sock.bind(("", 0))
            self._ip = ipaddress()
            self._port = self._cport = self.getsockname()[1]
            self._cip = get_broker()
            if not self._cip:
                self._cip = self._ip

            self._status = SocketStatus.BOUND

        # Now do a broker connection
        if self._type == SocketType.TCP:
            try:
                self._connectTCP(address)
            except BrokerError:
                self._connectDirect(address)

        else:
            raise NotImplementedError("UDP not supported")

    def listen(self, backlog):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported listen, using direct socket.")
            self._sock.listen(backlog)

        logging.debug("Listening with %s", backlog)

        # Check if the socket has already been bound
        if self._status != SocketStatus.BOUND:
            raise socket.error(
                errno.EINVAL,
                "Socket not BOUND, it may have be listening elsewhere")

        # If there's no manager socket, remove it
        manager = self._manager
        if manager == ManagerStatus.NO_MNG:
            self._type = SocketType.NOT_DEFINED

            return self._sock.listen(backlog)

        # Send LISTEN to broker
        if (manager != ManagerStatus.NO_MNG and
            manager != ManagerStatus.NEED_RECOVER):
            if not put_operation(
                manager, OperationCode.LISTEN,
                cip=self._cip, cport=self._cport):
                delete_manager(manager)
                self._manager = ManagerStatus.NEED_RECOVER
                raise socket.error(errno.EPIPE, "Failed to listen")

            # Receive the result
            operation = get_operation(manager)
            if operation is None:
                delete_manager(manager)
                self._manager = ManagerStatus.NEED_RECOVER
                raise socket.error(
                    errno.ENOTCONN, "Failed to receive response from broker")

            if operation["code"] == OperationCode.ACK:
                pass
            elif operation["code"] == OperationCode.NAK:
                self._manager = delete_manager(manager)
                raise socket.error(
                    operation["errcode"], "LISTEN  sent; NAK received")
            else:
                self._manager = delete_manager(manager)
                raise socket.error("Received unexpected message")

        try:
            self._sock.listen(backlog)
        except socket.error:
            self._manager = delete_manager(manager)
            raise
        else:
            self._status = SocketStatus.LISTENING

    def _isBlocking(self, flags):
        if flags & socket.MSG_DONTWAIT:
            return False

        current = fcntl.fcntl(self._sock.fileno(), fcntl.F_GETFL)
        if current & os.O_NONBLOCK:
            return False

        return True

    def _recvNonBlocking(self, size, flags):
        sock_flags = fcntl.fcntl(self.fileno(), fcntl.F_GETFL)

        while True:
            try:
                return self._sock.recv(size, flags | socket.MSG_DONTWAIT)
            except socket.error, error:
                if error.errno != errno.EAGAIN:
                    raise

            if (sock_flags & os.O_NONBLOCK or
                sock_flags & socket.MSG_DONTWAIT):
                raise socket.error(errno.EAGAIN, "Requested nonlocking recv")

            selector = Selector()
            selector.setTimeout(10)
            selector.add_fd(self.fileno(), SelectorIO.READ)
            selector.execute()
            if selector.has_failed:
                raise socket.error(selector.get_errno(), "Failed to select")

        raise Exception("Should not get here")

    def _recvTCP(self, size, flags):
        if not self._isBlocking(flags):
            return self._sock.recv(size, flags)

        return self._recvNonBlocking(size, flags)

    def recv(self, size, flags=0):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported recv, using direct socket.")
            return self._sock.recv(size, flags)

        if self._type == SocketType.TCP:
            return self._recvTCP(size, flags)
        elif self._type == SocketType.UDP:
            raise NotImplementedError("UDP not supported")
        else:
            raise socket.error(
                errno.EBADF, "Failed sending to internal socket")

    def _sendNonBlocking(self, string, flags):
        while True:
            try:
                return self._sock.send(string, flags | socket.MSG_DONTWAIT)
            except socket.error, error:
                if error.errno != errno.EAGAIN:
                    raise

            self.select(SelectorIO.WRITE)

        raise Exception("Should not get here")

    def send(self, string, flags=0):
        if self._type == SocketType.NOT_DEFINED:
            logging.info("Unsupported send, using direct socket.")
            return self._sock.send(string, flags)

        logging.debug("Sending %s bytes with %s", len(string), flags)

        if not self._isBlocking(flags):
            return self._sock.send(string, flags)

        return self._sendNonBlocking(string, flags)

    def _handleAsyncEvent(self, selector):
        if selector.check_fd(self._manager.fileno(), SelectorIO.READ):
            self._handleCommand()

        elif self._type == SocketType.TCP:
            if self._status == SocketStatus.LISTENING:
                self._handleListener()
            elif self._status == SocketStatus.PASSIVE_CONNECTING:
                self._handlePassiveBlockedConnection()
            elif self._status == SocketStatus.ACTIVE_CONNECTING:
                self._handleActiveBlockedConnection()

        elif self._type == SocketType.UDP:
            raise NotImplementedError("UDP not supported")

    def _handleCommand(self):
        manager = self._manager
        operation = get_operation(manager)
        if operation is None:
            delete_manager(manager)
            self._manager = ManagerStatus.NEED_RECOVER
            raise socket.error(
                errno.ENOTCONN, "Failed to receive operation from broker")

        if operation["code"] in (
            OperationCode.BIND_OK,
            OperationCode.ACK,
            OperationCode.NAK,
            OperationCode.ACTIVE_CONNECT,
            OperationCode.PASSIVE_CONNECT,
            OperationCode.ACTIVE_INIT,
            OperationCode.PASSIVE_INIT,
            ):
            logging.debug("Delayed synchronous message received; ignored")
            return

        elif operation == OperationCode.CONTACT:
            return self._handleContact(operation["ip"], operation["port"])

        elif operation == OperationCode.TRIGGER:
            # TODO: check for UDP
            return self._handleTrigger(
                operation["cip"], operation["cport"], operation["ip"],
                operation["port"])

        else:
            raise socket.error("Unsupported operation received")

    def _handleContact(self, address):
        # Do sanity checking
        if self._status != SocketStatus.LISTENING:
            put_operation(
                self._manager, OperationCode.REMOTE_REPORT,
                cip=address[0], cport=address[1], active=2, errcode=1)
            raise socket.error(errno.EINVAL, "Socket not LISTENING")

        sock = SockTCP()
        try:
            do_connect(sock, address)
        except socket.error:
            sock.close()
            put_operation(
                self._manager, OperationCode.REMOTE_REPORT,
                cip=address[0], cport=address[1], active=2, errcode=1)
            raise

        # Queue the connection
        self._queue.append(sock, address)
        put_operation(
            self._manager, OperationCode.REMOTE_REPORT,
            cip=address[0], cport=address[1], active=2, errcode=0)

    def _handleListener(self):
        (sock, address) = self._sock.accept()
        self._queue.append((sock, address))
        logging.debug("Connection accepted from %s:%s" % address)

    def _handleTrigger(self, cip, cport, ip, port):
        raise NotImplementedError

    def select(self, interest, timeout=None):
        first = True
        if timeout:
            remained = timeout
        else:
            remained = KEEP_EYE_RATE

        while True:
            # Trigger heartbeat
            current = time()
            if current - Sock._last_heart_beat > HEARTBEAT_RATE:
                # TODO: finish this
                #self._triggerHeartBeat(current)
                Sock._last_heart_beat = current

            # TODO: add support for broker going down temporarily

            # Prepare descriptors to be selected
            selector = Selector()
            selector.add_fd(self._manager.fileno(), SelectorIO.READ)
            if (self._type == SocketType.UDP or
                (self._type == SocketType.TCP and
                    (self._status == SocketStatus.LISTENING or
                    self._status == SocketStatus.PASSIVE_CONNECTING))):
                if interest == SelectorIO.READ:
                    selector.add_fd(self._sock.fileno(), SelectorIO.READ)
                elif self._status == SocketStatus.PASSIVE_CONNECTING:
                    if interest == SelectorIO.WRITE:
                        selector.add_fd(self._sock.fileno(), SelectorIO.WRITE)
                    elif interest == SelectorIO.EXCEPT:
                        selector.add_fd(
                            self._sock.fileno(), SelectorIO.EXCEPT)

            elif (self._type == SocketType.TCP and
                self._status == SocketStatus.ACTIVE_CONNECTING):
                if interest == SelectorIO.WRITE:
                    selector.add_fd(self._sock.fileno(), SelectorIO.WRITE)

            # Non-blocking connect for the first time
            if first:
                selector.setTimeout(0)
            elif remained >= HEARTBEAT_RATE:
                selector.setTimeout(HEARTBEAT_RATE)
            else:
                selector.setTimeout(remained)

            # Execute select
            before = time()
            selector.execute()
            after = time()
            first = False

            if timeout:
                remained -= after - before

            if selector.has_failed or selector.has_signaled:
                raise socket.error(selector.get_errno(), "Failed to select")
            elif selector.has_ready:
                self._handleAsyncEvent(selector)

            if self._type == SocketType.TCP:
                if self._status in (
                    SocketStatus.NONE,
                    SocketStatus.CREATED,
                    SocketStatus.BOUND,
                    SocketStatus.PASSIVE_CONNECTING,
                    ):
                    break
                elif self._status == SocketStatus.LISTENING:
                    if self._queue:
                        break
                elif self._status in (
                    SocketStatus.ACTIVE_CONNECTING,
                    SocketStatus.CONNECTED,
                    SocketStatus.CONN_FAILED,
                    ):
                    raise NotImplementedError
                else:
                    raise Exception("Should not get here")

            # TODO: adjust descriptors

            if remained <= 0:
                return
