#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from checkbox.io.pipe import Pipe

from checkbox.connector.operation import (
    OperationCode,
    OperationError,
    get_operation,
    put_operation,
    )

from checkbox.tests.cases import TestCase
from checkbox.tests.unique import UniqueMixin


class TestOperation(UniqueMixin, TestCase):

    def setUp(self):
        super(TestOperation, self).setUp()

        reader_fd, writer_fd = os.pipe()
        self.reader = Pipe()
        self.reader.listen(reader_fd)
        self.writer = Pipe()
        self.writer.connect(writer_fd)

        self.cip = self.getUniqueIPAddress()
        self.cport = self.getUniqueInteger(16)
        self.ip = self.getUniqueIPAddress()
        self.active = True
        self.port = self.getUniqueInteger(16)
        self.dip = self.getUniqueIPAddress()
        self.dport = self.getUniqueInteger(16)
        self.errcode = self.getUniqueInteger(8)

    def assertOperation(self, code, **kwargs):
        self.assertTrue(put_operation(self.writer, code, **kwargs))
        operation = get_operation(self.reader)
        self.assertNotEquals(operation, None)
        for key, value in kwargs.iteritems():
            self.assertEquals(operation[key], value)

    def test_CODE(self):
        for code in (
            OperationCode.ACK,
            OperationCode.HEARTBEAT,
            OperationCode.NEW_SOCK,
            OperationCode.PASSIVE_CONNECT,
            OperationCode.PASSIVE_INIT,
            ):
            self.assertOperation(code)

    def test_ADDRESS(self):
        for code in (
            OperationCode.ACTIVE_CONNECT,
            OperationCode.ACTIVE_INIT,
            OperationCode.CONTACT,
            OperationCode.RELAY_SERVER,
            ):
            self.assertOperation(code, ip=self.ip, port=self.port)

    def test_CANONICAL_CLIENT(self):
        for code in (
            OperationCode.TRIGGER_FWD,
            ):
            self.assertOperation(
                code, cip=self.cip, cport=self.cport, ip=self.ip,
                port=self.port, dip=self.dip, dport=self.dport)

    def test_CANONICAL(self):
        for code in (
            OperationCode.BIND_OK,
            OperationCode.LISTEN,
            OperationCode.TCP_DEL,
            OperationCode.UDP_DEL,
            ):
            self.assertOperation(code, cip=self.cip, cport=self.cport)

    def test_CANONICAL_DESTINATION(self):
        for code in (
            OperationCode.CONNECT,
            OperationCode.INIT_COMM,
            ):
            self.assertOperation(
                code, cip=self.cip, cport=self.cport, active=self.active,
                ip=self.ip, port=self.port, dip=self.dip, dport=self.dport)

    def test_CANONICAL_PHYSICAL(self):
        for code in (
            OperationCode.CONTACT_FWD,
            OperationCode.TRIGGER,
            ):
            self.assertOperation(
                code, cip=self.cip, cport=self.cport,
                ip=self.ip, port=self.port)

    def test_DUMMY(self):
        for code in (
            OperationCode.DUMMY,
            OperationCode.DUMMY_ACK,
            ):
            self.assertOperation(
                code, cip=self.cip, cport=self.cport,
                dip=self.dip, dport=self.dport)

    def test_ERROR(self):
        for code in (
            OperationCode.NAK,
            ):
            self.assertOperation(code, errcode=self.errcode)

    def test_OFFICIAL_LOCAL(self):
        for code in (
            OperationCode.TCP_BIND,
            OperationCode.TCP_CLIENT,
            OperationCode.UDP_BIND,
            OperationCode.UDP_CLIENT,
            ):
            self.assertOperation(
                code, cip=self.cip, cport=self.cport, active=self.active,
                ip=self.ip, port=self.port)

    def test_QUERY(self):
        for code in (
            OperationCode.DATA_QUERY,
            ):
            self.assertOperation(code, cip=self.cip)

    def test_QUERY_RESPONSE(self):
        for code in (
            OperationCode.DATA_RESPONSE,
            ):
            self.assertOperation(code, ip=self.ip, cip=self.cip)

    def test_invalidCode(self):
        self.assertRaises(
            OperationError, put_operation, self.writer,
            len(OperationCode.__dict__))

    def test_invalidKwargs(self):
        self.assertRaises(
            OperationError, put_operation, self.writer,
            OperationCode.CONTACT)

    def test_invalidIPAddress(self):
        self.assertRaises(
            OperationError, put_operation, self.writer,
            OperationCode.CONTACT, ip=".")
