#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "do_connect",
    ]

import os
import errno
import fcntl
import socket

from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )


def do_connect(sock, address, timeout=None):
    fileno = sock.fileno()

    # If nonblocking socket, call connect and return the result
    flags = fcntl.fcntl(fileno, fcntl.F_GETFL)
    if flags & os.O_NONBLOCK:
        return sock.connect(address)

    # Do nonblocking connect
    fcntl.fcntl(fileno, fcntl.F_SETFL, flags | os.O_NONBLOCK)
    try:
        sock.connect(address)
    except socket.error, error:
        if error.errno != errno.EINPROGRESS:
            fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
            raise
    else:
        fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
        return

    # Wait until connection finished or failed
    while True:
        selector = Selector()
        selector.setTimeout(timeout)
        selector.add_fd(fileno, SelectorIO.WRITE)
        selector.add_fd(fileno, SelectorIO.READ)

        selector.execute()
        if selector.has_signaled:
            continue
        elif selector.has_failed:
            fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
            raise socket.error(
                selector.get_errno(), "Failed to select")
        elif not selector.has_ready:
            fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
            raise socket.error(
                errno.ETIMEDOUT, "Timed out after %s seconds" % timeout)
        else:
            break

    # Check if the connection established
    error = sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
    if error:
        fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
        raise socket.error(error, "getsockopt reported connection failed")

    fcntl.fcntl(fileno, fcntl.F_SETFL, flags)


def do_accept(sock, timeout=None):
    fileno = sock.fileno()

    # If nonblocking socket, call accept and return the result
    flags = fcntl.fcntl(fileno, fcntl.F_GETFL)
    if flags & os.O_NONBLOCK:
        return sock.accept()

    # Do nonblocking accept and go back to blocking mode
    fcntl.fcntl(fileno, fcntl.F_SETFL, flags | os.O_NONBLOCK)
    try:
        result = sock.accept()
    except socket.error, error:
        if error.errno != errno.EAGAIN:
            result = (None, (None, 0))
        else:
            result = None

    fcntl.fcntl(fileno, fcntl.F_SETFL, flags)
    if result:
        return result

    # Wait until socket becomes ready for accept
    while True:
        selector = Selector()
        selector.setTimeout(timeout)
        selector.add_fd(fileno, SelectorIO.READ)

        selector.execute()
        if selector.has_signaled:
            continue
        elif selector.has_failed:
            raise socket.error(
                selector.get_errno(), "Failed to select")
        elif not selector.has_ready:
            raise socket.error(
                errno.ETIMEDOUT, "Timed out after %s seconds" % timeout)
        else:
            return sock.accept()
