#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Child",
    "ChildError",
    ]

import os
import errno
import fcntl
import logging

from checkbox.lib.enum import Enum
from checkbox.lib.file import (
    fd_open,
    fd_close,
    )
from checkbox.lib.sigset import Sigset

from checkbox.process.args import Args
from checkbox.process.env import Env
from checkbox.process.pidenv import (
    PidEnvId,
    PidEnvIdStatus,
    )

from checkbox.daemon.server import server


ChildError = Enum(
    ("EXEC_AS_ROOT", 666666),
    ("PID_COLLISION", 666667),
    ("REGISTRATION_FAILED", 666668),
    )


class Child:

    __slots__ = (
        "_error_pipe",
        "_executable",
        "_executable_fullpath",
        "_args",
        "_env",
        "_inherit_env",
        "_inherit_buffer",
        "_ppid",
        "_directory",
        "_family_info",
        "_std",
        "_fds",
        "_renice",
        "_command_port",
        "_sigset",
        "_child_args",
        "_child_env",
        )

    def __init__(
        self, error_pipe, executable, executable_fullpath, args=[],
        env={}, inherit_env=False, inherit_buffer="", ppid=0,
        directory=None, family_info=None, std=[], fds=[], renice=0,
        command_port=True, sigset=None):

        self._error_pipe = error_pipe
        self._args = args
        self._env = env
        self._inherit_env = inherit_env
        self._inherit_buffer = inherit_buffer
        self._ppid = ppid
        self._family_info = family_info
        self._directory = directory
        self._executable = executable
        self._executable_fullpath = executable_fullpath
        self._std = std
        self._fds = fds
        self._renice = renice
        self._command_port = command_port
        self._sigset = sigset

        self._child_args = Args()
        self._child_env = Env()

    def forkExec(self):
        """Fork a new process and run the executable in the child."""
        pid = os.fork()
        if pid == 0:
            # never returns
            self.doExec()

        return pid

    def doExec(self):
        """Do the actual running of the executable in the current process."""
        fd_close(self._error_pipe[0])
        fcntl.fcntl(self._error_pipe[1], fcntl.F_SETFD, fcntl.FD_CLOEXEC)

        pid = os.getpid()
        ppid = os.getppid()

        if pid in server._pid_table:
            os.write(self._error_pipe[1], "%08X" % ChildError.PID_COLLISION)
            os._exit(4)

        # Prepare child's environment
        if self._inherit_env:
            self._child_env.mergeEnv(os.environ)

        if self._env:
            self._child_env.mergeEnv(self._env)

        if self._inherit_env:
            self._child_env["DAEMON_INHERIT"] = self._inherit_buffer

        pid_env_id = PidEnvId()

        if not self._inherit_env:
            pid_env_id.filter_and_insert(os.environ)

            # Propogate the ancestor history to the child's environment
            for ancestor in pid_env_id.ancestors:
                if not ancestor.active:
                    break

                self._child_env.setString(ancestor.envid)

        # create the new ancestor entry for the child's environment
        envid = PidEnvId.format_to_envid(self._ppid, pid)
        if pid_env_id.append(envid) == PidEnvIdStatus.OK:
            self._child_env.setString(envid)
        else:
            logging.info("Failed to insert envid: %s", envid)
            os.write(self._error_pipe[1], "%08X" % 0)
            os._exit(0)

        # Prepare child's arguments
        if not self._args:
            self._child_args.append(self._executable)
        else:
            self._child_args.extend(self._args)

        if self._family_info:
            # Create a new process group
            try:
                os.setsid()
            except OSError, e:
                logging.info("Setting process group failed: %d", e.errno)
                os.write(self._error_pipe[1], "%08X" % e.errno)
                os._exit(e.errno)

            if not server._manager:
                os.write(
                    self._error_pipe[1],
                    "%08X" % ChildError.REGISTRATION_FAILED)
                os._exit(4)

            if not server._manager.is_direct:
                if not server._registerFamily(
                    pid, ppid, self._family_info.max_snapshot_interval,
                    pid_env_id, self._family_info.login):

                    os.write(
                        self._error_pipe[1],
                        "%08X" % ChildError.REGISTRATION_FAILED)
                    os._exit(4)

        # Get descriptor table size
        try:
            open_max = os.sysconf('SC_OPEN_MAX')
        except (AttributeError, ValueError):
            open_max = 1024

        if self._std:
            for i, fd in enumerate(self._std):
                if fd > -1:
                    try:
                        os.dup2(fd, i)
                    except OSError, e:
                        logging.info("Dup2 of %d failed: %d", fd, e.errno)
                else:
                    fd_close(i)

        else:
            closed_fds = []

            # Close descriptors unless in inherit fds
            for fd in xrange(3):
                if fd not in self._fds:
                    if fd_close(fd) != -1:
                        closed_fds.append(fd)

            # Re-open then to point at /dev/null
            if closed_fds:
                fd_null = fd_open("/dev/null", os.O_RDWR)
                if fd_null < 0:
                    logging.info("Unable to open /dev/null")
                else:
                    for fd in closed_fds:
                        if fd != fd_null:
                            try:
                                os.dup2(fd_null, fd)
                            except OSError, e:
                                logging.info(
                                    "Dup2 of %d failed: %d", fd, errno)

                    if fd_null > 2:
                        fd_close(fd_null)

        # Apply renice
        if self._renice > 0:
            self._renice = min(self._renice, 19)

            try:
                newnice = os.nice(self._renice)
            except OSError, e:
                logging.info(
                    "Renice to %d returned %d: errno %d %s",
                    self._renice, newnice, e.errno, e.strerror)

        # No logging after this point

        for fd in xrange(3, open_max):
            # Don't close the error pipe
            if fd != self._error_pipe[1] and fd not in self._fds:
                fd_close(fd)

        # Check for permissions
        if os.getuid() == 0:
            os.write(
                self._error_pipe[1],
                "%08X" % ChildError.EXEC_AS_ROOT)
            os._exit(4)

        # Switch to the directory
        if self._directory:
            try:
                os.chdir(self._directory)
            except OSError, e:
                os.write(self._error_pipe[1], "%08X" % e.errno)
                os._exit(e.errno)

        # Modify signal mask unless starting a daemon server
        if not self._command_port:
            if self._sigset is None:
                self._sigset = Sigset()

            if self._sigset.setmask() is None:
                os.write(
                    self._error_pipe[1],
                    "%08X" % self._sigset.errno)
                os._exit(self._sigset.errno)

        # Finally, exec
        try:
            os.execve(
                self._executable_fullpath, self._child_args, self._child_env)
        except OSError, e:
            os.write(self._error_pipe[1], "%08X" % e.errno)
            os._exit(e.errno)
