#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Client",
    ]

import logging

from checkbox.lib import param
from checkbox.lib.address import (address_is_valid, address_from_host_and_port,
    address_to_host_and_port, address_to_host, address_to_port)
from checkbox.lib.daemon import daemon_build, daemon_host, daemon_name
from checkbox.lib.enum import enum_value_to_name
from checkbox.lib.hostname import fullhostname
from checkbox.lib.net import is_ipaddress, getfullhostname, gethostbyaddr

from checkbox.daemon.type import DaemonType


CLIENT_INFOS = {}


def register_client_info(client_class):
    """Register a class to be an implementation for L{Client}s."""
    if "type" not in client_class.__dict__:
        return

    if client_class.type is not None:
        existent_client_info = CLIENT_INFOS.get(client_class.type)
        if existent_client_info is not None:
            raise RuntimeError("%r is the same type as %r" %
                               (client_class, existent_client_info))
        CLIENT_INFOS[client_class.type] = client_class


class ClientMeta(type):

    def __init__(cls, name, bases, dict):
        super(ClientMeta, cls).__init__(name, bases, dict)
        register_client_info(cls)


class Client:

    __slots__ = (
        "info",
        )

    def __init__(self, type, *args, **kwargs):
        if type is not None:
            self.info = CLIENT_INFOS[type](*args, **kwargs)
        else:
            self.info = None

    def copy(self):
        client = Client()
        client.copy_from(self)
        return client

    def copy_from(self, other):
        if other.info:
            self.info = other.info.copy()
        else:
            self.info = None


class ClientInfo:

    __slots__ = (
        "_name",
        "_pool",
        "_hostname",
        "_fullhostname",
        "_address",
        "_port",
        "_initialized_hostname",
        "_is_configured",
        "_is_local",
        "_tried_locate",
        "_error",
        )

    __metaclass__ = ClientMeta

    def __init__(self, name=None, pool=None):
        """Create a new daemon client info instance.

        :param name: Name expressed as a daemon name, hostname, full
                     hostname, address with port or empty for local.
        :param pool: Pool identifier or empty for local.
        """
        if address_is_valid(name):
            self._address = name
            self._name = None
        else:
            self._address = None
            self._name = name

        self._pool = pool

        self._port = -1
        self._is_local = False
        self._tried_locate = False
        self._initialized_hostname = False
        self._is_configured = True
        self._error = None
        self._hostname = None
        self._fullhostname = None
 
    def copy(self):
        client = CLIENT_INFOS[self.type]()
        client.copy_from(self)
        return client

    def copy_from(self, other):
        self._name = other._name
        self._pool = other._pool

    def get_name(self):
        """Get the daemon name. This is usually the full hostname,
           but it might be something else."""
        if not self._name:
            self.locate()

        return self._name

    def get_pool(self):
        """Get the remote pool this daemon reports to. This function
           returns None if the daemon is reporting to a local pool."""
        if not self._pool:
            self.locate()

        return self._pool

    def get_hostname(self):
        """Get the hostname where the daemon is running. This is
           just the hostname without the domain."""
        if not self._hostname and not self._initialized_hostname:
            self._initialize_hostname()

        return self._hostname

    def get_fullhostname(self):
        """Get the full hostname where the daemon is running."""
        if not self._fullhostname and not self._initialized_hostname:
            self._initialize_hostname()

        return self._fullhostname

    def get_address(self):
        """Get the address of the daemon."""
        if not self._address:
            self.locate()

        return self._address

    def get_port(self):
        """Get the daemon the daemon is listening on. It will always
           be a positive, non-zero integer, unless there was an error
           in locate()."""
        if self._port < 0:
            self.locate()

        return self._port

    def is_local(self):
        """Return whether the client is local."""
        return self._is_local

    def locate(self):
        """Determine all information about the daemon. Returns True
           on success and False on error, at which point the error can be
           retrieved using get_error()."""

        # Only need to call locate once
        if self._tried_locate:
            if self._address:
                return True
            else:
                return False

        self._tried_locate = True

        if not self._do_locate():
            return False

        self._initialize_hostname_from_full()

        if self._port <= 0 and self._address:
            self._port = address_to_port(self._address)

        if not self._name and self._is_local:
            self._name = self._get_local_name()

        return True

    def create_sock_tcp(self, timeout=0, blocking=True):
        """Create a new SockTCP object, connected to the daemon.

        :param timeout: Number of for the timeout on connect.
        :param blocking: Whether to connect blocking or non-blocking.
        """
        if not self._check_address():
            return None

        sock = SockTCP()
        if timeout:
            sock.set_timeout(timeout)

        ret = sock.connect_to_port(self._address, 0, blocking)
        if not ret or (not blocking and ret == StreamError.WOULDBLOCK):
            return sock

        else:
            # TODO: report error
            return None

    def create_sock_udp(self, timeout=0, blocking=True):
        """Create a new SockUDP object, connected to the daemon.

        :param timeout: Number of for the timeout on connect.
        :param blocking: Whether to connect blocking or non-blocking.
        """
        if not self._check_address():
            return None

        sock = SockUDP()
        if timeout:
            sock.set_timeout(timeout)

        ret = sock.connect_to_port(self._address, 0, blocking)
        if not ret or (not blocking and ret == StreamError.WOULDBLOCK):
            return sock

        else:
            # TODO: report error
            return None

    def _do_locate(self):
        return True

    def _initialize_hostname(self):
        if self._initialized_hostname:
            return True

        self._initialized_hostname = True

        if self._hostname and self._fullhostname:
            return True

        if not self._tried_locate:
            self.locate()

        if not self._address:
            return False

        host = address_to_host(self._address)
        _fullhostname = gethostbyaddr(host)[0]
        if not _fullhostname:
            self._hostname = None
            self._fullhostname = None
            self._set_error("Can't find host info for %s" % self._address)
            return False

        self._fullhostname = _fullhostname
        self._initialize_hostname_from_full()
        return True

    def _initialize_hostname_from_full(self):
        if self._fullhostname:
            self._hostname = self._fullhostname.split(".")[0]
            return True

        return False

    def get_error(self):
        return self._error

    def _set_error(self, error, *args):
        self._error = error % args

    def _get_local_name(self):
        param_name = "%s_NAME" % enum_value_to_name(DaemonType, self.type)
        param_value = param.get(param_name)
        if param_value:
            return daemon_build(param_value)

        else:
            return fullhostname()

    def _get_default_port(self):
        # TODO: should probably raise exception
        return 0

    def _get_daemon_info(self, query_collector=True):
        if self._address and address_is_valid(self._address):
            logging.debug("Already have address, no info to locate.")
            self._is_local = False
            return True

        if not self._name and not self._pool:
            param_name = "%s_HOST" % enum_value_to_name(DaemonType, self.type)
            param_value = param.get(param_name)
            if param_value:
                self._name = param_value
                logging.debug("No name given, but %s defined to %s",
                    param_name, param_value)

        name_has_port = False
        if self._name:
            host, self._port = address_to_host_and_port(self._name)
            if host and self._port >= 0:
                name_has_port = True

        if name_has_port:
            logging.debug("Port %d specified in name", self._port)

            if is_ipaddress(host)[0]:
                self._address = address_from_host_and_port(host, self._port)
                logging.debug("Host info is an IP address: %s", host)

            else:
                logging.debug("Host info is a hostname: %s", host)
                _fullhostname, ipaddress = getfullhostname(host)
                if not _fullhostname:
                    self._set_error("Unknown host: %s", host)

                    # This is probably a transient DNS failure
                    self._tried_locate = False
                    return False

                self._address = address_from_host_and_port(ipaddress, self._port)
                self._fullhostname = _fullhostname

            self._is_local = False
            return True

        elif self._name:
            name = daemon_name(self._name)
            if not name:
                self._set_error("Unknown host: %s", self._name)
                return False

            self._name = name
            logging.debug("Using '%s' for name of daemon client.",
                name)

            self._fullhostname = daemon_host(name)
            logging.debug("Using '%s' for full hostname of daemon client.",
                self._fullhostname)

            if self._pool:
                logging.debug("Pool was specified, forcing collector query.")

            else:
                name = self._get_local_name()
                logging.debug("Local daemon name would be '%s'", name)

                if name == self._name:
                    logging.debug("Name '%s' matches local namd and no pool given, "
                        "treating as a local daemon.")
                    self._is_local = True

        else:
            self._is_local = True
            self._name = self._get_local_name()
            self._fullhostname = fullhostname()
            logging.debug("Neither name nor address specified, using local "
                "name %s and full host %s'", self._name, self._fullhostname)

        if not self._address and not query_collector:
            return False

        if not self._address:
            # TODO: If there is still no address, query the collector
            return False

        self._port = address_to_port(self._address)
        logging.debug("Using port %d based on address '%s'",
            self._port, self._address)

        return True

    def _get_connector_info(self):
        if self._address and is_ipaddress(self._address)[0]:
            self._port = address_to_port(self._address)
            if self._port > 0:
                logging.debug("Already have address, no info to locate.")
                self._is_local = False
                return True

        self._is_local = True

        if self._name and not self._pool:
            self._pool = self._name

        elif not self._name and self._pool:
            self._name = self._pool

        elif self._name and self._pool:
            if self._name != self._pool:
                raise Exception, "Pool '%s' and name '%s' conflict" % (
                    self._name, self._pool)

        if self._name:
            host = self._name
            self._is_local = False

        else:
            host = client_host_from_config(enum_value_to_name(DaemonType, self.type))

        if not host:
            self._set_error("Address or hostname not specified in config file.")
            self._is_configured = False
            return False

        logging.debug("Using name '%s' to find daemon.", host)

        self._port = address_to_port(host)
        if self._port < 0:
            self._port = self._get_default_port()
            logging.debug("Port not specified, using default %d.", self._port)

        else:
            logging.debug("Port %d specified in name.", self._port)
            

        if not self._name:
            self._name = host

        host = address_to_host(host)
        if not host:
            self._set_error("Address or hostname not specified in config file.")
            self._is_configured = False
            return False

        if is_ipaddress(host)[0]:
            self._address = address_from_host_and_port(host, self._port)
            logging.debug("Host info '%s' is an IP address", host)

        else:
            logging.debug("Host info '%s' is a hostname, finding IP address", host)

            fullhostname, ipaddress = getfullhostname(host)
            if not fullhostname:
                self._set_error("Unknown host '%s'", host)

                # This is probably a transient DNS failure
                self._tried_locate = False
                return False

            self._address = address_from_host_and_port(ipaddress, self._port)
            logging.debug("Found IP address and port %s", self._address)
            self._fullhostname = fullhostname

        # If the pool was set, use name instead
        if self._pool:
            self._pool = self._name

        return True

    def _check_address(self):
        if not self._address:
            self.locate()
            tried_locate = True
        else:
            tried_locate = False

        if not self._address:
            return False

        if self._port == 0:
            # Try again, just in case
            if tried_locate:
                self.set_error("Port is still 0 after locate, address invalid.")
                return False

            self._tried_locate = False
            if self._is_local:
                self._name = None

            self.locate()
            if self._port == 0:
                self.set_error("Port is still 0 after locate, address invalid.")
                return False

        return True

def client_host_from_config(name):
    # Try the config file for a hostname
    param_name = "%s_HOST" % name
    host = param.get(param_name)
    if host:
        logging.debug("%s is set to '%s'", param_name, host)
        return host

    # Try the config file for an IP address
    param_name = "%s_IP_ADDR" % name
    host = param.get(param_name)
    if host:
        logging.debug("%s is set to '%s'", param_name, host)
        return host

    return None
