#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Message",
    "MessageClosure",
    "MessageDelivery",
    ]

import logging

from checkbox.lib.enum import Enum

from checkbox.io.sock_tcp import SockTCP


MESSAGE_TIMEOUT = 20

MessageClosure = Enum(
    "FINISHED",
    "CONTINUING",
    )

MessageDelivery = Enum(
    "PENDING",
    "SUCCEEDED",
    "FAILED",
    "CANCELED",
    )


class Message:

    __slots__ = (
        "_command",
        "_delivery",
        "_timeout",
        "_callback",
        "_error",
        "_messenger",
        )

    def __init__(self, command):
        self._command = command
        self._delivery = MessageDelivery.PENDING
        self._stream = SockTCP
        self._timeout = MESSAGE_TIMEOUT

        self._callback = None
        self._error = None
        self._messenger = None

    @property
    def name(self):
        return "command %d" % self.command

    def get_delivery(self):
        """Get the message delivery status.

        :return: Message delivery name.
        """
        return self._delivery

    def set_delivery(self, delivery):
        """Set the message delivery status.

        :param delivery: Message delivery name.
        """
        if self._delivery != MessageDelivery.CANCELED:
            self._delivery = delivery

    def get_stream_type(self):
        """Get the message stream type.

        :return: Stream class name.
        """
        return self._stream_type

    def set_stream_type(self, stream_type):
        """Set the message stream type.

        :param stream_type: Stream class name.
        """
        self._stream_type = stream_type

    def get_timeout(self):
        """Get the message timeout.

        :return: Timeout in seconds.
        """
        return self._timeout

    def set_timeout(self, timeout):
        """Set the message timeout.

        :param timeout: Timeout in seconds.
        """
        self._timeout = timeout

    def set_callback(self, callback):
        """Set the callback function to call when FINISHED is returned
        by one of the message closure functions.

        :param callback: Callback message.
        """
        if callback.get():
            callback.set_message(self)

        self._callback = callback

    def do_callback(self):
        """Do the callback message."""
        if self._callback.get():
            callback = self._callback
            self._callback = None

            callback.do_callback()

    def set_messenger(self, messenger):
        self._messenger = messenger

    def cancel_message(self):
        """Cancel any pending operations related to this message."""
        self.set_delivery(MessageDelivery.CANCELED)
        self.add_error("operation was canceled")

        if self._messenger.get():
            self._messenger.cancel_message(self)

    def set_error(self, format, *args):
        self._error = format % args

    def write_message(self, messenger, sock):
        """Called by the messenger once the sock is ready to send data.

        :param messenger: Messenger caller.
        :param sock: Sock where to write.
        :return: True on success, false on failure.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def read_message(self, messenger, sock):
        """Called by the messenger once the sock is ready to receive data.

        :param messenger: Messenger caller.
        :param sock: Sock from which to read.
        :return: True on success, false on failure.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def message_sent(self, messenger, sock):
        """Called by the messenger upon successful delivery of the message.

        :param messenger: Messenger caller.
        :param sock: Sock where the message was delivered.
        :return: FINISHED if the sock is closed, CONTINUING otherwise.
        """
        self.report_success(messenger)

        return MessageClosure.FINISHED

    def message_received(self, messenger, sock):
        """Called by the messenger upon successful receipt of the message.

        :param messenger: Messenger caller.
        :param sock: Sock where the message was delivered.
        :return: FINISHED if the sock is closed, CONTINUING otherwise.
        """
        self.report_success(messenger)

        return MessageClosure.FINISHED

    def message_send_failed(self, messenger):
        """Called by the messenger when message delivery failed.

        :param messenger: Messenger caller.
        """
        self.report_failure(messenger)

    def message_receive_failed(self, messenger):
        """Called by the messenger when message receipt failed.

        :param messenger: Messenger caller.
        """
        self.report_failure(messenger)

    def report_success(self, messenger):
        """Log successful delivery of the message.

        :param messenger: Messenger caller.
        """
        logging.debug("Sent %s to %s",
            self.name,
            messenger.get_peer_description())

    def report_failure(self, messenger):
        """Log failure to deliver the message.

        :param messenger: Messenger caller.
        """
        logging.debug("Failed to send %s to %s: %s",
            self.name,
            messenger.get_peer_description(),
            self._error)

    def _call_message_send_failed(self, messenger):
        self.set_delivery(MessageDelivery.FAILED)
        self.message_send_failed(messenger)
        self.do_callback()

    def _call_message_receive_failed(self, messenger):
        self.set_delivery(MessageDelivery.FAILED)
        self.message_receive_failed(messenger)
        self.do_callback()

    def _call_message_sent(self, messenger, sock):
        self.set_delivery(MessageDelivery.SUCCEEDED)
        closure = self.message_sent(messenger, sock)
        if closure == MessageClosure.FINISHED:
            self.do_callback()

        return closure

    def _call_message_received(self, messenger, sock):
        self.set_delivery(MessageDelivery.SUCCEEDED)
        closure = self.message_received(messenger, sock)
        if closure == MessageClosure.FINISHED:
            self.do_callback()

        return closure
