#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "PidEntry",
    ]

import errno
import logging

from checkbox.lib.file import (
    File,
    fd_close,
    )


class PidEntry:

    __slots__ = (
        "pid",
        "is_local",
        "parent_is_local",
        "reaper_id",
        "pipe_buffer",
        "std_pipes",
        "stdin_offset",
        "pid_env_id",
        "hung_tid",
        "was_not_responding",
        )

    max_pipe_buffer = 10240

    def __init__(self):
        self.pipe_buffer = ["" for i in xrange(3)]
        self.std_pipes = [-1 for i in xrange(3)]
        self.stdin_offset = 0
        self.pid_env_id = None
        self.hung_tid = -1
        self.was_not_responding = False

    def __del__(self):
        for i in xrange(3):
            pipe = self.std_pipes[i]
            if pipe != -1:
                fd_close(pipe)

    def pipe_handler(self, pipe_fd):
        if self.std_pipes[1] == pipe_fd:
            pipe_index = 1
        elif self.std_pipes[2] == pipe_fd:
            pipe_index = 2
        else:
            raise Exception(
                "Pipe handler for pid %d with unknown fd %d"
                % (self.pid, pipe_fd))

        # Read until all data is consumed
        max_read_bytes = (
            self.max_pipe_buffer - len(self.pipe_buffer[pipe_index]))

        file = File.create_from_fd(pipe_fd, "r")
        buffer = file.read(max_read_bytes)
        if buffer:
            self.pipe_buffer[pipe_index] += buffer
            if len(buffer) >= self.max_pipe_buffer:
                fd_close(pipe_fd)
                self.std_pipes[pipe_index] = -1

        elif buffer is None \
             and file.errno != errno.EWOULDBLOCK \
             and file.errno != errno.EAGAIN:
            logging.info("Read failed for pid %d: %d", pipe_fd, file.errno)
            return False

        return True

    def pipe_full_write(self, pipe_fd):
        buffer = self.pipe_buffer[0]
        if buffer:
            data_left = buffer[self.stdin_offset:]
            total_len = len(buffer)

            file = File.create_from_fd(pipe_fd, "w")
            bytes_written = file.write(data_left)

        else:
            bytes_written = 0
            total_len = 0

        if bytes_written >= 0:
            self.stdin_offset += bytes_written
            if self.stdin_offset == total_len or not buffer:
                fd_close(pipe_fd)

        elif file.errno != errno.EINTR and file.errno != errno.EAGAIN:
            logging.info(
                "Unable to write to fd %d: %d. Aborting write attempts",
                pipe_fd, file.errno)
            # TODO: this should use Close_Stdin_Pipe
            file.close()

        else:
            logging.debug(
                "Unable to write to fd %d: %d. Will try again",
                pipe_fd, file.errno)
