#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Server",
    "server",
    ]

import os
import errno
import fcntl
import logging
import signal
import socket

from time import time

from checkbox.lib.enum import Enum
from checkbox.lib.file import File, fd_close, fd_get_flags, fd_set_flags
from checkbox.lib.sigset import Sigset
from checkbox.lib.stat import Stat

from checkbox.io.error import StreamError
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.sock import (
    SockPort,
    SockState,
    )
from checkbox.io.socktcp import (
    SockTCP,
    SockTCPState,
    )
from checkbox.io.sockudp import SockUDP

from checkbox.process.pidenv import (
    PidEnvId,
    PidEnvIdStatus,
    )

from checkbox.daemon.command import CommandType
from checkbox.daemon.handler import (
    CommandHandler,
    HandlerType,
    PipeHandler,
    ReapHandler,
    SignalHandler,
    SockHandler,
    )
from checkbox.daemon.pidentry import PidEntry
from checkbox.daemon.timer import (
    TimerManager,
    TIMER_NEVER,
    )
from checkbox.daemon.waitpid import (
    WaitPid,
    WaitPidQueue,
    )


DEFAULT_MAX_COMMAND_SIZE = 255
DEFAULT_MAX_PID_SIZE = 11
DEFAULT_MAX_PID_COLLISIONS = 9
DEFAULT_MAX_PIPE_SIZE = 8
DEFAULT_MAX_REAP_SIZE = 100
DEFAULT_MAX_SIGNAL_SIZE = 99
DEFAULT_MAX_SOCKET_SIZE = 8

DAEMON_FD_PIPE = -10
DAEMON_FD_NO_PIPE = -11


ServerSignal = Enum(
    ("RAISE", 1),
    "BLOCK",
    "UNBLOCK",
    )

ServerStream = Enum(
    ("KEEP", 100),
    "CLOSE",
    )


class Server:

    __slots__ = (
        "_max_command",
        "_max_signal",
        "_max_sock",
        "_max_reap",
        "_max_pipe",
        "_pid",
        "_ppid",
        "_manager",
        "_command_table",
        "_pid_table",
        "_pipe_table",
        "_reap_table",
        "_signal_table",
        "_sock_table",
        "_sock_tcp",
        "_sock_udp",
        "_default_reaper",
        "_in_service_command_socket",
        "_sent_signal",
        "_wait_pid_queue",
        )

    num_pid_collisions = 0

    timer_manager = TimerManager()

    def __init__(self, pid_size=DEFAULT_MAX_PID_SIZE,
            command_size=DEFAULT_MAX_COMMAND_SIZE,
            signal_size=DEFAULT_MAX_SIGNAL_SIZE,
            sock_size=DEFAULT_MAX_SOCKET_SIZE,
            reap_size=DEFAULT_MAX_REAP_SIZE,
            pipe_size=DEFAULT_MAX_PIPE_SIZE):
        if pid_size < 0 or command_size < 0 or signal_size < 0 \
           or sock_size < 0 or reap_size < 0 or pipe_size < 0:
            raise Exception("Invalid arguments to daemon")

        self._pid = os.getpid()
        self._ppid = 0

        self._manager = None

        self._max_command = command_size
        self._max_signal = signal_size
        self._max_sock = sock_size
        self._max_reap = reap_size
        self._max_pipe = pipe_size

        self._command_table = {}
        self._pid_table = {}
        self._pipe_table = {}
        self._signal_table = {}
        self._sock_table = []

        self._reap_table = []
        self._default_reaper = -1

        self._sock_tcp = None
        self._sock_udp = None
        self._in_service_command_socket = False

        self._wait_pid_queue = WaitPidQueue()

        Server.timer_manager.cancelTimers()

    def initialize_command_socket(self, port):
        if port == 0:
            logging.info("No command port requested.")
            return

        logging.info("Setting up command socket")

        # First, try to inherit the sockets from the parent
        self._inherit()

        if self._sock_tcp is None or self._sock_udp is None:
            self._sock_tcp = SockTCP()
            self._sock_udp = SockUDP()
            initialize_command_sockets(
                port, self._sock_tcp, self._sock_udp, True)

        # Register the command sockets
        self.registerCommandSocket(self._sock_tcp)
        self.registerCommandSocket(self._sock_udp)

        # Register daemon default handlers
        server.registerCommand(
            CommandType.RAISE_SIGNAL, server._handleSignalCommand)

        server.registerCommand(
            CommandType.CHILD_ALIVE, server._handleChildAlive)

    def pid_is_alive(self, pid):
        """Return whether a pid is alive.

        :param pid: Pid to check.
        """
        if self.pid_exited_but_not_reaped(pid):
            status = True

        else:
            try:
                os.kill(pid, 0)
                status = True
            except OSError, error:
                if error.errno == errno.EPERM:
                    logging.debug(
                        "Kill returned EPERM, assuming pid %d is alive.", pid)
                    status = True
                else:
                    logging.debug(
                        "Kill returned errno %d, assuming pid %d is dead.",
                        error.errno, pid)
                    status = False

        return status

    def pid_exited_but_not_reaped(self, pid):
        """Return whether a pid has exited but has not been reaped.

        :param pid: Pid to check.
        """
        wait_pid = WaitPid(pid)

        return wait_pid in self._wait_pid_queue

    def service_command_socket(self):
        if self._in_service_command_socket:
            return 0

        # Return if there is no command socket
        if not self._sock_table:
            return 0

        sock_handler = self._sock_table[0]
        if not sock_handler.sock:
            return 0

        # Setting timeout to 0 to not block
        selector = Selector()
        selector.set_timeout(0)
        selector.add_fd(sock_handler.sock.fileno(), SelectorIO.READ)

        commands_served = 0
        self._in_service_command_socket = True
        while True:
            selector.execute()
            if selector.has_failed():
                raise Exception("Selector, errno %d" % selector.get_errno())

            if selector.has_ready():
                commands_served += 1
                self._handleRequest(sock_handler.sock)

            # Loop until no more commands are waiting on the sock
            if not selector.has_ready():
                break

        self._in_service_command_socket = False
        return commands_served

    def in_service_command_socket(self):
        return self._in_service_command_socket

    def registerCommand(self, command, function):
        if not function:
            logging.info("Can't register an empty command function.")
            return -1

        if len(self._command_table) >= self._max_command:
            raise Exception("Max command count reached")

        if command in self._command_table:
            raise Exception("Same command registered twice: %d" % command)

        command_handler = CommandHandler()
        command_handler.number = command
        command_handler.function = function

        self._command_table[command] = command_handler

        return command

    def unregisterCommand(self, command):
        if command not in self._command_table:
            logging.info("Failed to unregister command: %d", command)
            return False

        del self._command_table[command]
        return True

    def registerSignal(self, sig, function):
        if sig in (signal.SIGKILL, signal.SIGSTOP, signal.SIGCONT):
            raise Exception(
                "Trying to register signal %d which cannot be caught" % sig)

        elif sig == signal.SIGCHLD:
            self.unregisterSignal(sig)

        if len(self._signal_table) >= self._max_signal:
            raise Exception("Max signal count reached")

        # Check that index was not already registered
        if sig in self._signal_table:
            raise Exception("Signal already registered")

        signal_handler = SignalHandler()
        signal_handler.number = sig
        signal_handler.function = function
        signal_handler.is_blocked = False
        signal_handler.is_pending = False

        self._signal_table[sig] = signal_handler

        return sig

    def unregisterSignal(self, sig):
        if sig not in self._signal_table:
            logging.info("Failed to unregister signal")
            return False

        del self._signal_table[sig]
        return True

    def blockSignal(self, sig):
        return self._hangleSignal(ServerSignal.BLOCK, sig)

    def unblockSignal(self, sig):
        return self._hangleSignal(ServerSignal.UNBLOCK, sig)

    # TODO: implement send_signal and other methods using it

    def set_default_reaper(self, reaper_id):
        self._default_reaper = reaper_id

    def registerReaper(self, reaper_id=-1, function=None):
        if reaper_id == -1:
            for index, reap_handler in enumerate(self._reap_table):
                if reap_handler.number == 0:
                    break
                elif reap_handler.number != index + 1:
                    raise Exception("Reaper table corrupted")
            else:
                if len(self._reap_table) >= self._max_reap:
                    raise Exception("Max reap count reached")

                reap_handler = ReapHandler()
                self._reap_table.append(reap_handler)

                reaper_id = len(self._reap_table)
                reap_handler.number = reaper_id
        else:
            # Invalid reaper id
            if reaper_id < 0 or reaper_id > self._max_reap:
                return False

            # Trying to register non-existant entry
            reap_handler = self._reap_table[reaper_id - 1]
            if reap_handler.number != reaper_id:
                return False

        reap_handler.function = function

        return reaper_id

    def unregisterReaper(self, reaper_id):
        return True

    def registerSock(self, sock, function=None, data=None):
        if not sock:
            logging.warn("Can't register empty sock.")
            return -1

        if len(self._sock_table) >= self._max_sock:
            raise Exception("Max sock count reached")

        # Check that sock was not already registered
        fileno = sock.fileno()
        for sock_handler in self._sock_table:
            if sock_handler.sock \
               and (sock_handler.sock == sock
                    or sock_handler.sock.fileno() == fileno):
                logging.info("Attempt to register a sock twice")
                return -2

        # Add the new data
        sock_handler = SockHandler()
        sock_handler.sock = sock
        sock_handler.function = function
        sock_handler.data = data
        sock_handler.call_handler = False

        protocol = sock.get_protocol()
        if protocol == "tcp":
            sock_handler.is_connect_pending = sock.is_connect_pending()

        elif protocol == "udp":
            sock_handler.is_connect_pending = False

        else:
            raise Exception("Unknown sock protocol: %s" % protocol)

        self._sock_table.append(sock_handler)

        return len(self._sock_table)

    def unregisterSock(self, sock):
        for index, sock_handler in enumerate(self._sock_table):
            if sock_handler.sock == sock:
                break
        else:
            logging.info("Failed to unregister sock")
            return False

        del self._sock_table[index]
        return True

    def unregisterSocks(self):
        while self._sock_table:
            sock_handler = self._sock_table[0]
            self.unregisterSock(sock_handler.sock)
            if sock_handler.sock == self._sock_tcp:
                self._sock_tcp = None
            elif sock_handler.sock == self._sock_udp:
                self._sock_udp = None

    def call_sock_handler(self, sock, handle_command=False):
        for sock_handler in self._sock_table:
            if sock_handler.sock == sock:
                break
        else:
            logging.info("Called on unregistered sock")
            return

        if sock_handler.function:
            logging.info("Calling handler for sock")

            result = sock_handler.function(sock)
            logging.info("Return from handler")

        elif handle_command:
            result = self._handleRequest(sock)

        else:
            result = False

        if result != ServerStream.KEEP:
            self.unregisterSock(sock)

    def registerPipe(self, pipe_end, function, handler_type=HandlerType.READ):
        print 'register pipe: %d' % pipe_end
        if len(self._pipe_table) >= self._max_pipe:
            raise Exception("Max pipe count reached")

        # Check that index was not already registered
        if pipe_end in self._pipe_table:
            raise Exception("Pipe already registered")

        # Add the new data
        pipe_handler = PipeHandler()
        pipe_handler.pipe_end = pipe_end
        pipe_handler.function = function
        pipe_handler.type = handler_type

        self._pipe_table[pipe_end] = pipe_handler

        return pipe_end

    def unregisterPipe(self, pipe_end):
        print 'unregister pipe: %d, %s' % (pipe_end, self._pipe_table)
        if pipe_end not in self._pipe_table:
            logging.info("Failed to unregister pipe")
            return False

        del self._pipe_table[pipe_end]
        return True

    def createPipe(
        self, pipe_ends, nonblocking_read=False, nonblocking_write=False):
        try:
            handles = os.pipe()
        except OSError:
            logging.info("Call to pipe() failed.")
            return False

        failed = False
        if nonblocking_read:
            flags = fd_get_flags(handles[0])
            if flags < 0:
                failed = True

            else:
                flags |= os.O_NONBLOCK
                if fd_set_flags(handles[0], flags) < 0:
                    failed = True

        if nonblocking_write:
            flags = fd_get_flags(handles[1])
            if flags < 0:
                failed = True

            else:
                flags |= os.O_NONBLOCK
                if fd_set_flags(handles[1], flags) < 0:
                    failed = True

        if failed:
            fd_close(handles[0])
            fd_close(handles[1])
            logging.info("Failed to set pipe in non-blocking mode.")
            return False

        pipe_ends[0] = handles[0]
        pipe_ends[1] = handles[1]

        return True

    def closePipe(self, pipe_end):
        # Ignore return value when pipe_end is not in _pipe_table
        if pipe_end in self._pipe_table:
            self.unregisterPipe(pipe_end)

        if fd_close(pipe_end) < 0:
            logging.info("Failed to close pipe")
            return False

        return True

    def readPipe(self, pipe_end, size):
        if size < 0:
            raise Exception("Invalid size: %d" % size)

        pipe_file = File.create_from_fd(pipe_end, "r")
        return pipe_file.read(size)

    def writePipe(self, pipe_end, buffer):
        if not buffer:
            raise Exception("Invalid buffer, empty")

        pipe_file = File.create_from_fd(pipe_end, "w")
        return pipe_file.write(buffer)

    @classmethod
    def registerTimer(
        cls, function, delta=0, period=0, timeslice=None, data=None):
        return cls.timer_manager.createTimer(
            function, delta, period, timeslice, data)

    @classmethod
    def unregisterTimer(cls, id):
        return cls.timer_manager.cancelTimer(id)

    @classmethod
    def resetTimer(cls, id, delta=0, period=0):
        return cls.timer_manager.resetTimer(id, delta, period)

    def createThread(self, function, args=[], sock=None, reaper_id=1):
        from checkbox.daemon.child import ChildError

        # Check reapder_id validity
        if (reaper_id < 1
            or reaper_id > self._max_reap
            or self._reap_table[reaper_id - 1].number == 0):
            logging.info("Invalid reaper id")
            return False

        # Check for pid collision
        try:
            error_pipe = os.pipe()
        except OSError, error:
            logging.info("Pipe failed with errno %d", error.errno)
            return False

        try:
            tid = os.fork()
        except OSError, error:
            logging.info("Failed to fork, errno %d", error.errno)
            Server.num_pid_collisions = 0
            fd_close(error_pipe[0])
            fd_close(error_pipe[1])
            return False

        if tid == 0:
            # Close the read end of the error pipe
            fd_close(error_pipe[0])
            fcntl.fcntl(error_pipe[1], fcntl.F_SETFD, fcntl.FD_CLOEXEC)

            pid = os.getpid()
            if pid in self._pid_table:
                # Pid already in table
                os.write(error_pipe[1], "%08X" % ChildError.PID_COLLISION)
                fd_close(error_pipe[1])
                self._exit(4)

            # No need for the error pipe anymore
            fd_close(error_pipe[1])
            status = function(args, sock)
            self._exit(status)

        elif tid > 0:
            # Close the write end of the error pipe
            fd_close(error_pipe[1])

            # Check the error pipe for any errno
            buffer = os.read(error_pipe[0], 8)
            fd_close(error_pipe[0])
            if buffer:
                child_errno = int(buffer, 16)
                status = os.waitpid(tid, 0)[1]
                if child_errno != ChildError.PID_COLLISION:
                    raise Exception("Unsupported errno, %d" % child_errno)

                logging.info(
                    "Failed to create child because pid %d is still used",
                    tid)

                Server.num_pid_collisions += 1
                if Server.num_pid_collisions > DEFAULT_MAX_PID_COLLISIONS:
                    logging.info(
                        "Maximum number of PID collisions encountered")
                    Server.num_pid_collisions = 0
                    return False

                # Try again
                return self.createThread(function, args, sock, reaper_id)

        Server.num_pid_collisions = 0

        logging.info("Created new thread, tid %d", tid)

        pid_entry = PidEntry()
        pid_entry.pid = tid
        pid_entry.is_local = True
        pid_entry.parent_is_local = True
        pid_entry.reaper_id = reaper_id

        self._pid_table[tid] = pid_entry
        return tid

    def kill_thread(self, tid):
        """Kill a thread by id.

        :param tid: Thread id.
        """
        logging.info("Killing thread %d", tid)
        status = self.kill(tid, signal.SIGKILL)
        return status >= 0

    def suspend_thread(self, tid):
        logging.info("Suspending thread %d", tid)

        # Check tid validity
        if tid not in self._pid_table:
            logging.info("Failed to suspend thread, bad tid %d", tid)
            return False

        return self.suspendProcess(tid)

    def continue_thread(self, tid):
        logging.info("Continuing thread %d", tid)

        # Check tid validity
        if tid not in self._pid_table:
            logging.info("Failed to continue thread, bad tid %d", tid)
            return False

        return self.continueProcess(tid)

    def createProcess(
        self, executable, args=[], reaper_id=0, command_port=True,
        env={}, inherit_env=False, directory=None, family_info=None,
        std=None, fds=[], renice=0, sigset=None):
        """Create a process by forking and running the executable.

        :param executable: Full path name of the executable.
        :param args: List of args passed to the executable.
        :param reaper_id: Reaper number to use.
        :param command_port: Port on which to create a command socket.
        :param env: Dictionary of environment variables.
        :param inherit_env: Inherit the parent environment.
        :param directory: Current working directory.
        :param family_info: Process family information.
        :param std: List of the three standard file descriptors.
        :param fds: List of file descriptors to inherit.
        :param renice: Nice value to be passed to nice().
        :param sigset: Signal set.
        """
        from checkbox.daemon.child import (
            Child,
            ChildError,
            )

        new_pid = -1
        daemon_pipe_fds = [[-1, -1]] * 3

        def cleanup_process():
            # Clear out static counter
            Server.num_pid_collisions = 0

            # Close pipes
            for i in xrange(3):
                for j in xrange(2):
                    if daemon_pipe_fds[i][j] != -1:
                        self.closePipe(daemon_pipe_fds[i][j])

            return new_pid

        # Check reaper_id validity
        if (reaper_id < 0
            or reaper_id > self._max_reap
            or (reaper_id > 0 and reaper_id - 1 >= len(self._reap_table))):
            logging.info("Invalid reaper id")
            return cleanup_process()

        # Check executable validity
        if not executable:
            logging.info("Invalid executable")
            return cleanup_process()

        inherit_buffers = []
        inherit_buffers.append("%lu" % os.getpid())

        sock_tcp = SockTCP()
        sock_udp = SockUDP()

        inherit_fds = list(fds)
        if command_port:
            if not initialize_command_sockets(
                command_port, sock_tcp, sock_udp, False):
                return cleanup_process()

            inherit_buffers.append(sock_tcp.serialize())
            inherit_buffers.append(sock_udp.serialize())

            inherit_fds.append(sock_tcp.fileno())
            inherit_fds.append(sock_udp.fileno())

        # Terminate inherit buffers
        inherit_buffers.append("0")

        if family_info and not self._manager:
            from checkbox.process.manager import create_manager

            # TODO: use proper subsystem name
            self._manager = create_manager("MASTER")

        for i in xrange(3):
            if std and std[i] == DAEMON_FD_PIPE:
                if i == 0:
                    if not self.createPipe(daemon_pipe_fds[i], False, True):
                        logging.info("Failed to create pipe for stdin")
                        return cleanup_process()

                    std[i] = daemon_pipe_fds[i][0]

                else:
                    if not self.createPipe(daemon_pipe_fds[i], True, False):
                        logging.info(
                            "Failed to create pipe for ",
                            "stdout" if i == 1 else "stderr")
                        return cleanup_process()

                    std[i] = daemon_pipe_fds[i][1]

        # First, check that the executable exists.
        if os.access(executable, os.F_OK | os.X_OK) < 0:
            logging.info("Cannot access executable: %s", executable)
            return cleanup_process()

        # Next, check that the directory exists.
        if directory:
            stat = Stat.create_from_path(directory)
            if stat.stat() == -1:
                logging.info(
                    "Cannot access directory '%s': %d",
                    directory, stat.get_errno())
                return cleanup_process()

            if not os.path.isabs(executable):
                cwd = os.getcwd()
                executable_fullpath = os.path.join(cwd, executable)

        else:
            executable_fullpath = executable

        try:
            error_pipe = os.pipe()
        except OSError, error:
            logging.info("Pipe failed with errno %d", error.errno)
            return cleanup_process()

        forker_pid = os.getpid()

        if sigset:
            # Can't set signal set for daemon server process
            assert(not command_port)

            saved_sigset = sigset.block()
            if saved_sigset is None:
                logging.info("Failed to set signal mask.")
                return cleanup_process()

        inherit_buffer = " ".join(inherit_buffers)
        child = Child(
            error_pipe, executable, executable_fullpath, args, env,
            inherit_env, inherit_buffer, forker_pid, directory,
            family_info, std, inherit_fds, renice, command_port, sigset)

        new_pid = child.forkExec()
        if new_pid > 0:
            if sigset is not None:
                if saved_sigset.setmask() is None:
                    raise Exception("Failed to reset signal mask.""")

            fd_close(error_pipe[1])

            # Check the error pipe for any problems
            buffer = os.read(error_pipe[0], 8)
            if buffer:
                child_errno = int(buffer, 16)
                os.waitpid(new_pid, 0)[1]
                if child_errno == ChildError.EXEC_AS_ROOT:
                    logging.info("Child failed because process "
                        "was still root before exec()")

                elif child_errno == ChildError.REGISTRATION_FAILED:
                    logging.info("Child failed to register itself")

                elif child_errno == ChildError.PID_COLLISION:
                    logging.info("Child failed because PID is still in use")

                    Server.num_pid_collisions += 1
                    if Server.num_pid_collisions > DEFAULT_MAX_PID_COLLISIONS:
                        logging.info(
                            "Maximum number of PID collisions encountered")

                    else:
                        fd_close(error_pipe[0])

                        sock_tcp.close()
                        sock_udp.close()

                        logging.info(
                            "Retrying create process to avoid PID collision")

                        return self.createProcess(
                            executable, args, reaper_id, env, directory,
                            family_info, std, renice, sigset)

                else:
                    logging.info("Child failed with errno: %d", child_errno)

                fd_close(error_pipe[0])
                new_pid = -1

                return cleanup_process()

            fd_close(error_pipe[0])

        elif new_pid < 0:
            logging.info("Failed to fork %s", executable)
            fd_close(error_pipe[0])
            fd_close(error_pipe[1])
            return cleanup_process()

        pid_entry = PidEntry()
        pid_entry.pid = new_pid
        pid_entry.is_local = True
        pid_entry.parent_is_local = True
        pid_entry.reaper_id = reaper_id

        # Manage daemon managed pipes
        for i in xrange(3):
            if daemon_pipe_fds[i][0] != -1:
                if i == 0:
                    self.closePipe(daemon_pipe_fds[i][0])
                    pid_entry.std_pipes[i] = daemon_pipe_fds[i][1]

                else:
                    self.closePipe(daemon_pipe_fds[i][1])
                    pid_entry.std_pipes[i] = daemon_pipe_fds[i][0]

                    self.registerPipe(
                        daemon_pipe_fds[i][0], pid_entry.pipe_handler)

                daemon_pipe_fds[i][0] = daemon_pipe_fds[i][1] = -1

        pid_env_id = PidEnvId()
        if pid_env_id.filter_and_insert(os.environ) != PidEnvIdStatus.OK:
            raise Exception("Failed to insert environment")

        if (pid_env_id.append_direct(forker_pid, new_pid)
            == PidEnvIdStatus.OVERSIZED):
            raise Exception("Failed to append direct entry")

        pid_entry.pid_env_id = pid_env_id
        assert(new_pid not in self._pid_table)
        self._pid_table[new_pid] = pid_entry

        # Register the new process with the process family
        if family_info and self._manager.is_direct:
            self._registerFamily(
                new_pid, os.getpid(), family_info.max_snapshot_interval,
                pid_env_id.pid_env_id, family_info.login)

        logging.info("Child process %d", new_pid)

        return cleanup_process()

    def killProcess(self, pid, sig=signal.SIGTERM):
        logging.info("Killing process %d", pid)

        # Cannot kill parent
        if pid == self._ppid:
            return False

        status = os.kill(pid, sig)
        return status >= 0

    def suspendProcess(self, pid):
        logging.info("Suspending process %d", pid)

        # Cannot suspend parent
        if pid == self._ppid:
            return False

        status = os.kill(pid, signal.SIGSTOP)
        return status >= 0

    def continueProcess(self, pid):
        logging.info("Continuing process %d", pid)

        status = os.kill(pid, signal.SIGCONT)
        return status >= 0

    def run(self):
        selector = Selector()

        # Prepare signal sets
        emptyset = Sigset()
        fullset = Sigset()
        fullset.fill()
        fullset = fullset.difference([
            signal.SIGABRT,
            signal.SIGBUS,
            signal.SIGFPE,
            signal.SIGILL,
            signal.SIGPROF,
            signal.SIGSEGV,
            signal.SIGTRAP,
            ])

        while True:
            # Call signal handlers for any pending signals
            self._sent_signal = False
            for signal_handler in self._signal_table.values():
                if signal_handler.is_pending and not signal_handler.is_blocked:
                    signal_handler.is_pending = False
                    logging.info("Calling handler for signal %d",
                        signal_handler.number)
                    signal_handler.function()

            # TODO: implement async pipe

            # Prepare to enter main select()
            timeout = Server.timer_manager.timeout()
            if self._sent_signal:
                timeout = 0

            elif timeout < 0:
                timeout = TIMER_NEVER

            # Setup socket descriptors to select on
            selector.reset()
            min_deadline = 0
            for sock_handler in self._sock_table:
                sock = sock_handler.sock
                if sock_handler.is_connect_pending:
                    # Wake when a non-blocking connect is ready to write
                    selector.add_fd(sock.fileno(), SelectorIO.WRITE)
                    selector.add_fd(sock.fileno(), SelectorIO.EXCEPT)
                else:
                    selector.add_fd(sock.fileno(), SelectorIO.READ)

                deadline = sock.get_deadline()
                if deadline:
                    if min_deadline == 0 or min_deadline > deadline:
                        min_deadline = deadline

            if min_deadline:
                deadline_timeout = min_deadline - time() + 1
                if deadline_timeout < timeout:
                    timeout = deadline_timeout

            # Add the registered pipe descriptors
            for pipe_handler in self._pipe_table:
                if pipe_handler.type == HandlerType.READ:
                    selector.add_fd(pipe_handler.pipe_end, SelectorIO.READ)
                elif pipe_handler.type == HandlerType.WRITE:
                    selector.add_fd(pipe_handler.pipe_end, SelectorIO.WRITE)

                elif pipe_handler.type == HandlerType.READ_WRITE:
                    selector.add_fd(pipe_handler.pipe_end, SelectorIO.READ)
                    selector.add_fd(pipe_handler.pipe_end, SelectorIO.WRITE)

            # TODO: add async pipe to selector

            # Unblock all signals so that they can be caught during select
            emptyset.setmask()

            selector.set_timeout(timeout)
            selector.execute()

            # Block all signals until the next select
            fullset.setmask()

            if selector.has_failed():
                raise Exception(
                    "Select failed with errno %d" % selector.get_errno())

            elif selector.has_ready() \
                 or (selector.has_timed_out() \
                     and min_deadline \
                     and min_deadline < time()):
                # Either an operation has happened or timed out
                now = time()
                recheck_status = False

                # Scan through the socket table to find which ones are callable
                for sock_handler in self._sock_table:
                    sock_handler.call_handler = False
                    sock = sock_handler.sock
                    sock_timed_out = deadline and deadline < now
                    if sock_handler.is_pending:
                        if (selector.is_fd_ready(
                                sock.fileno(), SelectorIO.WRITE)
                            or selector.is_fd_ready(
                                sock.fileno(), SelectorIO.EXCEPT)
                            or sock_timed_out):
                            if (sock_timed_out.sock._finish_connect()
                                != StreamError.WOULDBLOCK):
                                sock_timed_out.call_handler = True
                    else:
                        if (selector.is_fd_ready(
                                sock.fileno(), SelectorIO.READ)
                            or sock_timed_out):
                            sock_timed_out.call_handler = True

                # Scan through the pipe table to find which ones are callable
                for pipe_handler in self._pipe_table:
                    pipe_end = pipe_handler.pipe_end
                    if selector.is_fd_ready(pipe_end, SelectorIO.READ):
                        pipe_handler.call_handler = True

                    if selector.is_fd_ready(pipe_end, SelectorIO.WRITE):
                        pipe_handler.call_handler = True

                # Now loop through all pipe handlers, calling the
                # functions if required
                for index in xrange(len(self._pipe_table)):
                    pipe_handler = self._pipe_table[index]
                    if pipe_handler.call_handler:
                        pipe_end = pipe_handler.pipe_end
                        pipe_handler.call_handler = False
                        saved_pid_entry = pipe_handler.pid_entry
                        if recheck_status or saved_pid_entry:
                            selector.reset()
                            selector.set_timeout(0)
                            selector.add_fd(pipe_end, SelectorIO.READ)
                            selector.execute()
                            if selector.has_timed_out():
                                # Nothing is avilable, try the next handler
                                continue

                        pipe_handler.in_handler = True
                        logging.info(
                            "Calling pipe handler on pipe end %d", pipe_end)

                        recheck_status = True
                        pipe_handler.function(pipe_end)
                        logging.info("Returned from pipe handler")

                        pipe_handler.in_handler = False

                        if pipe_handler.call_handler == True:
                            # Looks like the handler called unregisterPipe()
                            index -= 1

                # Now loop through all sock handlers, calling the
                # functions if required
                for sock_handler in self._sock_table[:]:
                    if sock_handler.call_handler:
                        sock_handler.call_handler = False
                        sock = sock_handler.sock
                        if recheck_status or not sock_handler.is_pending:
                            selector.reset()
                            selector.set_timeout(0)
                            selector.add_fd(sock.fileno(), SelectorIO.READ)
                            selector.execute()
                            if selector.has_timed_out():
                                # Nothing is avilable, try the next handler
                                continue

                        # Enqueue this packet into the buffers if this
                        # is a udp sock
                        if sock.get_protocol() == "udp":
                            if not sock.handle_incoming_packet():
                                # There is not yet a complete message ready
                                continue

                        recheck_status = True
                        self.call_sock_handler(sock, True)

    def _inherit(self):
        inherit_buffer = os.environ.pop("DAEMON_INHERIT", "")
        inherit_list = inherit_buffer.split(" ")
        if inherit_list:
            ppid = int(inherit_list.pop(0))

            # TODO: check PidEntry attributes
            pid_entry = PidEntry()
            pid_entry.pid = ppid
            pid_entry.description = inherit_list.pop(0)
            pid_entry.is_local = True
            pid_entry.parent_is_local = True
            pid_entry.reaper_id = 0
            pid_entry.hug_tid = -1
            pid_entry.was_not_responding = False

            assert(ppid not in self._pid_table)
            self._pid_table[ppid] = pid_entry

            # Inherit command socks
            self._sock_tcp = None
            self._sock_udp = None
            # TODO: revisit this after creating the inherit_buffer
            # in createProcess
            if inherit_list and inherit_list.pop(0) != "0":
                self._sock_tcp = SockTCP()
                string = inherit_list.pop(0)
                self._sock_tcp.unserialize(string)

            if inherit_list and inherit_list.pop(0) != "0":
                self._sock_udp = SockUDP()
                string = inherit_list.pop(0)
                self._sock_udp.unserialize(string)

    def _handleRequest(self, sock):
        protocol = sock.get_protocol()
        if protocol == "tcp":
            is_tcp = True
        elif protocol == "udp":
            is_tcp = False
        else:
            raise Exception("Unknown sock protocol: %s" % protocol)

        # Setup a connection for a TCP socket
        if is_tcp:
            if (sock._state == SockState.SPECIAL
                and sock._special_state == SockTCPState.LISTEN):
                stream = sock.accept()
                if not stream:
                    logging.info("Failed to accept from sock.")
                    return ServerStream.KEEP

                if stream.get_bytes_available() < 4:
                    # Set a timer for 200 seconds in case nothing ever arrives
                    tid = server.registerTimer(
                        self._handleRequestSockTimerHandler, 200, data=stream)
                    result = server.registerSock(
                        stream, self._handleRequestSockHandler, tid)
                    if result < 0:
                        server.unregisterTimer(tid)
                    else:
                        return ServerStream.KEEP

            else:
                stream = sock

        # Setup a connection for a UDP socket
        else:
            stream = sock

        # Set the timeout to just 1 second since there's data waiting.
        stream.setTimeout(1)
        command = stream.getInteger()

        # For now, set the timeout to 20 seconds
        stream.setTimeout(20)
        if command is None:
            host, port = sock.get_peer_host_and_port()
            logging.info(
                "Can't receive command reques tfrom %s, perhaps timeout?",
                host)

            result = False

        else:
            # Get handler function
            if command in self._command_table:
                logging.info(
                    "Received %s command %d",
                    "TCP" if is_tcp else "UDP", command)

                handler_start_time = time()
                command_handler = self._command_table[command]
                result = command_handler.function(command, stream)
                handler_end_time = time()

                logging.info(
                    "Return from handle request, %d seconds",
                    handler_end_time - handler_start_time)
            else:
                logging.info(
                    "Received %s unregistered command %d",
                    "TCP" if is_tcp else "UDP", command)

                result = False
                if not is_tcp:
                    stream.getEnd()

        if result != ServerStream.KEEP:
            stream.putEnd()
            if not is_tcp:
                result = ServerStream.KEEP

        elif not is_tcp:
            stream.getEnd()

        # Return KEEP only if the user said so or if sock is a listening
        # socket
        if result == ServerStream.KEEP or sock != stream:
            return ServerStream.KEEP
        else:
            return True

    def _handleRequestSockTimerHandler(self, sock):
        assert(sock.get_protocol() == "tcp")

        self.unregisterSock(sock)

        logging.info("Closing socket, no data received")
        return True

    def _handleRequestSockHandler(self, sock, timeout_tid):
        self.unregisterTimer(timeout_tid)
        self.unregisterSock(sock)

        self._handleRequest(sock)

        return ServerStream.KEEP

    def _hangleSignal(self, command, sig):
        if sig not in self._signal_table:
            logging.info("Received request for unregistered signal %d", sig)
            return False

        signal_handler = self._signal_table[sig]
        if command == ServerSignal.RAISE:
            logging.info("Received signal %d, raising event", sig)
            signal_handler.is_pending = True

        elif command == ServerSignal.BLOCK:
            signal_handler.is_blocked = True

        elif command == ServerSignal.UNBLOCK:
            signal_handler.is_blocked = False
            # Check if this signal is already pending
            if signal_handler.is_pending:
                self._sent_signal = True

        else:
            logging.warn("Unrecognized command: %d", command)
            return False

        return True

    def _handleSignalCommand(self, command, stream):
        assert(command == ServerSignal.RAISE)

        # Read the signal number
        sig = stream.getInteger()
        if sig is None:
            return False

        stream.getEnd()

        return self._hangleSignal(ServerSignal.RAISE, sig)

    def _handleChildAlive(self, command, stream):
        child_pid = stream.getInteger()
        timeout = stream.getInteger()
        if not stream.getEnd():
            logging.info("Failed to get child alidve packet.")
            return False

        pid_entry = self._pid_table.get(child_pid)
        if pid_entry is None:
            logging.info(
                "Received child alive command from unknown pid %d", child_pid)
            return False

        if pid_entry.hung_tid != -1:
            ret = server.resetTimer(pid_entry.hung_tid, timeout)
            assert(ret != -1)

        else:
            pid_entry.hung_tid = self.registerTimer(
                self._hung_child_timeout, timeout, data=pid_entry.pid)
            assert(pid_entry.hung_tid != -1)

        pid_entry.was_not_responding = False

        logging.debug(
            "Received child alive, pid %d, timeout %d", child_pid, timeout)

        return True

    def _hung_child_timeout(self, hung_child_pid):
        pid_entry = self._pid_table.get(hung_child_pid)
        if pid_entry is None:
            return False

        # Reset tid to -1 for _handleChildAlive
        pid_entry.hung_tid = -1

        if self.pid_exited_but_not_reaped(hung_child_pid):
            logging.debug(
                "Cancelling hung child timer for pid %d, "
                "because it has exited but has not yet been reaped.",
                hung_child_pid)
            return False

        # Set a flag in the pid entry for the reaper
        pid_entry.was_not_responding = True

        # Give the child one last chance to save itself
        self.service_command_socket()

        # Make sure this pid did not exit
        if hung_child_pid not in self._pid_table:
            return False

        # Now check was_not_responding
        if not pid_entry.was_not_responding:
            return False

        logging.info("Child pid %d appears hung, killing it.", hung_child_pid)

        self.killProcess(hung_child_pid, signal.SIGKILL)

        return True

    def _registerFamily(
        self, child_pid, parent_pid, max_snapshot_interval, pid_env_id,
        login):

        success = False
        family_registered = False

        def cleanup_register():
            if family_registered and not success:
                if not self._manager.unregisterFamily(child_pid):
                    logging.info(
                        "Error unregistering family for pid %d", child_pid)

            return success

        if not self._manager.registerSubfamily(
            child_pid, parent_pid, max_snapshot_interval):
            logging.info("Error registering family for pid %d", child_pid)
            return cleanup_register()

        family_registered = True
        if pid_env_id:
            if not self._manager.trackEnvironment(child_pid, pid_env_id):
                logging.info(
                    "Error tracking family from environment for pid %d",
                    child_pid)
                return cleanup_register()

        if login:
            if not self._manager.trackLogin(child_pid, login):
                logging.info(
                    "Error tracking family from login for pid %d", child_pid)
                return cleanup_register()

        success = True

        return cleanup_register()


def bind_any_command_port(sock_tcp, sock_udp):
    for i in xrange(1000):
        # Bind to an incoming connection
        if not sock_tcp.bind(port=SockPort.INBOUND):
            logging.info("Failed to bind to command socket.")
            return False

        # Open a socket on the same port chosen above
        host, port = sock_tcp.get_host_and_port()
        if sock_udp and not sock_udp.bind(port=port):
            sock_tcp.close()
            continue

        return True

    logging.info("Failed to bind to any command port.")
    return False


def initialize_command_sockets(port, sock_tcp, sock_udp, is_fatal):
    """Initialize both the command sockets.

    :param port: Port to bind to.
    :param sock_tcp: SockTCP instance.
    :param sock_udp: SockUDP instance.
    :param is_fatal: Whether to raise an exception or log.
    """
    assert(port > 0)

    def cleanup(format, *args):
        message = format % args
        if is_fatal:
            raise Exception(message)

        else:
            logging.info(message)
            return False

    if port == 1:
        if not bind_any_command_port(sock_tcp, sock_udp):
            return cleanup("Failed to call bind_any_command_port.")

        if not sock_tcp.listen():
            return cleanup("Failed to post listen on command port.")

    else:
        # Set SO_REUSEADDR to bind to the same address after a crash
        if not sock_tcp.setOptions(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1):
            return cleanup("Failed to set SO_REUSEADDR on TCP command port.")

        if sock_udp and not sock_udp.setOptions(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1):
            return cleanup("Failed to set SO_REUSEADDR on UDP command port.")

        # Set TCP_NODELAY to disable Nagle
        if not sock_tcp.setOptions(
            socket.IPPROTO_TCP, socket.TCP_NODELAY, 1):
            logging.info("Failed to set TCP_NODELAY on TCP command port.")

        if not sock_tcp.listen_on_port(port):
            return cleanup(
                "Failed to listen on port %d for TCP command socket.", port)

        if sock_udp and not sock_udp.bind(port=port):
            return cleanup(
                "Failed to bind to port %d for UDP command socket.", port)

    return True


server = Server()
