#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Subsystem",
    "SubsystemClass",
    "SubsystemType",
    ]

from checkbox.lib.enum import (
    Enum,
    enum_value_to_name_string,
    )


SubsystemClass = Enum(
    "NONE",
    "DAEMON",
    "CLIENT",
    "JOB",
    )

SubsystemType = Enum(
    "INVALID",
    "MASTER",
    "RUNNER",
    "AUTO",
    )


class SubsystemEntry:

    __slots__ = (
        "_cls",
        "_name",
        "_substring",
        "_type",
        )

    def __init__(self, type, cls, name, substring):
        self._type = type
        self._cls = cls
        self._name = name
        self._substring = substring

    def get_class(self):
        """Get subsystem entry class."""
        return self._cls

    def get_type(self):
        """Get subsystem entry type."""
        return self._type

    def get_name(self):
        """Get subsystem entry name."""
        return self._name

    def get_substring(self):
        """Get subsystem entry substring."""
        return self._substring

    def match_by_class(self, cls):
        """Match subsystem entry by class."""
        return self._cls == cls

    def match_by_type(self, type):
        """Match subsystem entry by type."""
        return self._type == type

    def match_by_name(self, name):
        """Match subsystem entry by name."""
        return self._name.lower() == name.lower()

    def match_by_substring(self, substring):
        """Match subsystem entry by substring."""
        name = self._substring if self._substring else self._name
        return substring.lower() in name.lower()

    def is_valid(self):
        """Return whether the subsystem entry type is valid."""
        return self._type != SubsystemType.INVALID


class SubsystemTable:

    __slots__ = (
        "_entries",
        "_invalid",
        )

    def __init__(self):
        self._entries = []
        self._invalid = None

        self.add_entry(SubsystemType.MASTER, SubsystemClass.DAEMON, "MASTER")
        self.add_entry(SubsystemType.RUNNER, SubsystemClass.DAEMON, "RUNNER")
        self.add_entry(SubsystemType.INVALID, SubsystemClass.NONE, "INVALID")

    def add_entry(self, type, cls, name, substring=None):
        """Add an entry to the subsystem table.

        :param type: Subsystem type.
        :param cls: Subsystem class.
        :param name: Subsystem name.
        :param substring: Subsystem substring, optional.
        """
        entry = SubsystemEntry(type, cls, name, substring)
        self._entries.append(entry)

        if type == SubsystemType.INVALID:
            self._invalid = entry

    def get_entries(self):
        """Get all entries in the subsystem table."""
        return self._entries

    def get_valid_entries(self):
        """Get valid entries in the subsystem table."""
        for entry in self._entries:
            if entry.is_valid():
                yield entry

    def lookup_by_type(self, type):
        """Lookup valid entries by type.

        :param type: Type to lookup.
        """
        for entry in self.get_valid_entries():
            if entry.match_by_type(type):
                return entry

        return self._invalid

    def lookup_by_class(self, cls):
        """Lookup valid entries by class.

        :param cls: Class to lookup.
        """
        for entry in self.get_valid_entries():
            if entry.match_by_class(cls):
                return entry

        return self._invalid

    def lookup_by_name(self, name):
        """Lookup valid entries by name, which might be an exact or
           substring match.

        :param name: Name to lookup.
        """
        # Lookup exact match
        for entry in self.get_valid_entries():
            if entry.match_by_name(name):
                return entry

        # Lookup substring match
        for entry in self.get_valid_entries():
            if entry.match_by_substring(name):
                return entry

        return self._invalid


class Subsystem:

    __slots__ = (
        "_name",
        "_name_valid",
        "_temp_name",
        "_local_name",
        "_cls",
        "_cls_name",
        "_type",
        "_type_name",
        "_entry",
        "_table",
        )

    def __init__(self, name, type=SubsystemType.AUTO):
        self._name = None
        self._name_valid = False
        self._temp_name = None
        self._local_name = None

        self._cls = 0
        self._cls_name = None

        self._type = 0
        self._type_name = None

        self._entry = None
        self._table = SubsystemTable()

        self.set_name(name)
        if type == SubsystemType.AUTO:
            self.set_type_from_name(name)

        else:
            self.set_type(type)

    def is_class(self, cls):
        """Return whether the subsystem class is the same.

        :param cls: Subsystem class to compare.
        """
        return self._cls == cls

    def is_client(self):
        """Return whether the subsystem is a client."""
        return self._cls == SubsystemClass.CLIENT

    def is_daemon(self):
        """Return whether the subsystem is a daemon."""
        return self._cls == SubsystemClass.DAEMON

    def is_job(self):
        """Return whether the subsystem is a job."""
        return self._cls == SubsystemClass.JOB

    def is_name_valid(self):
        """Return whether the name is valid."""
        return self._name_valid

    def is_type(self, type):
        """Return whether the susbystem type is the same.

        :param type: Subsystem type to compare.
        """
        return self._type == type

    def is_valid(self):
        """Return whether the type and name are valid."""
        return self._type != SubsystemType.INVALID and self._name_valid

    def get_class_name(self):
        """Get the subsystem class name."""
        return self._cls_name

    def get_name(self):
        """Get the subsystem name."""
        return self._temp_name if self._temp_name else self._name

    def set_name(self, name):
        """Set the subsystem name."""
        if name:
            self._name = name
            self._name_valid = True

        else:
            self._name = "UNKNOWN"
            self._name_valid = False

    def get_local_name(self, default=None):
        """Get the subsystem local name.

        :param default: Default name if not found, optional.
        """
        if self._local_name is None:
            return default

        else:
            return self._local_name

    def set_local_name(self, local_name):
        """Set the subsystem local name.

        :param local_name: Local name to set.
        """
        self._local_name = local_name

    def has_local_name(self):
        """Return whether the subsystem has a local name."""
        return self._local_name is not None

    def set_temp_name(self, temp_name):
        """Set the subsystem temporary name, which will be returned
           by get_name().

        :param temp_name: Temporary name to set.
        """
        self._temp_name = temp_name

    def unset_temp_name(self):
        """Unset the subsystem temporary name."""
        self._temp_name = None

    def get_type(self):
        """Get the subsystem type."""
        return self._type

    def get_type_name(self):
        """Get the subsystem type name."""
        return self._type_name

    def set_type(self, type):
        """Set the subsystem type.

        :param type: Type to set.
        """
        self._set_type_by_name(type, None)

    def set_type_from_name(self, name=None):
        """Set the subsystem type by name.

        :param name: Type name to set, optional.
        """
        if name is None:
            name = self._name

        if name is None:
            self.set_type(SubsystemType.DAEMON)

        else:
            entry = self._table.lookup_by_name(name)
            if entry:
                self._set_entry_by_name(entry, name)

            else:
                self._set_type_by_name(SubsystemType.DAEMON, name)

    def _set_type_by_name(self, type, name):
        entry = self._table.lookup_by_type(type)
        self._set_entry_by_name(entry, name)

    def _set_entry_by_name(self, entry, name):
        self._entry = entry
        self._type = entry.get_type()
        self._set_class(entry)

        if name is not None:
            self._type_name = name

        else:
            self._type_name = entry.get_name()

    def _set_class(self, entry):
        self._cls = entry.get_class()
        self._cls_name = enum_value_to_name_string(SubsystemClass, self._cls)
