#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from unittest import TestCase

from tempfile import mktemp

from checkbox.process.family import FamilyInfo

from checkbox.daemon.child import Child, ChildError


class TestChild(TestCase):

    def setUp(self):
        self.input_pipe = os.pipe()
        self.output_pipe = os.pipe()
        self.error_pipe = os.pipe()

    def check_error(self, pipe, error=""):
        value = self.readPipe(pipe)
        if value:
            value = int(value, 16)

        self.assertEquals(error, value)

    def check_status(self, pid, value=0):
        status = os.waitpid(pid, 0)[1]
        self.assertEquals(status, value)

    def forkExec(self, pipe, *args, **kwargs):
        child = Child(pipe, *args, **kwargs)
        pid = child.forkExec()
        self.assertTrue(pid > 0)

        return pid

    def readPipe(self, pipe):
        os.close(pipe[1])
        file = os.fdopen(pipe[0], "r")
        return file.read()

    def writePipe(self, pipe, text):
        os.close(pipe[0])
        os.write(pipe[1], text)
        os.close(pipe[1])

    def test_executable_existing(self):
        pid = self.forkExec(self.error_pipe, "true", "/bin/true")
        self.check_error(self.error_pipe)
        self.check_status(pid)

    def test_executable_non_existing(self):
        executable_fullpath = mktemp()
        child = Child(self.error_pipe, "true", executable_fullpath)
        self.assertTrue(child.forkExec() > 0)
        self.check_error(self.error_pipe, 2)

    def test_args(self):
        std = [-1, self.output_pipe[1], -1]
        pid = self.forkExec(self.error_pipe, "echo", "/bin/echo",
            args=["echo", "-n", "foo"], std=std)
        self.check_error(self.error_pipe)
        self.check_status(pid)
        self.assertEquals(self.readPipe(self.output_pipe), "foo")

    def test_env(self):
        std = [-1, self.output_pipe[1], -1]

        pid = self.forkExec(self.error_pipe, "env", "/usr/bin/env",
            env={"FOO": "bar"}, std=std)
        self.check_error(self.error_pipe)
        self.check_status(pid)
        self.assertTrue("FOO=bar" in self.readPipe(self.output_pipe))

    def test_inherit_env_existing(self):
        std = [-1, self.output_pipe[1], -1]

        os.environ["FOO"] = "bar"
        pid = self.forkExec(self.error_pipe, "env", "/usr/bin/env",
            inherit_env=True, std=std)
        self.check_status(pid)
        self.assertTrue("FOO=bar" in self.readPipe(self.output_pipe))

    def test_inherit_env_non_existing(self):
        std = [-1, self.output_pipe[1], -1]

        del os.environ["FOO"]
        pid = self.forkExec(self.error_pipe, "env", "/usr/bin/env",
            inherit_env=True, std=std)
        self.check_status(pid)
        self.assertTrue("FOO=bar" not in self.readPipe(self.output_pipe))

    def test_directory_existing(self):
        directory = "/tmp"
        pid = self.forkExec(self.error_pipe, "true", "/bin/true",
            directory=directory)
        self.check_error(self.error_pipe)

    def test_directory_non_existing(self):
        directory = mktemp()
        pid = self.forkExec(self.error_pipe, "true", "/bin/true",
            directory=directory)
        self.check_error(self.error_pipe, 2)

    def test_family_info(self):
        family_info = FamilyInfo()
        pid = self.forkExec(
            self.error_pipe, "true", "/bin/true", family_info=family_info)

        self.check_error(self.error_pipe, ChildError.REGISTRATION_FAILED)

    def test_std(self):
        std = [self.input_pipe[0], self.output_pipe[1], -1]
        pid = self.forkExec(self.error_pipe, "cat", "/bin/cat",
            std=std)
        self.check_error(self.error_pipe)

        self.writePipe(self.input_pipe, "foo")
        self.assertEquals(self.readPipe(self.output_pipe), "foo")
