#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from unittest import TestCase

from checkbox.daemon.subsystem import (
    Subsystem,
    SubsystemClass,
    SubsystemEntry,
    SubsystemTable,
    SubsystemType,
    )


class TestSubsystemEntry(TestCase):

    def setUp(self):
        self.entry = SubsystemEntry(
            SubsystemType.MASTER, SubsystemClass.DAEMON, "foo", "bar")

    def test_get_class(self):
        self.assertEquals(self.entry.get_class(), SubsystemClass.DAEMON)

    def test_get_type(self):
        self.assertEquals(self.entry.get_type(), SubsystemType.MASTER)

    def test_get_name(self):
        self.assertEquals(self.entry.get_name(), "foo")

    def test_get_substring(self):
        self.assertEquals(self.entry.get_substring(), "bar")

    def test_match_by_class(self):
        self.assertTrue(self.entry.match_by_class(SubsystemClass.DAEMON))
        self.assertFalse(self.entry.match_by_class(SubsystemClass.CLIENT))

    def test_match_by_type(self):
        self.assertTrue(self.entry.match_by_type(SubsystemType.MASTER))
        self.assertFalse(self.entry.match_by_type(SubsystemType.INVALID))

    def test_match_by_name(self):
        self.assertTrue(self.entry.match_by_name("foo"))
        self.assertTrue(self.entry.match_by_name("FOO"))
        self.assertFalse(self.entry.match_by_name("bar"))

    def test_match_by_substring(self):
        self.assertTrue(self.entry.match_by_substring("bar"))
        self.assertTrue(self.entry.match_by_substring("BAR"))
        self.assertFalse(self.entry.match_by_substring("foo"))

    def test_is_valid(self):
        self.assertTrue(self.entry.is_valid())


class TestSubsystemEntry(TestCase):

    def setUp(self):
        self.table = SubsystemTable()

    def test_lookup_by_type(self):
        entry = self.table.lookup_by_type(SubsystemType.MASTER)
        self.assertNotEquals(entry, None)
        self.assertEquals(entry.get_type(), SubsystemType.MASTER)

    def test_lookup_by_class(self):
        entry = self.table.lookup_by_class(SubsystemClass.DAEMON)
        self.assertNotEquals(entry, None)
        self.assertEquals(entry.get_class(), SubsystemClass.DAEMON)

    def test_lookup_by_name(self):
        entry = self.table.lookup_by_name("master")
        self.assertNotEquals(entry, None)
        self.assertEquals(entry.get_name(), "MASTER")


class TestSubsystem(TestCase):

    def test_init(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_name(), "master")
        self.assertEquals(subsystem.get_type(), SubsystemType.MASTER)

        subsystem = Subsystem("master", SubsystemType.RUNNER)
        self.assertEquals(subsystem.get_name(), "master")
        self.assertEquals(subsystem.get_type(), SubsystemType.RUNNER)

    def test_is_class(self):
        subsystem = Subsystem("master")
        self.assertTrue(subsystem.is_class(SubsystemClass.DAEMON))
        self.assertFalse(subsystem.is_class(SubsystemClass.CLIENT))

    def test_is_client(self):
        subsystem = Subsystem("master")
        self.assertFalse(subsystem.is_client())

    def test_is_daemon(self):
        subsystem = Subsystem("master")
        self.assertTrue(subsystem.is_daemon())

    def test_is_job(self):
        subsystem = Subsystem("master")
        self.assertFalse(subsystem.is_job())

    def test_is_type(self):
        subsystem = Subsystem("master")
        self.assertTrue(subsystem.is_type(SubsystemType.MASTER))
        self.assertFalse(subsystem.is_type(SubsystemType.RUNNER))

    def test_is_valid(self):
        subsystem = Subsystem("master")
        self.assertTrue(subsystem.is_valid())

    def test_get_class_name(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_class_name(), "DAEMON")

    def test_get_name(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_name(), "master")

        subsystem.set_temp_name("foo")
        self.assertEquals(subsystem.get_name(), "foo")

        subsystem.unset_temp_name()
        self.assertEquals(subsystem.get_name(), "master")

    def test_set_name(self):
        subsystem = Subsystem("master")
        self.assertTrue(subsystem.is_name_valid())
        self.assertEquals(subsystem.get_name(), "master")

        subsystem.set_name("foo")
        self.assertTrue(subsystem.is_name_valid())
        self.assertEquals(subsystem.get_name(), "foo")

        subsystem.set_name(None)
        self.assertFalse(subsystem.is_name_valid())
        self.assertEquals(subsystem.get_name(), "UNKNOWN")

    def test_local_name(self):
        subsystem = Subsystem("master")
        self.assertFalse(subsystem.has_local_name())
        self.assertEquals(subsystem.get_local_name("foo"), "foo")

        subsystem.set_local_name("foo")
        self.assertTrue(subsystem.has_local_name())
        self.assertEquals(subsystem.get_local_name("bar"), "foo")

    def test_get_type(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_type(), SubsystemType.MASTER)

    def test_get_type_name(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_type_name(), "master")

    def test_set_type(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_type(), SubsystemType.MASTER)

        subsystem.set_type(SubsystemType.RUNNER)
        self.assertEquals(subsystem.get_type(), SubsystemType.RUNNER)

    def test_set_type_from_name(self):
        subsystem = Subsystem("master")
        self.assertEquals(subsystem.get_type(), SubsystemType.MASTER)

        subsystem.set_type_from_name("runner")
        self.assertEquals(subsystem.get_type(), SubsystemType.RUNNER)
