#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import time

from unittest import TestCase

from checkbox.process.timeslice import Timeslice

from checkbox.daemon.timer import (
    Timer,
    TimerManager,
    )


class TestTimer(TestCase):

    def test_lt(self):
        timer1 = Timer()
        timer2 = Timer()
        self.assertFalse(timer1 < timer2)

        timer2.when = time.time()
        self.assertTrue(timer1 < timer2)

        timer1.when = timer2.when
        self.assertFalse(timer1 < timer2)


class TestTimerManager(TestCase):

    def setUp(self):
        self.manager = TimerManager()
        self.call_count = 0

    def tearDown(self):
        self.manager.cancelTimers()

    def test_singleton(self):
        manager1 = TimerManager()
        manager2 = TimerManager()
        self.assertEquals(manager1, manager2)

    def handler(self, data):
        self.call_count += 1

    def test_createTimer(self):
        id = self.manager.createTimer(self.handler)
        self.assertTrue(id >= 0)
        self.assertEquals(self.call_count, 0)

        self.assertTrue(self.manager.timeout() == -1)
        self.assertEquals(self.call_count, 1)

        self.assertTrue(self.manager.timeout() == -1)
        self.assertEquals(self.call_count, 1)

    def test_createTimerDelta(self):
        id = self.manager.createTimer(self.handler, delta=1)
        self.assertTrue(id >= 0)
        self.assertEquals(self.call_count, 0)

        seconds = self.manager.timeout()
        self.assertTrue(seconds > 0)
        self.assertEquals(self.call_count, 0)

        time.sleep(seconds)
        self.assertTrue(self.manager.timeout() == -1)
        self.assertEquals(self.call_count, 1)

    def test_createTimerPeriod(self):
        id = self.manager.createTimer(self.handler, period=1)
        self.assertTrue(id >= 0)
        self.assertEquals(self.call_count, 0)

        seconds = self.manager.timeout()
        self.assertTrue(seconds > 0)
        self.assertEquals(self.call_count, 1)

        time.sleep(seconds)
        self.assertTrue(self.manager.timeout() > 0)
        self.assertEquals(self.call_count, 2)

    def test_createTimerTimeslice(self):
        timeslice = Timeslice()
        id = self.manager.createTimer(self.handler, timeslice=timeslice)
        self.assertTrue(id >= 0)
        self.assertEquals(self.call_count, 0)

        self.assertTrue(self.manager.timeout() == 0)
        self.assertEquals(self.call_count, 1)

        self.assertTrue(self.manager.timeout() == 0)
        self.assertEquals(self.call_count, 2)

    def test_createTimerTimesliceInitialInterval(self):
        timeslice = Timeslice()
        timeslice.set_initial_interval(2)
        id = self.manager.createTimer(self.handler, timeslice=timeslice)
        self.assertTrue(id >= 0)
        self.assertEquals(self.call_count, 0)

        seconds = self.manager.timeout()
        self.assertTrue(seconds >= 0)
        self.assertEquals(self.call_count, 0)

        time.sleep(seconds)
        self.assertTrue(self.manager.timeout() == 0)
        self.assertEquals(self.call_count, 1)

    def test_resetTimer(self):
        self.assertFalse(self.manager.resetTimer(0))
        id = self.manager.createTimer(self.handler, delta=1)

        self.assertTrue(self.manager.resetTimer(id))
        self.assertFalse(self.manager.resetTimer(id + 1))

    def test_cancelTimer(self):
        self.assertFalse(self.manager.cancelTimer(0))
        id = self.manager.createTimer(self.handler, period=1)

        seconds = self.manager.timeout()
        self.assertTrue(seconds > 0)
        self.assertEquals(self.call_count, 1)

        time.sleep(seconds)
        self.assertTrue(self.manager.cancelTimer(id))
        self.assertFalse(self.manager.cancelTimer(id + 1))
        self.assertTrue(self.manager.timeout() == -1)
        self.assertEquals(self.call_count, 1)
