#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "TimerManager",
    "TIMER_NEVER",
    ]

import logging
import select
import time

from heapq import (
    heapify,
    heappop,
    heappush,
    )

from checkbox.lib.singleton import Singleton


MAX_CHECKS_PER_TIMEOUT = 10
MAX_FIRES_PER_TIMEOUT = 3
TIMER_NEVER = 0xffffffff


class Timer:

    __slots__ = (
        "when",
        "period",
        "id",
        "data",
        "function",
        "timeslice",
        "release",
        )

    def __init__(self):
        self.when = 0.0
        self.period = 0
        self.id = -1
        self.data = None
        self.function = None
        self.timeslice = None

    def __lt__(self, other):
        return self.when < other.when


class TimerManager(Singleton):

    __slots__ = (
        "_ids",
        "_timers",
        "_in_timeout",
        "_did_reset",
        "_did_cancel",
        )

    _initialized = False

    def __init__(self):
        """Instantiate a timer manager singleton."""
        if TimerManager._initialized:
            return

        TimerManager._initialized = True
        self._ids = 0
        self._timers = []
        self._in_timeout = None
        self._did_reset = False
        self._did_cancel = False

    def __del__(self):
        self.cancelTimers()
        TimerManager._initialized = False

    def createTimer(
        self, function, delta=0, period=0, timeslice=None, data=None):
        """Create an event in the timer list. If period is 0, this event
           is a one time event instead of periodical.

        :param function: Function to call.
        :param delta: Delta when making the next call.
        :param period: Period between calls, 0 for single call.
        :param timeslice: Time slice between calls.
        :param data: Data to pass as argument to the function
        :return: The assigned timer identifier.
        """
        timer = Timer()
        timer.function = function
        timer.period = period
        timer.data = data

        if timeslice:
            timer.timeslice = timeslice.copy()
            delta = timeslice.get_time_to_next_run()

        if delta == TIMER_NEVER:
            timer.when = TIMER_NEVER
        else:
            timer.when = delta + time.time()

        timer.id = self._ids
        self._ids += 1

        heappush(self._timers, timer)

        return timer.id

    def resetTimer(self, id, delta=0, period=0):
        """Reset the given timer.

        :param id: Identifier of the timer.
        :param delta: New delta when making the next call.
        :param period: New period between calls.
        :return: True on success, false on failure.
        """
        if not self._timers:
            logging.info("Resetting timer from empty list.")
            return False

        for timer in self._timers:
            if timer.id == id:
                break
        else:
            logging.info("Timer not found: %d", id)
            return False

        if timer.timeslice:
            logging.info("Cannot reset timer with timeslice: %d", id)
            return True

        if delta == TIMER_NEVER:
            timer.when = TIMER_NEVER
        else:
            timer.when = delta + time.time()

        timer.period = period

        heapify(self._timers)

        if self._in_timeout == timer:
            self._did_reset = True

        return True

    def cancelTimer(self, id):
        """Cancel the given timer.

        :param id: Identifier of the timer.
        :return: True on success, false on failure.
        """
        if not self._timers:
            logging.info("Cancelling timer from empty list.")
            return False

        for index, timer in enumerate(self._timers):
            if timer.id == id:
                break
        else:
            logging.info("Timer not found: %d", id)
            return False

        del self._timers[index]

        if self._in_timeout == timer:
            self._did_cancel = True

        return True

    def cancelTimers(self):
        """Cancel all timers."""
        while self._timers:
            timer = heappop(self._timers)
            if self._in_timeout == timer:
                self._did_cancel = True

    def timeout(self):
        """Called when a select() times out.

        :return: Number of seconds until the next timeout event, 0
        if there are no more events, or -1 if called while a function
        is active.
        """
        if self._in_timeout:
            if self._timers:
                result = self._timers[0].when - time.time()
                if result < 0:
                    result = 0
            else:
                result = 0

            return result

        if not self._timers:
            logging.info("Empty timer list, nothing to do.")

        now = time.time()
        check_count = fire_count = 0
        while True:
            if not self._timers or self._timers[0].when > now:
                break

            if fire_count >= MAX_FIRES_PER_TIMEOUT:
                break

            timer = self._timers[0]
            self._in_timeout = timer

            check_count += 1
            fire_count += 1

            # Only check the time once in a while
            if check_count > MAX_CHECKS_PER_TIMEOUT:
                check_count = 0

                sample = time.time()
                if now > sample:
                    logging.info("Clock skew detected.")
                    now = sample

            # Initialize flags
            self._did_reset = False
            self._did_cancel = False

            if timer.timeslice:
                timer.timeslice.set_start_time_now()

            timer.function(timer.data)

            if timer.timeslice:
                timer.timeslice.set_finish_time_now()

            if self._did_cancel:
                # Timer was canceled inside its function
                self._timers.remove(timer)

            elif not self._did_reset:
                if timer.period > 0 or timer.timeslice:
                    timer.when = time.time()
                    if timer.timeslice:
                        timer.when += timer.timeslice.get_time_to_next_run()
                    else:
                        timer.when += timer.period

                    heapify(self._timers)

                else:
                    self._timers.remove(timer)

        if self._timers:
            result = self._timers[0].when - time.time()
            if result < 0:
                result = 0

        else:
            result = -1

        self._in_timeout = None
        return result

    def start(self):
        """Start the main timeout loop.

        :return: Never.
        """
        while True:
            seconds = self.timeout()
            if seconds == 0:
                # No timer events registered, waiting on signal
                select.select([], [], [])

            else:
                select.select([], [], [], seconds)
