#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Bucket",
    "BucketChain",
    ]

import logging

from collections import deque


BUCKET_SIZE = 2 ** 12


class Bucket:

    __slots__ = (
        "_buffer",
        "_buffers",
        "_max_size",
        "_touched",
        "_used",
        )

    def __init__(self, size=BUCKET_SIZE):
        self._max_size = size
        self.reset()

    def reset(self):
        """Reset the bucket."""
        self._buffer = ""
        self._buffers = []
        self._touched = 0
        self._used = 0

    @property
    def max_size(self):
        """Return the maximum bucket size."""
        return self._max_size

    @property
    def touched(self):
        """Return the number of bytes touched."""
        return self._touched

    @property
    def untouched(self):
        """Return the number of bytes untouched."""
        return self._used - self._touched

    @property
    def used(self):
        """Return the number of bytes used."""
        return self._used

    @property
    def unused(self):
        """Return the number of bytes unused."""
        return self._max_size - self._used

    @property
    def is_empty(self):
        """Return whether the bucket is empty."""
        return self._used == 0

    @property
    def is_full(self):
        """Return whether the bucket is full."""
        return self._used == self._max_size

    @property
    def is_consumed(self):
        """Return whether the bucket has been consumed."""
        return self._touched == self._used

    def rewind(self):
        """Rewind the bucket."""
        self._touched = 0

    def read(self, sock, size=-1):
        """Read bytes from the given socket into the bucket.

        :param sock: Socket to read from.
        :param size: Number of bytes to read or -1 for all.
        :return: Number of bytes read.
        """
        if size < 0 or size > self.unused:
            size = self.unused

        bytes = sock.getBytesRaw(size)
        if bytes is None:
            logging.info("Bucket read failed.")
            return None

        return self.putMax(bytes)

    def write(self, sock, size=-1):
        """Write bytes from the bucket to the given socket.

        :param sock: Socket to write to.
        :param size: Number of bytes to write or -1 for all.
        :return: Number of bytes written.
        """
        if size < 0 or size > self.untouched:
            size = self.untouched

        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        new_position = self._touched + size
        size = sock.putBytesRaw(self._buffer[self._touched:new_position])
        if size < 0:
            logging.info("Bucket write failed.")
            return -1

        self._touched += new_position

        return size

    def flush(self, sock):
        """Flush all the bytes from the bucket to the given socket and
           reset the bucket.

        :param sock: Socket to write to.
        :return: Number of bytes written.
        """
        self.rewind()
        size = self.write(sock)

        self.reset()
        return size

    def putMax(self, bytes):
        """Put the maximum number of bytes to fill the bucket.

        :param bytes: Bytes to put into the bucket.
        :return: Number of bytes put.
        """
        size = min(len(bytes), self.unused)
        if size == 0:
            return 0

        self._buffers.append(bytes[:size])
        self._used += size

        return size

    def getMax(self, size=-1):
        """Get untouched bytes from the bucket.

        :param size: Maximum number of bytes to get.
        :return: Bytes gotten.
        """
        if size < 0 or size > self.untouched:
            size = self.untouched

        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        bytes = self._buffer[self._touched:self._touched + size]
        self._touched += size

        return bytes

    def find(self, sub):
        """Find a substring in the untouched bytes of the bucket.

        :param sub: Substring to find.
        :return: Offset starting at untouched bytes.
        """
        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        index = self._buffer.find(sub, self._touched)
        if index < 0:
            return -1

        return index - self._touched

    def peek(self):
        """Peek the next byte in the bucket without touching it."""
        if self._touched == self._used:
            return None

        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        return self._buffer[self._touched]

    def seek(self, position=0):
        """Seek the untouched bytes in the bucket at the given
           position.

        :param position: Position where to seek.
        :return: Previous position.
        """
        old_position = self._touched
        if position < 0:
            self._touched = 0
        elif position < self._max_size:
            self._touched = position
        else:
            self._touched = self._max_size

        if self._touched > self._used:
            self.putMax("\0" * (self._touched - self._used))
            self._used = self._touched

        return old_position

    def replace(self, bytes, position, size=-1):
        """Replace bytes in the bucket.

        :param bytes: New bytes to replace.
        :param position: Position where to replace.
        :param size: Number of bytes to replace or -1 for all.
        :return: Number of bytes replaced.
        """
        if size < 0:
            size = len(bytes)

        if position > self._used:
            position = self._used
            size = 0

        elif position + size > self._used:
            size = self._used - position

        if self._buffers:
            self._buffer += "".join(self._buffers)

        self._buffers = [self._buffer[:position],
            bytes, self._buffer[position + size:]]
        self._buffer = ""

        return size


class BucketChain:

    __slots__ = (
        "_buckets",
        )

    def __init__(self):
        self.reset()

    @property
    def is_consumed(self):
        """Return whether the bucket chain has been consumed."""
        return not self._buckets or self._buckets[-1].is_consumed

    def reset(self):
        """Reset the bucket chain."""
        self._buckets = deque()

    def get(self, size):
        """Get untouched bytes from the bucket chain.

        :param size: Maximum number of bytes to get.
        :return: Bytes gotten.
        """
        total = 0
        buffers = []
        while self._buckets:
            bucket = self._buckets[0]
            buffer = bucket.getMax(size - total)

            buffers.append(buffer)
            total += len(buffer)
            if total == size:
                break

            self._buckets.popleft()

        return "".join(buffers)

    def put(self, bucket):
        """Put a bucket onto the chain.

        :param bucket: Bucket to put.
        :return: Whether the operation succeeded.
        """
        self._buckets.append(bucket)
        return True

    def search(self, sub):
        """Search for bytes in the bucket chain until a substring is
           encountered.

        :param sub: Substring to search for.
        :return: Bytes searched.
        """
        if not self._buckets:
            return None

        # Delim in one bucket
        bucket = self._buckets[0]
        index = bucket.find(sub)
        if index >= 0:
            buffer = bucket.getMax(index)
            position = bucket.seek(0)
            bucket.seek(position + len(sub))
            return buffer

        # Delim in more than one bucket
        if len(self._buckets) == 1:
            return None

        untouched = bucket.untouched
        for bucket in self._buckets[1:]:
            index = bucket.find(sub)
            if index < 0:
                untouched += bucket.untouched

            else:
                untouched += index
                return self.get(untouched + len(sub))

        return None

    def peek(self):
        """Peek the next byte in the bucket chain without touching it."""
        if not self._buckets:
            return None

        c = self._buckets[0].peek()
        if c is None and len(self._buckets) > 1:
            c = self._buckets[1].peek()

        return c
