#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Fifo",
    ]

import os
import logging

from itertools import count

from checkbox.lib.enum import Enum
from checkbox.lib.fifo import (
    FIFO_BUF,
    fifo_create,
    )
from checkbox.lib.file import (
    File,
    fd_open,
    fd_get_flags,
    fd_set_flags
    )

from checkbox.io.buffer import Buffer
from checkbox.io.pipe import Pipe
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.watchdog import (
    WatchdogClient,
    WatchdogServer,
    )


FifoState = Enum(
    "VIRGIN",
    "CONNECT",
    "ACCEPT",
    "LISTEN",
    )


class Fifo(Pipe):

    __slots__ = (
        "_serial_number",
        "_pid",
        "_path",
        "_dummy",
        "_state",
        )

    _serial_number_counter = count(0)
    _buffer_size = FIFO_BUF

    def __init__(self):
        super(Fifo, self).__init__()
        self._path = None
        self._pid = 0
        self._serial_number = -1

        self._dummy = File()
        self._state = FifoState.VIRGIN

    def __del__(self):
        if self._state != FifoState.VIRGIN:
            if self._state == FifoState.CONNECT:
                path = self._makeClientPath(
                    self._path, self._pid, self._serial_number)
            else:
                path = self._path

            if os.path.exists(path):
                os.unlink(path)

    def setPrincipal(self, id):
        """Set the principal that is allowed to connect to this server.

        :param id: String representation of the user identifier.
        """
        if self._state != FifoState.LISTEN:
            return False

        # Determine the client uid
        euid = os.geteuid()
        if id:
            client_uid = int(id)
        elif euid == 0:
            client_uid = os.getuid()
        else:
            client_uid = euid

        # Return if the client uid is the same as the effective uid
        if client_uid == euid:
            return True

        # Set permissions if running as root
        if euid == 0:
            os.chown(self._path, client_uid, -1)
            os.chown(self._watchdog_server.getPath(), client_uid, -1)
            return True

        # Otherwise, can't allow connections
        logging.info(
            "Running as uid %u, can't allow connections from uid %u",
            euid, client_uid)

        return False

    def connect(self, path):
        if self._state != FifoState.VIRGIN:
            return False

        # Connect to the server watchdog fifo
        self._watchdog_client = WatchdogClient()
        if not self._watchdog_client.setPath(path):
            self._watchdog_client = None
            return False

        # Connect to the server command fifo
        if not self._openWriter(path):
            self._watchdog_client = None
            return False

        # Store the path, pid and next serial number
        self._path = path
        self._pid = os.getpid()
        self._serial_number = Fifo._serial_number_counter.next()

        self._state = FifoState.CONNECT

        return True

    def listen(self, path):
        """Initialize a local server instance to accept commands on the
           given path.

        :param path: Path pointing to a fifo.
        """
        if self._state != FifoState.VIRGIN:
            return False

        # Connect to the server watchdog fifo
        self._watchdog_server = WatchdogServer()
        if not self._watchdog_server.setPath(path):
            self._watchdog_server = None
            return False

        # Create the reader to get client requests
        if not self._openReader(path):
            self._watchdog_server = None
            return False

        self._path = path
        self._state = FifoState.LISTEN

        return True

    def accept(self):
        """Accept a connection on this fifo."""
        if self._state != FifoState.LISTEN:
            return False

        # Check if a connection arrived within the timeout
        selector = Selector()
        selector.setTimeout(self._timeout)
        selector.add_fd(self._reader.fileno(), SelectorIO.READ)

        selector.execute()

        if not selector.has_ready:
            return 0

        # Fifo is ready for reading
        pid = self.getInteger()
        if pid is None:
            logging.info("Failed to get client pid")
            return -1

        serial_number = self.getInteger()
        if serial_number is None:
            logging.info("Failed to get client serial number")
            return -1

        # Create the fifo for responding to the client
        path = self._makeClientPath(self._path, pid, serial_number)
        if not self._openWriter(path):
            return -1

        self._state = FifoState.ACCEPT

        return 1

    def getEnd(self):
        """See Stream.getEnd"""
        self._bytes = ""

        if self._state == FifoState.CONNECT:
            self._reader.reset()
            return True
        elif self._state == FifoState.ACCEPT:
            return True
        else:
            return False

    def putBytes(self, bytes):
        """See Stream.putBytes"""
        if len(bytes) > self._buffer_size:
            bytes = bytes[:self._buffer_size]

        if self._state == FifoState.CONNECT:
            self._bytes += bytes
            return len(bytes)
        elif self._state == FifoState.ACCEPT:
            return self.putBytesRaw(bytes)
        else:
            return -1

    def putEnd(self):
        """See Stream.putEnd"""
        if self._state == FifoState.CONNECT:
            # Create a fifo reader to receive responses
            path = self._makeClientPath(
                self._path, self._pid, self._serial_number)
            if not self._openReader(path):
                logging.info("Error initializing fifo reader.")
                self._reader = None
                return False

            # Write pid, serial number and buffer
            buffer = Buffer()
            buffer.putInteger(self._pid)
            buffer.putInteger(self._serial_number)
            buffer.putBytes(self._bytes)
            self._bytes = ""

            bytes = buffer.getBytes()
            if self.putBytesRaw(bytes) != len(bytes):
                logging.info("Error sending message to server.")
                return False
        elif self._state == FifoState.ACCEPT:
            self._state = FifoState.LISTEN
            self._writer.reset()
        else:
            return False

        return True

    def _openReader(self, path):
        # Keep dummy writer open so that reader never receives EOF
        if not fifo_create(path, self._reader, self._dummy):
            logging.info("Failed to set fifo path at %s", path)
            return False

        return True

    def _openWriter(self, path):
        # Open write connection to the server
        write_fd = fd_open(path, os.O_WRONLY | os.O_NONBLOCK)
        if write_fd == -1:
            logging.info("Failed to open write-only %s", path)
            return False

        # Set the connection back to blocking mode
        flags = fd_get_flags(write_fd)
        if flags < 0:
            return False

        flags &= ~os.O_NONBLOCK
        if fd_set_flags(write_fd, flags) < 0:
            return False

        self._writer.initialize_from_fd(write_fd, "w")

        return True

    def _makeClientPath(self, path, pid, serial_number):
        return "%s.%u.%u" % (path, pid, serial_number)
