#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "MessageID",
    "MessageIncoming",
    "MessageOutgoing",
    "MessagePacket",
    "MESSAGE_MAX_PACKET_SIZE",
    ]

import logging
import socket
import struct
import time

from collections import deque

from checkbox.io.buffer import Buffer


MESSAGE_NO_OF_DIR_ENTRY = 41

MESSAGE_MAX_PACKET_SIZE = 60000
MESSAGE_HEADER_SIZE = 30
MESSAGE_FRAGMENT_SIZE = 1000
MESSAGE_PACKET_SIZE = MESSAGE_FRAGMENT_SIZE - MESSAGE_HEADER_SIZE

MESSAGE_MAGIC = "MaGiC"


class MessageID:

    __slots__ = (
        "address",
        "pid",
        "time",
        "number",
        )

    def __init__(self, address=None, pid=0, time=0, number=0):
        self.address = address
        self.pid = pid
        self.time = time
        self.number = number

    def __cmp__(self, other):
        if not self:
            return 1 if other else 0
        elif not other:
            return -1
        else:
            return (
                cmp(self.address, other.address) or
                cmp(self.pid, other.pid) or
                cmp(self.time, other.time) or
                cmp(self.number, other.number))

    def __hash__(self):
        ip = struct.unpack(">i", socket.inet_aton(self.address))[0]
        return abs(ip + self.time + self.number)

    def copy(self):
        """Create a copy of the current message ID.

        :return: New message ID.
        """
        id = MessageID()
        id.copyFrom(self)
        return id

    def copyFrom(self, other):
        """Copy another message ID into this one."""
        self.address = other.address
        self.pid = other.pid
        self.time = other.time
        self.number = other.number


class MessageDirectory:

    __slots__ = (
        "number",
        "previous",
        "next",
        "entries",
        )

    def __init__(self, number=0, previous=None):
        self.number = number
        self.previous = previous
        self.next = None
        self.entries = [None] * MESSAGE_NO_OF_DIR_ENTRY


class MessageHeader:

    __slots__ = (
        "sequence",
        "id",
        "data",
        "length",
        "is_full",
        "is_last",
        )

    def __init__(self):
        self.is_last = False
        self.sequence = 0
        self.id = None

        self.data = None
        self.length = 0
        self.is_full = False


class MessagePacket:

    __slots__ = (
        "_buffer",
        "_buffers",
        "_touched",
        "_used",
        "_max_size",
        )

    def __init__(self, size=MESSAGE_PACKET_SIZE):
        self._max_size = size
        self.reset()

    @property
    def is_consumed(self):
        """Return whether the bucket is consumed."""
        return self._touched == self._used

    @property
    def is_full(self):
        """Return whether the bucket is full."""
        return self._used == self._max_size

    @property
    def is_empty(self):
        """Return whether the bucket is empty."""
        return not self._used

    def reset(self):
        """Reset the packet."""
        self._buffer = ""
        self._buffers = []
        self._touched = 0
        self._used = 0

    def rewind(self):
        """Rewind the bucket."""
        self._touched = 0

    def getBytes(self, size=-1):
        """Get a fixed number of bytes from the packet.

        :param size: Number of bytes to get.
        :return: Bytes gotten.
        """
        unused = self._used - self._touched
        if size < 0 or size > unused:
            size = unused

        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        bytes = self._buffer[self._touched:self._touched + size]
        self._touched += size

        return bytes

    def getHeader(self):
        """Get the header from the packet and move the position to the
           beginning of the data."""
        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        header = MessageHeader()
        buffer = Buffer(self._buffer)
        magic = buffer.getBytes(len(MESSAGE_MAGIC))

        if magic != MESSAGE_MAGIC:
            header.length = self._used
            header.data = self._buffer

            header.is_full = True
            self._touched = 0
        else:
            header.is_last = buffer.getBoolean()
            header.sequence = buffer.getInteger()
            header.length = buffer.getInteger()

            header.id = MessageID()
            header.id.address = buffer.getAddress()
            header.id.pid = buffer.getInteger()
            header.id.time = buffer.getInteger()
            header.id.number = buffer.getInteger()
            header.data = buffer.getBytes()

            header.is_full = False
            self._touched = MESSAGE_HEADER_SIZE

        return header

    def peek(self):
        """Peek the next character in the packet without consuming it."""
        if self._touched == self._used:
            return None

        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        return self._buffer[self._touched]

    def search(self, sub):
        """Search for bytes in the packet until a substring is encountered.

        :param sub: Substring to search for.
        :return: Bytes searched.
        """
        if self._buffers:
            self._buffer += "".join(self._buffers)
            self._buffers = []

        index = self._buffer.find(sub, self._touched)
        if index < 0:
            return None

        bytes = self._buffer[self._touched:index]
        self._touched = index + len(sub)

        return bytes

    def putBytes(self, bytes):
        """Put bytes onto the packet.

        :param bytes: Bytes to put.
        :return: Number of bytes put.
        """
        size = min(len(bytes), self._max_size - self._used)
        if not size:
            return 0

        self._buffers.append(bytes[:size])
        self._used += size

        return size

    def putHeader(self, is_last, sequence, id):
        """Put the header into the packet.

        :param is_last: Whether this is the last packet.
        :param sequence: Sequence number.
        :param id: Message ID.
        :return: Number of bytes in the header.
        """
        buffer = Buffer()
        buffer.putBytes(MESSAGE_MAGIC)
        buffer.putBoolean(is_last)
        buffer.putInteger(sequence)
        buffer.putInteger(self._used)
        buffer.putAddress(id.address)
        buffer.putInteger(id.pid)
        buffer.putInteger(id.time)
        buffer.putInteger(id.number)

        if self._buffers:
            self._buffer += "".join(self._buffers)

        # Check if header has already been put
        if self._buffer.startswith(MESSAGE_MAGIC):
            self._buffer = self._buffer[MESSAGE_MAGIC:]

        self._buffers = [buffer.getBytes(), self._buffer]
        self._buffer = ""

        self._touched += MESSAGE_HEADER_SIZE
        self._used += MESSAGE_HEADER_SIZE

        return MESSAGE_HEADER_SIZE


class MessageOutgoing:

    __slots__ = (
        "_packets",
        )

    def __init__(self):
        self._packets = deque([MessagePacket()])

    def putBytes(self, bytes):
        """Put bytes onto the outgoing message.

        :param bytes: Bytes to put.
        :return: Number of bytes put.
        """
        size = len(bytes)

        total = 0
        packet = self._packets[-1]
        while total < size:
            if packet.is_full:
                packet = MessagePacket()
                self._packets.append(packet)
            total += packet.putBytes(bytes[total:])

        return total

    def send(self, sock, id):
        """Send bytes from the outgoing message to the given socket.

        :param sock: Socket to write to.
        :param id: Message ID of the message.
        :return: Number of bytes sent.
        """
        if self._packets[0].is_empty:
            return 0

        total = 0
        last_packet = self._packets[-1]
        for sequence, packet in enumerate(self._packets):
            # Short messages are sent without a magic header
            if sequence != 0 or packet != last_packet:
                packet.putHeader(packet == last_packet, sequence, id)

            packet.rewind()
            bytes = packet.getBytes()
            size = sock.putBytesRaw(bytes)
            if size < 0:
                logging.info("Failed to send packet.")
                total = -1
                break

            total += size

        self._packets = [self._packets[0]]
        self._packets[0].reset()

        return total


class MessageIncoming:

    __slots__ = (
        "_id",
        "_length",
        "_last_number",
        "_last_time",
        "_passed",
        "_received",
        "_current_packet",
        "_current_data",
        "_head_directory",
        "_current_directory",
        "previous",
        "next",
        )

    def __init__(self, id, is_last, sequence, bytes, previous=None):
        self._id = id.copy()

        self._length = len(bytes)
        self._last_number = sequence if is_last else 0
        self._last_time = time.time()
        self._passed = 0
        self._received = 1

        # Make directory entries as needed
        self._head_directory = self._current_directory = MessageDirectory()
        dest_directory_number = int(sequence / MESSAGE_NO_OF_DIR_ENTRY)
        while dest_directory_number != self._current_directory.number:
            self._current_directory.next = MessageDirectory(
                self._current_directory.number + 1,
                self._current_directory)
            self._current_directory = self._current_directory.next

        # Make a packet and add to the appropriate directory entry
        index = sequence % MESSAGE_NO_OF_DIR_ENTRY
        self._current_directory.entries[index] = bytes
        self._current_packet = 0
        self._current_data = 0

        self.previous = previous
        self.next = None

    @property
    def is_consumed(self):
        """Return whether the bucket is consumed."""
        return self._length != 0 and self._length == self._passed

    def put(self, is_last, sequence, bytes):
        """Put bytes in sequence onto the incoming message.

        :param is_last: Whether this is the last message.
        :param sequence: Sequence number.
        :param bytes: Bytes to put.
        :return: Whether the last message was put.
        """
        if self._last_number != 0 and self._last_number + 1 == self._received:
            logging.info("Duplicated packet, the message is defragmented.")
            return False

        dest_directory_number = int(sequence / MESSAGE_NO_OF_DIR_ENTRY)
        while dest_directory_number != self._current_directory.number:
            # Move forward, while appending new directory entries
            if dest_directory_number < self._current_directory.number:
                if not self._current_directory.next:
                    self._current_directory.next = MessageDirectory(
                        self._current_directory.number + 1,
                        self._current_directory)

                self._current_directory = self._current_directory.next

            # Move backward, directory entries have already been made
            else:
                self._current_directory = self._current_directory.previous

        index = sequence % MESSAGE_NO_OF_DIR_ENTRY
        entries = self._current_directory.entries

        # New packet
        if entries[index] is None:
            entries[index] = bytes

            self._length += len(bytes)
            self._received += 1
            if is_last:
                self._last_number = sequence

            # Every packet has been received
            if self._last_number + 1 == self._received:
                self._current_directory = self._head_directory
                self._current_packet = 0
                self._current_data = 0
                return True
            else:
                self._last_time = time.time()
                return False

        # Duplicated packet
        else:
            return False

    def getBytes(self, size):
        """Get bytes from the incoming message.

        :param size: Number of bytes to get.
        :return: Bytes gotten.
        """
        if self._passed + size > self._length:
            logging.info(
                "More data (%d) than queued (%d) requested",
                size, self._length - self._passed)
            return None

        bytes = []
        total = 0
        while total < size:
            current_entry = (
                self._current_directory.entries[self._current_packet])
            length = min(
                size - total, len(current_entry) - self._current_data)
            bytes.append(current_entry[self._current_data:])

            self._incrementCurrentData(length)
            total += length

        return "".join(bytes)

    def peek(self):
        """
        Peek the next character in the incoming message without consuming it.
        """
        current_entry = self._current_directory.entries[self._current_packet]
        if current_entry is None:
            return None

        return current_entry[self._current_data]

    def search(self, sub):
        """Search for bytes in the incoming message until a substring
           is encountered.

        :param sub: Substring to search for.
        :return: Bytes searched.
        """
        number = 0

        temp_directory = self._current_directory
        temp_packet = self._current_packet
        temp_data = self._current_data

        copy_needed = False
        while True:
            entry = temp_directory.entries[temp_packet]
            if entry is None:
                return None

            data = entry[temp_data:]
            size = len(data) - temp_data
            index = data.find(sub)
            if index >= 0:
                number += index
                if number == size:
                    copy_needed = True

                if not copy_needed:
                    self._incrementCurrentData(number + len(sub))
                    return data[:number]

                # Found substring
                break

            # Data spans multiple buffers
            copy_needed = True

            number += size
            temp_packet += 1
            temp_data = 0

            if temp_packet == MESSAGE_NO_OF_DIR_ENTRY:
                if not temp_directory.next:
                    return None

                temp_directory = temp_directory.next
                temp_packet = 0

            # Last packet
            elif temp_directory.entries[temp_packet] is None:
                return None

        bytes = self.getBytes(number)
        self._incrementCurrentData(len(sub))

        return bytes

    def _incrementCurrentData(self, number):
        self._current_data += number
        self._passed += number

        entries = self._current_directory.entries
        if self._current_data == len(entries[self._current_packet]):
            entries[self._current_packet] = None
            self._current_packet += 1

            # Last of the current directory
            if self._current_packet == MESSAGE_NO_OF_DIR_ENTRY:
                self._head_directory \
                  = self._current_directory \
                  = self._head_directory.next
                if self._head_directory:
                    self._head_directory.previous = None

                self._current_packet = 0

            self._current_data = 0
