#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Pipe",
    ]

import logging

from checkbox.lib.file import File

from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.stream import Stream


# Number of bytes in atomic write to a pipe.
PIPE_BUF = 2 ** 12


class Pipe(Stream):

    __slots__ = (
        "_fd",
        "_bytes",
        "_reader",
        "_writer",
        "_watchdog_client",
        "_watchdog_server",
        )

    _buffer_size = PIPE_BUF

    def __init__(self):
        super(Pipe, self).__init__()
        self._fd = None
        self._bytes = ""
        self._reader = File()
        self._writer = File()
        self._watchdog_client = None
        self._watchdog_server = None

    def fileno(self):
        return self._fd

    def get_description(self):
        if self._fd is None:
            return None

        return "fd:%d" % self._fd

    def connect(self, fd):
        self._fd = fd
        return self._writer.initialize_from_fd(fd, "w") == 0

    def listen(self, fd):
        self._fd = fd
        return self._reader.initialize_from_fd(fd, "r") == 0

    def handleIncomingMessage(self):
        """Receive messages from the reader and buffer them.

        :return: Whether a message becomes ready.
        """
        bytes = self.getBytesRaw(PIPE_BUF)
        if bytes is None:
            return False

        self._bytes += bytes

        return True

    def getBytes(self, size=-1):
        """See Stream.getBytes"""
        if size < 0:
            size = self._buffer_size
        else:
            size = min(size, self._buffer_size)
        logging.debug("getting: %s", size)

        while size > len(self._bytes):
            if not self.handleIncomingMessage():
                break

        size = min(size, len(self._bytes))
        bytes = self._bytes[:size]
        self._bytes = self._bytes[size:]

        logging.debug("got bytes: %s", len(bytes))
        return bytes

    def getBytesRaw(self, size):
        """Get bytes directly from the source.

        :param size: Number of bytes to get.
        :return: Bytes gotten.
        """
        read_fd = self._reader.fileno()

        selector = Selector()
        selector.setTimeout(self._timeout)
        selector.add_fd(read_fd, SelectorIO.READ)

        # Select with both the real and watchdog descriptors
        if self._watchdog_client:
            selector.add_fd(self._watchdog_client.fileno(), SelectorIO.READ)

        selector.execute()
        if not selector.has_ready:
            return None

        if (self._watchdog_client and
            selector.check_fd(
                self._watchdog_client.fileno(), SelectorIO.READ) and
            not selector.check_fd(read_fd, SelectorIO.READ)):
            logging.info("Error reading from source, watchdog side closed")
            return None

        bytes = self._reader.read(size)
        if bytes is None:
            logging.info(
                "Error reading from source, errno %d", self._reader.errno)

        return bytes

    def getEnd(self):
        return True

    def peek(self):
        """See Stream.peek"""
        while not self._bytes:
            if not self.handleIncomingMessage():
                return None

        return self._bytes[0]

    def search(self, sub):
        """See Stream.search"""
        bytes = ""
        position = 0
        while True:
            index = self._bytes.find(sub, position)
            if index >= 0:
                bytes = self._bytes[:index]
                self._bytes = self._bytes[index + len(sub):]
                break

            position = len(self._bytes)

            if not self.handleIncomingMessage():
                return None

        return bytes

    def putBytes(self, bytes):
        """See Stream.putBytes"""
        if len(bytes) > self._buffer_size:
            bytes = bytes[:self._buffer_size]

        return self.putBytesRaw(bytes)

    def putBytesRaw(self, bytes):
        """Put bytes directly onto the destination.

        :param bytes: Bytes to put.
        :return: Number of bytes put.
        """
        # Select with both the real and watchdog descriptors
        if self._watchdog_client:
            writer_fd = self._writer.fileno()
            watchdog_fd = self._watchdog_client.fileno()

            selector = Selector()
            selector.add_fd(writer_fd, SelectorIO.WRITE)
            selector.add_fd(watchdog_fd, SelectorIO.READ)

            selector.execute()

            if not selector.has_ready:
                logging.info("Put bytes failed")
                return -1

            if selector.check_fd(watchdog_fd, SelectorIO.READ):
                logging.info(
                    "Error writing to descriptor, watchdog side closed")
                return -1

        size = self._writer.write(bytes)
        if size < 0:
            logging.info(
                "Error writing to descriptor, errno %d", self._writer.errno)

        return size

    def putEnd(self):
        return True
