#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Sock",
    "SockAddress",
    "SockPort",
    "SockState",
    ]

import os
import errno
import socket
import struct
import logging

from ctypes import c_int
from datetime import datetime
from fcntl import ioctl
from termios import FIONREAD
from time import time

from checkbox.lib import (
    hostname,
    net,
    )
from checkbox.lib.enum import Enum
from checkbox.lib.file import (
    fd_get_flags,
    fd_set_flags,
    fd_close,
    )

from checkbox.io.error import StreamError
from checkbox.io.port import get_port_range
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.stream import Stream


SOCK_CONNECT_TIMEOUT = 10

SockAddress = type("SocketAddress", (object,), dict(
    LOOPBACK=socket.inet_ntoa(struct.pack("!L", socket.INADDR_LOOPBACK)),
    ANY=socket.inet_ntoa(struct.pack("!L", socket.INADDR_ANY))))

SockPort = Enum(
    ("OUTBOUND", -2),
    ("INBOUND", -1),
    ("ANY", 0),
    )

SockState = Enum(
    "VIRGIN",
    "ASSIGNED",
    "BOUND",
    "CONNECT",
    "SPECIAL",
    "CONNECT_PENDING",
    "CONNECT_PENDING_RETRY",
    )


class SockConnectState:

    __slots__ = (
        "retry_timeout_interval",
        "connect_failed",
        "failed_once",
        "connect_refused",
        "first_try_start_time",
        "this_try_timeout_time",
        "retry_timeout_time",
        "retry_wait_timeout_time",
        "old_timeout_value",
        "blocking",
        "host",
        "port",
        "failure_reason",
        )

    def __init__(self):
        self.retry_timeout_interval = 0
        self.connect_failed = False
        self.failed_once = False
        self.connect_refused = False
        self.first_try_start_time = 0
        self.this_try_timeout_time = 0
        self.retry_timeout_time = 0
        self.retry_wait_timeout_time = 0
        self.old_timeout_value = 0
        self.blocking = True
        self.host = None
        self.port = 0
        self.failure_reason = None


class Sock(Stream):

    __slots__ = (
        "_sock",
        "_state",
        "_connect_state",
        "_peer_host",
        "_peer_port",
        "_allow_empty_message",
        "_socket_factory",
        )

    def __init__(self, socket_factory=socket.socket):
        super(Sock, self).__init__()
        self._sock = None
        self._state = SockState.VIRGIN
        self._connect_state = SockConnectState()
        self._peer_host = None
        self._peer_port = 0
        self._allow_empty_message = False
        self._socket_factory = socket_factory

    @property
    def is_connect_pending(self):
        """Return whether a non-blocking connection is pending."""
        return self._state in (
            SockState.CONNECT_PENDING,
            SockState.CONNECT_PENDING_RETRY,
            )

    @property
    def is_connected(self):
        """Return whether the socket is connected."""
        return self._state == SockState.CONNECT

    def close(self):
        """Close the socket connection and reset state."""
        if self._state == SockState.VIRGIN:
            return False

        if self._sock is not None:
            try:
                self._sock.close()
            except socket.error:
                return False

        self._sock = None
        self._state = SockState.VIRGIN
        self._connect_state.host = None
        self._peer_host = None
        self._peer_port = 0

        return True

    def fileno(self):
        if self._sock is None:
            return None

        return self._sock.fileno()

    def get_family(self):
        """Return the communication family for this socket."""
        return socket.AF_INET

    def get_type(self):
        """Return the communication type for this socket."""
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def get_protocol_name(self):
        """Return the protocol name for this socket."""
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def get_host_and_port(self):
        """Return the host name and port number of the local connection."""
        if self._sock is None:
            return (None, 0)

        # TODO: cache host and port

        try:
            host, port = self._sock.getsockname()
        except socket.error, error:
            logging.info("Failed to get socket name, errno %d", error.errno)
            return (None, 0)

        # Returning 0.0.0.0 is not a good idea
        host_address = struct.unpack(">i", socket.inet_aton(host))[0]
        if host_address == socket.INADDR_ANY:
            host = hostname.ipaddress()

        return (host, port)

    def get_description(self):
        host, port = self.get_host_and_port()
        if host is None:
            return None

        return "<%s:%s>" % (host, port)

    def get_peer_host_and_port(self):
        """Return the host name and port number of the peer connection."""
        return (self._peer_host, self._peer_port)

    def get_peer_description(self):
        """Return a descriptive string of the peer connection."""
        host, port = self.get_peer_host_and_port()
        if host is None:
            return None

        return "<%s:%s>" % (host, port)

    def get_peer_description_default(self):
        description = self.get_peer_description()
        if description is None:
            description = "(unconnected)"

        return description

    def get_port_by_service(self, service):
        """Return the port number corresponding to the given service name.

        :param service: Service name.
        """
        if not service:
            return -1

        protocol = self.get_protocol_name()

        try:
            return socket.getservbyname(service, protocol)
        except socket.error:
            return -1

    def get_bytes_available(self):
        if self._state == SockState.VIRGIN:
            self.assign()

        if self._state not in (
            SockState.ASSIGNED, SockState.CONNECT, SockState.BOUND):
            return -1

        size_int = c_int()

        try:
            ioctl(self._sock.fileno(), FIONREAD, size_int)
        except IOError:
            return -1

        return size_int.value

    def setOptions(self, level, name, value):
        """Set low level socket options, see the manpage for more details."""
        if self._state == SockState.VIRGIN:
            self.assign()

        try:
            self._sock.setsockopt(level, name, value)
        except socket.error:
            return False

        return True

    def getDeadline(self):
        """See Stream.getDeadline"""
        deadline = super(Sock, self).getDeadline()
        if self.is_connect_pending:
            if self._state == SockState.CONNECT_PENDING_RETRY:
                connect_deadline = self._connect_state.retry_wait_timeout_time
            else:
                connect_deadline = self._connect_state.this_try_timeout_time

            if connect_deadline:
                if not deadline or deadline >= connect_deadline:
                    return connect_deadline

        return deadline

    def setTimeout(self, timeout):
        """See Stream.setTimeout"""
        old_timeout = super(Sock, self).setTimeout(timeout)

        if self._state == SockState.VIRGIN:
            self.assign()

        if self._state not in (
           SockState.ASSIGNED,
           SockState.CONNECT,
           SockState.BOUND,
           ):
            return -1

        if timeout == 0:
            # Put socket into blocking mode
            flags = fd_get_flags(self._sock)
            if flags < 0:
                return -1

            flags &= ~os.O_NONBLOCK
            if fd_set_flags(self._sock, flags) < 0:
                return -1

        return old_timeout

    def assign(self, sock=None):
        """Assign or create a socket for this socket.

        :param sock: Optional sock to assign.
        """
        if self._state != SockState.VIRGIN:
            return False

        if sock is not None:
            self._sock = sock
            self._state = SockState.ASSIGNED

            if self._timeout > 0:
                self.setTimeout(self._timeout)

            return True

        try:
            self._sock = self._socket_factory(
                self.get_family(), self.get_type())
        except socket.error, error:
            if error.errno == errno.EMFILE:
                raise Exception("Too many files open")

            return False

        self._state = SockState.ASSIGNED

        if self._timeout > 0:
            self.setTimeout(self._timeout)

        return True

    def bind(self, address=SockAddress.ANY, port=SockPort.ANY):
        """Bind a socket to a local port.

        :param address: IP address or SockAddress value.
        :param port: Local port number or service name to bind to.
        """
        if self._state == SockState.VIRGIN:
            self.assign()

        if self._state != SockState.ASSIGNED:
            logging.info("State in sock is not correct")
            return False

        if isinstance(port, str):
            port = self.get_port_by_service(port)
        elif port < SockPort.ANY:
            low_port, high_port = get_port_range(port == SockPort.OUTBOUND)
            port = 0

        if (port == 0 and
            address != SockAddress.LOOPBACK and
            (low_port or high_port)):
            # Bind within a specific port range
            if not self._bindWithin(address, low_port, high_port):
                return False

        else:
            # Bind to a dynamic port
            try:
                self._sock.bind((address, port))
            except socket.error, error:
                logging.info(
                    "Failed to bind to (%s, %d), errno %d %s",
                    address, port, error.errno, error.strerror)
                return False

        self._state = SockState.BOUND

        return True

    def connect(self, host, port=SockPort.ANY, blocking=True):
        """Connect the socket to a remote peer.

        :param host: Either a domain name or an IP address.
        :param port: Remote port number or service name to connect to.
        :param blocking: Whether to connect blocking or non-blocking.
        """
        if not host:
            return False

        if isinstance(port, str):
            port = self.get_port_by_service(port)
        elif port < SockPort.ANY:
            return False

        return self._doConnect(host, port, blocking)

    def serialize(self):
        """Serialize the socket into a string."""
        return "%u*%d*%d*%d*%s" % (
            self._sock.fileno(), self._state, self._timeout,
            self._peer_port, self._peer_host)

    def unserialize(self, string):
        """Unserialize the socket from a string.

        :param string: Serialized string.
        """
        parts = string.split("*")
        if len(parts) < 5:
            raise Exception(
                "Failed to parse serialized socket information: %s"
                % string)

        passed_fd = int(parts[0])
        self._state = int(parts[1])
        self._timeout = int(parts[2])
        self._peer_port = int(parts[3])
        self._peer_host = parts[4]

        if not self._sock:
            if passed_fd >= Selector.get_fd_select_size():
                try:
                    fd = os.dup(passed_fd)
                except OSError, error:
                    raise Exception(
                        "Failed to dup high descriptor %d, errno %d %s"
                        % (passed_fd, error.errno, error.errstr))
                else:
                    if fd >= Selector.get_fd_select_size():
                        raise Exception(
                            "Dup of high descriptor %d resulted "
                            "in another high descriptor %d"
                            % (passed_fd, fd))

                fd_close(passed_fd)
                passed_fd = fd

            self._sock = socket.fromfd(passed_fd,
                self.get_family(), self.get_type())

        self.setTimeout(self._timeout)

        return "*".join(parts[5:])

    def _bindWithin(self, address, low_port, high_port):
        # Hash the start port to begin iteration.
        now = datetime.now()
        range = high_port - low_port + 1
        start_port = current_port = low_port + (now.microsecond * 73 % range)

        while True:
            try:
                self._sock.bind((address, current_port))
            except socket.error, error:
                logging.info(
                    "Failed to bind to (%s, %d), errno %d %s",
                    address, current_port, error.errno, error.strerror)
            else:
                return True

            current_port += 1
            if current_port > high_port:
                current_port = low_port
            elif current_port == start_port:
                break

        logging.info(
            "Failed to bind to any port within (%d - %d)",
            low_port, high_port)

        return False

    def _doConnect(self, host, port, blocking):
        # Set peer information
        name, aliases, addresses = net.gethostbyname(host)
        self._peer_port = port
        self._peer_host = addresses[0]

        if self._state in (SockState.VIRGIN, SockState.ASSIGNED):
            self.bind(port=SockPort.OUTBOUND)

        if self._state != SockState.BOUND:
            return False

        self._connect_state.retry_timeout_interval = min(self._timeout,
            SOCK_CONNECT_TIMEOUT)
        self._connect_state.first_try_start_time = time()
        self._connect_state.retry_timeout_time = time() \
            + self._connect_state.retry_timeout_interval

        if self._timeout == 0:
            # Do not timeout on calls to connect()
            self._connect_state.this_try_timeout_time = 0
        else:
            self._connect_state.this_try_timeout_time = time() \
                + self._timeout

        self._connect_state.connect_failed = False
        self._connect_state.failed_once = False
        self._connect_state.connect_refused = False
        self._connect_state.blocking = blocking
        self._connect_state.host = host
        self._connect_state.port = port
        self._connect_state.old_timeout_value = self._timeout

        self._setFailureReason(None)

        return self._finishConnect()

    def _finishConnect(self):
        while True:
            if self._state == SockState.CONNECT_PENDING_RETRY:
                self._state = SockState.BOUND

            if self._state == SockState.BOUND:
                if self._tryConnect():
                    self._state = SockState.CONNECT
                    return True

                if not self._connect_state.connect_failed:
                    self._state = SockState.CONNECT_PENDING

                if (not self._connect_state.blocking and
                    self._state == SockState.CONNECT_PENDING):
                    # Expect to be called back later
                    return StreamError.WOULDBLOCK

            while self._state == SockState.CONNECT_PENDING:
                selector = Selector()
                time_left = self._connect_state.this_try_timeout_time - time()

                if not self._connect_state.blocking:
                    time_left = 0

                elif time_left < 0:
                    time_left = 0

                elif time_left > self._timeout:
                    time_left = self._timeout

                selector.reset()
                selector.setTimeout(time_left)
                selector.add_fd(self._sock.fileno(), SelectorIO.WRITE)
                selector.add_fd(self._sock.fileno(), SelectorIO.EXCEPT)
                selector.execute()

                if selector.has_timed_out:
                    if self._connect_state.blocking:
                        self._cancelConnect()
                    break

                if selector.has_signaled:
                    continue

                if selector.has_failed:
                    self._setFailureErrno(selector.get_errno(), "select")
                    self._connect_state.connect_failed = True
                    self._connect_state.connect_refused = True
                    self._cancelConnect()
                    break

                if not self._testConnect():
                    self._state = SockState.BOUND
                    self._connect_state.connect_failed = True
                    self._cancelConnect()
                    break

                elif selector.check_fd(
                     self._sock.fileno(), SelectorIO.EXCEPT):
                    self._state = SockState.BOUND
                    self._connect_state.connect_failed = True
                    self._setFailureReason("select() detected failure")
                    self._cancelConnect()
                    break

                else:
                    self._state = SockState.CONNECT
                    if self._connect_state.old_timeout_value != self._timeout:
                        self.setTimeout(
                            self._connect_state.old_timeout_value)

                    return True

            timed_out = (self._connect_state.retry_timeout_time and
                time() >= self._connect_state.retry_timeout_time)

            if timed_out or self._connect_state.connect_refused:
                if self._state != SockState.BOUND:
                    self._cancelConnect()

                self._reportFailure(timed_out)
                return False

            if self._connect_state.connect_failed:
                if not self._connect_state.failed_once:
                    self._connect_state.failed_once = True
                    self._reportFailure(timed_out)

            if (not self._connect_state.blocking and
                self._state == SockState.CONNECT_PENDING):
                return StreamError.WOULDBLOCK

            if not self._connect_state.blocking:
                if self._state == SockState.BOUND:
                    self._cancelConnect()

                self._state = SockState.CONNECT_PENDING_RETRY
                self._connect_state.retry_wait_timeout_time = time() + 1

                return StreamError.WOULDBLOCK

            os.sleep(1)

        raise Exception("Should not get here")

    def _tryConnect(self):
        self._connect_state.connect_failed = False
        self._connect_state.connect_refused = False

        if not self._connect_state.blocking:
            if self.setTimeout(1) < 0:
                # Failed to set socket to non-blocking
                self._connect_state.connect_refused = True
                self._setFailureReason("Failed to set timeout.")
                return False

        try:
            self._sock.connect((self._peer_host, self._peer_port))
        except socket.error, error:
            if error.errno == errno.EINPROGRESS:
                self._connect_state.connect_failed = True
                self._setFailureErrno(error.errno, "connect")
                self._cancelConnect()

            return False

        if not self._connect_state.blocking:
            # Pretend that we haven't connected yet so that there is
            # only one code path for all non-blocking connects.
            return False

        self._state = SockState.CONNECT
        return True

    def _testConnect(self):
        try:
            error = self._sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        except socket.error, error:
            logging.info("Failed to get socket option")
            self._connect_state.connect_failed = True
            self._setFailureErrno(error.errno, "getsockopt")
            return False

        if error:
            self._connect_state.connect_failed = True
            self._setFailureErrno(error, "connect")
            return False

        else:
            return True

    def _cancelConnect(self):
        # In some cases, we may be cancelling a non-blocking connect
        # attempt that has timed out. In others, we may be cleaning
        # up after a failed connect attempt.

        # Close the underlying socket
        fd = self._sock.fileno()
        self._sock.close()
        self._sock = None
        self._state = SockState.VIRGIN

        # Create a new socket
        if not self.assign():
            self._connect_state.connect_refused = True
            return

        # Make sure the descriptor number has not changed
        if self._sock.fileno() != fd:
            try:
                os.dup2(self._sock.fileno(), fd)
            except OSError, error:
                logging.info(
                    "dup2 failed after connect, errno %d", error.errno)
                self._connect_state.connect_refused = True
                return

            family = self._sock.family
            type = self._sock.type

            self._sock.close()
            self._sock = socket.fromfd(fd, family, type)

        # Finally, bind the socket
        if not self.bind(port=SockPort.OUTBOUND):
            self._connect_state.connect_refused = True

        # Restore old timeout
        if self._connect_state.old_timeout_value != self._timeout:
            self.setTimeout(self._connect_state.old_timeout_value)

    def _setFailureReason(self, reason):
        self._connect_state.failure_reason = reason

    def _setFailureErrno(self, _errno, syscall):
        if _errno == errno.ECONNREFUSED:
            self._connect_state.connect_refused = True

        reason = "%.80s (%.15s errno = %d)" % (
            os.strerror(_errno), syscall, _errno)
        self._setFailureReason(reason)

    def _reportFailure(self, timed_out):
        reason = self._connect_state.failure_reason
        if not reason and timed_out:
            reason = "timed out after %d seconds" % (
                self._connect_state.retry_timeout_interval)

        if reason is None:
            reason = ""

        if self._connect_state.connect_refused and not timed_out:
            will_keep_trying = (
                " Will keep trying for %ld seconds (%ld to go).\n" %
                (self._connect_state.retry_timeout_interval,
                self._connect_state.retry_timeout_time - time()))
        else:
            will_keep_trying = ""

        host = self._connect_state.host
        if not host:
            host = ""

        if host.startswith("<"):
            host = ""

        logging.info(
            "Attempt to connect to %s%s%s failed%s%s.%s",
            host,
            " " if host else "",
            self.get_peer_description(),
            ": " if reason else "",
            reason,
            will_keep_trying)
