#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "SockTCP",
    "SockTCPState",
    ]

import os
import errno
import logging
import socket
import struct

from time import time

from checkbox.lib.enum import Enum
from checkbox.lib.file import (
    fd_get_flags,
    fd_set_flags,
    )

from checkbox.io.bucket import (
    Bucket,
    BucketChain,
    )
from checkbox.io.buffer import Buffer
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.sock import (
    Sock,
    SockAddress,
    SockState,
    )


SOCK_TCP_HEADER_SIZE = 5
SOCK_TCP_PAGE_SIZE = 65536


SockTCPState = Enum(
    "NONE",
    "LISTEN",
    )


class SockTCP(Sock):

    __slots__ = (
        "_special_state",
        "_receiver_is_ready",
        "_receiver_chain",
        "_sender_bucket",
        "_ignore_next_put_end",
        "_ignore_next_get_end",
        )

    def __init__(self, *args, **kwargs):
        super(SockTCP, self).__init__(*args, **kwargs)
        self._receiver_is_ready = False
        self._receiver_chain = BucketChain()
        self._sender_bucket = Bucket()
        self.initialize()

    def initialize(self):
        """Initialize the TCP socket."""
        self._ignore_next_put_end = False
        self._ignore_next_get_end = False
        self._receiver_chain.reset()
        self._sender_bucket.reset()
        self._special_state = SockTCPState.NONE

    def get_type(self):
        """See Sock.get_type"""
        return socket.SOCK_STREAM

    def get_protocol_name(self):
        """See Sock.get_protocol_name"""
        return "tcp"

    def setTimeout(self, timeout):
        """See StreamBase.setTimeout"""
        old_timeout = super(SockTCP, self).setTimeout(timeout)

        if timeout != 0:
            # Put socket into non-blocking mode
            flags = fd_get_flags(self._sock)
            if flags < 0:
                return -1

            flags |= os.O_NONBLOCK
            if fd_set_flags(self._sock, flags) < 0:
                return -1

        return old_timeout

    def accept(self):
        """Accept a connection on this socket."""
        sock = SockTCP(socket_factory=self._socket_factory)
        if not self.acceptSock(sock):
            return None

        return sock

    def acceptSock(self, sock):
        """Accept a connection on a socket.

        :param sock: Sock from which to accept.
        """
        if (self._state != SockState.SPECIAL or
            self._special_state != SockTCPState.LISTEN or
            sock._state != SockState.VIRGIN):
            return False

        if self._timeout > 0:
            selector = Selector()
            selector.setTimeout(self._timeout)
            selector.add_fd(self._sock.fileno(), SelectorIO.READ)

            selector.execute()

            if selector.has_timed_out:
                return False

            elif not selector.has_ready:
                logging.info("Connect failed")
                return False

        try:
            sock._sock, peer_host_and_port = self._sock.accept()
        except socket.error, error:
            if error.errno == errno.EMFILE:
                raise Exception("Should not get here")
            return False

        sock._state = SockState.CONNECT
        sock._peer_host, sock._peer_port = peer_host_and_port
        sock.setOptions(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        # Set no delay to disable Nagle
        sock.setOptions(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        return True

    def bind(self, *args, **kwargs):
        """See Sock.bind"""
        ret = super(SockTCP, self).bind(*args, **kwargs)
        if ret:
            self.setOptions(
                socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 0, 0))
            self.setOptions(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.setOptions(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        return ret

    def connect(self, *args, **kwargs):
        """See Sock.connect"""
        self.initialize()
        return super(SockTCP, self).connect(*args, **kwargs)

    def connectSock(self, sock):
        """Connect this socket to another socket. This is equivalent to
           the socketpair system call using the loopback interface.

        :param sock: The other socket.
        """
        temp_sock = SockTCP(socket_factory=self._socket_factory)

        if not self.bind(SockAddress.LOOPBACK):
            logging.info("Failed to bind to loopback")
            return False

        if not temp_sock.bind(SockAddress.LOOPBACK):
            logging.info("Failed to bind temporary socket to loopback")
            return False

        if not temp_sock.listen():
            logging.info("Failed to listen on temporary socket")
            return False

        if not self.connect(*temp_sock.get_host_and_port()):
            logging.info("Failed to connect to temporary socket")
            return False

        if not temp_sock.acceptSock(sock):
            logging.info("Failed to accept from temporary socket")
            return False

        return temp_sock

    def listen(self, port=None):
        """Listen for connections on the socket.

        :param port: Optional local port on which to listen.
        """
        if port is not None:
            if not self.bind(port=port):
                return False

        if self._state != SockState.BOUND:
            logging.info("Failed to listen because not bound to a port")
            return False

        try:
            self._sock.listen(500)
        except socket.error, error:
            logging.info(
                "Failed to listen on TCP socket, errno %d %s",
                error.errno, error.strerror)
            return False

        self._state = SockState.SPECIAL
        self._special_state = SockTCPState.LISTEN

        return True

    def handleIncomingMessage(self):
        """Receive messages from the socket and buffer them.

        :return: Whether a message becomes ready.
        """
        if (self._state == SockState.SPECIAL and
            self._special_state == SockTCPState.LISTEN):
            return True

        if self._receiver_is_ready:
            return True

        if not self._receiveMessage():
            return False

        return True

    def serialize(self):
        """See Sock.serialize"""
        string = super(SockTCP, self).serialize()
        return "%s*%d" % (string, self._special_state)

    def unserialize(self, string):
        """See Sock.unserialize"""
        string = super(SockTCP, self).unserialize(string)

        parts = string.split("*")
        if len(parts) < 1:
            raise Exception(
                "Failed to parse serialized TCP socket information: %s"
                % string)

        self._special_state = int(parts[0])

        return "*".join(parts[1:])

    def getBytes(self, size):
        """See Stream.getBytes"""
        self._ignore_next_put_end = False

        while not self._receiver_is_ready:
            if not self.handleIncomingMessage():
                return None

        return self._receiver_chain.get(size)

    def getBytesUnbuffered(self, max_size, receive_size=True):
        """Get bytes from the socket but do not buffer anything.

        :param size: Number of bytes to get.
        :param receive_size: Whether to receive the size.
        :return: Bytes gotten.
        """
        assert(max_size > 0)

        if receive_size:
            size = self.getInteger()
            self.getEnd()

        else:
            size = max_size

        # Prepare for no buffering
        if not self._ignore_next_get_end:
            if self._receiver_is_ready:
                ret = self._receiver_chain.is_consumed
                self._receiver_is_ready = False
                self._receiver_chain.reset()

                if not ret:
                    return None

            self._ignore_next_get_end = True

        if size > max_size:
            logging.info("Data too large for buffer.")
            return None

        result = self.getBytesRaw(size)
        if result is None:
            logging.info("Receive failed.")
            return None

        return result

    def getBytesRaw(self, size, flags=0):
        """Get bytes directly from the socket.

        :param size: Number of bytes to get.
        :param flags: Optional flag for the socket.
        :return: Bytes gotten.
        """
        assert(size > 0)
        assert(self._sock)

        fd = self._sock.fileno()

        selector = Selector()
        selector.add_fd(fd, SelectorIO.READ)

        if self._timeout > 0:
            start_time = current_time = time()

        total = 0
        buffers = []
        while total < size:
            if self._timeout > 0:
                if current_time == 0:
                    current_time = time()

                if start_time + self._timeout > current_time:
                    selector.setTimeout(
                        (start_time + self._timeout) - current_time)

                else:
                    logging.info(
                        "Timeout reading %d bytes from %s",
                        size, self.get_peer_description())
                    return None

                current_time = 0
                selector.execute()
                if selector.has_timed_out:
                    logging.info(
                        "Timeout reading %d bytes from %s",
                        size, self.get_peer_description())
                    return None

                elif selector.has_signaled:
                    continue

                elif not selector.has_ready:
                    logging.info(
                        "Failed selecting %d bytes from %s, errno %d",
                        size, self.get_peer_description(),
                        selector.get_errno())

                    return None

            try:
                buffer = self._sock.recv(size - total, flags)
            except socket.error, error:
                if _errno_is_temporary(error.errno):
                    logging.debug(
                        "recv returned temporary error %d %s",
                        error.errno, error.strerror)
                    continue

                logging.info(
                    "recv returned %d %s when reading %d bytes from %s",
                    error.errno, error.strerror, size,
                    self.get_peer_description())
                return None

            if not buffer:
                logging.info(
                    "Socket closed when trying to read %d bytes from %s",
                    size, self.get_peer_description())
                return None

            buffers.append(buffer)
            total += len(buffer)

        assert(total == size)
        return "".join(buffers)

    def getEnd(self):
        """See Stream.getEnd"""
        ret = False

        if self._ignore_next_get_end == True:
            self._ignore_next_get_end = False
            return True

        if self._receiver_is_ready:
            if self._receiver_chain.is_consumed:
                ret = True

            elif not self._allow_empty_message:
                logging.debug("Failed to get end of message.")

            self._receiver_is_ready = False
            self._receiver_chain.reset()

        if self._allow_empty_message:
            self._allow_empty_message = False
            return True

        return ret

    def peek(self):
        """See Stream.peek"""
        while not self._receiver_is_ready:
            if not self.handleIncomingMessage():
                return None

        return self._receiver_chain.peek()

    def search(self, sub):
        """See Stream.search"""
        while not self._receiver_is_ready:
            if not self.handleIncomingMessage():
                return False

        return self._receiver_chain.search(sub)

    def putBytes(self, bytes):
        """See Stream.putBytes"""
        self._ignore_next_put_end = False

        size = len(bytes)
        position = 0
        while position < size:
            if self._sender_bucket.is_full:
                if not self._sendMessage(False):
                    return -1

            if self._sender_bucket.is_empty:
                self._sender_bucket.seek(SOCK_TCP_HEADER_SIZE)

            size = self._sender_bucket.putMax(bytes[position:])
            if size < 0:
                return -1

            position += size

        return size

    def putBytesUnbuffered(self, bytes, send_size=True):
        """Put bytes onto the socket but do not buffer anything.

        :param bytes: Bytes to put.
        :param send_size: Whether to send the size.
        :return: Number of bytes put.
        """
        if send_size:
            self.putInteger(len(bytes))
            self.putEnd()

        # Prepare for no buffering
        if not self._ignore_next_put_end:
            if (not self._sender_bucket.is_empty and
                not self._sendMessage(True)):
                return -1

            self._ignore_next_put_end = True

        size = len(bytes)
        position = 0
        while position < size:
            # If there's less than a page left
            if size - position < SOCK_TCP_PAGE_SIZE:
                result = self.putBytesRaw(bytes[position:])
                position += size - position

            # Send another page
            else:
                result = self.putBytesRaw(
                    bytes[position:position + SOCK_TCP_PAGE_SIZE])
                position += SOCK_TCP_PAGE_SIZE

            if result < 0:
                logging.info("Send failed.")
                return -1

        return position

    def putBytesRaw(self, bytes, flags=0):
        """Put bytes directly onto the socket.

        :param bytes: Bytes to put.
        :param flags: Optional flag for the socket.
        :return: Number of bytes put.
        """
        assert(self._sock)

        if not bytes:
            return 0

        fd = self._sock.fileno()

        selector = Selector()
        selector.add_fd(fd, SelectorIO.READ)
        selector.add_fd(fd, SelectorIO.WRITE)
        selector.add_fd(fd, SelectorIO.EXCEPT)

        if self._timeout > 0:
            start_time = current_time = time()

        total_write = 0
        select_for_read = True
        while total_write < len(bytes):
            if self._timeout > 0:
                while True:
                    if current_time == 0:
                        current_time = time()

                    if start_time + self._timeout > current_time:
                        selector.setTimeout(
                            (start_time + self._timeout) - current_time)

                    else:
                        logging.info(
                            "Timeout writing %d bytes to %s",
                            len(bytes), self.get_peer_description())
                        return -1

                    current_time = 0

                    if select_for_read:
                        selector.add_fd(fd, SelectorIO.READ)

                    else:
                        selector.remove_fd(fd, SelectorIO.READ)

                    selector.execute()

                    if selector.has_timed_out:
                        logging.info(
                            "Timeout writing %d bytes to %s",
                            len(bytes), self.get_peer_description())
                        return -1

                    elif selector.has_signaled:
                        continue

                    elif selector.has_ready:
                        if selector.check_fd(fd, SelectorIO.READ):
                            try:
                                data = self._sock.recv(1, socket.MSG_PEEK)
                            except socket.error, error:
                                if _errno_is_temporary(error.errno):
                                    continue

                                logging.info(
                                    "Socket closed when trying to "
                                    "write %d bytes to %d, errno %d %s",
                                    len(bytes), self.get_peer_description(),
                                    error.errno, error.strerror)
                                return -1

                            if not data:
                                logging.info(
                                    "Socket closed when trying to "
                                    "write %d bytes to %s",
                                    len(bytes), self.get_peer_description())
                                return -1

                            select_for_read = False
                            continue

                    else:
                        logging.info(
                            "Failed selecting %d bytes from %s, errno %d",
                            len(bytes), self.get_peer_description(),
                            selector.get_errno())
                        return -1

                    break

            try:
                write_size = self._sock.send(bytes[total_write:], flags)
            except socket.error, error:
                if _errno_is_temporary(error.errno):
                    logging.debug(
                        "send returned temporary error %d %s",
                        error.errno, error.strerror)
                    continue

                logging.info(
                    "send returned %d %s when writing %d bytes to %s",
                    error.errno, error.strerror, len(bytes),
                    self.get_peer_description())
                return -1

            if not write_size:
                logging.info(
                    "Socket closed when trying to write %d bytes to %s",
                    len(bytes), self.get_peer_description())
                return -2

            total_write += write_size

        assert(total_write == len(bytes))
        return total_write

    def putEnd(self):
        """See Stream.putEnd"""
        if self._ignore_next_put_end == True:
            self._ignore_next_put_end = False
            return True

        elif not self._sender_bucket.is_empty:
            return self._sendMessage(True)

        elif self._allow_empty_message:
            self._allow_empty_message = False
            return True

        else:
            return False

    def _receiveMessage(self):
        header = self.getBytesRaw(SOCK_TCP_HEADER_SIZE)
        if header is None:
            logging.info("Failed to read message header.")
            return False

        buffer = Buffer(header)
        end = buffer.getBoolean()
        content_size = buffer.getInteger()

        if content_size <= 0:
            logging.info(
                "Incoming message improperly sized: %d", content_size)
            return False

        bucket = Bucket()
        if content_size > bucket.max_size:
            logging.info("Incoming message too big.")
            return False

        bucket_size = bucket.read(self, content_size)
        if bucket_size != content_size:
            logging.info(
                "Failed reading %d of %d messages", bucket_size, content_size)
            return False

        self._receiver_chain.put(bucket)

        if end:
            self._receiver_is_ready = True

        return True

    def _sendMessage(self, end):
        content_size = self._sender_bucket.used - SOCK_TCP_HEADER_SIZE

        buffer = Buffer()
        buffer.putBoolean(end)
        buffer.putInteger(content_size)

        self._sender_bucket.replace(buffer.getBytes(), 0)
        total_size = self._sender_bucket.flush(self)
        if total_size != content_size + SOCK_TCP_HEADER_SIZE:
            return False

        return True


def _errno_is_temporary(_errno):
    # Returns true if the given error number indicates
    # a temporary condition that may immediately be retried.
    return _errno in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINTR)
