#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "SockUDP",
    ]

import os
import logging
import random
import socket
import time

from checkbox.lib import hostname
from checkbox.lib.enum import Enum

from checkbox.io.message import (
    MessageID,
    MessageIncoming,
    MessageOutgoing,
    MessagePacket,
    MESSAGE_MAX_PACKET_SIZE,
    )
from checkbox.io.selector import (
    Selector,
    SelectorIO,
    )
from checkbox.io.sock import (
    Sock,
    SockPort,
    SockState,
    )


SOCK_UDP_BUCKET_SIZE = 7
SOCK_UDP_PACKET_TIMEOUT = 10


SockUDPState = Enum(
    "NONE",
    "LISTEN",
    )


class SockUDP(Sock):

    __slots__ = (
        "_special_state",
        "_is_long",
        "_is_ready",
        "_message_outgoing",
        "_message_long",
        "_message_short",
        "_messages_incoming",
        "_local_host",
        "_local_port",
        )

    _message_id = MessageID()

    def __init__(self, *args, **kwargs):
        super(SockUDP, self).__init__(*args, **kwargs)
        self._message_outgoing = MessageOutgoing()
        self._message_short = MessagePacket()
        self._local_host = None
        self._local_port = 0

        self.initialize()

    def initialize(self):
        """Initialize the UDP socket."""
        self._special_state = SockUDPState.NONE
        self._is_ready = False
        self._message_long = None
        self._messages_incoming = [None] * SOCK_UDP_BUCKET_SIZE

        if SockUDP._message_id.number == 0:
            SockUDP._message_id.address = hostname.ipaddress()
            SockUDP._message_id.pid = os.getpid()
            SockUDP._message_id.time = time.time()
            SockUDP._message_id.number = int(random.getrandbits(16))

    def get_type(self):
        """See Sock.get_type"""
        return socket.SOCK_DGRAM

    def get_protocol_name(self):
        """See Sock.get_protocol_name"""
        return "udp"

    def get_host_and_port(self):
        """See Sock.get_host_and_port"""
        if self._state != SockState.CONNECT:
            logging.info("Get host called on socket that is not connected")
            return (None, 0)

        if self._local_host:
            return (self._local_host, self._local_port)

        sock = SockUDP(socket_factory=self._socket_factory)
        sock.bind(port=SockPort.OUTBOUND)

        if sock._state != SockState.BOUND:
            logging.info("Failed to bind, state = %d", sock._state)
            return (None, 0)

        try:
            sock._sock.connect((self._peer_host, self._peer_port))
        except socket.error, error:
            logging.info("Failed to connect, errno %d %s",
                error.errno, error.errstr)
            return (None, 0)

        self._local_host, self._local_port = super(SockUDP,
            sock).get_host_and_port()

        return (self._local_host, self._local_port)

    def bind(self, *args, **kwargs):
        """See Sock.bind"""
        self.setOptions(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        return super(SockUDP, self).bind(*args, **kwargs)

    def handleIncomingMessage(self):
        """Receive messages from the socket and buffer them.

        :return: Whether a message becomes ready.
        """
        if self._is_ready:
            if self._message_long:
                existing_message_type = "long"
                existing_is_consumed = self._message_long.is_consumed

            else:
                existing_message_type = "short"
                existing_is_consumed = self._message_short.is_consumed

            logging.info(
                "Receiving new UDP message but found a %s "
                "message still waiting to be closed (consumed=%d).",
                existing_message_type, existing_is_consumed)

            # If a short message is received and a long one is still
            # unclosed, the long message will continue to act as the
            # source for all read operations.
            self.getEnd()

        bytes = self.getBytesRaw(MESSAGE_MAX_PACKET_SIZE)
        if not bytes:
            return False

        self._message_short.reset()
        self._message_short.putBytes(bytes)
        header = self._message_short.getHeader()

        if header.length <= 0 or header.length > MESSAGE_MAX_PACKET_SIZE:
            logging.info("Incoming datagram improperly sized.")
            return False

        if header.is_full:
            self._is_ready = True
            self._message_short.rewind()
            logging.debug("Full message, %d bytes", header.length)
            return True

        logging.debug("Fragmented message, %d bytes", header.length)

        # Long message
        current_time = time.time()
        index = hash(header.id) % SOCK_UDP_BUCKET_SIZE
        temp_message = self._messages_incoming[index]
        while temp_message and temp_message._id != header.id:
            previous_message = temp_message
            temp_message = temp_message.next

            # Delete timed out message
            if (current_time - previous_message._last_time >
                SOCK_UDP_PACKET_TIMEOUT):
                logging.debug(
                    "Timeout message, current %u, message %u",
                    current_time, previous_message._last_time)

                delete_message = previous_message
                previous_message = delete_message.previous
                if previous_message:
                    previous_message.next = delete_message.next
                else:
                    self._messages_incoming[index] = temp_message

                if temp_message:
                    temp_message.previous = previous_message
        else:
            previous_message = None

        # Found
        if temp_message:
            if temp_message.put(header.is_last, header.sequence, header.data):
                self._message_long = temp_message
                self._is_ready = True
                return True

        # Not found
        else:
            if previous_message:
                previous_message.next = MessageIncoming(header.id,
                    header.is_last, header.sequence, header.data,
                    previous_message)

            else:
                self._messages_incoming[index] = MessageIncoming(header.id,
                    header.is_last, header.sequence, header.data)

        return False

    def serialize(self):
        """See Sock.serialize"""
        string = super(SockUDP, self).serialize()
        return "%s*%d" % (string, self._special_state)

    def unserialize(self, string):
        """See Sock.unserialize"""
        string = super(SockUDP, self).unserialize(string)

        parts = string.split("*")
        if len(parts) < 1:
            raise Exception(
                "Failed to parse serialized UDP socket information: %s"
                % string)

        self._special_state = int(parts[0])

        return "*".join(parts[1:])

    def _doConnect(self, host, port, blocking):
        """See Sock._doConnect"""
        # TODO: uncache host and port

        # Bind here so that the sock may be assigned to the stream
        if self._state in (SockState.VIRGIN, SockState.ASSIGNED):
            self.bind(port=SockPort.OUTBOUND)

        if self._state != SockState.BOUND:
            logging.info("Failed to bind, state = %d", self._state)

        self._peer_host = host
        self._peer_port = port
        self._state = SockState.CONNECT

        return True

    def getBytes(self, size):
        """See Stream.getBytes"""
        while not self._is_ready:
            if self._timeout > 0:
                selector = Selector()
                selector.setTimeout(self._timeout)
                selector.add_fd(self._sock.fileno(), SelectorIO.READ)

                selector.execute()

                if selector.has_timed_out:
                    return ""

                elif not selector.has_ready:
                    logging.debug("Failed to receive  bytes.")
                    return ""

            self.handleIncomingMessage()

        if self._message_long:
            return self._message_long.getBytes(size)

        else:
            return self._message_short.getBytes(size)

    def getBytesRaw(self, size, flags=0):
        """Get bytes directly from the socket.

        :param size: Number of bytes to get.
        :param flags: Optional flag for the socket.
        :return: Bytes gotten.
        """
        assert(self._sock)

        try:
            data, address = self._sock.recvfrom(size, flags)
        except socket.error, error:
            logging.info(
                "Receive from failed, errno %d %s",
                error.errno, error.strerror)
            return None

        return data

    def getEnd(self):
        """See Stream.getEnd"""
        ret = False

        if self._is_ready:
            # Long message is ready
            if self._message_long:
                message = self._message_long
                if message.is_consumed:
                    ret = True

                # Unlink the message
                if message.previous:
                    message.previous.next = message.next
                else:
                    index = hash(message._id) % SOCK_UDP_BUCKET_SIZE
                    self._messages_incoming[index] = message.next

                if message.next:
                    message.next.previous = message.previous

                self._message_long = None

            # Short message is ready
            else:
                if self._message_short.is_consumed:
                    ret = True
                self._message_short.reset()

            self._is_ready = False

        else:
            ret = True

        if self._allow_empty_message:
            self._allow_empty_message = False
            ret = True

        return ret

    def peek(self):
        """See Stream.peek"""
        while not self._is_ready:
            if self._timeout > 0:
                selector = Selector()
                selector.setTimeout(self._timeout)
                selector.add_fd(self._sock.fileno(), SelectorIO.READ)

                selector.execute()

                if selector.has_timed_out:
                    return ""

                elif not selector.has_ready:
                    logging.debug("Failed to receive  bytes.")
                    return ""

            self.handleIncomingMessage()

        if self._message_long:
            return self._message_long.peek()

        else:
            return self._message_short.peek()

    def search(self, sub):
        """See Stream.search"""
        while not self._is_ready:
            if self._timeout > 0:
                selector = Selector()
                selector.setTimeout(self._timeout)
                selector.add_fd(self._sock.fileno(), SelectorIO.READ)

                selector.execute()

                if selector.has_timed_out:
                    return ""

                elif not selector.has_ready:
                    logging.debug("Failed to receive  bytes.")
                    return ""

            self.handleIncomingMessage()

        if self._message_long:
            return self._message_long.search(sub)

        else:
            return self._message_short.search(sub)

    def putBytes(self, bytes):
        """See Stream.putBytes"""

        return self._message_outgoing.putBytes(bytes)

    def putBytesRaw(self, bytes, flags=0):
        """Put bytes directly onto the socket.

        :param bytes: Bytes to put.
        :param flags: Optional flag for the socket.
        :return: Number of bytes put.
        """
        assert(self._sock)

        try:
            size = self._sock.sendto(bytes, flags,
                (self._peer_host, self._peer_port))
        except socket.error, error:
            logging.info(
                "Failed to send packet, errno %d %s",
                error.errno, error.strerror)
            return -1

        if size != len(bytes):
            return -1

        return size

    def putEnd(self):
        """See Stream.putEnd"""
        size = self._message_outgoing.send(self, SockUDP._message_id)

        # It doesn't hurt to increment the message id number even if it fails
        SockUDP._message_id.number += 1

        return size >= 0
