#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Stream",
    ]

import socket
import struct

from time import time


STREAM_STRING_NULL = "\xFF"


class Stream:

    __slots__ = (
        "_deadline",
        "_timeout",
        )

    def __init__(self):
        self._deadline = 0
        self._timeout = 0

    @property
    def deadline_expired(self):
        """Return whether the deadline timeout has expired."""
        return self._deadline != 0 and time() > self._deadline

    def getDeadline(self):
        """Get the deadline on the stream."""
        return self._deadline

    def setDeadline(self, deadline):
        """
        Set the deadline for all stream operations. Returns the previous
        deadline.

        :param deadline: Number of seconds for the deadline.
        """
        old_deadline = self._deadline
        self._deadline = deadline

        return old_deadline

    def getTimeout(self):
        """Get the timeout on the stream."""
        return self._timeout

    def setTimeout(self, timeout):
        """Set the timeout on the stream. Returns the previous timeout.

        :param timeout: Number of seconds to wait until timing out.
        """
        old_timeout = self._timeout
        self._timeout = timeout

        return old_timeout

    def getAddress(self):
        """Get an IP address value from the stream."""
        size = 4
        bytes = self.getBytes(size)
        if bytes is None or len(bytes) != size:
            return None

        return socket.inet_ntoa(bytes)

    def getBoolean(self):
        """Get a boolean value from the stream."""
        size = 1
        bytes = self.getBytes(size)
        if bytes is None or len(bytes) != size:
            return None

        return bool(ord(bytes))

    def getInteger(self):
        """Get an integer value from the stream."""
        size = struct.calcsize("i")
        bytes = self.getBytes(size)
        if bytes is None or len(bytes) != size:
            return None

        return struct.unpack(">i", bytes)[0]

    def getReal(self):
        """Get a real value from the stream."""
        size = struct.calcsize("d")
        bytes = self.getBytes(size)
        if bytes is None or len(bytes) != size:
            return None

        return struct.unpack(">d", bytes)[0]

    def getString(self):
        """Get a string value from the stream."""
        c = self.peek()
        if c is None:
            return None

        if c == STREAM_STRING_NULL:
            c = self.getBytes(1)
            # Ignore empty return value

            return None

        else:
            return self.search("\0")

    def getBytes(self, size):
        """Get the maximum number of bytes from the stream.

        :param size: Number of bytes to get.
        :return: Bytes gotten.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def getEnd(self):
        """Discard bytes up until the next record delimiter."""
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def peek(self):
        """Peek the next character in the stream without consuming it."""
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def search(self, sub):
        """Search for bytes in the stream until a substring is encountered.

        :param sub: Substring to search for.
        :return: Bytes searched.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def putAddress(self, address):
        """Put an IP address value onto the stream."""
        bytes = socket.inet_aton(address)
        return self.putBytes(bytes) == len(bytes)

    def putBoolean(self, boolean):
        """Put a boolean value onto the stream."""
        bytes = chr(boolean)
        return self.putBytes(bytes) == len(bytes)

    def putInteger(self, integer):
        """Put an integer value onto the stream."""
        bytes = struct.pack(">i", integer)
        return self.putBytes(bytes) == len(bytes)

    def putReal(self, real):
        """Put a real value onto the stream."""
        bytes = struct.pack(">d", real)
        return self.putBytes(bytes) == len(bytes)

    def putString(self, string):
        """Put a string value onto the stream."""
        if not string:
            return self.putBytes(STREAM_STRING_NULL) == 1

        else:
            if self.putBytes(string) != len(string):
                return False

            return self.putBytes("\0") == 1

    def putBytes(self, bytes):
        """Put bytes onto the stream.

        :param bytes: Bytes to put.
        :return: Number of bytes put.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def putEnd(self):
        """Flush stream and send record delimiter."""
        raise NotImplementedError(
            "this function must be overridden by subclasses")
