#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from checkbox.tests.cases import TestCase

from checkbox.io.bucket import (
    Bucket,
    BucketChain,
    )
from checkbox.io.buffer import Buffer


class SockFake(Buffer):

    getBytesRaw = Buffer.getBytes
    putBytesRaw = Buffer.putBytes


class TestBucket(TestCase):

    def test_reset(self):
        bucket = Bucket()
        self.assertTrue(bucket.is_empty)

        bucket.putMax("foo")
        self.assertFalse(bucket.is_empty)

        bucket.reset()
        self.assertTrue(bucket.is_empty)

    def test_rewind(self):
        bucket = Bucket()
        self.assertEquals(bucket.touched, 0)

        self.assertEquals(bucket.seek(1), 0)
        self.assertEquals(bucket.touched, 1)

        bucket.rewind()
        self.assertEquals(bucket.touched, 0)

    def test_max_size(self):
        bucket = Bucket()
        self.assertNotEquals(bucket.max_size, 0)

        bucket = Bucket(10)
        self.assertEquals(bucket.max_size, 10)

    def test_touched_and_untouched(self):
        bucket = Bucket(2)
        self.assertEquals(bucket.touched, 0)
        self.assertEquals(bucket.untouched, 0)

        self.assertEquals(bucket.seek(1), 0)
        self.assertEquals(bucket.touched, 1)
        self.assertEquals(bucket.untouched, 0)

        bucket.rewind()
        self.assertEquals(bucket.touched, 0)
        self.assertEquals(bucket.untouched, 1)

    def test_used_and_unused(self):
        bucket = Bucket(2)
        self.assertEquals(bucket.used, 0)
        self.assertEquals(bucket.unused, 2)

        self.assertEquals(bucket.seek(1), 0)
        self.assertEquals(bucket.used, 1)
        self.assertEquals(bucket.unused, 1)

    def test_is_empty(self):
        bucket = Bucket()
        self.assertTrue(bucket.is_empty)

        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertFalse(bucket.is_empty)

    def test_is_full(self):
        bucket = Bucket(3)
        self.assertFalse(bucket.is_full)

        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertTrue(bucket.is_full)

    def test_is_consumed(self):
        bucket = Bucket(3)
        self.assertTrue(bucket.is_consumed)

        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertFalse(bucket.is_consumed)

        self.assertEquals(bucket.getMax(3), "foo")
        self.assertTrue(bucket.is_consumed)

    def test_read(self):
        sock = SockFake("foobar")
        bucket = Bucket(3)
        self.assertEquals(bucket.read(sock), 3)
        self.assertEquals(bucket.getMax(3), "foo")

    def test_write(self):
        sock = SockFake()
        bucket = Bucket(3)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.write(sock), 3)
        self.assertFalse(bucket.is_empty)
        self.assertEquals(sock.getBytesRaw(), "foo")

    def test_flush(self):
        sock = SockFake()
        bucket = Bucket(3)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.flush(sock), 3)
        self.assertTrue(bucket.is_empty)

    def test_putAndGetMax(self):
        bucket = Bucket(3)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.getMax(3), "foo")

        bucket = Bucket(3)
        self.assertEquals(bucket.putMax("foobar"), 3)
        self.assertEquals(bucket.getMax(3), "foo")

        bucket = Bucket(6)
        self.assertEquals(bucket.seek(3), 0)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.getMax(6), "foo")

        bucket.rewind()
        self.assertEquals(bucket.getMax(6), "\0\0\0foo")

        bucket = Bucket(6)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.putMax("bar"), 3)
        self.assertEquals(bucket.getMax(6), "foobar")

    def test_find(self):
        bucket = Bucket(6)
        self.assertEquals(bucket.putMax("foobar"), 6)
        self.assertEquals(bucket.seek(3), 0)
        self.assertEquals(bucket.find("o"), -1)
        self.assertEquals(bucket.find("a"), 1)

    def test_peek(self):
        bucket = Bucket(3)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.peek(), "f")
        self.assertEquals(bucket.getMax(3), "foo")

    def test_seek(self):
        bucket = Bucket(6)
        self.assertEquals(bucket.seek(3), 0)
        self.assertEquals(bucket.putMax("foo"), 3)

        bucket.rewind()
        self.assertEquals(bucket.getMax(6), "\0\0\0foo")

        bucket = Bucket(6)
        self.assertEquals(bucket.putMax("foo"), 3)
        self.assertEquals(bucket.seek(3), 0)
        self.assertEquals(bucket.getMax(3), "")

        self.assertEquals(bucket.seek(6), 3)

        bucket.rewind()
        self.assertEquals(bucket.getMax(6), "foo\0\0\0")

    def test_replace(self):
        bucket = Bucket(6)
        self.assertEquals(bucket.putMax("foobar"), 6)
        self.assertEquals(bucket.replace("bar", 0), 3)
        self.assertEquals(bucket.getMax(6), "barbar")

        bucket = Bucket(6)
        self.assertEquals(bucket.putMax("foobar"), 6)
        self.assertEquals(bucket.replace("foobar", 3), 3)
        self.assertEquals(bucket.getMax(6), "foofoo")


class TestBucketChain(TestCase):

    def setUp(self):
        self.bucket1 = Bucket()
        self.bucket1.putMax("foo")

        self.bucket2 = Bucket()
        self.bucket2.putMax("bar")

        self.chain = BucketChain()

    def test_is_consumed(self):
        self.assertTrue(self.chain.is_consumed)

        self.assertTrue(self.chain.put(self.bucket1))
        self.assertFalse(self.chain.is_consumed)

        self.assertEquals(self.bucket1.getMax(3), "foo")
        self.assertTrue(self.chain.is_consumed)

        self.assertTrue(self.chain.put(self.bucket2))
        self.assertFalse(self.chain.is_consumed)

    def test_putAndGet(self):
        self.assertTrue(self.chain.put(self.bucket1))
        self.assertTrue(self.chain.put(self.bucket2))
        self.assertEquals(self.chain.get(6), "foobar")

    def test_peek(self):
        self.assertEquals(self.chain.peek(), None)

        self.assertTrue(self.chain.put(self.bucket1))
        self.assertEquals(self.chain.peek(), "f")

        self.assertTrue(self.chain.put(self.bucket2))
        self.assertEquals(self.chain.peek(), "f")

        self.assertEquals(self.chain.get(3), "foo")
        self.assertEquals(self.chain.peek(), "b")
