#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from checkbox.tests.cases import TestCase

from checkbox.io.message import (
    MessageID,
    MessageIncoming,
    MessageOutgoing,
    MessagePacket,
    MESSAGE_HEADER_SIZE,
    )
from checkbox.io.buffer import Buffer


class SockFake(Buffer):

    getBytesRaw = Buffer.getBytes
    putBytesRaw = Buffer.putBytes


class TestMessageID(TestCase):

    def test_cmp(self):
        id1 = MessageID("127.0.0.1", 1, 2, 3)
        self.assertEquals(cmp(id1, None), -1)
        self.assertEquals(cmp(None, id1), 1)

        id2 = MessageID("127.0.0.1", 1, 2, 3)
        self.assertEquals(cmp(id1, id2), 0)

        id2 = MessageID("127.0.0.1", 1, 2, 4)
        self.assertEquals(cmp(id1, id2), -1)

        id2 = MessageID("127.0.0.1", 1, 3, 3)
        self.assertEquals(cmp(id1, id2), -1)

        id2 = MessageID("127.0.0.1", 2, 2, 3)
        self.assertEquals(cmp(id1, id2), -1)

        id2 = MessageID("127.0.0.2", 1, 2, 3)
        self.assertEquals(cmp(id1, id2), -1)

    def test_hash(self):
        id1 = MessageID("127.0.0.1", 1, 2, 3)
        self.assertNotEquals(hash(id1), 0)

        id2 = MessageID("127.0.0.1", 1, 2, 4)
        self.assertNotEquals(hash(id1), hash(id2))

        id2 = MessageID("127.0.0.1", 1, 3, 3)
        self.assertNotEquals(hash(id1), hash(id2))

        # PID does not need to change the hash value
        id2 = MessageID("127.0.0.2", 1, 2, 3)
        self.assertNotEquals(hash(id1), hash(id2))

    def test_copy(self):
        id1 = MessageID("127.0.0.1", 1, 2, 3)
        id2 = id1.copy()
        self.assertEquals(id1, id2)

    def test_copyFrom(self):
        id1 = MessageID("127.0.0.1", 1, 2, 3)
        id2 = MessageID("127.0.0.1", 1, 2, 4)
        self.assertNotEquals(id1, id2)

        id2.copyFrom(id1)
        self.assertEquals(id1, id2)


class TestMessagePacket(TestCase):

    def test_getHeader(self):
        packet = MessagePacket()
        self.assertEquals(packet.putBytes("foobar"), 6)

        header = packet.getHeader()
        self.assertNotEquals(header, None)
        self.assertEquals(header.is_full, True)
        self.assertEquals(header.length, 6)
        self.assertEquals(header.data, "foobar")
        self.assertEquals(header.sequence, 0)
        self.assertEquals(header.id, None)

        self.assertEquals(packet.getBytes(6), "foobar")

    def test_putHeader(self):
        id = MessageID("127.0.0.1", 1, 2, 3)

        packet = MessagePacket()
        self.assertEquals(packet.putBytes("foobar"), 6)
        self.assertEquals(
            packet.putHeader(True, 10, id), MESSAGE_HEADER_SIZE)

        header = packet.getHeader()
        self.assertNotEquals(header, None)
        self.assertEquals(header.is_full, False)
        self.assertEquals(header.length, 6)
        self.assertEquals(header.data, "foobar")
        self.assertEquals(header.sequence, 10)

        self.assertNotEquals(header.id, None)
        self.assertEquals(header.id.address, "127.0.0.1")
        self.assertEquals(header.id.pid, 1)
        self.assertEquals(header.id.time, 2)
        self.assertEquals(header.id.number, 3)

        self.assertEquals(packet.getBytes(6), "foobar")

    def test_putAndGetBytes(self):
        packet = MessagePacket(3)
        self.assertEquals(packet.putBytes("foo"), 3)
        self.assertEquals(packet.putBytes("foo"), 0)
        self.assertEquals(packet.getBytes(3), "foo")
        self.assertEquals(packet.getBytes(3), "")

        packet = MessagePacket(3)
        self.assertEquals(packet.putBytes("foobar"), 3)
        self.assertEquals(packet.getBytes(3), "foo")

        packet = MessagePacket(6)
        self.assertEquals(packet.putBytes("foo"), 3)
        self.assertEquals(packet.putBytes("bar"), 3)
        self.assertEquals(packet.getBytes(6), "foobar")

    def test_peek(self):
        packet = MessagePacket(3)
        self.assertEquals(packet.putBytes("foo"), 3)
        self.assertEquals(packet.peek(), "f")
        self.assertEquals(packet.getBytes(3), "foo")


class TestMessageIncoming(TestCase):

    def setUp(self):
        self.id = MessageID("127.0.0.1", 1, 2, 3)

    def test_is_consumed(self):
        incoming = MessageIncoming(self.id, True, 0, "")
        self.assertFalse(incoming.is_consumed)

        incoming = MessageIncoming(self.id, True, 0, "a")
        self.assertFalse(incoming.is_consumed)

    def test_put(self):
        incoming = MessageIncoming(self.id, False, 0, "a")
        self.assertFalse(incoming.put(False, 0, "a"))
        self.assertFalse(incoming.put(True, 2, "c"))
        self.assertTrue(incoming.put(False, 1, "b"))

        self.assertEquals(incoming.getBytes(3), "abc")
        self.assertEquals(incoming.getBytes(1), None)

    def test_getBytes(self):
        incoming = MessageIncoming(self.id, True, 0, "")
        self.assertTrue(incoming.put(True, 1, "a"))
        self.assertFalse(incoming.is_consumed)

        self.assertEquals(incoming.getBytes(1), "a")
        self.assertTrue(incoming.is_consumed)

        self.assertEquals(incoming.getBytes(1), None)
        self.assertTrue(incoming.is_consumed)

    def test_peek(self):
        incoming = MessageIncoming(self.id, False, 0, "a")
        self.assertTrue(incoming.put(True, 1, "b"))

        self.assertEquals(incoming.peek(), "a")
        self.assertEquals(incoming.getBytes(1), "a")

        self.assertEquals(incoming.peek(), "b")
        self.assertEquals(incoming.getBytes(1), "b")

        self.assertEquals(incoming.peek(), None)

    def test_search(self):
        # One buffer, one message
        incoming = MessageIncoming(self.id, True, 0, "a\0b\0")
        self.assertEquals(incoming.search("\0"), "a")
        self.assertEquals(incoming.search("\0"), "b")
        self.assertEquals(incoming.search("\0"), None)

        # One buffer, multiple messages
        incoming = MessageIncoming(self.id, False, 0, "a\0")
        self.assertTrue(incoming.put(True, 1, "b\0"))

        self.assertEquals(incoming.search("\0"), "a")
        self.assertEquals(incoming.search("\0"), "b")
        self.assertEquals(incoming.search("\0"), None)

        # Multiple buffers, multiple messages
        incoming = MessageIncoming(self.id, False, 0, "a" * 1000)
        self.assertFalse(incoming.put(False, 1, "b" * 1000))
        self.assertFalse(incoming.put(False, 2, "\0"))
        self.assertTrue(incoming.put(True, 3, "c\0"))
        self.assertEquals(incoming.search("\0"), "a" * 1000 + "b" * 1000)
        self.assertEquals(incoming.search("\0"), "c")


class TestMessageOutgoing(TestCase):

    def setUp(self):
        self.id = MessageID("127.0.0.1", 1, 2, 3)

    def test_putBytes(self):
        outgoing = MessageOutgoing()
        self.assertEquals(outgoing.putBytes("a" * 1), 1)
        self.assertEquals(outgoing.putBytes("a" * 10), 10)
        self.assertEquals(outgoing.putBytes("a" * 100), 100)
        self.assertEquals(outgoing.putBytes("a" * 1000), 1000)
        self.assertEquals(outgoing.putBytes("a" * 10000), 10000)

    def test_send(self):
        sock = SockFake()

        # No bytes
        outgoing = MessageOutgoing()
        self.assertEquals(outgoing.send(sock, self.id), 0)

        # One byte, one packet
        outgoing = MessageOutgoing()
        self.assertEquals(outgoing.putBytes("a"), 1)
        self.assertEquals(outgoing.send(sock, self.id), 1)
        self.assertEquals(sock.getBytes(), "a")

        # More bytes, one packet
        outgoing = MessageOutgoing()
        self.assertEquals(outgoing.putBytes("a" * 100), 100)
        self.assertEquals(outgoing.send(sock, self.id), 100)
        self.assertEquals(sock.getBytes(), "a" * 100)

        # Even more bytes, two packets
        outgoing = MessageOutgoing()
        self.assertEquals(outgoing.putBytes("a" * 1000), 1000)
        self.assertEquals(outgoing.send(sock, self.id), 1060)
