#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "WatchdogClient",
    "WatchdogServer",
    ]

import os
import logging

from checkbox.lib.fifo import fifo_create
from checkbox.lib.file import (
    File,
    fd_open,
    fd_close,
    )


WATCHDOG_SUFFIX = ".watchdog"


class WatchdogClient:

    __slots__ = (
        "_fileno",
        )

    def __init__(self):
        self._fileno = -1

    def __del__(self):
        self.reset()

    def fileno(self):
        """Get the watchdog client file descriptor."""
        return self._fileno

    def reset(self):
        if self._fileno != -1:
            fd_close(self._fileno)
            self._fileno = -1

        return True

    def setPath(self, path):
        """Initialize a new watchdog client to watch the server with
           the given path.

        :param path: Watchdog server path.
        """
        if not self.reset():
            return False

        path += WATCHDOG_SUFFIX
        self._fileno = fd_open(path, os.O_RDONLY | os.O_NONBLOCK)
        if self._fileno == -1:
            logging.info("Failed to open read-only %s", path)
            return False

        return True


class WatchdogServer:

    __slots__ = (
        "_path",
        "_reader",
        "_writer",
        )

    def __init__(self):
        self._path = None
        self._reader = File()
        self._writer = File()

    def __del__(self):
        self.reset()

    def reset(self):
        if self._path:
            if os.path.exists(self._path):
                os.unlink(self._path)

            self._path = None

        return True

    def setPath(self, path):
        if not self.reset():
            return False

        path += WATCHDOG_SUFFIX
        if not fifo_create(path, self._reader, self._writer):
            logging.info("Failed to set watched fifo path at %s", path)
            return False

        self._path = path

        return True

    def getPath(self):
        return self._path
