#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Event",
    "EventOutcome",
    "EventType",
    ]

import re
import time
import logging

from checkbox.lib.enum import Enum


HEADER_LENGTH = 35
HEADER_PATTERN = (
    r" \((?P<cluster>\-?\d{2,3})"
    r"\.(?P<proc>\-?\d{2,3})"
    r"\.(?P<subproc>\-?\d{2,3})\)"
    r" (?P<tm_year>\d{4})/(?P<tm_mon>\d{2})/(?P<tm_mday>\d{2})"
    r" (?P<tm_hour>\d{2}):(?P<tm_min>\d{2}):(?P<tm_sec>\d{2}) ")
HEADER_RE = re.compile(HEADER_PATTERN)


EventOutcome = Enum(
    "OK",             # Event is valid
    "NO_EVENT",       # No event occured (like EOF)
    "READ_ERROR",     # Error reading journal file
    "MISSED_EVENT",   # Missed event
    "UNKNOWN_ERROR",  # Unknown Error
    )

EventType = Enum(
    "SUBMIT",         # Job submitted
    "EXECUTE",        # Job now executing
    "SUSPEND",        # Job was suspended
    "RESUME",         # Job was resumed
    "TERMINATE",      # Job was terminated
    "GENERIC",        # Generic journal event
    )


class Event:

    __slots__ = (
        "info",
        )

    def __init__(self, type=None):
        self.info = None

        if type is not None:
            self.initialize(type)

    def __getattr__(self, name):
        return getattr(self.info, name)

    def __setattr__(self, name, value):
        if name in Event.__slots__:
            return super(Event, self).__setattr__(name, value)

        return setattr(self.info, name, value)

    def __nonzero__(self):
        return self.info is not None

    def initialize(self, type):
        if type not in EVENT_INFOS:
            logging.debug("Invalid event type: %d", type)
            self.info = None

        else:
            self.info = EVENT_INFOS[type]()

    def _initializeFromRecord(self, record):
        if not record:
            return

        event_type = record.evaluateInteger("event_type")
        if event_type is not None:
            self.initialize(event_type)

        self.info._initializeFromRecord(record)

    def reset(self):
        self.info = None


EVENT_INFOS = {}


def register_event_info(event_class):
    """Register a class to be an implementation for L{EventInfo}s."""
    if "type" not in event_class.__dict__:
        return

    if event_class.type is not None:
        existent_event_info = EVENT_INFOS.get(event_class.type)
        if existent_event_info is not None:
            raise RuntimeError(
                "%r is the same type as %r"
                % (event_class, existent_event_info))
        EVENT_INFOS[event_class.type] = event_class


class EventMeta(type):

    def __init__(cls, name, bases, dict):
        super(EventMeta, cls).__init__(name, bases, dict)
        register_event_info(cls)


class EventInfo:

    __slots__ = (
        "event_clock",
        "event_time",
        "cluster",
        "proc",
        "subproc",
        )

    __metaclass__ = EventMeta

    type = None

    def __init__(self):
        self.event_clock = time.time()
        self.event_time = time.localtime(self.event_clock)
        self.cluster = -1
        self.proc = -1
        self.subproc = -1

    def _initializeFromRecord(self, record):
        self.cluster = record.evaluateInteger("cluster", -1)
        self.proc = record.evaluateInteger("proc", -1)
        self.subproc = record.evaluateInteger("subproc", -1)

    def getEvent(self, file):
        """Get the next event from the journal.  Reads the journal header
        and body.

        :param file: Readable journal file.
        """
        if not file:
            return False

        return self._readHeader(file) and self._readEvent(file)

    def putEvent(self, file):
        """Put the currently stored event values to the journal file. Writes
           the journal header and body.

        :param file: Writable journal file.
        """
        if not file:
            return False

        return self._writeHeader(file) and self._writeEvent(file)

    def _readHeader(self, file):
        """Protected method to read the header of an event from the
           journal and fill the fields of the event object. It does not read
           the event type, the caller is expected to read this under to
           instantiate the appropriate event type.

           :param file: Readable journal file.
           """
        header = file.read(HEADER_LENGTH)
        if not header:
            return False

        match = HEADER_RE.match(header)
        if not match:
            return False

        self.cluster = int(match.group("cluster"))
        self.proc = int(match.group("proc"))
        self.subproc = int(match.group("subproc"))

        self.event_time = time.mktime((
            int(match.group("tm_year")),
            int(match.group("tm_mon")),
            int(match.group("tm_mday")),
            int(match.group("tm_hour")),
            int(match.group("tm_min")),
            int(match.group("tm_sec")),
            0,
            0,
            -1))

        return True

    def _readEvent(self, file):
        """Protected method to read the body of the next event.

        :param file: Readable journal file.
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def _writeHeader(self, file):
        """Protected method to write the header of an event to the journal.

        :param file: Writable journal file.
        """
        file.write(
            "%03d (%03d.%03d.%03d) %04d/%02d/%02d %02d:%02d:%02d " % (
            self.type,
            self.cluster,
            self.proc,
            self.subproc,
            self.event_time.tm_year,
            self.event_time.tm_mon,
            self.event_time.tm_mday,
            self.event_time.tm_hour,
            self.event_time.tm_min,
            self.event_time.tm_sec))

        return True

    def _writeEvent(self, file):
        """Protected method to write the body of the next event.

        :param file: Writable journal file
        """
        raise NotImplementedError(
            "this function must be overridden by subclasses")

    def _insert_common_identifiers(self, record):
        if not record:
            return

        record.insert_integer("cluster_id", self.cluster)
        record.insert_integer("proc_id", self.proc)
        record.insert_integer("subproc_id", self.subproc)


class GenericEvent(EventInfo):

    __slots__ = (
        "data",
        )

    type = EventType.GENERIC

    def __init__(self):
        super(GenericEvent, self).__init__()
        self.data = ""

    def _initializeFromRecord(self, record):
        super(GenericEvent, self)._initializeFromRecord(record)

        if not record:
            return

        self.data = record.evaluateString("generic_data", "")

    def _writeEvent(self, file):
        file.write("%s\n" % self.data)
        return True

    def _readEvent(self, file):
        self.data = file.readline().rstrip("\n")
        return True


class ExecuteEvent(EventInfo):

    __slots__ = (
        "host",
        )

    type = EventType.EXECUTE

    def __init__(self):
        super(ExecuteEvent, self).__init__()
        self.host = ""

    def _initializeFromRecord(self, record):
        super(ExecuteEvent, self)._initializeFromRecord(record)

        if not record:
            return

        self.host = record.evaluateString("execute_host", "")

    def _writeEvent(self, file):
        file.write("Job executing on host: %s\n" % self.host)
        return True

    def _readEvent(self, file):
        line = file.readline()
        match = re.match(r"Job executing on host: (?P<host>[^\s]*)\n", line)
        if not match:
            return False

        self.host = match.group("host")
        return True


class SuspendEvent(EventInfo):

    __slots__ = (
        "pid_count",
        )

    type = EventType.SUSPEND

    def __init__(self):
        super(SuspendEvent, self).__init__()
        self.pid_count = 0

    def _initializeFromRecord(self, record):
        super(SuspendEvent, self)._initializeFromRecord(record)

        if not record:
            return

        self.pid_count = record.evaluateString("pid_count", 0)

    def _writeEvent(self, file):
        file.write("Job was suspended\n")
        file.write("Number of processes suspended: %d\n" % self.pid_count)
        return True

    def _readEvent(self, file):
        line = file.readline()
        if line != "Job was suspended\n":
            return False

        line = file.readline()
        match = re.match(
            r"Number of processes suspended: (?P<pid_count>\d+)\n", line)
        if not match:
            return False

        self.pid_count = int(match.group("pid_count"))
        return True


class ResumeEvent(EventInfo):

    type = EventType.RESUME

    def _writeEvent(self, file):
        file.write("Job was resumed\n")
        return True

    def _readEvent(self, file):
        line = file.readline()
        if line != "Job was resumed\n":
            return False

        return True


class TerminateEvent(EventInfo):

    __slots__ = (
        "sent_bytes",
        "received_bytes",
        "total_sent_bytes",
        "total_received_bytes",
        "return_value",
        "signal_number",
        "normal",
        "core_file",
        )

    type = EventType.TERMINATE

    attribute_strings = (
        ("sent_bytes", "Run Bytes Sent"),
        ("received_bytes", "Run Bytes Received"),
        ("total_sent_bytes", "Total Bytes Sent"),
        ("total_received_bytes", "Total Bytes Received"))

    def __init__(self):
        super(TerminateEvent, self).__init__()
        self.sent_bytes = self.received_bytes = 0.0
        self.total_sent_bytes = self.total_received_bytes = 0.0
        self.return_value = self.signal_number = -1
        self.normal = True
        self.core_file = None

    def _writeEvent(self, file):
        file.write("Job terminated\n")

        if self.normal:
            file.write(
                "\t(1) Normal termination (return value %d)\n"
                % self.return_value)
        else:
            file.write(
                "\t(0) Abnormal termination (signal %d)\n"
                % self.signal_number)

            if self.core_file:
                file.write("\t(1) Core file in: %s\n" % self.core_file)
            else:
                file.write("\t(0) No core file\n")

        # Write bytes
        for attribute, string in self.attribute_strings:
            file.write("\t%.0f  -  %s\n" % (getattr(self, attribute), string))

        return True

    def _readEvent(self, file):
        line = file.readline()
        if line != "Job terminated\n":
            return False

        line = file.readline()
        match = re.match(r"\t\((?P<normal_status>\d)\) ", line)
        if not match:
            return False

        # Strip normal status
        line = line[5:]
        if int(match.group("normal_status")):
            self.normal = True
            match = re.match(
                r"Normal termination"
                " \(return value (?P<return_value>\-?\d+)\)\n",
                line)
            if not match:
                return False

            self.return_value = int(match.group("return_value"))
        else:
            self.normal = False
            match = re.match(
                r"Abnormal termination"
                " \(signal (?P<signal_number>\-?\d+)\)\n",
                line)
            if not match:
                return False

            self.signal_number = int(match.group("signal_number"))

            line = file.readline()
            match = re.match(r"\t\((?P<has_core>\d)\) ", line)
            if not match:
                return False

            # Strip core status
            line = line[5:]
            if int(match.group("has_core")):
                match = re.match(r"Core file in: (?P<core_file>.*)\n", line)
                if not match:
                    return False

                self.core_file = match.group("core_file")
            else:
                if line != "No core file\n":
                    return False

        # Read bytes
        for attribute, string in self.attribute_strings:
            line = file.readline()
            match = re.match(r"\t(?P<bytes>\d+)  -  %s\n" % string, line)
            if not match:
                return False

            bytes = float(match.group("bytes"))
            setattr(self, attribute, bytes)

        return True
