#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Header",
    "HeaderReader",
    "HeaderWriter",
    ]

import re

from checkbox.journal.event import (
    Event,
    EventOutcome,
    EventType,
    )


JOURNAL_PATTERN = (
    r"Global Journal:"
     " ctime=(?P<ctime>\d+)"
     " id=(?P<id>[^ ]+)"
     " sequence=(?P<sequence>\d+)"
     " size=(?P<size>\d+)L?"
     " events=(?P<events>\d+)L?"
     " offset=(?P<offset>\d+)L?"
     " event_offset=(?P<event_offset>\d+)L?")
JOURNAL_RE = re.compile(JOURNAL_PATTERN)


class Header:
    """Simple class to extract info from a journal header event."""

    __slots__ = (
        "id",
        "sequence",
        "ctime",
        "size",
        "num_events",
        "file_offset",
        "event_offset",
        "_valid",
        )

    def __init__(self, other=None):
        if other is None:
            self.id = None
            self.sequence = 0
            self.ctime = 0
            self.size = 0
            self.num_events = 0
            self.file_offset = 0
            self.event_offset = 0
            self._valid = False
        else:
            self.copyFrom(other)

    def copyFrom(self, other):
        """Copy the parameters from another header object into this one.

        :param other: Other header object.
        """
        self.id = other.id
        self.sequence = other.sequence
        self.ctime = other.ctime
        self.size = other.size
        self.num_events = other.num_events
        self.file_offset = other.file_offset
        self.event_offset = other.event_offset
        self._valid = other._valid

    @property
    def is_valid(self):
        """Return true if the header is valid."""
        return self._valid

    def incrementSequence(self):
        """Increment the sequence number."""
        self.sequence += 1

    def addFileOffset(self, offset):
        """Add to the file offset.

        :param offset: Offset to add.
        """
        self.file_offset += offset

    def addEventOffset(self, offset):
        """Add to the event offset.

        :param offset: Offset to add.
        """
        self.event_offset += offset

    def extractEvent(self, event):
        """Extract data from an event and update the header.

        :param event: Event from which to extract.
        """
        if event.type != EventType.GENERIC:
            return EventOutcome.NO_EVENT

        match = JOURNAL_RE.match(event.data)
        if not match:
            return EventOutcome.NO_EVENT

        self.ctime = int(match.group("ctime"))
        self.id = match.group("id")
        self.sequence = int(match.group("sequence"))
        self.size = int(match.group("size"))
        self.num_events = int(match.group("events"))
        self.file_offset = int(match.group("offset"))
        self.event_offset = int(match.group("event_offset"))
        self._valid = True

        return EventOutcome.OK


class HeaderReader(Header):

    def read(self, reader):
        """Read the header from a file.

        :param reader: Journal reader object.
        """
        event = Event()
        outcome = reader.readEvent(event)

        if outcome != EventOutcome.OK:
            return outcome

        if event.type != EventType.GENERIC:
            return EventOutcome.NO_EVENT

        return self.extractEvent(event)


class HeaderWriter(Header):

    def write(self, writer, file=None):
        """Write the header to a file.

        :param reader: Journal writer object.
        """
        event = Event(EventType.GENERIC)
        if not self.generateEvent(event):
            return EventOutcome.UNKNOWN_ERROR

        return writer.writeGlobalEvent(event, file, True)

    def generateEvent(self, event):
        """Generate a header event.

        :param event: Generic event to contain the header data.
        """
        event.data = (
            "Global Journal:"
            " ctime=%d"
            " id=%s"
            " sequence=%d"
            " size=%d"
            " events=%d"
            " offset=%d"
            " event_offset=%d" % (
            self.ctime,
            self.id,
            self.sequence,
            self.size,
            self.num_events,
            self.file_offset,
            self.event_offset))

        return True
