#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "MatchResult",
    "Match",
    ]

import logging

from checkbox.lib.enum import Enum

from checkbox.journal.event import EventOutcome
from checkbox.journal.header import HeaderReader
from checkbox.journal.score import (
    Score,
    ScoreFactor,
    )


MatchResult = Enum(
    ("ERROR", -1),
    "MATCH",
    "UNKNOWN",
    "NOMATCH",
    )


class Match:

    __slots__ = (
        "_state",
        )

    def __init__(self, state):
        self._state = state

    def matchRotation(self, rotation, match_thresh, score=None):
        """Compare the specified rotation number with the cached state.

        :param rotation: Rotation to compare.
        :param match_thresh: Match threshold.
        :param score: If defined, the score is assigned to this object.
        """
        if score is None:
            score = Score()

        score.value = self._state.getScore(rotation)

        return self._match(rotation, None, match_thresh, score)

    def matchPath(self, path, rotation, match_thresh, score=None):
        """Compare the specified path with the cached state.

        :param path: Path to compare.
        :param rotation: Rotation for retrieving path from state.
        :param match_thresh: Match threshold.
        :param score: If defined, the score is assigned to this object.
        """
        if score is None:
            score = Score()

        score.value = self._state.getPathScore(path, rotation)

        return self._match(rotation, path, match_thresh, score)

    def matchStatResult(self, stat_result, rotation, match_thresh, score=None):
        """Compare the specified stat result with the cached state.

        :param stat_result: Stat result to compare.
        :param rotation: Rotation for retrieving path from state.
        :param match_thresh: Match threshold.
        :param score: If defined, the score is assigned to this object.
        """
        if score is None:
            score = Score()

        score.value = self._state.getStatResultScore(stat_result, rotation)

        return self._match(rotation, None, match_thresh, score)

    def _match(self, rotation, path, match_thresh, score):
        from checkbox.journal.reader import Reader

        value = score.value

        if not path:
            path = self._state.generatePath(rotation)

        logging.debug("Match score of '%s' = %d", path, value)

        # Quick look at the score passed in from the state comparison
        result = self._evaluateScore(match_thresh, score)
        if result != MatchResult.UNKNOWN:
            return result

        # Read the file header info
        reader = Reader()
        logging.debug("Reading file '%s'", path)

        if not reader.initializeFromPath(path):
            return MatchResult.ERROR

        header_reader = HeaderReader()
        outcome = header_reader.read(reader)
        if outcome == EventOutcome.OK:
            # Do nothing
            pass

        elif outcome == EventOutcome.NO_EVENT:
            return self._evaluateScore(match_thresh, value)

        else:
            return MatchResult.ERROR

        # Extract and store the ID
        result_id = self._state.compareUniqueId(header_reader.id)
        if result_id > 0:
            value += ScoreFactor.UNIQ_MATCH
            result_str = "match"

        elif result_id < 0:
            value = 0
            result_str = "no match"

        else:
            result_str = "unknown"

        logging.debug(
            "Read ID from '%s' as '%s': %d (%s)",
            path, header_reader.id, result, result_str)

        logging.debug("Match final score is %d", value)
        return self._evaluateScore(match_thresh, value)

    def _evaluateScore(self, match_thresh, value):
        # < 0 is an error
        if value < 0:
            return MatchResult.ERROR

        # Less than the min threshold, no match.
        elif value < ScoreFactor.MIN_MATCH:
            return MatchResult.NOMATCH

        # Above the min match threshold, match.
        elif value >= match_thresh:
            return MatchResult.MATCH

        else:
            return MatchResult.UNKNOWN
