#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Multi",
    "MultiException",
    "MultiReader",
    ]

import os
import re
import logging

from checkbox.lib.directory import Directory
from checkbox.lib.file import (
    File,
    fd_open,
    fd_close,
    )

from checkbox.message.lexer import Lexer
from checkbox.message.parser import Parser
from checkbox.message.token import (
    Token,
    TokenType,
    )
from checkbox.message.unparser import Unparser


CONTINUATION_CHARACTER = "\\"


class MultiException(Exception):

    pass


class Multi:

    @classmethod
    def truncateFilenames(cls, filenames):
        """Static method to truncate the given journal files to zero length."""
        for filename in filenames:
            if os.acess(filename, os.F_OK):
                logging.info("Truncating older version of %s", filename)
                fd = fd_open(filename, os.O_WRONLY | os.O_TRUNC)
                if fd == -1:
                    logging.info("Can't truncate %s", filename)
                else:
                    if fd_close(fd) == -1:
                        logging.info("Can't close %s", filename)

    @classmethod
    def filenameToLogicalLines(cls, filename):
        """Static method to read the specified file, break it into lines,
        and combine them into logical lines.

        :param filename: Filename to read.
        """
        file = File.create_from_path(filename, "r")
        if not file:
            logging.debug(
                "Failed to open '%s' with errno %d", filename, file.errno)
            raise MultiException(
                "Unable to read file: %s" % filename)

        content = file.read()
        file.close()

        logical_lines = []
        if content:
            index = 0
            physical_lines = re.split(r"\r?\n", content)
            while index < len(physical_lines):
                logical_line = physical_lines[index]
                while logical_line[-1] == CONTINUATION_CHARACTER:
                    logical_line = logical_line[:-1]

                    index += 1
                    if index < len(physical_lines):
                        logical_line += physical_lines[index]
                    else:
                        result = (
                            "Invalid file syntax, continuation character"
                            + "with no trailing line in file: %s"
                            % filename)
                        logging.info(result)
                        raise MultiException(result)

                logical_lines.append(logical_line)

        return logical_lines

    @classmethod
    def loadFilenameFromSubmitFilename(cls, submit_filename, directory):
        """Static method to get the journal file from a submit file.

        :param submit_filename: The submit file name.
        :param directory: The directory containing the submit file.
        """
        temp = Directory()
        if directory:
            if not temp.push(directory):
                logging.info(
                    "Error pushing directory '%s': %d", directory, temp.errno)
                return ""

        cwd = temp.cwd()
        if not cwd:
            logging.info(
                "Error getting working directory: %d", temp.errno)
            return ""

        try:
            logical_lines = cls.filenameToLogicalLines(submit_filename)
        except MultiException, error:
            logging.info(error.message)
            return ""

        # Look through the submit file to find the journal and initialdir
        journal = initialdir = ""
        for logical_line in logical_lines:
            temp_journal = cls._getParamFromSubmitLine(
                logical_line, "journal")
            if temp_journal:
                journal = temp_journal

            temp_initialdir = cls._getParamFromSubmitLine(
                logical_line, "initialdir")
            if temp_initialdir:
                initialdir = temp_initialdir

        # Check for macros in the journal file -- not supported yet
        if journal and "$" in journal:
            logging.info("Macros not allowed in journal: %s", journal)
            journal = ""

        if journal:
            if initialdir and not os.path.isabs(journal):
                journal = os.path.join(initialdir, journal)

            if not os.path.isabs(journal):
                if cwd:
                    journal = os.path.join(cwd, journal)
                else:
                    logging.info("Unable to get current directory.")
                    journal = ""

        if directory:
            if not temp.pop():
                logging.info(
                    "Error popping directory '%s': %d", directory, temp.errno)
                return ""

        return journal

    @classmethod
    def loadFilenamesFromStoreSubmitFilename(cls, submit_filename, directory):
        """ Gets the journal files from a stored submit file.

        :param submit_filename: The submit file name.
        :param directory: The directory containing the submit file.
        """
        logging.debug(
            "Loading filenames from %s in %s", submit_filename, directory)

        if directory:
            submit_filename = os.path.join(directory, submit_filename)

        file = File.create_from_path(submit_filename, "r")
        if not file:
            logging.debug(
                "Failed to open '%s' with errno %d",
                submit_filename, file.errno)
            raise MultiException("Unable to read file: %s" % submit_filename)

        parser = Parser()
        journal_files = set()

        while True:
            record = parser.parseRecord(file)
            if not record:
                break

            # Check for no journal
            journal = record.evaluateName("journal")
            if not journal:
                continue

            # Check for empty journal file
            journal_file = journal.data
            if not journal_file:
                raise MultiException(
                    "Data specifies null journal file: %s"
                    % Unparser.unparseRecord(record))

            # Check for macros in the journal file
            if "$" in journal_file:
                raise MultiException(
                    "Macros not supported in journal file: %s"
                    % Unparser.unparseRecord(record))

            # Journal file must be an absolute path
            if not os.path.isabs(journal_file):
                directory = Directory()
                cwd = directory.cwd()
                if not cwd:
                    raise MultiException(
                        "Failed to get working directory: %d"
                        % directory.errno)

                journal_file = os.path.join(cwd, journal_file)

            # Add the file to the journal file list
            journal_files.add(journal_file)

        return list(journal_files)

    @classmethod
    def getQueueCountFromSubmitFile(cls, submit_filename, directory):
        """Gets the number of messages queued by a submit file.

        :param submit_filename: The submit file name.
        :param directory: The directory containing the submit file.
        """
        logging.debug(
            "Getting queue count from %s in %s", submit_filename, directory)

        if directory:
            submit_filename = os.path.join(directory, submit_filename)

        # Let exception percolate
        logical_lines = cls.filenameToLogicalLines(submit_filename)

        queue_count = 0
        for logical_line in logical_lines:
            lexer = Lexer(logical_line)

            token = Token()
            token_type = lexer.consumeToken(token)
            if (token_type == TokenType.IDENTIFIER
                and token.value.lower() == "queue"):
                token_type = lexer.consumeToken(token)
                if token_type == TokenType.INTEGER:
                    queue_count += token.value
                else:
                    queue_count += 1

        return queue_count

    @classmethod
    def getValuesFromFile(cls, filename, keyword, skip_tokens=0):
        # Let exception percolate
        logical_lines = cls.filenameToLogicalLines(filename)

        values = set()
        for logical_line in logical_lines:
            if logical_line:
                lexer = Lexer(logical_line)

                token = Token()
                token_type = lexer.consumeToken(token)
                if (token_type == TokenType.IDENTIFIER
                    and token.value == keyword):
                    # Skip over unwanted tokens
                    for skipped in xrange(skip_tokens):
                        token_type = lexer.consumeToken(token)
                        if token_type == 1:
                            raise MultiException(
                                "Improperly formatted graph file: "
                                "value missing after keyword '%s'"
                                % keyword)

                    # Add the value to the list
                    values.add(token.value)

        return list(values)

    @classmethod
    def _getParamFromSubmitLine(cls, line, param):
        lexer = Lexer(line)
        token = Token()

        value = ""
        token_type = lexer.consumeToken(token)
        if (token_type == TokenType.IDENTIFIER
            and token.value.lower() == param):
            token_type = lexer.consumeToken(token)
            if token_type == TokenType.BOUND_TO:
                token_type = lexer.consumeToken(token)
                if token_type == TokenType.STRING:
                    value = token.value

        return value


class MultiReader:

    # TODO: finish this
    pass
