#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Reader",
    "ReaderStatus",
    ]

import os
import logging

from time import sleep

from checkbox.lib import param
from checkbox.lib.enum import Enum
from checkbox.lib.file import (
    File,
    fd_open,
    fd_close,
    )
from checkbox.lib.lock import (
    Lock,
    LockType,
    )

from checkbox.journal.event import (
    EventOutcome,
    EventType,
    )
from checkbox.journal.file import FileStatus
from checkbox.journal.header import HeaderReader
from checkbox.journal.match import (
    Match,
    MatchResult,
    )
from checkbox.journal.score import (
    Score,
    ScoreFactor,
    ScoreThresh,
    )
from checkbox.journal.state import (
    State,
    StateScore,
    )


# Threshold to consider file stat's as recent (seconds)
SCORE_RECENT_THRESH = 60

SYNC_DELIMITER = "...\n"

ReaderStatus = Enum(
    "NONE",             # No error
    "NOT_INITIALIZED",  # Reader not initialized
    "RE_INITIALIZE",    # Attempt to re-initialize
    "FILE_NOT_FOUND",   # Journal file not found
    "FILE_OTHER",       # Other file error
    "STATE_ERROR",      # Invalid state
    )


class Reader:

    __slots__ = (
        "_initialized",
        "_missed_event",
        "_state",
        "_match",
        "_fd",
        "_file",
        "_lock",
        "_lock_rot",
        "_close_file",
        "_enable_close",
        "_handle_rot",
        "_lock_enable",
        "_max_rotations",
        "_read_header",
        "_read_only",
        "_status",
        )

    def __init__(self, is_event_journal=False):
        """Default constructor.

        :param is_event_journal: True to read the global event journal,
            defaults to false.
        """
        self._initialized = False
        self._missed_event = False
        self._state = None
        self._match = None
        self._fd = -1
        self._file = None
        self._lock = Lock()
        self._lock_rot = -1

        self._close_file = False
        self._enable_close = True
        self._handle_rot = False
        self._lock_enable = False
        self._max_rotations = 0
        self._read_header = False

        self._status = ReaderStatus.NONE

        if is_event_journal:
            self.initialize()

    @classmethod
    def createFromFileState(cls, file_state, read_only=False):
        """Create a journal reader from a file state.

        :param file_tate: File state to restore from
        :param read_only: True to access the file for reading, defaults
            to false.
        """
        reader = cls()

        if not reader.initializeFromFileState(file_state, read_only):
            logging.debug("Failed to initialize from state")

        return reader

    @classmethod
    def createFromPath(cls, path, read_only=False):
        """Create a journal reader from a path.

        :param path: Path to read from
        :param read_only: True to access the file for reading, defaults
            to false.
        """
        reader = cls()

        if not reader.initializeFromPath(path, read_only=read_only):
            logging.debug("Failed to initialize from path")

        return reader

    @classmethod
    def createFromFile(cls, file, enable_close=False):
        """Create a limited journal reader from a file object. The resulting
        reader cannot handle rotating nor locking.

        :param file: File object to read from.
        :param enable_close: True to close the file when done, defaults
        to false.
        """
        reader = cls()
        if file:
            reader._file = file
            reader._fd = file.fileno()
            reader._enable_close = enable_close

            reader._state = State()
            reader._match = Match(reader._state)

            reader._initialized = True

        return reader

    @property
    def is_initialized(self):
        """Detect whether the reader has been initialized."""
        return self._initialized

    def initialize(self):
        """
        Initialize the reader from the EVENT_JOURNAL environment variable.
        """
        path = param.get("EVENT_JOURNAL")
        if not path:
            self._status = ReaderStatus.FILE_NOT_FOUND
            return False

        max_rotations = param.getInteger("EVENT_JOURNAL_MAX_ROTATIONS", 1)
        return self.initializeFromPath(path, max_rotations, True)

    def initializeFromPath(
        self, path, max_rotations=0, check_for_old=False, read_only=False):
        """Initialize the reader from a path.

        :param path: Path to read from.
        :param max_rotations: Sets the maximum number of rotations,
            defaults to 0.
        :param check_for_old: Try to open the rotated files first.
        :param read_only: True to access the file for reading, defaults
            to false.
        """
        if self._initialized:
            self._status = ReaderStatus.RE_INITIALIZE
            return False

        handle_rotation = max_rotations > 0
        self._state = State.createFromPath(
            path, max_rotations, SCORE_RECENT_THRESH)

        if not self._state.is_initialized:
            self._status = ReaderStatus.NOT_INITIALIZED
            return False

        self._match = Match(self._state)

        if not self._initialize(
            max_rotations, check_for_old, False, handle_rotation, read_only):
            return False

        return True

    def initializeFromFileState(
        self, file_state, max_rotations=0, read_only=False):
        """Initialize the reader from a file state.

        :param path: File state to restore from.
        :param max_rotations: Sets the maximum number of rotations,
            defaults to 0.
        :param read_only: True to access the file for reading, defaults
            to false.
        """
        if self._initialized:
            self._status = ReaderStatus.RE_INITIALIZE
            return False

        self._state = State.createFromFileState(
            file_state, SCORE_RECENT_THRESH)
        if (self._state.initialized_error
            or not self._state.is_initialized):
            self._status = ReaderStatus.STATE_ERROR
            return False

        if max_rotations > 0:
            self._state.setMaxRotations(max_rotations)
        else:
            max_rotations = self._state.getMaxRotations()

        self._match = Match(self._state)
        return self._initialize(max_rotations, False, True, True, read_only)

    def _initialize(
        self, max_rotations, check_for_rotation, restore,
        enable_read_header, read_only):
        if self._initialized:
            self._status = ReaderStatus.RE_INITIALIZE
            return False

        self._handle_rot = max_rotations > 0
        self._max_rotations = max_rotations
        self._read_header = enable_read_header
        self._read_only = read_only
        self._lock.reset()

        # Set the score factor in the file state manager
        self._state.setScoreFactor(StateScore.CTIME, ScoreFactor.CTIME)
        self._state.setScoreFactor(StateScore.INODE, ScoreFactor.INODE)
        self._state.setScoreFactor(StateScore.SAME_SIZE, ScoreFactor.SAME_SIZE)
        self._state.setScoreFactor(StateScore.GROWN, ScoreFactor.GROWN)
        self._state.setScoreFactor(StateScore.SHRUNK, ScoreFactor.SHRUNK)

        if restore:
            # Do nothing
            pass

        elif self._handle_rot and check_for_rotation:
            if not self._findPrevFile(self._max_rotations, 0, True):
                self._uninitialize()
                self._status = ReaderStatus.FILE_NOT_FOUND
                return False

        else:
            self._max_rotations = 0
            if self._state.setRotation(0, True):
                self._uninitialize()
                self._status = ReaderStatus.FILE_NOT_FOUND
                return False

        # Set locking
        if read_only:
            self._lock_enable = False
        else:
            self._lock_enable = param.getBoolean(
                "ENABLE_USERJOURNAL_LOCKING", True)

        self._close_file = param.getBoolean(
            "ALWAYS_CLOSE_USERJOURNAL", False)

        if restore:
            logging.debug(
                "Reopening journal file '%s'", self._state.getCurrentPath())
            outcome = self._reopenJournalFile(True)
            if outcome == EventOutcome.MISSED_EVENT:
                self._missed_event = True
                logging.debug("Missed event")

            elif outcome != EventOutcome.OK:
                logging.debug("Error re-opening file")
                self._uninitialize()
                self._status = ReaderStatus.FILE_NOT_FOUND
                return False

        else:
            logging.debug(
                "Opening journal file '%s'", self._state.getCurrentPath())
            outcome = self._openJournalFile(restore)
            if outcome != EventOutcome.OK:
                logging.debug("Error opening journal file")
                self._uninitialize()
                self._status = ReaderStatus.FILE_NOT_FOUND
                return False

        self._closeJournalFile(False)
        self._initialized = True

        return True

    def _uninitialize(self):
        self._match = None
        self._state = None

        self._closeJournalFile(True)

        self._lock.reset()

    def synchronize(self):
        """
        Synchronize the journal file if the last event read was an
        error.
        """
        if not self._initialized:
            self._status = ReaderStatus.NOT_INITIALIZED
            return False

        while True:
            line = self._file.readline()
            if not line:
                break

            if line == SYNC_DELIMITER:
                return True

        return False

    def readEvent(self, event, store_state=True):
        """Read the next event from the journal file.

        :param event: Event object updated with the new event.
        :param store_state: False to skip storing the state, defaults
            to true.
        """
        if not self._initialized:
            self._status = ReaderStatus.NOT_INITIALIZED
            return EventOutcome.READ_ERROR

        # Previous operation (initialization) detected a missed event
        if self._missed_event:
            self._missed_event = False
            return EventOutcome.MISSED_EVENT

        # If the file was closed, try to reopen it
        if not self._file:
            outcome = self._reopenJournalFile()
            if outcome != EventOutcome.OK:
                return outcome

        if not self._file:
            return EventOutcome.NO_EVENT

        outcome = self._readEvent(event)
        if not self._handle_rot:
            try_again = False
        else:
            try_again = outcome == EventOutcome.NO_EVENT

        # If the end of a rotated file is reached, try the previous one
        if try_again:
            # Reached the end of file and file has been closed
            if self._state.getRotation() < 0:
                return EventOutcome.MISSED_EVENT

            # Reached the end of a non-rotated file
            elif self._state.getRotation() == 0:
                result = self._match.matchPath(
                    self._state.getCurrentPath(), self._state.getRotation(),
                    ScoreThresh.NONROT)

                if result == MatchResult.NOMATCH:
                    self._closeJournalFile(True)

                else:
                    try_again = False

            # Reached the end of a ".old" or ".1", ".2" ... file
            else:
                self._closeJournalFile(True)
                found = self._findPrevFile(
                    self._state.getRotation() - 1, 1, True)
                logging.debug(
                    "Checking for previous file (%d): %s",
                    self._state.getRotation(),
                    "Found" if found else "Not found")

                if found:
                    self._closeJournalFile(True)

                else:
                    try_again = False

        # Finally, one more attempt to read an event
        if try_again:
            outcome = self._reopenJournalFile()
            if outcome == EventOutcome.OK:
                outcome = self._readEvent(event)

        # Store the current offset
        if outcome == EventOutcome.OK and store_state:
            pos = self._file.tell()
            if pos > 0:
                self._state.setOffset(pos)

            self._state.getStatus(self._fd)

        # Close the file between operations
        self._closeJournalFile(False)

        return outcome

    def _readEvent(self, event):
        # First, obtain a write lock
        if self._lock.is_unlocked:
            self._lock.acquire(LockType.WRITE_LOCK)

        # Store file position so that we can rewind
        if not self._file:
            logging.debug("Invalid file object")
            if self._lock.is_locked:
                self._lock.release()
            return EventOutcome.UNKNOWN_ERROR

        filepos = self._file.tell()
        if filepos == -1:
            logging.debug("Failed to tell file position")
            if self._lock.is_locked:
                self._lock.release()
            return EventOutcome.UNKNOWN_ERROR

        event_string = self._file.read(3)
        if not event_string:
            event.reset()
            if self._lock.is_locked:
                self._lock.release()
            return EventOutcome.NO_EVENT

        try:
            event_type = int(event_string)
            ret1 = True
        except ValueError:
            logging.debug("Error reading event number: %s", event_string)
            event_type = EventType.EXECUTE
            ret1 = False

        # Initialize event object
        event.initialize(event_type)
        if not event:
            logging.debug("Unable to initialize event")
            if self._lock.is_locked:
                self._lock.release()
            return EventOutcome.UNKNOWN_ERROR

        ret2 = event.getEvent(self._file)

        if not ret1 or not ret2:
            logging.debug("Error reading event, retrying")

            if self._lock.is_locked:
                self._lock.release()

            # Just in case locking over NFS
            sleep(1)
            if self._lock.is_unlocked:
                self._lock.acquire(LockType.WRITE_LOCK)

            if self._file.seek(filepos, os.SEEK_SET):
                logging.debug("Seek failed")
                if self._lock.is_locked:
                    self._lock.release()
                return EventOutcome.UNKNOWN_ERROR

            if self.synchronize():
                # Reset file position
                if self._file.seek(filepos, os.SEEK_SET):
                    logging.debug("Seek failed")
                    if self._lock.is_locked:
                        self._lock.release()
                    return EventOutcome.UNKNOWN_ERROR

                # Attempt to read the event again
                old_event_type = event_type
                try:
                    event_string = self._file.read(3)
                    event_type = int(event_string)
                    ret1 = True
                except ValueError:
                    event_type = -1
                    ret1 = False

                if ret1 is True:
                    if event_type != old_event_type:
                        if event:
                            event.reset()

                        event.initialize(event_type)
                        if not event:
                            logging.debug("Unable to initialize event")
                            if self._lock.is_locked:
                                self._lock.release()

                    ret2 = event.getEvent(self._file)

                # Failed again, parse error
                if not ret1 or not ret2:
                    logging.debug("Error reading event on second try")
                    event.reset()
                    self.synchronize()
                    if self._lock.is_locked:
                        self._lock.release()

                    return EventOutcome.READ_ERROR

                # Got the event successfully
                else:
                    # Synchronize the journal
                    if self.synchronize():
                        if self._lock.is_locked:
                            self._lock.release()

                        return EventOutcome.OK

                    # Could not synchronize
                    else:
                        logging.debug("Got event on second try")
                        event.reset()

                        if self._lock.is_locked:
                            self._lock.release()

                        return EventOutcome.NO_EVENT

            else:
                logging.debug("Synchronize failed")
                if self._file.seek(filepos, os.SEEK_SET):
                    logging.debug("Seek failed")
                    if self._lock.is_locked:
                        self._lock.release()

                    return EventOutcome.UNKNOWN_ERROR

                event.reset()
                if self._lock.is_locked:
                    self._lock.release()

                return EventOutcome.NO_EVENT

        else:
            # Got event successfully
            if self.synchronize():
                if self._lock.is_locked:
                    self._lock.release()

                return EventOutcome.OK

            else:
                logging.debug("Got event on first try but synchronize failed")
                event.reset()

                if self._lock.is_locked:
                    self._lock.release()

                return EventOutcome.NO_EVENT

        if self._lock.is_locked:
            self._lock.release()

        return EventOutcome.UNKNOWN_ERROR

    def checkFileStatus(self):
        """Check the status of the file - has it grown, shrunk, etc."""
        if self._state:
            return self._state.getFileStatus(self._fd)

        else:
            return FileStatus.ERROR

    def _findPrevFile(self, start, num, store_stat):
        if not self._handle_rot:
            return True

        if num == 0:
            end = 0

        else:
            end = start - num + 1
            if end < 0:
                end = 0

        for rot in xrange(start, end - 1, -1):
            if self._state.setRotation(rot, store_stat) == 0:
                logging.debug("Found '%s'", self._state.getCurrentPath())
                return True

        self._status = ReaderStatus.FILE_NOT_FOUND
        return False

    def _openJournalFile(self, do_seek, read_header=True):
        is_lock_current = self._state.getRotation() == self._lock_rot

        logging.debug(
            "Opening journal file #%d '%s'",
            self._state.getRotation(), self._state.getCurrentPath())

        if self._state.getRotation() < 0:
            if self._state.setRotation(-1) < 0:
                return EventOutcome.READ_ERROR

        self._fd = fd_open(
            self._state.getCurrentPath(),
            os.O_RDONLY if self._read_only else os.O_RDWR, 0)
        if self._fd < 0:
            return EventOutcome.READ_ERROR

        try:
            self._file = File.create_from_fd(self._fd, "r")
        except OSError:
            self._closeJournalFile(True)
            return EventOutcome.READ_ERROR

        # Seek to the previous location
        if do_seek and self._state.getOffset():
            if self._file.seek(self._state.getOffset(), os.SEEK_SET):
                self._closeJournalFile(True)
                return EventOutcome.READ_ERROR

        # Prepare to lock the file
        if self._lock_enable:
            if not is_lock_current and self._lock:
                self._lock.reset()
                self._lock_rot = -1

            if not self._lock:
                logging.debug(
                    "Creating file lock '%s'", self._state.getCurrentPath())

                self._lock.initialize(self._fd, self._file)
                self._lock_rot = self._state.getRotation()

            else:
                # TODO: look into this
                self._lock.initialize(self._fd, self._file)

        else:
            if self._lock:
                self._lock_rot = -1

            self._lock.reset()

        # Read the file's header event
        if (read_header
            and self._read_header
            and not self._state.valid_unique_id):
            path = self._state.getCurrentPath()
            if not path:
                path = self._state.generatePath(self._state.getRotation())

            reader = Reader()
            header_reader = HeaderReader()
            if (path
                and reader.initializeFromPath(path, read_only=True)
                and header_reader.read(reader) == EventOutcome.OK):
                self._state.setUniqueId(header_reader.id)
                self._state.setSequence(header_reader.sequence)
                self._state.setPosition(header_reader.file_offset)
                self._state.setRecord(header_reader.event_offset)
                logging.debug(
                    "%s: set unique id to '%s', sequence to %d",
                    self._state.getCurrentPath(), header_reader.id,
                    header_reader.sequence)

            else:
                logging.debug(
                    "%s: failed to read file header",
                    self._state.getCurrentPath())

        return EventOutcome.OK

    def _reopenJournalFile(self, restore=False):
        # Done if the file is already open
        if self._file:
            return EventOutcome.OK

        if not self._handle_rot:
            return self._openJournalFile(True)

        # If state info is not valid, find previous file
        if not self._state.is_valid:
            if self._handle_rot:
                logging.debug("Looking for previous file...")
                if not self._findPrevFile(self._max_rotations, 0, True):
                    self._status = ReaderStatus.FILE_NOT_FOUND
                    return EventOutcome.NO_EVENT

            else:
                if self._state.setRotation(0, True):
                    self._status = ReaderStatus.FILE_NOT_FOUND
                    return EventOutcome.NO_EVENT

            return self._openJournalFile(False)

        new_rot = -1
        max_score_value = -1
        max_score_rot = -1
        rot = self._state.getRotation()
        thresh = ScoreThresh.RESTORE if restore else ScoreThresh.FWSEARCH
        while True:
            if rot > self._max_rotations or new_rot >= 0:
                break
            score = Score()
            result = self._match.match_rotation(rot, thresh, score)
            if result == MatchResult.ERROR:
                pass

            elif result == MatchResult.MATCH:
                new_rot = rot

            elif result == MatchResult.UNKNOWN:
                if score > max_score_value:
                    max_score_rot = rot
                    max_score_value = score.value

        # No good match found, fall back to highest score
        if new_rot < 0 and max_score_value > 0:
            if restore:
                return EventOutcome.MISSED_EVENT

            new_rot = max_score_rot

        # Good match found or a high score, do it
        if new_rot >= 0:
            if self._state.getRotation(new_rot):
                self._status = ReaderStatus.FILE_NOT_FOUND
                return EventOutcome.READ_ERROR

            return self._openJournalFile(True)

        # No match found
        self._state.reset()
        return EventOutcome.MISSED_EVENT

    def _closeJournalFile(self, force):
        if force or self._close_file:
            if self._lock and self._lock.is_locked:
                self._lock.release()
                self._lock_rot = -1

            if self._enable_close:
                if self._file:
                    self._file.close()
                    self._file = None
                    self._fd = -1

                elif self._fd >= 0:
                    fd_close(self._fd)
                    self._fd = -1

        return True
