#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "StateReset",
    "StateScore",
    "State",
    ]

import time

import logging

from checkbox.lib.enum import Enum
from checkbox.lib.stat import (
    Stat,
    StatResult,
    )

from checkbox.journal.file import (
    FileStateInfo,
    FileStatus,
    FILE_STATE_SIGNATURE,
    FILE_STATE_VERSION,
    )


StateReset = Enum(
    "FILE",
    "FULL",
    "INIT",
    )

StateScore = Enum(
    "CTIME",      # ctime match
    "INODE",      # inodes match
    "SAME_SIZE",  # File is the same size
    "GROWN",      # File has grown
    "SHRUNK",     # File has shrunk
    )


class State:
    """Journal state."""

    __slots__ = (
        "_initialized",
        "_init_error",
        "_base_path",
        "_max_rotations",
        "_recent_thresh",
        "_score_fact_ctime",
        "_score_fact_inode",
        "_score_fact_same_size",
        "_score_fact_grown",
        "_score_fact_shrunk",
        "_base_path",
        "_current_path",
        "_current_rotation",
        "_unique_id",
        "_sequence",
        "_stat_result",
        "_stat_size",
        "_stat_valid",
        "_stat_time",
        "_offset",
        "_position",
        "_record",
        "_update_time",
        )

    def __init__(self):
        self.reset(StateReset.INIT)

    @classmethod
    def createFromPath(cls, path, max_rotations=0, recent_thresh=0):
        state = cls()
        state._base_path = path
        state._max_rotations = max_rotations
        state._recent_thresh = recent_thresh
        state._initialized = True

        return state

    @classmethod
    def createFromFileState(cls, file_state, recent_thresh=0):
        state = cls()
        state._recent_thresh = recent_thresh
        if not state.setFileState(file_state):
            logging.debug("Failed to set file state from buffer")
            state._init_error = True

        return state

    @property
    def is_valid(self):
        """Return true if there is a current rotatoin."""
        return self._current_rotation >= 0

    @property
    def is_initialized(self):
        """Return true if the initialization was done."""
        return self._initialized

    @property
    def initialized_error(self):
        """Return true if the initialization encountered an error."""
        return self._init_error

    def reset(self, reset=StateReset.FILE):
        """Reset the parameters about the current journal file

        :param reset: The state reset type, defaults to FILE.
        """
        if reset == StateReset.INIT:
            self._initialized = False
            self._init_error = False
            self._base_path = ""
            self._max_rotations = 0

            self._recent_thresh = 0
            self._score_fact_ctime = 0
            self._score_fact_inode = 0
            self._score_fact_same_size = 0
            self._score_fact_grown = 0
            self._score_fact_shrunk = 0

        elif reset == StateReset.FULL:
            self._base_path = ""

        self._current_path = ""
        self._current_rotation = -1
        self._unique_id = ""
        self._sequence = 0

        self._stat_result = StatResult()
        self._stat_size = -1
        self._stat_valid = False
        self._stat_time = 0

        self._offset = 0

        self._position = 0
        self._record = 0

        self._update_time = 0

    def update(self):
        """Update the last update time."""
        self._update_time = time.time()

    def generatePath(self, rotation, initializing=False):
        """Return a generated journal path.

        :param rotation: Rotation number for the path
        :param initializing: True if initializing the state, defaults to false.
        """
        if not initializing and not self._initialized:
            return ""

        if rotation < 0 or rotation > self._max_rotations:
            return ""

        if not self._base_path:
            return ""

        path = self._base_path
        if rotation:
            if self._max_rotations > 1:
                path += ".%d" % rotation
            else:
                path += ".old"

        return path

    def getBasePath(self, file_state=None):
        """Get the base journal path.

        :param file_state: If defined, get the base path from file state.
        """
        if file_state:
            info = file_state.info
            return info.base_path if info and info.version else None
        else:
            return self._base_path

    def getCurrentPath(self, file_state=None):
        """Get the current journal path.

        :param file_state: If defined, get the current path from file state.
        """
        if file_state:
            info = file_state.info
            return (self.generatePath(info.rotation, True)
                if info and info.version else "")
        else:
            return self._current_path

    def getPosition(self, file_state=None):
        """Get the position in the whole journal.

        :param file_state: If defined, get the position from file state.
        """
        if file_state:
            info = file_state.info
            return info.rotation if info and info.version else -1
        else:
            return self._position

    def setPosition(self, position):
        """Set the journal position.

        :param position: New journal position.
        """
        self.update()
        self._position = position

    def getRecord(self, file_state=None):
        """Get the record in the whole journal.

        :param file_state: If defined, get the record from file state.
        """
        if file_state:
            info = file_state.info
            return info.record if info and info.version else -1
        else:
            return self._record

    def setRecord(self, record):
        """Set the journal record.

        :param record: New journal record.
        """
        self.update()
        self._record = record

    def getOffset(self, file_state=None):
        """Get the current journal offset.

        :param file_state: If defined, get the offset from file state.
        """
        if file_state:
            info = file_state.info
            return info.offset if info and info.version else -1
        else:
            return self._offset

    def setOffset(self, offset):
        """Set the current offset.

        :param offset: New offset.
        """
        self.update()
        self._offset = offset

    def getMaxRotations(self):
        """Get the maximum number of rotations."""
        return self._max_rotations

    def setMaxRotations(self, max_rotations):
        """Set the maximum number of rotations.

        :param max_rotations: New maximum number of rotations."""
        self._max_rotations = max_rotations

    def getSequence(self):
        """Get the journal sequence number."""
        return self._sequence

    def setSequence(self, sequence):
        """Set the journal sequence number.

        :param sequence: New sequence number."""
        self._sequence = sequence

    def getUniqueId(self):
        """Get the journal unique identifier."""
        return self._unique_id

    def setUniqueId(self, unique_id):
        """Set the journal unique identifier.

        :param unique_id: New unique identifier."""
        self._unique_id = unique_id

    @property
    def valid_unique_id(self):
        """Return true if the unique id is valid."""
        return len(self._unique_id) != 0

    def compareUniqueId(self, unique_id):
        """Compare the ID to the one stored.

        :param unique_id: Unique ID to compare.
        """
        if self._unique_id == "" or unique_id == "":
            return 0
        elif self._unique_id == unique_id:
            return 1
        else:
            return -1

    def getStatus(self, fd=None):
        """Get updated status from the journal file.

        :param fd: If defined, get status from the file descriptor.
        """
        if fd is None:
            current_path = self.getCurrentPath()
            status = self._getStatResultStatus(
                self._stat_result, current_path)
        else:
            stat = Stat()
            status = stat.get_status() if stat.stat_fd(fd) else 0
            if status == 0:
                stat.set_result(self._stat_result)

        if status == 0:
            self._stat_time = time.time()
            self._stat_valid = True
            self.update()

        return status

    def _getStatResultStatus(self, stat_result, path=""):
        if not path:
            path = self._current_path

        stat = Stat()
        status = stat.get_status() if stat.stat_path(path) else 0
        if status == 0:
            stat.set_result(stat_result)

        return status

    def getFileStatus(self, fd=-1):
        """Get file status to determine if the journal has grown.

        :param fd: If defined, get file status from the file descriptor.
        """
        stat = Stat()
        if fd >= 0:
            stat.stat_fd(fd)
        else:
            stat.stat_path(self._current_path)

        if stat.get_status():
            logging.debug("Failed to stat; errno = %s" % stat.get_errno())
            return FileStatus.ERROR

        size = stat.get_result().st_size
        if self._stat_size < 0 or size > self._stat_size:
            file_status = FileStatus.GROWN
        elif size == self._stat_size:
            file_status = FileStatus.NO_CHANGE
        else:
            file_status = FileStatus.SHRUNK

        self._stat_size = size
        self.update()

        return file_status

    @staticmethod
    def initializeFileState(file_state):
        """Class method to initialize file state information."""
        file_state.info = FileStateInfo()
        return True

    @staticmethod
    def uninitializeFileState(file_state):
        """Class method to uninitialize file state information."""
        file_state.info = None
        return True

    def getFileState(self, file_state):
        """Get the file state of the journal state.

        :param file_state: Destination for the file state.
        """
        info = file_state.info
        if not info:
            return False

        # Verify that the signature and version matches
        if (info.signature != FILE_STATE_SIGNATURE
            or info.version != FILE_STATE_VERSION):
            return False

        # The paths shouldn't change... copy them only the first time
        if not info.base_path:
            info.base_path = self._base_path

        info.rotation = self._current_rotation
        if self._unique_id:
            info.unique_id = self._unique_id
        else:
            info.unique_id = ""

        info.sequence = self._sequence
        info.max_rotations = self._max_rotations

        info.inode = self._stat_result.st_ino
        info.ctime = self._stat_result.st_ctime
        info.size = self._stat_result.st_size

        info.offset = self._offset

        info.position = self._position
        info.record = self._record

        info.update_time = self._update_time

        return True

    def setFileState(self, file_state):
        """Set the journal state from the file state.

        :param file_state: Source for the file state.
        """
        info = file_state.info
        if not info:
            return False

        # Verify that the signature and version matches
        if (info.signature != FILE_STATE_SIGNATURE
            or info.version != FILE_STATE_VERSION):
            self._init_error = True
            return False

        self._base_path = info.base_path

        # Set the rotation and path
        self._max_rotations = info.max_rotations
        self.setRotation(info.rotation, False, True)

        self._unique_id = info.unique_id
        self._sequence = info.sequence

        self._stat_result.st_ino = info.inode
        self._stat_result.st_ctime = info.ctime
        self._stat_result.st_size = info.size
        self._stat_valid = True

        self._offset = info.offset

        self._position = info.position
        self._record = info.record

        self._update_time = info.update_time

        self._initialized = True

        return True

    def getRotation(self, file_state=None):
        """Get the current journal rotation.

        :param file_state: If defined, get the rotation from file state.
        """
        if file_state:
            info = file_state.info
            return info.rotation if info and info.version else -1
        else:
            return self._current_rotation

    def setRotation(self, rotation, store=False, initializing=False):
        """Set the current journal rotation.

        :param rotation: New rotation number.
        :param store: True to store the result, defaults to false.
        :param initializing: True if initializing the state, defaults to false.
        """
        if not initializing and not self._initialized:
            return -1

        if rotation > self._max_rotations:
            return -1

        if store:
            self.reset()
            status = self.setStatResultRotation(rotation, self._stat_result)
            if status == 0:
                self._stat_valid = True

            return status
        else:
            stat_result = StatResult()
            return self.setStatResultRotation(
                rotation, stat_result, initializing)

    def setStatResultRotation(self, rotation, stat_result, initializing=False):
        """Set the current journal rotation and update the stat result.

        :param rotation: New rotation number.
        :param stat_result: The stat result to update.
        :param initializing: True if initializing the state, defaults to false.
        """
        if not initializing and not self._initialized:
            return -1

        if rotation < 0 or rotation > self._max_rotations:
            return -1

        if rotation == self._current_rotation:
            return 0

        self._unique_id = ""
        self._current_path = self.generatePath(rotation, initializing)
        self._current_rotation = rotation
        self.update()

        return self._getStatResultStatus(stat_result)

    def getScore(self, rotation=-1):
        """Get the score for a rotation.

        :param rotation: The rotation number, defaults to the current rotation.
        """
        if rotation > self._max_rotations:
            return -1

        elif rotation < 0:
            rotation = self._current_rotation

        path = self.generatePath(rotation)
        if not path:
            return -1

        return self.getPathScore(path, rotation)

    def getPathScore(self, path="", rotation=-1):
        """Get the score for a rotation for a given path.

        :param path: The path to get the score, defaults to the current path.
        :param rotation: The rotation number, defaults to the current rotation.
        """
        if not path:
            path = self.getCurrentPath()

        if rotation < 0:
            rotation = self._current_rotation

        if self._getStatResultStatus(self._stat_result, path):
            logging.debug("Stat error: %s", path)
            return -1

        return self.getStatResultScore(self._stat_result, rotation)

    def getStatResultScore(self, stat_result, rotation=-1):
        """Get the score for a rotation for a given stat result.

        :param stat_result: The stat result to get the score.
        :param rotation: The rotation number, defaults to the current rotation.
        """
        score = 0

        if rotation < 0:
            rotation = self._current_rotation

        if self._stat_result.st_ino == stat_result.st_ino:
            score += self._score_fact_inode

        if self._stat_result.st_ctime == stat_result.st_ctime:
            score += self._score_fact_ctime

        if stat_result.st_size == self._stat_result.st_size:
            score += self._score_fact_same_size

        elif (time.time() < (self._update_time + self._recent_thresh)
            and rotation == self._current_rotation
            and stat_result.st_size > self._stat_result.st_size):
            score += self._score_fact_grown

        if self._stat_result.st_size > stat_result.st_size:
            score += self._score_fact_shrunk

        if score < 0:
            score = 0

        return score

    def setScoreFactor(self, score_factor, factor):
        """Set the factors for calculating the score.

        :param score_factor: The score factor type.
        "param factor: The factor value for the given type.
        """
        if score_factor == StateScore.CTIME:
            self._score_fact_ctime = factor

        elif score_factor == StateScore.INODE:
            self._score_fact_inode = factor

        elif score_factor == StateScore.SAME_SIZE:
            self._score_fact_same_size = factor

        elif score_factor == StateScore.GROWN:
            self._score_fact_grown = factor

        elif score_factor == StateScore.SHRUNK:
            self._score_fact_shrunk = factor

        self.update()
