#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from unittest import TestCase

from tempfile import mktemp
from StringIO import StringIO

from checkbox.journal.event import (
    Event,
    EventOutcome,
    EventType,
    )
from checkbox.journal.header import (
    Header,
    HeaderReader,
    HeaderWriter,
    )
from checkbox.journal.reader import Reader
from checkbox.journal.writer import Writer


class TestHeader(TestCase):

    def getEventData(self):
        return "Global Journal:" \
            " ctime=1" \
            " id=foo" \
            " sequence=2" \
            " size=3" \
            " events=4" \
            " offset=5" \
            " event_offset=6"

    def checkEventData(self, header):
        self.assertEquals(header.ctime, 1)
        self.assertEquals(header.id, "foo")
        self.assertEquals(header.sequence, 2)
        self.assertEquals(header.size, 3)
        self.assertEquals(header.num_events, 4)
        self.assertEquals(header.file_offset, 5)
        self.assertEquals(header.event_offset, 6)

    def test_init(self):
        header = Header()

    def test_copyFrom(self):
        header1 = Header()
        self.assertEquals(header1.file_offset, 0)

        header1.addFileOffset(1)
        self.assertEquals(header1.file_offset, 1)

        header2 = Header()
        self.assertEquals(header2.file_offset, 0)

        header2.copyFrom(header1)
        self.assertEquals(header2.file_offset, 1)

        header2 = Header(header1)
        self.assertEquals(header2.file_offset, 1)

    def test_incrementSequence(self):
        header = Header()
        self.assertEquals(header.sequence, 0)

        header.incrementSequence()
        self.assertEquals(header.sequence, 1)

    def test_addFileOffset(self):
        header = Header()
        self.assertEquals(header.file_offset, 0)

        header.addFileOffset(2)
        self.assertEquals(header.file_offset, 2)

    def test_addEventOffset(self):
        header = Header()
        self.assertEquals(header.event_offset, 0)

        header.addEventOffset(2)
        self.assertEquals(header.event_offset, 2)

    def test_extractEvent(self):
        header = Header()
        self.assertFalse(header.is_valid)

        event = Event(EventType.GENERIC)
        outcome = header.extractEvent(event)
        self.assertEquals(outcome, EventOutcome.NO_EVENT)
        self.assertFalse(header.is_valid)

        event.data = self.getEventData()
        outcome = header.extractEvent(event)
        self.assertEquals(outcome, EventOutcome.OK)
        self.assertTrue(header.is_valid)

        self.checkEventData(header)


class TestHeaderReader(TestHeader):

    def setUp(self):
        self.path = mktemp()

    def tearDown(self):
        if os.path.exists(self.path):
            os.unlink(self.path)

    def test_read(self):
        reader = Reader.createFromPath(self.path)
        header_reader = HeaderReader()
        outcome = header_reader.read(reader)
        self.assertEquals(outcome, EventOutcome.READ_ERROR)

        # Write event
        event = Event(EventType.GENERIC)
        event.data = self.getEventData()
        writer = Writer.createFromPath(self.path)
        writer.writeEvent(event)

        # Read event
        reader = Reader.createFromPath(self.path)
        outcome = header_reader.read(reader)
        self.assertEquals(outcome, EventOutcome.OK)
        self.checkEventData(header_reader)


class TestHeaderWriter(TestHeader):

    def setUp(self):
        self.path = mktemp()

    def tearDown(self):
        if os.path.exists(self.path):
            os.unlink(self.path)

    def test_write(self):
        event = Event(EventType.GENERIC)
        event.data = self.getEventData()

        header_writer = HeaderWriter()
        header_writer.extractEvent(event)

        writer = Writer.createFromPath(self.path)

        file = StringIO()
        header_writer.write(writer, file)

        file.seek(0, os.SEEK_SET)
        line = file.readline().strip()
        self.assertTrue(line.endswith(self.getEventData()))

    def test_generateEvent(self):
        header_writer = HeaderWriter()
        header_writer.ctime = 1
        header_writer.id = "foo"
        header_writer.sequence = 2
        header_writer.size = 3
        header_writer.num_events = 4
        header_writer.file_offset = 5
        header_writer.event_offset = 6

        event = Event(EventType.GENERIC)
        self.assertTrue(header_writer.generateEvent(event))
        self.assertEquals(event.data, self.getEventData())

