#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from unittest import TestCase

from tempfile import mkstemp

from checkbox.lib.file import File

from checkbox.journal.event import (
    Event,
    EventOutcome,
    EventType,
    )
from checkbox.journal.file import (
    FileState,
    FileStateInfo,
    )
from checkbox.journal.reader import Reader
from checkbox.journal.writer import Writer


class TestReader(TestCase):

    def setUp(self):
        (fd, self.path) = mkstemp()
        self.file = File.create_from_fd(fd)

        os.environ["EVENT_JOURNAL"] = self.path

    def tearDown(self):
        if os.path.exists(self.path):
            os.unlink(self.path)

        del os.environ["EVENT_JOURNAL"]

    def test_init(self):
        reader = Reader()
        self.assertFalse(reader.is_initialized)

        event_log = os.environ.pop("EVENT_JOURNAL")
        reader = Reader(True)
        self.assertFalse(reader.is_initialized)

        os.environ["EVENT_JOURNAL"] = event_log
        reader = Reader(True)
        self.assertTrue(reader.is_initialized)

    def test_createFromFileState(self):
        file_state = FileState()
        reader = Reader.createFromFileState(file_state)
        self.assertFalse(reader.is_initialized)

        file_state.info = FileStateInfo()
        reader = Reader.createFromFileState(file_state)
        self.assertFalse(reader.is_initialized)

        file_state.base_path = self.path
        reader = Reader.createFromFileState(file_state)
        self.assertTrue(reader.is_initialized)

    def test_createFromPath(self):
        reader = Reader.createFromPath("")
        self.assertFalse(reader.is_initialized)

        reader = Reader.createFromPath(self.path)
        self.assertTrue(reader.is_initialized)

    def test_createFromFile(self):
        reader = Reader.createFromFile(None)
        self.assertFalse(reader.is_initialized)

        reader = Reader.createFromFile(self.file)
        self.assertTrue(reader.is_initialized)

    def test_readEvent(self):
        event = Event()
        reader = Reader.createFromPath(self.path)
        outcome = reader.readEvent(event)
        self.assertEquals(outcome, EventOutcome.NO_EVENT)

        event = Event(EventType.GENERIC)
        event.data = "foo"
        writer = Writer.createFromPath(self.path)
        ret = writer.writeEvent(event)
        self.assertTrue(ret)

        event = Event()
        reader = Reader.createFromPath(self.path)

        # Header event
        outcome = reader.readEvent(event)
        self.assertEquals(outcome, EventOutcome.OK)
        self.assertEquals(event.type, EventType.GENERIC)

        # Global event
        outcome = reader.readEvent(event)
        self.assertEquals(outcome, EventOutcome.OK)
        self.assertEquals(event.type, EventType.GENERIC)
        self.assertEquals(event.data, "foo")

        # User event
        outcome = reader.readEvent(event)
        self.assertEquals(outcome, EventOutcome.OK)
        self.assertEquals(event.type, EventType.GENERIC)
        self.assertEquals(event.data, "foo")
