#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from unittest import TestCase

from tempfile import mkstemp

from checkbox.lib.stat import StatResult

from checkbox.journal.state import (
    State,
    StateScore,
    )
from checkbox.journal.file import (
    FileStatus,
    FileState,
    )


class TestState(TestCase):

    def setUp(self):
        (fd, self.path) = mkstemp()
        self.file = os.fdopen(fd, "w")

    def tearDown(self):
        self.file.close()
        os.unlink(self.path)

    def test_init(self):
        state = State()
        self.assertFalse(state.is_initialized)

    def test_createFromPath(self):
        state = State.createFromPath(self.path, 1)
        self.assertTrue(state.is_initialized)

    def test_createFromFileState(self):
        file_state = FileState()
        state = State.createFromFileState(file_state)
        self.assertFalse(state.is_initialized)

        State.initializeFileState(file_state)
        state = State.createFromFileState(file_state)
        self.assertTrue(state.is_initialized)

    def update(self):
        state = State()
        state.update()

    def test_position(self):
        state = State()
        self.assertEquals(state.getPosition(), 0)

        file_state = FileState()
        self.assertEquals(state.getPosition(file_state), -1)

        State.initializeFileState(file_state)
        self.assertEquals(state.getPosition(file_state), 0)

        state.setPosition(1)
        self.assertEquals(state.getPosition(), 1)
        self.assertEquals(state.getPosition(file_state), 0)

    def test_record(self):
        state = State()
        self.assertEquals(state.getRecord(), 0)

        file_state = FileState()
        self.assertEquals(state.getRecord(file_state), -1)

        State.initializeFileState(file_state)
        self.assertEquals(state.getRecord(file_state), 0)

        state.setRecord(1)
        self.assertEquals(state.getRecord(), 1)
        self.assertEquals(state.getRecord(file_state), 0)

    def test_maxRotations(self):
        state = State()
        self.assertEquals(state.getMaxRotations(), 0)

        state.setMaxRotations(1)
        self.assertEquals(state.getMaxRotations(), 1)

    def test_offset(self):
        state = State()
        self.assertEquals(state.getOffset(), 0)

        state.setOffset(1)
        self.assertEquals(state.getOffset(), 1)

    def test_sequence(self):
        state = State()
        self.assertEquals(state.getSequence(), 0)

        state.setSequence(1)
        self.assertEquals(state.getSequence(), 1)

    def test_unique_id(self):
        state = State()
        self.assertEquals(state.getUniqueId(), "")
        self.assertFalse(state.valid_unique_id)

        state.setUniqueId("fake id")
        self.assertEquals(state.getUniqueId(), "fake id")
        self.assertTrue(state.valid_unique_id)

    def test_generatePath(self):
        state = State()
        self.assertEquals(state.generatePath(0), "")

        state = State.createFromPath("foo")
        self.assertEquals(state.generatePath(0), "foo")
        self.assertEquals(state.generatePath(1), "")

        state = State.createFromPath("foo", 1)
        self.assertEquals(state.generatePath(0), "foo")
        self.assertEquals(state.generatePath(1), "foo.old")
        self.assertEquals(state.generatePath(2), "")

        state = State.createFromPath("foo", 2)
        self.assertEquals(state.generatePath(0), "foo")
        self.assertEquals(state.generatePath(1), "foo.1")
        self.assertEquals(state.generatePath(2), "foo.2")
        self.assertEquals(state.generatePath(3), "")

    def test_status(self):
        state = State()
        self.assertEquals(state.getStatus(), -3)

        state = State.createFromPath(self.path)
        self.assertNotEquals(state.setRotation(0), -1)
        self.assertEquals(state.getStatus(), 0)

        file_state = FileState()
        State.initializeFileState(file_state)
        file_state.base_path = self.path

        state = State.createFromFileState(file_state)
        self.assertEquals(state.getStatus(), 0)

        self.assertEquals(state.getStatus(0), 0)
        self.assertEquals(state.getStatus(1), 0)
        self.assertEquals(state.getStatus(2), 0)

    def test_fileStatus(self):
        state = State()
        self.assertEquals(state.getFileStatus(), FileStatus.ERROR)

        state = State.createFromPath(self.path)
        self.assertNotEquals(state.setRotation(0), -1)
        self.assertEquals(state.getFileStatus(), FileStatus.GROWN)
        self.assertEquals(state.getFileStatus(), FileStatus.NO_CHANGE)

        self.file.write("some text\n")
        self.file.flush()
        self.assertEquals(state.getFileStatus(), FileStatus.GROWN)

        self.file.truncate(0)
        self.file.flush()
        self.assertEquals(state.getFileStatus(), FileStatus.SHRUNK)

    def test_fileState(self):
        state = State()
        file_state = FileState()

        self.assertFalse(state.getFileState(file_state))
        self.assertFalse(state.setFileState(file_state))

        State.initializeFileState(file_state)
        file_state.offset = 1
        self.assertTrue(state.getFileState(file_state))
        self.assertEquals(file_state.offset, 0)

        file_state.offset = 1
        self.assertTrue(state.setFileState(file_state))
        self.assertEquals(state.getOffset(), 1)

    def test_rotation(self):
        state = State()
        self.assertEquals(state.getRotation(), -1)
        self.assertEquals(state.setRotation(0), -1)

        state = State.createFromPath(self.path)
        self.assertEquals(state.getRotation(), -1)
        self.assertEquals(state.setRotation(0), 0)
        self.assertEquals(state.setRotation(1), -1)

        state = State.createFromPath(self.path)
        stat_result = StatResult()
        self.assertEquals(state.setStatResultRotation(0, stat_result), 0)
        self.assertNotEquals(stat_result.st_ino, 0)

    def test_score(self):
        state = State()
        self.assertEquals(state.getScore(), -1)

        state = State.createFromPath(self.path)
        self.assertNotEquals(state.setRotation(0), -1)
        self.assertEquals(state.getScore(), 0)

        state.setScoreFactor(StateScore.INODE, 1)
        self.assertEquals(state.getScore(), 1)

        state.setScoreFactor(StateScore.CTIME, 1)
        self.assertEquals(state.getScore(), 2)

        state.setScoreFactor(StateScore.SAME_SIZE, 1)
        self.assertEquals(state.getScore(), 3)
