#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Writer",
    ]

import os
import logging

from datetime import datetime

from checkbox.lib.param import (
    getBoolean,
    getInteger,
    get as getString,
    )
from checkbox.lib.file import (
    File,
    fd_open,
    fd_close,
    )
from checkbox.lib.lock import (
    Lock,
    LockType,
    )
from checkbox.lib.stat import (
    Stat,
    StatOpType,
    )

from checkbox.journal.event import (
    Event,
    EventOutcome,
    )
from checkbox.journal.header import (
    Header,
    HeaderReader,
    HeaderWriter,
    )
from checkbox.journal.reader import Reader


DEV_NULL = "/dev/null"

SYNC_DELIMITER = "...\n"


class Writer:

    __slots__ = (
        "_configured",
        "_user_enable",
        "_path",
        "_file",
        "_lock",
        "_enable_fsync",
        "_enable_locking",
        "_global_path",
        "_global_file",
        "_global_lock",
        "_global_stat",
        "_rotation_lock",
        "_rotation_lock_fd",
        "_rotation_lock_path",
        "_global_enable",
        "_global_path",
        "_global_count_events",
        "_global_max_filesize",
        "_global_max_rotations",
        "_global_filesize",
        "_global_locking",
        "_global_fsync",
        "_global_uniq_base",
        "_global_sequence",
        )

    def __init__(self):
        self._configured = False

        self._user_enable = True
        self._path = ""
        self._file = File()
        self._lock = Lock()
        self._enable_fsync = True
        self._enable_locking = True

        self._global_path = ""
        self._global_file = File()
        self._global_lock = Lock()
        self._global_stat = None

        self._rotation_lock = Lock()
        self._rotation_lock_fd = -1
        self._rotation_lock_path = ""

        self._global_enable = True
        self._global_path = ""
        self._global_count_events = False
        self._global_max_filesize = 1000000
        self._global_max_rotations = 1
        self._global_filesize = 0
        self._global_locking = True
        self._global_fsync = False

        utc = datetime.utcnow()
        self._global_sequence = 0
        self._global_uniq_base = "%d.%d.%d.%d" % (
            os.getuid(),
            os.getpid(),
            utc.second,
            utc.microsecond)

    def __del__(self):
        if os.path.exists(self._rotation_lock_path):
            os.unlink(self._rotation_lock_path)

    @classmethod
    def createFromPath(cls, path=""):
        """Class method to create a journal writer initialized from a path.

        :param path: Path name of the journal file to be written.
        """
        writer = cls()
        writer.initializeFromPath(path)

        return writer

    def getGlobalPath(self):
        """Get the global journal path name."""
        return self._global_path

    def getGlobalSize(self):
        """Get the global journal size."""
        if not self._global_stat or self._global_stat.stat():
            return -1

        return self._global_stat.get_result().st_size

    def setUserWrite(self, enable):
        """Set the flag to write to the user journal. Enabled by default.

        :param enable: True to enable, false to disable.
        """
        self._user_enable = enable

    def setGlobalWrite(self, enable):
        """Set the flag to write to the global journal. Enabled by default.

        :param enable: True to enable, false to disable.
        """
        self._global_enable = enable

    def initialize(self):
        """Initialize the journal with information which will be filled for
           each event passed to L{writeEvent}."""
        self.configure(False)

        if self._global_enable and not self._global_file:
            header = Header()
            self._initializeGlobal(header)

        return True

    def initializeFromPath(self, path=""):
        """Initialize the journal with path information.

        :param path: Path name of the journal file to be written.
        """
        self._path = path

        # Reset local resources
        self._file.reset()
        self._lock.reset()

        if (self._user_enable
            and not self._openPath(path, self._file, lock=self._lock)):
            return False

        return self.initialize()

    def _initializeGlobal(self, header):
        if self._global_lock:
            self._global_lock.reset()

        if self._global_file:
            self._global_file.reset()

        if not self._global_path:
            return True

        if not self._openPath(
            self._global_path, self._global_file, lock=self._global_lock):
            return False

        stat = Stat()
        if (not stat.stat_path(self._global_path)
            and not stat.get_result().st_size):
            writer = HeaderWriter(header)

            self._global_sequence = writer.incrementSequence()
            if not self._global_sequence:
                self._global_sequence = 1

            utc = datetime.utcnow()
            writer.id = "%s.%d.%d.%d" % (
                self._global_uniq_base,
                self._global_sequence,
                utc.second,
                utc.microsecond)

            writer.addFileOffset(writer.size)
            writer.size = 0

            writer.addEventOffset(writer.num_events)
            writer.num_events = 0

            return writer.write(self)

        else:
            return True

    def _checkGlobalRotation(self):
        """Return true if the file was rotated."""
        if not self._global_file:
            return False

        if not self._global_path:
            return False

        if not self._global_max_rotations:
            return False

        # Check the size of the journal file
        current_filesize = self.getGlobalSize()
        if current_filesize < 0:
            return False

        header_reader = HeaderReader()

        # File has shrunk? Another process rotated it
        if current_filesize < self._global_filesize:
            self._setGlobalRotated(header_reader)
            return True

        # Less than the size limit, so do nothing
        elif current_filesize <= self._global_max_filesize:
            return False

        # Get the rotation lock
        if not self._rotation_lock.acquire(LockType.WRITE_LOCK):
            logging.debug("Failed to get rotation lock")
            return False

        # Here, it appears that the file is over the limit
        # Grab the rotation lock and check again

        # Check the size of the journal file
        current_filesize = self.getGlobalSize()
        if current_filesize < 0:
            return False

        # File has shrunk? Another process rotated it
        if current_filesize < self._global_filesize:
            self._setGlobalRotated(header_reader)
            return True

        # Less than the size limit, so do nothing
        elif current_filesize <= self._global_max_filesize:
            self._rotation_lock.release()
            return False

        # Here, we have the rotation lock and the file is over the limit
        # Rotate the journal

        # Read the old header, use it to write an updated one
        file = File.create_from_path(self._global_path, "r")
        if not file:
            logging.debug("Failed to open '%s'", self._global_path)

        else:
            reader = Reader(file, False)
            if header_reader.read(reader) != EventOutcome.OK:
                logging.debug(
                    "Error reading header of '%s'", self._global_path)

            if self._global_count_events:
                events = 0
                while True:
                    event = Event()
                    if reader.readEvent(event) != EventOutcome.OK:
                        break

                    events += 1

                header_reader.num_events = events

            file.reset()

        header_reader.size = current_filesize

        # Craft a header writer object from the header reader
        if not self._openPath(self._global_path, file, False):
            logging.debug(
                "Failed to open '%s' for header rewrite", self._global_path)

        header_writer = HeaderWriter(header_reader)

        if file:
            file.seek(0, os.SEEK_SET)
            header_writer.write(self, file)
            file.close()

        self._doRotation(
            self._global_path, self._global_file, self._global_max_rotations)
        self._setGlobalRotated(header_reader)
        self._rotation_lock.release()

        return True

    def _setGlobalRotated(self, header_reader):
        # log was rotated, so it needs to be reopened or created
        # also, recreate the lock.
        self._initializeGlobal(header_reader)
        if self._global_lock:
            self._global_lock.acquire(LockType.WRITE_LOCK)
            self._global_file.seek(0, os.SEEK_END)
            self._global_filesize = self._global_file.tell()

        return True

    def _doRotation(self, path, file, max_rotations):
        num_rotations = 0
        rotated = path
        if max_rotations == 1:
            rotated += ".old"
        else:
            rotated += ".1"
            for i in xrange(max_rotations, 1, -1):
                old1 = "%s.%d" % (path, i - 1)

                s = Stat.create_from_path(old1)
                if s.get_status() == 0:
                    old2 = "%s.%d" % (path, i)
                    os.rename(old1, old2)
                    num_rotations += 1

        try:
            os.rename(path, rotated)
        except OSError, error:
            logging.error(
                "Calling rename(%s, %s) failed with errno %d",
                path, rotated, error.errno)
        else:
            num_rotations += 1

        return num_rotations

    def configure(self, force=True):
        """Read configuration information.

        :param force: False to return if configured already, defaults to True.
        """
        if self._configured and not force:
            return True

        # Reset global resources
        self._global_file.reset()
        self._global_lock.reset()
        self._global_stat = None

        self._rotation_lock_path = ""
        self._rotation_lock_fd = -1
        self._rotation_lock.reset()

        self._configured = True

        self._enable_fsync = getBoolean("ENABLE_USERJOURNAL_FSYNC", True)
        self._enable_locking = getBoolean("ENABLE_USERJOURNAL_LOCKING", True)

        self._global_path = getString("EVENT_JOURNAL")
        if not self._global_path:
            return True

        self._global_stat = Stat.create_from_path(
            self._global_path, StatOpType.OP_NONE)
        self._rotation_lock_path = getString("EVENT_JOURNAL_ROTATION_LOCK")
        if not self._rotation_lock_path:
            self._rotation_lock_path = self._global_path
            self._rotation_lock_path += ".lock"

        fd = fd_open(self._rotation_lock_path, os.O_WRONLY|os.O_CREAT)
        if fd < 0:
            self._rotation_lock_fd = -1
            self._rotation_lock.reset()
        else:
            self._rotation_lock_fd = fd
            self._rotation_lock.initialize(fd)

        self._global_count_events = getBoolean(
            "EVENT_JOURNAL_COUNT_EVENTS", False)
        self._global_max_rotations = getInteger(
            "EVENT_JOURNAL_MAX_ROTATIONS", 1)
        self._global_fsync = getBoolean("EVENT_JOURNAL_FSYNC", False)
        self._global_locking = getBoolean("EVENT_JOURNAL_LOCKING", True)

        self._global_max_filesize = getInteger("EVENT_JOURNAL_MAX_SIZE", -1)
        if self._global_max_filesize < 0:
            self._global_max_filesize = getInteger(
                "MAX_EVENT_JOURNAL", 1000000)

        if self._global_max_filesize == 0:
            self._global_max_rotations = 0

        return True

    def _openPath(self, path, file, append=True, lock=None):
        if path and path == DEV_NULL:
            file.reset()
            if lock is not None:
                lock.reset()
            return True

        flags = os.O_WRONLY|os.O_CREAT
        if append:
            flags |= os.O_APPEND

        fd = fd_open(path, flags, 0664)
        if fd < 0:
            return False

        mode = "a" if append else "w"
        try:
            file.initialize_from_fd(fd, mode)
        except OSError:
            fd_close(fd)
            return False

        if lock is not None:
            lock.initialize(fd, file)

        return True

    def writeEvent(self, event):
        """Write an event to the journal file.

        :param event: Event to be written.
        """
        if not self._file and not self._global_file:
            logging.debug("not initialized")
            return True

        if not event:
            return False

        if self._file:
            if not self._lock:
                logging.debug("User journal not locked")
                return False

        if self._global_file:
            if not self._global_lock:
                logging.debug("Global journal not locked")
                return False

        # Write global event
        if self._global_enable and self._global_file:
            if not self._doWriteEvent(event, True, False):
                logging.debug("Failed global write event")
                return False

        # Write user event
        if self._user_enable and self._file:
            if not self._doWriteEvent(event, False, False):
                logging.debug("Failed user write event")
                return False

        return True

    def writeGlobalEvent(self, event, file, is_header_event=True):
        """Write an event to the global journal file.

        :param event: Event to be written.
        :param file: File object where to write event, defaults to global file.
        :param is_header_event: Whether to seek to the start of the file
            before writing event, defaults to True.
        """
        if not file:
            file = self._global_file

        if is_header_event:
            file.seek(0, os.SEEK_SET)

        return self._doWriteEventFile(event, file)

    def _doWriteEvent(self, event, is_global_event, is_header_event):
        if is_global_event:
            file = self._global_file
            lock = self._global_lock

        else:
            file = self._file
            lock = self._lock

        lock.acquire(LockType.WRITE_LOCK)

        whence = os.SEEK_SET if is_header_event else os.SEEK_END
        file.seek(0, whence)

        if is_global_event and self._checkGlobalRotation():
            # if the journal has been rotated, there is a new file and lock
            file = self._global_file
            lock = self._global_lock

        success = self._doWriteEventFile(event, file)

        file.flush()
        if ((is_global_event and self._global_fsync)
            or (not is_global_event and self._enable_fsync)):
            os.fsync(file.fileno())

        lock.release()

        return success

    def _doWriteEventFile(self, event, file):
        success = event.putEvent(file)
        if not success:
            file.write("\n")

        file.write(SYNC_DELIMITER)

        return success
