#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "File",
    "fd_open",
    "fd_close",
    "fd_get_flags",
    "fd_set_flags",
    ]

import os
import fcntl


class CheckException:

    __slots__ = (
        "_function",
        "_default",
        "_instance",
        )

    def __init__(self, function, default):
        self._function = function
        self._default = default
        self._instance = None

    def __get__(self, instance, cls=None):
        self._instance = instance
        return self

    def __call__(self, *args, **kwargs):
        try:
            return self._function(self._instance, *args, **kwargs)

        except EnvironmentError, error:
            self._instance.errno = error.errno
            return self._default


def check_exception(default=None):

    def wrapper(func):
        return CheckException(func, default)

    return wrapper


class File:

    __slots__ = (
        "info",
        "errno",
        )

    def __init__(self):
        self.info = None
        self.errno = 0

    @classmethod
    def create_from_path(cls, *args, **kwargs):
        file = cls()
        file.initialize_from_path(*args, **kwargs)

        return file

    @classmethod
    def create_from_fd(cls, *args, **kwargs):
        file = cls()
        file.initialize_from_fd(*args, **kwargs)

        return file

    @check_exception(-1)
    def initialize_from_path(self, *args, **kwargs):
        self.info = open(*args, **kwargs)
        return 0

    @check_exception(-1)
    def initialize_from_fd(self, *args, **kwargs):
        self.info = os.fdopen(*args, **kwargs)
        return 0

    def copy_from(self, other):
        self.info = other.info

    def reset(self):
        self.info = None
        self.errno = 0

    @check_exception(None)
    def read(self, size):
        # Return something useful
        return os.read(self.info.fileno(), size)

    @check_exception(-1)
    def write(self, buffer):
        # Return something useful
        return os.write(self.info.fileno(), buffer)

    @check_exception(None)
    def __getattr__(self, name):
        return getattr(self.info, name)

    def __setattr__(self, name, value):
        if name in File.__slots__:
            return super(File, self).__setattr__(name, value)

        return setattr(self.info, name, value)

    def __nonzero__(self):
        return self.info is not None


def fd_open(path, flag, *args, **kwargs):
    try:
        return os.open(path, flag, **kwargs)
    except OSError:
        return -1


def fd_close(fd):
    try:
        os.close(fd)
        return 0
    except OSError:
        return -1


def fd_get_flags(fd):
    try:
        return fcntl.fcntl(fd, fcntl.F_GETFL)
    except IOError:
        return -1


def fd_set_flags(fd, flags):
    try:
        return fcntl.fcntl(fd, fcntl.F_SETFL, flags)
    except IOError:
        return -1
