#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Lock",
    "LockType",
    "lock_file",
    ]

import os
try:
    os.SEEK_SET
except AttributeError:
    os.SEEK_SET, os.SEEK_CUR, os.SEEK_END = range(3)

import errno
import fcntl
import struct

from checkbox.lib.enum import Enum


LockType = Enum(
    "READ_LOCK",
    "WRITE_LOCK",
    "UN_LOCK",
    )


class Lock:

    __slots__ = (
        "_blocking",
        "_fd",
        "_file",
        "_type",
        )

    def __init__(self, fd=-1, file=None):
        self._type = LockType.UN_LOCK
        self.initialize(fd, file)

    def __del__(self):
        if self._type != LockType.UN_LOCK:
            self.release()

    def __nonzero__(self):
        return self._fd != -1

    def initialize(self, fd=-1, file=None):
        """Initialize lock.

        :param fd: File descriptor.
        :param file: Optional file object.
        """
        self.reset()
        self._fd = fd
        self._file = file

    def reset(self):
        """Reset the lock."""
        if self._type != LockType.UN_LOCK:
            self.release()

        self._blocking = True
        self._type = LockType.UN_LOCK
        self._fd = -1
        self._file = None

    @property
    def is_locked(self):
        """Return false if type is UN_LOCK."""
        return self._type != LockType.UN_LOCK

    @property
    def is_unlocked(self):
        """Return true if type is UN_LOCK."""
        return self._type == LockType.UN_LOCK

    def getBlocking(self):
        """Get the blocking state of the lock."""
        return self._blocking

    def setBlocking(self, blocking):
        """Set the blocking state of the lock.

        :param blocking: New blocking state."""
        self._blocking = blocking

    def getType(self):
        """Get the current type of the lock."""
        return self._type

    def acquire(self, type):
        """Acquire a lock.

        :param type: Type of the lock to acquire.
        """
        if self._fd == -1:
            return False

        if self._file:
            position_before_lock = self._file.tell()

        status = lock_file(self._fd, type, self._blocking)

        if self._file:
            self._file.seek(position_before_lock)

        if status == 0:
            self._type = type
            return True
        else:
            return False

    def release(self):
        """Release a lock."""
        return self.acquire(LockType.UN_LOCK)


def lock_file(fd, type, blocking):
    opt = fcntl.F_SETLKW if blocking else fcntl.F_SETLK

    if type == LockType.READ_LOCK:
        l_type = fcntl.F_RDLCK
    elif type == LockType.WRITE_LOCK:
        l_type = fcntl.F_WRLCK
    elif type == LockType.UN_LOCK:
        l_type = fcntl.F_UNLCK
    else:
        # Unknown lock type
        return -1

    l_whence = os.SEEK_SET
    l_start = 0
    l_len = 0

    arg = struct.pack(
        "hhllhh", l_type, l_whence, l_start, l_len, 0, 0)

    while True:
        try:
            fcntl.fcntl(fd, opt, arg)
        except IOError, e:
            if e.errno == errno.EINTR:
                break
            else:
                return -1
        else:
            break

    return 0
