#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "is_ipaddress",
    "is_priv_net",
    "gethostbyaddr",
    "gethostbyname",
    "gethostname",
    "getfullhostname",
    ]

import socket
import struct
import logging

from checkbox.lib import param


def is_ipaddress(address, allow_wildcard=True):
    ip_addr = []
    mask_addr = []

    # Check length of address
    if len(address) < 1 or len(address) > 15:
        return (None, None)

    # Strip trailing wild card or dot
    if address[-1] in ("*", "."):
        if len(address) > 1 and address[-2] == ".":
            address = address[:-2]
        else:
            address = address[:-1]

    # Validate the address and count the parts
    for part in address.split("."):
        for c in part:
            if c < "0" or c > "9":
                return (None, None)

        number = int(part)
        if number < 0 or number > 255:
            return (None, None)

        ip_addr.append(part)
        mask_addr.append("255")

    if len(ip_addr) > 4:
        return (None, None)

    elif not allow_wildcard and len(ip_addr) != 4:
        return (None, None)

    # Pad the ip_addr and mask_addr lists
    ip_addr += ["255"] * (4 - len(ip_addr))
    mask_addr += ["0"] * (4 - len(mask_addr))

    return ".".join(ip_addr), ".".join(mask_addr)


def is_priv_net(address):
    net_ip = socket.inet_aton(address)
    int_ip = struct.unpack("!L", net_ip)[0]
    return ((int_ip & 0xFF000000) == 0x0A000000     # 10/8
           or (int_ip & 0xFFF00000) == 0xAC100000   # 172.16/12
           or (int_ip & 0xFFFF0000) == 0xC0A80000)  # 192.168/16


def _dns_enabled():
    return param.getBoolean("DNS_ENABLED", True)


def _convert_ip_to_hostname(ip):
    default_domain_name = param.get("DEFAULT_DOMAIN_NAME")
    if default_domain_name is None:
        logging.info("DEFAULT_DOMAIN_NAME must be defined")
        return None

    ip = ip.replace(".", "-")
    return ".".join((ip, default_domain_name))


def _convert_hostname_to_addr(name):
    default_domain_name = param.get("DEFAULT_DOMAIN_NAME")
    if default_domain_name is None:
        logging.info("DEFAULT_DOMAIN_NAME must be defined")
        return None

    index = name.find(default_domain_name)
    if index != -1:
        name = name[:index - 1]

    name = name.replace("-", ".")

    try:
        ip = socket.inet_aton(name)
    except socket.error:
        return None

    return socket.inet_ntoa(ip)


def gethostbyaddr(addr):
    if _dns_enabled():
        try:
            return socket.gethostbyaddr(addr)
        except socket.gaierror:
            return (None, [], [addr])

    else:
        return (_convert_ip_to_hostname(addr), [], [addr])


def gethostbyname(name):
    if _dns_enabled():
        try:
            return socket.gethostbyname_ex(name)
        except socket.gaierror, error:
            logging.debug(
                "Failed to get host by name %s, errno %d %s",
                name, error.errno, error.strerror)
            return (name, [], [])

    else:
        addresses = []
        address = _convert_hostname_to_addr(name)
        if address is not None:
            addresses.append(address)

        return (name, [], addresses)


def gethostname():
    if _dns_enabled():
        return socket.gethostname()

    # First, try NETWORK_INTERFACE
    network_interface = param.get("NETWORK_INTERFACE")
    if network_interface is not None:
        ip_addr, mask_addr = is_ipaddress(network_interface)
        if ip_addr is None:
            logging.info("NETWORK_INTERFACE is invalid: %s",
                network_interface)
            return None

        return _convert_ip_to_hostname(ip_addr)

    # Second, try COLLECTOR_HOST
    collector_host = param.get("COLLECTOR_HOST")
    if collector_host is not None:
        # Get only the name portion
        index = collector_host.find(":")
        if index != -1:
            collector_host = collector_host[:index]

        collector_addresses = gethostbyname(collector_host)[2]
        if not collector_addresses:
            logging.info("Failed to get address of collector host: %s",
                collector_host)
            return None

        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        except socket.error, error:
            logging.info("Failed to create socket, errno %d", error.errno)
            return None

        try:
            s.connect((collector_addresses[0], 1980))
        except socket.error, error:
            logging.info("Failed to bind socket, errno %d", error.errno)
            return None

        ip, port = s.getsockname()
        return _convert_ip_to_hostname(ip)

    # Last, try gethostname()
    tmp = socket.gethostname()
    name = socket.gethostbyname(tmp)
    return _convert_ip_to_hostname(name)


def getfullhostname(host):
    fullhostname, aliases, addresses = gethostbyname(host)
    if not addresses:
        return (None, None)

    found = False
    ip_address = None
    for address in addresses:
        if address == "127.0.0.1":
            if not found:
                ip_address = address
                found = True
            continue

        ip_address = address
        found = True
        if not is_priv_net(address):
            break

    if "." in fullhostname:
        return (fullhostname, ip_address)

    for alias in aliases:
        if "." in alias:
            return (alias, ip_address)

    default_domain_name = param.get("DEFAULT_DOMAIN_NAME")
    if default_domain_name is not None:
        if default_domain_name[0] == ".":
            default_domain_name = default_domain_name[1:]

        host = ".".join([host, default_domain_name])
        return (host, ip_address)

    logging.debug("Failed to find full hostname for %s", host)
    return (host, ip_address)
