/*
#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
*/
#include <Python.h>
#include <signal.h>


static PyObject *
sigprocmask_function(PyObject* self, PyObject* args)
{
    PyObject *sig, *old_sigobj, *new_sigobj = Py_None;
    sigset_t old_sigset, new_sigset, *new_sigset_ptr;
    int how, i;

    /* Parse arguments. */
    if (!PyArg_ParseTuple(args, "i|O", &how, &new_sigobj))
        return NULL;

    /* Prepare new signal set. */
    if (new_sigobj == Py_None)
        new_sigset_ptr = NULL;
    else {
        new_sigset_ptr = &new_sigset;
        sigemptyset(&new_sigset);
        Py_ssize_t size = 0;
        while (_PySet_Next(new_sigobj, &size, &sig))
            if (sigaddset(&new_sigset, (int)PyInt_AS_LONG(sig)) == -1) {
                PyErr_SetFromErrno(PyExc_OSError);
                return NULL;
            }
    }

    /* Run sigprocmask. */
    if (sigprocmask(how, new_sigset_ptr, &old_sigset) == -1) {
        PyErr_SetFromErrno(PyExc_OSError);
        return NULL;
    }

    /* Prepare old signal obj. */
    old_sigobj = PySet_New(NULL);
    for (i = 1; i < _NSIG; i++) {
        int ret = sigismember(&old_sigset, i);
        if (ret == -1) {
            PyErr_SetFromErrno(PyExc_OSError);
            return NULL;
        }

        if (ret == 1) {
            sig = PyInt_FromLong(i);
            Py_INCREF(sig);

            PySet_Add(old_sigobj, sig);
        }
    }

    Py_INCREF(old_sigobj);
    return old_sigobj;
}


static PyMethodDef sigprocmask_methods[] =
{
    {"sigprocmask", sigprocmask_function, METH_VARARGS,
     "Examine and change blocked signals."},
    {NULL, NULL, 0, NULL}
};


DL_EXPORT(void)
initsigprocmask(void)
{
    PyObject *dict, *module, *object;

    module = Py_InitModule3("sigprocmask", sigprocmask_methods, "");
    if (module == NULL)
        return;

    /* Add some symbolic constants to the module */
    dict = PyModule_GetDict(module);

#define REGISTER_NAME(name) \
    do { \
        object = PyInt_FromLong((long)name); \
        if (!object || PyDict_SetItemString(dict, #name, object) < 0) \
            goto finally; \
        Py_DECREF(object); \
    } while (0)

    REGISTER_NAME(SIG_BLOCK);
    REGISTER_NAME(SIG_UNBLOCK);
    REGISTER_NAME(SIG_SETMASK);

  finally:

    return;
}
