#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

__metaclass__ = type

__all__ = [
    "Sigset",
    ]

from signal import NSIG

from checkbox.lib.sigprocmask import (
    SIG_BLOCK,
    SIG_SETMASK,
    SIG_UNBLOCK,
    sigprocmask,
    )


class Sigset(set):

    __slots__ = (
        "_errno",
        )

    def __init__(self, *args):
        super(Sigset, self).__init__(*args)
        self._errno = 0

    @property
    def errno(self):
        """Get the errno return by the last process call."""
        return self._errno

    def add(self, sig):
        """Add the given signal to the set.

        :param sig: Signal to add.
        """
        if sig <= 0:
            raise Exception("Signal value too low: %d" % sig)

        elif sig >= NSIG:
            raise Exception("Signal value too high: %d" % sig)

        return super(Sigset, self).add(sig)

    def fill(self):
        """Fill the set with all signals."""
        for sig in xrange(1, NSIG):
            self.add(sig)

    def process(self, how):
        """Process the signal mask from the set.

        :param how: How to process the signal mask.
        """
        self._errno = 0

        try:
            old = sigprocmask(how, self)
        except OSError, e:
            self._errno = e.errno
            return None

        return Sigset(old)

    def block(self):
        """The signals in the set are added from the current mask."""
        return self.process(SIG_BLOCK)

    def unblock(self):
        """The signals in the set are removed from the current mask."""
        return self.process(SIG_UNBLOCK)

    def setmask(self):
        """Set the mask to the signal set."""
        return self.process(SIG_SETMASK)
