#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Stat",
    "StatOpType",
    ]

import os

from checkbox.lib.enum import Enum


StatOpType = Enum(
    "OP_NONE",              # No operation
    ("OP_MIN", "OP_NONE"),  # Min operation
    "OP_STAT",              # stat()
    "OP_LSTAT",             # lstat()
    "OP_BOTH",              # Both stat() and lstat()
    "OP_FSTAT",             # fstat()
    "OP_ALL",               # All of stat(), lstat() and fstat()
    "OP_LAST",              # Last operation
    ("OP_MAX", "OP_LAST"),  # Max operation
    )


class StatResult:

    __slots__ = (
        "st_mode",
        "st_ino",
        "st_dev",
        "st_nlink",
        "st_uid",
        "st_gid",
        "st_size",
        "st_atime",
        "st_mtime",
        "st_ctime",
        )

    def __init__(self, st_mode=0, st_ino=0, st_dev=0, st_nlink=0, st_uid=0,
            st_gid=0, st_size=0, st_atime=0, st_mtime=0, st_ctime=0):
        """Constructor calling the L{initialize()} method with all
           parameters set to 0 by default."""
        self.initialize(st_mode, st_ino, st_dev, st_nlink, st_uid,
            st_gid, st_size, st_atime, st_mtime, st_ctime)

    def initialize(self, st_mode, st_ino, st_dev, st_nlink, st_uid,
            st_gid, st_size, st_atime, st_mtime, st_ctime):
        """Initialize the stat result parameters.

        :param st_mode: protection
        :param st_ino: inode number
        :param st_dev: ID of device containing file
        :param st_nlink: number of hard links
        :param st_uid: user ID of owner
        :param st_gid: group ID of owner
        :param st_size: total size, in bytes
        :param st_atime: time of last access
        :param st_mtime: time of last modification
        :param st_ctime: time of last status change
        """
        self.st_mode = st_mode
        self.st_ino = st_ino
        self.st_dev = st_dev
        self.st_nlink = st_nlink
        self.st_uid = st_uid
        self.st_gid = st_gid
        self.st_size = st_size
        self.st_atime = st_atime
        self.st_mtime = st_mtime
        self.st_ctime = st_ctime

    def copy_from(self, other):
        """Copy the parameters from another stat result object into this one.

        :param other: Other stat result object.
        """
        self.st_mode = other.st_mode
        self.st_ino = other.st_ino
        self.st_dev = other.st_dev
        self.st_nlink = other.st_nlink
        self.st_uid = other.st_uid
        self.st_gid = other.st_gid
        self.st_size = other.st_size
        self.st_atime = other.st_atime
        self.st_mtime = other.st_mtime
        self.st_ctime = other.st_ctime


class StatBase:

    __slots__ = (
        "_name",
        "_result",
        "_result_valid",
        "_valid",
        "_status",
        "_errno",
        )

    def __init__(self, name):
        self._name = name
        self._result = StatResult()
        self._result_valid = False
        self._valid = False
        self._status = 0
        self._errno = 0

    def copy_from(self, other):
        self._name = other._name
        self._result = other._result
        self._result_valid = other._result_valid
        self._valid = other._valid
        self._status = other._status
        self._errno = other._errno

    def is_valid(self):
        return self._valid

    def is_result_valid(self):
        return self._result_valid

    def get_result(self):
        return self._result if self._valid else None

    def set_result(self, result):
        result.copy_from(self._result)
        return self._valid

    def get_name(self):
        return self._name

    def get_errno(self):
        return self._errno

    def get_status(self):
        return self._status

    def _set_status(self, status):
        self._status = status
        return status

    def _run_function(self, function, *args, **kwargs):
        try:
            result = list(function(*args, **kwargs))
            self._result.initialize(*result)
            self._result_valid = True
            self._errno = self._status = 0
        except OSError, e:
            self._result_valid = False
            self._errno = e.errno
            self._status = -1

        return self._status


class StatPath(StatBase):

    __slots__ = (
        "_function",
        "_path",
        )

    def __init__(self, name, function):
        super(StatPath, self).__init__(name)
        self._function = function
        self._path = None

    def copy_from(self, other):
        super(StatPath, self).copy_from(other)
        self._function = other._function
        self._path = other._path

    def set_path(self, path):
        self._path = path
        self._valid = path is not None
        self._result_valid = False
        self._status = 0
        return True

    def stat(self, force):
        if not self._function:
            return self._set_status(-2)

        if not self._path:
            return self._set_status(-3)

        if self._valid and not force:
            return self.get_status()

        return self._run_function(self._function, self._path)


class StatFd(StatBase):

    __slots__ = (
        "_function",
        "_fd",
        )

    def __init__(self, name, function):
        super(StatFd, self).__init__(name)
        self._function = function
        self._fd = -1

    def copy_from(self, other):
        super(StatFd, self).copy_from(other)
        self._function = other._function
        self._fd = other._fd

    def set_fd(self, fd):
        if fd != self._fd:
            self._result_valid = False
            self._status = 0

        self._valid = fd >= 0
        self._fd = fd
        return True

    def stat(self, force):
        if not self._function:
            return self._set_status(-2)

        if self._fd < 0:
            return self._set_status(-3)

        if self._valid and not force:
            return self.get_status()

        return self._run_function(self._function, self._fd)


class StatNop(StatBase):

    __slots__ = (
        "_function",
        )

    def __init__(self, name, function):
        super(StatNop, self).__init__(name)
        self._function = function

    def copy_from(self, other):
        super(StatNop, self).copy_from(other)
        self._function = other._function

    def stat(self, force):
        return 0


class StatOp:

    __slots__ = (
        "_stat",
        "_lstat",
        "_fstat",
        "_all",
        "_primary",
        )

    def __init__(self, stat, lstat, fstat, primary):
        self._stat = stat
        self._lstat = lstat
        self._fstat = fstat

        self._all = [stat, lstat, fstat]

        self._primary = primary

    def get_primary(self):
        return self._primary

    def get_all(self):
        return self._all

    def stat_all(self, force):
        self._stat.stat(force)
        self._lstat.stat(force)
        self._fstat.stat(force)

        if self._stat.get_status():
            return self._stat.get_status()

        elif self._lstat.get_status():
            return self._lstat.get_status()

        elif self._fstat.get_status():
            return self._fstat.get_status()

        else:
            return 0


class Stat:

    __slots__ = (
        "_nop",
        "_stat",
        "_lstat",
        "_fstat",
        "_ops",
        "_last_op",
        "_last_op_type",
        )

    def __init__(self):
        self._nop = nop = StatNop(None, None)
        self._stat = stat = StatPath("stat", os.stat)
        self._lstat = lstat = StatPath("lstat", os.lstat)
        self._fstat = fstat = StatFd("fstat", os.fstat)

        self._ops = (
            StatOp(nop,  nop,   nop,   nop),    # OP_NONE
            StatOp(stat, nop,   nop,   stat),   # OP_STAT
            StatOp(nop,  lstat, nop,   lstat),  # OP_LSTAT
            StatOp(stat, lstat, nop,   nop),    # OP_BOTH
            StatOp(nop,  nop,   fstat, fstat),  # OP_FSTAT
            StatOp(stat, lstat, fstat, nop),    # OP_ALL
            StatOp(nop,  nop,   nop,   nop))    # OP_LAST

        self._last_op = self._ops[StatOpType.OP_NONE]
        self._last_op_type = StatOpType.OP_NONE

    @classmethod
    def create_from_path(cls, path, op_type=StatOpType.OP_STAT):
        stat = cls()

        if op_type != StatOpType.OP_NONE:
            stat.stat_path(path, op_type, True)
        else:
            stat._stat.set_path(path)
            stat._lstat.set_path(path)

        return stat

    @classmethod
    def create_from_fd(cls, fd, op_type=StatOpType.OP_FSTAT):
        stat = cls()

        if op_type != StatOpType.OP_NONE:
            stat.stat_fd(fd, op_type, True)
        else:
            stat._fstat.set_fd(fd)

        return stat

    @property
    def last_op_type(self):
        return self._last_op_type

    def set_path(self, path):
        status = True
        if not self._stat.set_path(path):
            status = False

        if not self._lstat.set_path(path):
            status = False

        return status

    def set_fd(self, fd):
        return self._fstat.set_fd(fd)

    def stat(self, op_type=StatOpType.OP_STAT, force=True):
        op = self._ops[op_type]
        self._last_op = op
        self._last_op_type = op_type

        return op.stat_all(force)

    def stat_path(self, path, op_type=StatOpType.OP_STAT, force=True):
        if not self.set_path(path):
            return -1

        return self.stat(op_type, force)

    def stat_fd(self, fd, force=True):
        if not self.set_fd(fd):
            return -1

        return self.stat(StatOpType.OP_FSTAT, force)

    def retry(self):
        return self._last_op.stat_all(True)

    def is_valid(self):
        return self._stat.is_valid() or self._fstat.is_valid()

    def is_result_valid(self, op_type=StatOpType.OP_LAST):
        stat = self._get_stat(op_type)
        if not stat:
            return False

        return stat.is_result_valid()

    def get_result(self, op_type=StatOpType.OP_LAST):
        stat = self._get_stat(op_type)
        if not stat:
            return None

        return stat.get_result()

    def set_result(self, result, op_type=StatOpType.OP_LAST):
        stat = self._get_stat(op_type)
        if not stat:
            return False

        return stat.set_result(result)

    def get_status(self, op_type=StatOpType.OP_LAST):
        stat = self._get_stat(op_type)
        if not stat:
            return -1

        return stat.get_status()

    def get_errno(self, op_type=StatOpType.OP_LAST):
        stat = self._get_stat(op_type)
        if not stat:
            return -1

        return stat.get_errno()

    def _get_stat(self, op_type):
        if op_type == StatOpType.OP_LAST:
            op_type = self._last_op_type

        if op_type < StatOpType.OP_MIN or op_type > StatOpType.OP_MAX:
            op_type = StatOpType.OP_NONE

        return self._ops[op_type].get_primary()
