#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Template",
    ]

import re
import logging


EXTENDED_STRING = "_extended"

LINE_RE = re.compile(r"^([-_.A-Za-z0-9]*):\s?(.*)$")
EXTENDED_APPEND_RE = re.compile(r"^.*[^\n ]$", re.M)
EXTENDED_LINE_RE = re.compile(r"^\s(.*)$")
EXTENDED_NEWLINE_RE = re.compile(r"^\s\.$")
EXTENDED_SPACE_RE = re.compile(r"^\s(\s+.*)$")


class Template:

    def _reader(self, file, size=4096, delimiter=r"\n{2,}"):
        buffer_old = ""
        while True:
            buffer_new = file.read(size)
            if not buffer_new:
                break

            lines = re.split(delimiter, buffer_old + buffer_new)
            buffer_old = lines.pop(-1)

            for line in lines:
                yield line

        yield buffer_old

    def dumpFile(self, elements, file, filename="<stream>"):
        for element in elements:
            for long_key in element.keys():
                if long_key.endswith(EXTENDED_STRING):
                    short_key = long_key.replace(EXTENDED_STRING, "")
                    del element[short_key]

            for key, value in element.items():
                if key.endswith(EXTENDED_STRING):
                    key = key.replace(EXTENDED_STRING, "")
                    file.write("%s:\n" % key)
                    for line in value.split("\n"):
                        file.write(" %s\n" % line)
                elif isinstance(value, (list, tuple)):
                    file.write("%s:\n" % key)
                    for v in value:
                        file.write(" %s\n" % v)
                else:
                    file.write("%s: %s\n" % (key, value))

            file.write("\n")

    def dumpFilename(self, elements, filename):
        logging.info("Dumping elements to filename: %s", filename)

        file = open(filename, "w")
        return self.dumpFile(elements, file, filename)

    def loadFile(self, file, filename="<stream>"):
        elements = []
        for string in self._reader(file):
            if not string:
                break

            element = {}

            def _save(field, value, extended):
                extended = extended.rstrip("\n")
                if field:
                    if field in element:
                        raise Exception(
                            "Template %s has a duplicate field '%s'"
                            " with a new value '%s'."
                            % (filename, field, value))
                    element[field] = value
                    if extended:
                        element["%s%s" % (field, EXTENDED_STRING)] = extended

            string = string.strip("\n")
            field = value = extended = ""
            for line in string.split("\n"):
                line.strip()
                match = LINE_RE.match(line)
                if match:
                    _save(field, value, extended)
                    field = match.groups()[0].lower()
                    value = match.groups()[1].rstrip()
                    extended = ""
                    continue

                if EXTENDED_NEWLINE_RE.match(line):
                    extended += "\n\n"
                    continue

                match = EXTENDED_SPACE_RE.match(line)
                if match:
                    bit = match.groups()[0].rstrip()
                    if EXTENDED_APPEND_RE.match(extended):
                        extended += "\n"

                    extended += bit + "\n"
                    continue

                match = EXTENDED_LINE_RE.match(line)
                if match:
                    bit = match.groups()[0].rstrip()
                    if EXTENDED_APPEND_RE.match(extended):
                        if extended.endswith("\\"):
                            extended = extended[:-1].rstrip() + " "
                        else:
                            extended += "\n"

                    extended += bit
                    continue

                raise Exception(
                    "Template %s parse error at: %s"
                    % (filename, line))

            _save(field, value, extended)

            elements.append(element)

        return elements

    def loadFilename(self, filename):
        logging.info("Loading elements from filename: %s", filename)

        file = open(filename, "r")
        return self.loadFile(file, filename)
