#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "TemplateI18n",
    ]

import re
import locale

from os import environ
from gettext import gettext as _

from checkbox.lib.template import Template


class TemplateI18n(Template):

    def __init__(self, *args, **kwargs):
        super(TemplateI18n, self).__init__(*args, **kwargs)

        self._languages = self._getLanguages()

    def _addModifier(self, language, modifier):
        return "%s.%s" % (language, modifier)

    def _addTerritory(self, language, territory):
        return re.sub(r"([^_@.]+)", "\\1%s" % territory, language)

    def _addCharset(self, language, charset):
        return re.sub(r"([^@.]+)", "\\1%s" % charset, language)

    def _mergeLists(self, primaries, secondaries):
        for primary, secondary in zip(primaries, secondaries):
            yield primary
            yield secondary

    def _getLanguageList(self, language):
        regex = re.compile(r"(@[^.]+)")
        modifier_match = regex.search(language)
        language = regex.sub("", language)

        language_match = re.match(r"([^_@.]+)(_[^_@.]+)?(\..+)?", language)
        if not language_match:
            raise Exception("Unknown language format: %s" % language)

        ret = [language_match.group(1)]
        if modifier_match:
            modifier = modifier_match.group(1)
            modifiers = [self._addModifier(r, modifier) for r in ret]
            ret = self._mergeLists(modifiers, ret)

        if language_match.group(2):
            territory = language_match.group(2)
            territories = [self._addTerritory(r, territory) for r in ret]
            ret = self._mergeLists(territories, ret)

        if language_match.group(3):
            charset = language_match.group(3)
            charsets = [self._addCharset(r, charset) for r in ret]
            ret = self._mergeLists(charsets, ret)

        return ret

    def _getLanguages(self):
        languages = []
        if "LANGUAGE" in environ and environ["LANGUAGE"]:
            for language in environ["LANGUAGE"].split(":"):
                if language:
                    languages.extend(self._getLanguageList(language))

        language = locale.setlocale(locale.LC_MESSAGES)
        languages.extend(self._getLanguageList(language))

        return [l.lower() for l in languages]

    def _filterField(self, field):
        lines = []
        separator = "\n\n"
        for line in field.split(separator):
            lines.append(_(line))

        return separator.join(lines)

    def _filterLanguages(self, element):
        filter = {}
        basekeys = {}
        for key in element.keys():
            basekey = re.sub(r"^_?([^-]+).*$", "\\1", key)
            basekeys[basekey] = None

        for key in basekeys.keys():
            if self._languages:
                for language in self._languages:
                    field = "%s-%s" % (key, language)
                    if field in element:
                        filter[key] = element[field]
                        break
                if key in filter:
                    continue
            else:
                field = "%s-c" % key
                if field in element:
                    filter[key] = element[field]
                    continue

            if key in element:
                filter[key] = self._filterField(element[key])
                continue
            else:
                field = "_%s" % key
                if field in element:
                    filter[key] = self._filterField(element[field])
                    continue

            raise Exception("No language found for key: %s" % key)

        return filter

    def load_file(self, *args, **kwargs):
        elements = super(TemplateI18n, self).load_file(*args, **kwargs)

        return [self._filterLanguages(element) for element in elements]
