#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os

from checkbox.tests.cases import TestCase
from checkbox.tests.env import EnvMixin

from checkbox.lib import net


class TestNet(EnvMixin, TestCase):

    def setUp(self):
        super(TestNet, self).setUp()

        for name in ("COLLECTOR_HOST", "DNS_ENABLED",
                "DEFAULT_DOMAIN_NAME", "NETWORK_INTERFACE"):
            if name in os.environ:
                del os.environ[name]

    def test_is_ipaddress(self):
        ip_addr, mask_addr = net.is_ipaddress("-1")
        self.assertEquals(ip_addr, None)
        self.assertEquals(mask_addr, None)

        ip_addr, mask_addr = net.is_ipaddress("256")
        self.assertEquals(ip_addr, None)
        self.assertEquals(mask_addr, None)

        ip_addr, mask_addr = net.is_ipaddress("1.2.3.4")
        self.assertNotEquals(ip_addr, None)
        self.assertNotEquals(mask_addr, None)
        self.assertEquals(ip_addr, "1.2.3.4")
        self.assertEquals(mask_addr, "255.255.255.255")

        ip_addr, mask_addr = net.is_ipaddress("1.2.3.*", False)
        self.assertEquals(ip_addr, None)
        self.assertEquals(mask_addr, None)

        ip_addr, mask_addr = net.is_ipaddress("1.2.3.*")
        self.assertNotEquals(ip_addr, None)
        self.assertNotEquals(mask_addr, None)
        self.assertEquals(ip_addr, "1.2.3.255")
        self.assertEquals(mask_addr, "255.255.255.0")

    def test_is_priv_net(self):
        self.assertTrue(net.is_priv_net("10.0.0.0"))
        self.assertTrue(net.is_priv_net("10.0.0.1"))

        self.assertTrue(net.is_priv_net("172.16.0.0"))
        self.assertTrue(net.is_priv_net("172.16.1.0"))

        self.assertTrue(net.is_priv_net("192.168.0.0"))
        self.assertTrue(net.is_priv_net("192.168.1.0"))

    def test_dns_enabled(self):
        self.assertTrue(net._dns_enabled())

        os.environ["DNS_ENABLED"] = "true"
        self.assertTrue(net._dns_enabled())

        os.environ["DNS_ENABLED"] = "false"
        self.assertFalse(net._dns_enabled())

    def test_convert_ip_to_hostname(self):
        ip = "1.2.3.4"
        hostname = net._convert_ip_to_hostname(ip)
        self.assertEquals(hostname, None)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"
        hostname = net._convert_ip_to_hostname(ip)
        self.assertEquals(hostname, "1-2-3-4.foo.com")

    def test_convert_hostname_to_addr(self):
        hostname = "1-2-3-4"
        ip = net._convert_hostname_to_addr(hostname)
        self.assertEquals(ip, None)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"

        hostname = "1-2-3-4"
        ip = net._convert_hostname_to_addr(hostname)
        self.assertEquals(ip, "1.2.3.4")

        hostname = "1-2-3-4.foo.com"
        ip = net._convert_hostname_to_addr(hostname)
        self.assertEquals(ip, "1.2.3.4")

        hostname = "1-2-3-4.bar.com"
        ip = net._convert_hostname_to_addr(hostname)
        self.assertEquals(ip, None)

    def test_gethostbyaddr(self):
        hostname, aliases, addresses = net.gethostbyaddr("127.0.0.1")
        self.assertEquals(hostname, "localhost")

        os.environ["DNS_ENABLED"] = "true"
        hostname, aliases, addresses = net.gethostbyaddr("127.0.0.1")
        self.assertEquals(hostname, "localhost")

        os.environ["DNS_ENABLED"] = "false"
        hostname, aliases, addresses = net.gethostbyaddr("127.0.0.1")
        self.assertEquals(hostname, None)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"
        hostname, aliases, addresses = net.gethostbyaddr("127.0.0.1")
        self.assertEquals(hostname, "127-0-0-1.foo.com")

    def test_gethostbyname(self):
        name, aliases, addresses = net.gethostbyname("localhost")
        self.assertNotEquals(len(addresses), 0)
        self.assertEquals(addresses[0], "127.0.0.1")

        os.environ["DNS_ENABLED"] = "true"
        name, aliases, addresses = net.gethostbyname("localhost")
        self.assertEquals(addresses[0], "127.0.0.1")

        os.environ["DNS_ENABLED"] = "false"
        name, aliases, addresses = net.gethostbyname("localhost")
        self.assertEquals(len(addresses), 0)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"
        name, aliases, addresses = net.gethostbyname("127-0-0-1")
        self.assertNotEquals(len(addresses), 0)
        self.assertEquals(addresses[0], "127.0.0.1")

        name, aliases, addresses = net.gethostbyname("127-0-0-1.foo.com")
        self.assertEquals(addresses[0], "127.0.0.1")

    def test_gethostname(self):
        name = net.gethostname()
        self.assertNotEquals(name, None)

        os.environ["DNS_ENABLED"] = "true"
        name = net.gethostname()
        self.assertNotEquals(name, None)

        os.environ["DNS_ENABLED"] = "false"
        name = net.gethostname()
        self.assertEquals(name, None)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"
        name = net.gethostname()
        self.assertNotEquals(name, None)

        os.environ["COLLECTOR_HOST"] = "127.0.0.1"
        name = net.gethostname()
        self.assertNotEquals(name, None)
        self.assertEquals(name, "127-0-0-1.foo.com")

        # NETWORK_INTERFACE should take precedence over COLLECTOR_HOST
        os.environ["NETWORK_INTERFACE"] = "1.2.3.4"
        name = net.gethostname()
        self.assertNotEquals(name, None)
        self.assertEquals(name, "1-2-3-4.foo.com")

    def test_getfullhostname(self):
        name, address = net.getfullhostname("localhost")
        self.assertEquals(name, "localhost")
        self.assertEquals(address, "127.0.0.1")

        os.environ["DNS_ENABLED"] = "true"
        name, address = net.getfullhostname("localhost")
        self.assertEquals(name, "localhost")
        self.assertEquals(address, "127.0.0.1")

        os.environ["DNS_ENABLED"] = "false"
        name, address = net.getfullhostname("localhost")
        self.assertEquals(name, None)
        self.assertEquals(address, None)

        os.environ["DEFAULT_DOMAIN_NAME"] = "foo.com"
        name, address = net.getfullhostname("127-0-0-1")
        self.assertEquals(name, "127-0-0-1.foo.com")
        self.assertEquals(address, "127.0.0.1")
