#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Function",
    ]

from checkbox.message.expression import Expression
from checkbox.message.literal import Literal
from checkbox.message.types import (
    Error,
    Undefined,
    error,
    undefined,
    )


class Function(Expression):

    __slots__ = (
        "function",
        "function_name",
        "arguments",
        )

    def __init__(self, function_name=None, *arguments):
        super(Function, self).__init__()
        self.function = FUNCTIONS.get(function_name)
        self.function_name = function_name
        self.arguments = list(arguments)

    def _setParentScope(self, parent):
        for argument in self.arguments:
            argument.setParentScope(parent)

    def copyFrom(self, other):
        super(Function, self).copyFrom(other)
        self.function_name = other.function_name
        self.function = other.function

        self.arguments = []
        for argument in other.arguments:
            self.arguments.append(argument.copy())

    def sameAs(self, other):
        if not super(Function, self).sameAs(other):
            return False

        if (self.function_name != other.function_name or
            self.function != other.function or
            len(self.arguments) != len(other.arguments)):
            return False

        for s, o in zip(self.arguments, other.arguments):
            if not s.sameAs(o):
                return False

        return True

    def _evaluate(self, state):
        if not self.function:
            return error

        return self.function(self.function_name, self.arguments, state)

    def _evaluateTree(self, state, tree=None):
        value = self._evaluate(state)

        temp_sig = Function()
        temp_sig.function_name = self.function_name
        for argument in self.arguments:
            arg_sig = argument._evaluateTree(state)
            temp_sig.arguments.append(arg_sig)

        return (value, temp_sig)

    def _flatten(self, state, kind):
        if not self.function:
            return (error, kind, None)

        new_function = Function(self.function_name)

        fold = True
        for argument in self.arguments:
            arg_value, arg_tree = argument.flatten(state)
            if not arg_tree:
                arg_tree = Literal(arg_value)
            else:
                fold = False

            new_function.arguments.append(arg_tree)

        if fold:
            value = self.function(self.function_name, self.arguments, state)
            tree = None
        else:
            value = undefined
            tree = new_function

        return (value, tree, kind)


FUNCTIONS = {}


def register_function(*names):
    def decorator(method):
        def function(name, args, state):
            return method(name, args, state)

        for name in names:
            FUNCTIONS[name] = function

        return function

    return decorator


@register_function("int")
def convert_int(name, args, state):
    if len(args) != 1:
        return error

    arg = args[0].evaluate(state)
    # TODO: real conversion
    return int(arg)


@register_function("real")
def convert_real(name, args, state):
    if len(args) != 1:
        return error

    arg = args[0].evaluate(state)
    # TODO: real conversion
    return float(arg)


@register_function("string")
def convert_string(name, args, state):
    if len(args) != 1:
        return error

    arg = args[0].evaluate(state)
    if isinstance(arg, (basestring, Undefined, Error)):
        return arg
    elif isinstance(arg, (dict, list, bool, int, float)):
        from checkbox.message.unparser import Unparser
        return Unparser.unparseExpression(args[0])
    else:
        raise Exception("Should not get here")


@register_function("bool")
def convert_bool(name, args, state):
    if len(args) != 1:
        return error

    arg = args[0].evaluate(state)
    if isinstance(arg, (bool, Undefined, Error)):
        return arg
    elif isinstance(arg, (dict, list)):
        return error
    elif isinstance(arg, int):
        return arg != 0
    elif isinstance(arg, float):
        return arg != 0.0
    elif isinstance(arg, basestring):
        return arg.lower() != "false" and arg != ""
    else:
        raise Exception("Should not get here")
