#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Lexer",
    ]

from StringIO import StringIO
from string import (
    digits,
    hexdigits,
    letters,
    octdigits,
    )

from checkbox.lib.enum import Enum

from checkbox.message.convert import convert_escapes
from checkbox.message.literal import LiteralFactor
from checkbox.message.token import (
    TokenType,
    Token,
    )


CHAR_TO_FACTOR = {
    "B": LiteralFactor.BYTE,
    "K": LiteralFactor.KILO,
    "M": LiteralFactor.MEGA,
    "G": LiteralFactor.GIGA,
    "T": LiteralFactor.TERA}

ALPHA_CHARS = letters + "_"
NUMBER_CHARS = digits + "."


class Lexer:

    __slots__ = (
        "_file",
        "_ch",
        "_buffer",
        "_buffer_count",
        "_in_string",
        "_token_consumed",
        "_token_type",
        "_token",
        "_accumulating",
        "_initialized",
        )

    space_characters = (
        " ",
        "\f",
        "\r",
        "\t",
        "\v",
        )

    def __init__(self, file=None):
        self._token_type = TokenType.END_OF_INPUT
        self._token = Token()

        if file:
            self.initialize(file)
        else:
            self._ch = ""
            self._buffer = None
            self._buffer_count = 0
            self._in_string = False
            self._token_consumed = True
            self._accumulating = False
            self._initialized = False

    def _mark(self):
        self._buffer = self._ch
        self._buffer_count = 0
        self._accumulating = True

    def _cut(self):
        self._buffer = self._buffer[:self._buffer_count]
        self._accumulating = False

    def _wind(self):
        if self._ch == "":
            return

        # consume character from file
        self._ch = self._file.read(1)
        self._buffer_count += 1
        if self._ch == "":
            return

        if self._accumulating:
            self._buffer += self._ch

    @property
    def was_initialized(self):
        return self._initialized

    def initialize(self, file):
        if isinstance(file, basestring):
            file = StringIO(file)

        self._file = file
        self._ch = file.read(1)
        self._buffer = self._ch
        self._buffer_count = 0
        self._in_string = False
        self._token_consumed = True
        self._accumulating = False
        self._initialized = True

    def consumeToken(self, token=None):
        if token:
            token.copyFrom(self._token)

        if self._token_consumed:
            self.peekToken(token)

        self._token_consumed = True
        return self._token_type

    def peekToken(self, token=None):
        if not self._token_consumed:
            if token:
                token.copyFrom(self._token)
            return self._token_type

        self._token_consumed = False
        while self._ch in self.space_characters:
            self._wind()

        # TODO: add support for comments

        if self._ch == "":
            self._token_type = TokenType.END_OF_INPUT
            self._token.type = self._token_type
            return self._token_type

        elif self._ch in ALPHA_CHARS:
            self._tokenizeAlpha()

        elif self._ch in NUMBER_CHARS:
            self._tokenizeNumber()

        elif self._ch == "\"":
            self._tokenizeString("\"")

        elif self._ch == "'":
            self._tokenizeString("'")

        else:
            self._tokenizePunctuation()

        if token:
            token.copyFrom(self._token)

        self._token.type = self._token_type
        return self._token_type

    def _tokenizeNumber(self):
        number = Enum("NONE", "INTEGER", "REAL")
        number_type = number.NONE
        och = self._ch

        self._mark()
        self._wind()

        if och == "0":
            # number is octal, hex or real
            if self._ch.lower() == "x":
                number_type = number.INTEGER
                self._wind()
                if self._ch not in hexdigits:
                    self._cut()
                    self._token_type = TokenType.TOKEN_ERROR
                    return self._token_type

                while self._ch in hexdigits:
                    self._wind()

            else:
                number_type = number.INTEGER
                while self._ch.is_digit():
                    self._wind()
                    if self._ch not in octdigits:
                        number_type = number.REAL

                if self._ch == "." or self._ch.lower() == "e":
                    number_type = number.REAL

                elif number_type == number.REAL:
                    self._cut()
                    self._token_type = TokenType.TOKEN_ERROR
                    return self._token_type

        elif och.isdigit():
            while self._ch.isdigit():
                self._wind()

            if self._ch == "." or self._ch.lower() == "e":
                number_type = number.REAL
            else:
                number_type = number.INTEGER

        if och == "." or self._ch == ".":
            if self._ch == ".":
                self._wind()

            if self._ch.isdigit():
                number_type == number.REAL
                while self._ch.isdigit():
                    self._wind()

            else:
                if number_type != number.NONE:
                    self._cut()
                    self._token_type = TokenType.TOKEN_ERROR
                    return self._token_type

                self._cut()
                self._token_type = TokenType.SELECTION
                return self._token_type

        if number_type == number.REAL and self._ch.lower() == "e":
            self._wind()
            if self._ch == "+" or self == "-":
                self._wind()

            if self._ch.isdigit():
                self._cut()
                self._token_type == TokenType.TOKEN_ERROR
                return self._token_tpye

            while self._ch.isdigit():
                self._wind()

        if number_type == number.INTEGER:
            self._cut()
            self._token.value = int(self._buffer)
            self._token.type = self._token_type = TokenType.INTEGER

        elif number_type == number.REAL:
            self._cut()
            self._token.value = float(self._buffer)
            self._token.type = self._token_type = TokenType.REAL

        else:
            raise Exception("Unsupported number")

        if self._ch.upper() in CHAR_TO_FACTOR:
            self._token.factor = CHAR_TO_FACTOR[self._ch.upper()]
            self._wind()
        else:
            self._token.factor = LiteralFactor.NONE

        return self._token_type

    def _tokenizeAlpha(self):
        self._mark()
        while self._ch.isalpha():
            self._wind()

        # Check if the token is an identifier
        if self._ch.isdigit() or self._ch == "_":
            self._wind()
            while self._ch.isalnum() or self._ch == "_":
                self._wind()

            self._cut()

            self._token_type = TokenType.IDENTIFIER
            self._token.value = self._buffer

            return self._token_type

        # Check if the token is a reserved word
        self._cut()
        buffer_lower = self._buffer.lower()
        if buffer_lower == "true":
            self._token_type = TokenType.BOOLEAN
            self._token.value = True
        elif buffer_lower == "false":
            self._token_type = TokenType.BOOLEAN
            self._token.value = False
        elif buffer_lower == "null":
            self._token_type = TokenType.NULL
        elif buffer_lower == "undefined":
            self._token_type = TokenType.UNDEFINED
        elif buffer_lower == "error":
            self._token_type = TokenType.ERROR
        elif buffer_lower == "in":
            self._token_type = TokenType.IN
        elif buffer_lower == "is":
            self._token_type = TokenType.EQUAL
        elif buffer_lower == "isnt":
            self._token_type = TokenType.NOT_EQUAL
        else:
            self._token_type = TokenType.IDENTIFIER
            self._token.value = self._buffer

        return self._token_type

    def _tokenizeString(self, delimiter):
        self._in_string = True
        self._wind()
        self._mark()

        while True:
            backwards = False
            old_ch = ""

            # Consume the string literal
            while (self._ch != "" and
                (self._ch != delimiter or
                (self._ch == delimiter and old_ch == "\\" and backwards))):
                if not backwards and self._ch == "\\":
                    backwards = True
                else:
                    backwards = False

                old_ch = self._ch
                self._wind()

            if self._ch == delimiter:
                temp_ch = " "
                while temp_ch in self.space_characters:
                    temp_ch = self._file.read(1)

                # Check if a token exists after the string
                if temp_ch != delimiter:
                    if temp_ch != "":
                        self._file.seek(-1, 1)
                    break

                # Concatenate existing string to the adjacent one
                else:
                    self._buffer = self._buffer[:self._buffer_count]
                    self._buffer_count -= 1
                    self._wind()

            else:
                self._token_type = TokenType.TOKEN_ERROR
                return self._token_type

        self._cut()
        self._wind()

        self._buffer = convert_escapes(self._buffer)
        self._token.value = self._buffer
        if self._buffer:
            if delimiter == "\"":
                self._token_type = TokenType.STRING
            else:
                self._token_type = TokenType.IDENTIFIER
        else:
            self._token_type = TokenType.TOKEN_ERROR

        return self._token_type

    def _tokenizePunctuation(self):
        ch_to_type = {
            ".": TokenType.SELECTION,
            "*": TokenType.MULTIPLY,
            "/": TokenType.DIVIDE,
            "%": TokenType.MODULUS,
            "+": TokenType.PLUS,
            "-": TokenType.MINUS,
            "+": TokenType.SELECTION,
            "~": TokenType.BITWISE_NOT,
            "^": TokenType.BITWISE_XOR,
            "?": TokenType.QMARK,
            ":": TokenType.COLON,
            ";": TokenType.SEMICOLON,
            ",": TokenType.COMMA,
            "[": TokenType.OPEN_BRACKET,
            "]": TokenType.CLOSE_BRACKET,
            "(": TokenType.OPEN_PAREN,
            ")": TokenType.CLOSE_PAREN,
            "{": TokenType.OPEN_BRACE,
            "}": TokenType.CLOSE_BRACE}

        old_ch = self._ch
        self._mark()
        self._wind()

        if old_ch in ch_to_type:
            self._token_type = ch_to_type[old_ch]

        # The following cases need lookaheads
        elif old_ch == "\n":
            if self._ch == "\n":
                self._token_type = TokenType.NEW_PARAGRAPH
                while self._ch == "\n":
                    self._wind()
                if self._ch == "":
                    self._token_type = TokenType.END_OF_INPUT
            else:
                self._token_type = TokenType.NEW_LINE

        elif old_ch == "&":
            if self._ch == "&":
                self._token_type = TokenType.LOGICAL_AND
                self._wind()
            else:
                self._token_type = TokenType.BITWISE_AND

        elif old_ch == "|":
            if self._ch == "|":
                self._token_type = TokenType.LOGICAL_OR
                self._wind()
            else:
                self._token_type = TokenType.BITWISE_OR

        elif old_ch == "<":
            if self._ch == "=":
                self._token_type = TokenType.LESS_OR_EQUAL
                self._wind()

            elif self._ch == "<":
                self._token_type = TokenType.LEFT_SHIFT
                self._wind()

            else:
                self._token_type = TokenType.LESS_THAN

        elif old_ch == ">":
            if self._ch == "=":
                self._token_type = TokenType.GREATER_OR_EQUAL
                self._wind()

            elif self._ch == ">":
                self._token_type = TokenType.RIGHT_SHIFT
                self._wind()
                if self._ch == ">":
                    self._token_type = TokenType.URIGHT_SHIFT
                    self._wind()

            else:
                self._token_type = TokenType.GREATER_THAN

        elif old_ch == "=":
            if self._ch == "=":
                self._token_type = TokenType.EQUAL
                self._wind()

            elif self._ch == "~":
                self._token_type = TokenType.META_EQUAL
                self._wind()

            else:
                self._token_type = TokenType.BOUND_TO

        elif old_ch == "!":
            if self._ch == "=":
                self._token_type = TokenType.NOT_EQUAL
                self._wind()

            elif self._ch == "~":
                self._token_type = TokenType.META_NOT_EQUAL
                self._wind()

            else:
                self._token_type = TokenType.LOGICAL_NOT

        else:
            self._token_type = TokenType.TOKEN_ERROR

        self._cut()

        return self._token_type
