#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Operator",
    ]

from checkbox.lib.enum import Enum

from checkbox.message.expression import Expression
from checkbox.message.literal import Literal
from checkbox.message.types import (
    Error,
    Exceptional,
    Undefined,
    error,
    undefined,
    )


OperatorKind = Enum(
     "__NO_OP__",                               # No op
     "__FIRST_OP__",

    ("__COMPARISON_START__", "__FIRST_OP__"),

    # Strict comparison operators
    ("LESS_THAN_OP", "__COMPARISON_START__"),   # Less than operator
     "LESS_OR_EQUAL_OP",                        # Less or equal
     "NOT_EQUAL_OP",                            # Not equal (same as ISNT)
     "EQUAL_OP",                                # Equal (same as IS)
     "GREATER_OR_EQUAL_OP",                     # Greater or equal
     "GREATER_THAN_OP",                         # Greater than

    # Non-strict comparison operators
     "META_EQUAL_OP",                           # Meta-equal
     ("IS_OP", "META_EQUAL_OP"),                # Is
     "META_NOT_EQUAL_OP",                       # Meta-not-equal
     ("ISNT_OP", "META_NOT_EQUAL_OP"),          # Isnt

    ("__COMPARISON_END__", "ISNT_OP"),
     "__ARITHMETIC_START__",

    # Arithmetic operators
    ("UNARY_PLUS_OP", "__ARITHMETIC_START__"),  # Unary plus
     "UNARY_MINUS_OP",                          # Unary minus
     "ADDITION_OP",                             # Addition
     "SUBTRACTION_OP",                          # Subtraction
     "MULTIPLICATION_OP",                       # Multiplication
     "DIVISION_OP",                             # Division
     "MODULUS_OP",                              # Modulus

    ("__ARITHMETIC_END__", "MODULUS_OP"),
     "__LOGIC_START__",

    # Logical operators
    ("LOGICAL_NOT_OP", "__LOGIC_START__"),      # Logical not
     "LOGICAL_OR_OP",                           # Logical or
     "LOGICAL_AND_OP",                          # Logical and

    ("__LOGIC_END__", "LOGICAL_AND_OP"),
     "__BITWISE_START__",

    # Bitwise operators
    ("BITWISE_NOT_OP", "__BITWISE_START__"),    # Bitwise not
     "BITWISE_OR_OP",                           # Bitwise or
     "BITWISE_XOR_OP",                          # Bitwise xor
     "BITWISE_AND_OP",                          # Bitwise and
     "LEFT_SHIFT_OP",                           # Left shift
     "RIGHT_SHIFT_OP",                          # Right shift
     "URIGHT_SHIFT_OP",                         # Unsigned right shift

    ("__BITWISE_END__", "URIGHT_SHIFT_OP"),
     "__MEMBERSHIP_START__",

    # Membership operators
    ("MEMBERSHIP_IN", "__MEMBERSHIP_START__"),  # Membership in

    ("__MEMBERSHIP_END__", "MEMBERSHIP_IN"),
     "__MISC_START__",

    # Miscellaneous operators
    ("PARENTHESES_OP", "__MISC_START__"),       # Parentheses
     "SUBSCRIPT_OP",                            # Subscript
     "TERNARY_OP",                              # Conditional op

    ("__MISC_END__", "TERNARY_OP"),
    ("__LAST_OP__", "__MISC_END__"),
    )


OperatorSignal = Enum(
    ("SIG_NONE", 0),
    ("SIG_CHLD1", 1),
    ("SIG_CHLD2", 2),
    ("SIG_CHLD3", 4),
    )


class Operator(Expression):

    __slots__ = (
        "kind",
        "_child1",
        "_child2",
        "_child3",
        )

    def __init__(self, kind=OperatorKind.__NO_OP__,
            child1=None, child2=None, child3=None):
        super(Operator, self).__init__()
        self.kind = kind
        self._child1 = child1
        self._child2 = child2
        self._child3 = child3

    def copyFrom(self, other):
        super(Operator, self).copyFrom(other)
        self.kind = other.kind
        self._child1 = other._child1
        self._child2 = other._child2
        self._child3 = other._child3

    def sameAs(self, other):
        if not super(Operator, self).sameAs(other):
            return False

        if (self.kind != other.kind or
            not self._sameChild(self._child1, other._child1) or
            not self._sameChild(self._child2, other._child2) or
            not self._sameChild(self._child3, other._child3)):
            return False

        return True

    def _sameChild(self, tree1, tree2):
        if not tree1:
            if not tree2:
                return True
            else:
                return False
        elif not tree2:
            return False
        else:
            return tree1.sameAs(tree2)

    def _setParentScope(self, parent):
        for child in self._child1, self._child2, self._child3:
            if child:
                child.setParentScope(parent)

    def _evaluate(self, state):
        value1 = self._child1._evaluate(state) if self._child1 else None
        value2 = self._child2._evaluate(state) if self._child2 else None
        value3 = self._child3._evaluate(state) if self._child3 else None

        result, signal = Operator._doOperation(
            self.kind, value1, value2, value3, state=state)

        return result

    def _evaluateTree(self, state, tree=None):
        value1 = value2 = value3 = None
        tree1 = tree2 = tree3 = None

        if self._child1:
            value1, tree1 = self._child1._evaluateTree(state)
        if self._child2:
            value2, tree2 = self._child2._evaluateTree(state)
        if self._child3:
            value3, tree3 = self._child3._evaluateTree(state)

        result, signal = Operator._doOperation(
            self.kind, value1, value2, value3, state=state)

        if self._child1 and not (signal & OperatorSignal.SIG_CHLD1):
            tree1 = None
        if self._child2 and not (signal & OperatorSignal.SIG_CHLD2):
            tree2 = None
        if self._child3 and not (signal & OperatorSignal.SIG_CHLD3):
            tree3 = None

        if Operator.checkStringOperator(self.kind):
            # Strict unary operators
            if self.kind in (OperatorKind.UNARY_MINUS_OP,
               OperatorKind.UNARY_PLUS_OP, OperatorKind.LOGICAL_NOT_OP,
               OperatorKind.BITWISE_NOT_OP, OperatorKind.PARENTHESES_OP):
                if isinstance(value1, Exceptional):
                    tree = tree1

                else:
                    tree = Operator(self.kind, tree1)

            # Strict binary operators
            elif (isinstance(value1, Exceptional) or
                isinstance(value2, Exceptional)):

                if signal & OperatorSignal.SIG_CHLD1:
                    tree = tree1

                elif signal & OperatorSignal.SIG_CHLD2:
                    tree = tree2

                else:
                    raise Exception("Should not get here")

            else:
                tree = Operator(self.kind, tree1, tree2)

        else:
            # Non-strict operators
            if self.kind in (OperatorKind.IS_OP, OperatorKind.ISNT_OP):
                tree = Operator(self.kind, tree1, tree2)

            # Non-strict binary operators
            elif self.kind in (OperatorKind.LOGICAL_AND_OP,
               OperatorKind.LOGICAL_OR_OP, OperatorKind.MEMBERSHIP_IN):
                if ((signal & OperatorSignal.SIG_CHLD1) and
                    (signal & OperatorSignal.SIG_CHLD2)):
                    tree = Operator(self.kind, tree1, tree2)

                elif signal & OperatorSignal.SIG_CHLD1:
                    tree = tree1

                elif signal & OperatorSignal.SIG_CHLD2:
                    tree = tree2

                else:
                    raise Exception("Should not get here")

            # Non-strict ternary operators
            elif self.kind == OperatorKind.TERNARY_OP:
                tree = Literal(undefined)

                if signal & OperatorSignal.SIG_CHLD2:
                    tree = tree2

                elif signal & OperatorSignal.SIG_CHLD3:
                    tree = tree3

                else:
                    tree = tree1

        return (result, tree)

    def _flatten(self, state, kind):
        kind = self.kind
        kind1 = kind2 = OperatorKind.__NO_OP__

        # Kind is binary, but not associative or commutative, disallow
        # splitting
        if ((kind >= OperatorKind.__COMPARISON_START__ and
                kind <= OperatorKind.__COMPARISON_END__) or
            kind in (OperatorKind.SUBTRACTION_OP,
                OperatorKind.DIVISION_OP, OperatorKind.MODULUS_OP,
                OperatorKind.LEFT_SHIFT_OP, OperatorKind.RIGHT_SHIFT_OP,
                OperatorKind.URIGHT_SHIFT_OP)):

            value1, child1 = self._child1.flatten(state, kind)
            value2, child2 = self._child2.flatten(state, kind)

            if not child1 and not child2:
                value = Operator._doOperation(kind, value1, value2)[0]
                return (value, None, OperatorKind.__NO_OP__)

            elif child1 and child2:
                tree = Operator(kind, child1, child2)
                return (undefined, tree, OperatorKind.__NO_OP__)

            elif child1:
                tree = Operator._createFromTree(
                    kind, child1, value2)
                return (undefined, tree, OperatorKind.__NO_OP__)

            elif child2:
                tree = Operator._createFromValue(
                    kind, value1, child2)
                return (undefined, tree, OperatorKind.__NO_OP__)

        # Kind is non-binary
        elif kind in (OperatorKind.TERNARY_OP,
             OperatorKind.SUBSCRIPT_OP, OperatorKind.UNARY_PLUS_OP,
             OperatorKind.UNARY_MINUS_OP, OperatorKind.PARENTHESES_OP,
             OperatorKind.LOGICAL_NOT_OP, OperatorKind.BITWISE_NOT_OP):
            return self._flattenSpecials(state, kind)

        else:
            value1 = value2 = undefined
            child1 = child2 = None

        # Kind is binary, commutative and associative
        if self._child1:
            value1, child1, kind1 = self._child1._flatten(state, kind)
        if self._child2:
            value2, child2, kind2 = self._child2._flatten(state, kind)

        value, new_kind, tree = self._combine(
            kind, kind1, value1, child1, kind2, value2, child2)

        if new_kind != OperatorKind.__NO_OP__:
            tree = Operator._createFromValue(new_kind, value, tree)

        return (value, tree, new_kind)

    def _combine(self, kind, kind1, value1, tree1, kind2, value2, tree2):
        # Special case for logical operators with exactly one value
        if ((not tree1 or not tree2) and
            (tree1 or tree2) and
            (kind == OperatorKind.LOGICAL_OR_OP or
                kind == OperatorKind.LOGICAL_AND_OP)):
            value = Operator._doOperation(
                kind,
                value1 if not tree1 else undefined,
                value2 if not tree2 else undefined)[0]
            if isinstance(value, bool):
                return (value, None, OperatorKind.__NO_OP__)
        else:
            value = undefined

        if not tree1 and not tree2:
            value = Operator._doOperation(kind, value1, value2)[0]
            return (value, None, OperatorKind.__NO_OP__)

        elif not tree1 and (tree2 and kind2 == OperatorKind.__NO_OP__):
            return (value1, tree2, kind)

        elif not tree2 and (tree1 and kind1 == OperatorKind.__NO_OP__):
            return (value2, tree1, kind)

        elif ((tree1 and kind1 == OperatorKind.__NO_OP__) and
            (tree2 and kind2 == OperatorKind.__NO_OP__)):

            tree = Operator(kind, tree1, tree2)
            return (value, tree, OperatorKind.__NO_OP__)

        if ((kind1 != OperatorKind.__NO_OP__ or
                kind2 != OperatorKind.__NO_OP__) and
            kind != kind1 and kind != kind2):

            if kind1 != OperatorKind.__NO_OP__:
                tree1 = Operator(kind1, value1, tree1)
            elif not tree1:
                tree1 = Literal(value1)

            if kind2 != OperatorKind.__NO_OP__:
                tree2 = Operator(kind2, value2, tree2)
            elif not tree2:
                tree2 = Literal(value2)

            tree = Operator(kind, tree1, tree2)
            return (value, tree, OperatorKind.__NO_OP__)

        if kind == kind1 and kind == kind2:
            tree = Operator(kind, tree1, tree2)
            value = Operator._doOperation(kind, value1, value2)[0]
            return (value, tree, kind)

        elif kind == kind1:
            if not tree2:
                value = Operator._doOperation(kind, value1, value2)[0]
                return (value, tree1, kind)

            else:
                tree = Operator(kind, tree1, tree2)
                return (value1, tree, kind)

        elif kind == kind2:
            if not tree1:
                value = Operator._doOperation(kind, value1, value2)[0]
                return (value, tree2, kind)

            else:
                tree = Operator(kind, tree1, tree2)
                return (value2, tree, kind)

        raise Exception("Should not get here")

    def _flattenSpecials(self, state, kind):
        if self.kind in (OperatorKind.UNARY_PLUS_OP,
           OperatorKind.UNARY_MINUS_OP,
           OperatorKind.PARENTHESES_OP,
           OperatorKind.LOGICAL_NOT_OP,
           OperatorKind.BITWISE_NOT_OP):
            eval1, child1 = self._child1.flatten(state)
            if child1:
                tree = Operator(self.kind, child1)
                return (undefined, tree, kind)

            value = Operator._doOperation(self.kind, eval1)[0]
            return (value, None, kind)

        elif self.kind == OperatorKind.TERNARY_OP:
            # Flatten the selector expression
            eval1, child1 = self._child1.flatten(state)

            # Check if selector expression collapsed to a non-undefined value
            if not child1 and not isinstance(eval1, Undefined):
                if not isinstance(eval1, bool):
                    return (error, None, kind)

                if eval1 is True:
                    return self._child2.flatten(state)

                return self._child3.flatten(state)

            # Flatten arms of the if expression
            eval2, child2 = self._child2.flatten(state)
            eval3, child3 = self._child3.flatten(state)

            # if any arm collapsed into a value, make it a Literal
            if not child2:
                child2 = Literal(eval2)
            if not child3:
                child3 = Literal(eval3)

            # child1 may be NULL if child1 Flattened to UNDEFINED
            if not child1:
                child1.copyFrom(self.child1)

            tree = Operator(self.kind, child1, child2, child3)
            return (value, tree, kind)

        elif self.kind == OperatorKind.SUBSCRIPT_OP:
            # Flatten both arguments
            eval1, child1 = self._child1.flatten(state)
            eval2, child2 = self._child2.flatten(state)

            if not child1 or not child2:
                value = Operator._doOperation(self.kind, eval1, eval2)[0]
                return (value, None, kind)

            if not child1:
                child1 = Literal(eval1)
            if not child2:
                child2 = Literal(eval2)

            tree = Operator(self.kind, child1, child2)
            return (value, tree, kind)

        else:
            raise Exception("Should not get here")

    @staticmethod
    def _doOperation(kind, value1=None,
            value2=None, value3=None, state=None):
        if kind in (OperatorKind.__NO_OP__, OperatorKind.PARENTHESES_OP):
            return (value1, OperatorSignal.SIG_CHLD1)

        elif kind == OperatorKind.UNARY_PLUS_OP:
            if isinstance(value1, (bool, basestring, list, dict)):
                value1 = error

            return (value1, OperatorSignal.SIG_CHLD1)

        # Test for cases when evaluation is strict
        if Operator.checkStringOperator(kind):
            if isinstance(value1, Error):
                return (error, OperatorSignal.SIG_CHLD1)

            if value2 and isinstance(value2, Error):
                return (error, OperatorSignal.SIG_CHLD2)

            if value3 and isinstance(value3, Error):
                return (error, OperatorSignal.SIG_CHLD3)

            if value1 and isinstance(value1, Undefined):
                return (undefined, OperatorSignal.SIG_CHLD1)

            if value2 and isinstance(value2, Undefined):
                return (undefined, OperatorSignal.SIG_CHLD2)

            if value3 and isinstance(value3, Undefined):
                return (undefined, OperatorSignal.SIG_CHLD3)

        # Comparison operators (binary, one unary)
        if (kind >= OperatorKind.__COMPARISON_START__ and
            kind <= OperatorKind.__COMPARISON_END__):
            return Operator._doComparison(kind, value1, value2)

        # Arithmetic operators (binary)
        if (kind >= OperatorKind.__ARITHMETIC_START__ and
            kind <= OperatorKind.__ARITHMETIC_END__):
            return Operator._doArithmetic(kind, value1, value2)

        # Logical operators (binary, one unary)
        if (kind >= OperatorKind.__LOGIC_START__ and
            kind <= OperatorKind.__LOGIC_END__):
            return Operator._doLogical(kind, value1, value2)

        # Membership operators (binary, one unary)
        if (kind >= OperatorKind.__MEMBERSHIP_START__ and
            kind <= OperatorKind.__MEMBERSHIP_END__):
            return Operator._doMembership(kind, value1, value2)

        # Bitwise operators (binary, one unary)
        if (kind >= OperatorKind.__BITWISE_START__ and
            kind <= OperatorKind.__BITWISE_END__):
            return Operator._doBitwise(kind, value1, value2)

        # Miscellaneous
        if kind == OperatorKind.TERNARY_OP:
            if isinstance(value1, Undefined):
                return (undefined, OperatorSignal.SIG_CHLD1)

            if not isinstance(value1, bool):
                return (error, OperatorSignal.SIG_CHLD1)

            if value1 is True:
                return (value2, OperatorSignal.SIG_CHLD2)

            return (value3, OperatorSignal.SIG_CHLD3)

        if kind == OperatorKind.SUBSCRIPT_OP:
            if isinstance(value1, dict) and isinstance(value2, basestring):
                if not value1.lookup(value2):
                    return (error, OperatorSignal.SIG_CHLD2)

                return (value1.evaluateName(value2),
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            if isinstance(value1, list) and isinstance(value2, int):
                if value2 < 0 or value2 >= len(value1):
                    return (error, OperatorSignal.SIG_CHLD2)

                return (value1[value2],
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            return (error,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        raise Exception("Should not get here")

    @staticmethod
    def _doComparison(kind, value1, value2):
        value1, value2, coerce_result = Operator._coerceToNumber(
            value1, value2)
        exact = True

        if kind == OperatorKind.META_EQUAL_OP:
            if type(value1) != type(value2):
                return (False,
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            if isinstance(value1, Exceptional):
                return (True,
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            exact = False
            kind = OperatorKind.EQUAL_OP

        if kind == OperatorKind.META_NOT_EQUAL_OP:
            if type(value1) != type(value2):
                return (True,
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            if (isinstance(value1, Exceptional) or
                isinstance(value2, Exceptional)):

                return (False,
                    OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

            exact = False
            kind = OperatorKind.NOT_EQUAL_OP

        # At least one of v1, v2 is a string
        if coerce_result == basestring:
            if (not isinstance(value1, basestring) or
                not isinstance(value2, basestring)):

                result = error
            else:
                result = Operator._compareStrings(
                    kind, value1, value2, exact)

        elif coerce_result == int:
            result = Operator._compareIntegers(kind, value1, value2)

        elif coerce_result == float:
            result = Operator._compareIntegers(kind, value1, value2)

        elif coerce_result == bool:
            if isinstance(value1, bool) and isinstance(value2, bool):
                result = Operator._compareIntegers(kind, value1, value2)
            else:
                result = error

        elif coerce_result == list or coerce_result == dict:
            result = error

        else:
            raise Exception("Should not get here")

        return (result, OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

    @staticmethod
    def _doArithmetic(kind, value1, value2):
        # Ensure the operands have arithmetic types
        if (not isinstance(value1, (bool, int, float)) or
            (kind != OperatorKind.UNARY_MINUS_OP and
                not isinstance(value2, (bool, int, float)))):
            return (error,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        if kind == OperatorKind.UNARY_MINUS_OP:
            if isinstance(value1, bool):
                result = not value1

            elif isinstance(value1, (int, float)):
                result = -value1

            elif isinstance(value1, Exceptional):
                result = value1

            else:
                result = error

            return (result, OperatorSignal.SIG_CHLD1)

        value1, value2, coerce_result = Operator._coerceToNumber(
            value1, value2)
        if coerce_result == int:
            return Operator._doIntegerArithmetic(kind, value1, value2)

        elif coerce_result == float:
            return Operator._doRealArithmetic(kind, value1, value2)

        else:
            raise Exception("Should not get here")

    @staticmethod
    def _doLogical(kind, value1, value2):
        if not isinstance(value1, (Undefined, Error, bool)):
            return (error, OperatorSignal.SIG_CHLD1)

        if not isinstance(value2, (Undefined, Error, bool)):
            return (error, OperatorSignal.SIG_CHLD2)

        if kind == OperatorKind.LOGICAL_NOT_OP:
            if isinstance(value1, bool):
                value1 = not value1
            return (value1, OperatorSignal.SIG_CHLD1)

        if kind == OperatorKind.LOGICAL_OR_OP:
            if isinstance(value1, bool) and value1:
                return (value1, OperatorSignal.SIG_CHLD1)

            elif isinstance(value1, Error):
                return (error, OperatorSignal.SIG_CHLD1)

            elif isinstance(value1, bool) and not value1:
                result = value2

            elif not isinstance(value2, bool):
                result = value2

            elif value2:
                result = True

            else:
                result = undefined

            return (result,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        elif kind == OperatorKind.LOGICAL_AND_OP:
            if isinstance(value1, bool) and not value1:
                return (False, OperatorSignal.SIG_CHLD1)

            elif isinstance(value1, Error):
                return (error, OperatorSignal.SIG_CHLD1)

            elif isinstance(value1, bool) and value1:
                result = value2

            elif not isinstance(value2, bool):
                result = value2

            elif not value2:
                result = False

            else:
                result = undefined

            return (result,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        raise Exception("Should not get here")

    @staticmethod
    def _doMembership(kind, value1, value2):
        if not isinstance(value2, list):
            return (error, OperatorSignal.SIG_CHLD2)

        if kind == OperatorKind.MEMBERSHIP_IN:
            result = any(value1 == value for value in value2)

            return (result,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        raise Exception("Should not get here")

    @staticmethod
    def _doBitwise(kind, value1, value2):
        sign_mask = 1 << (32 - 1)

        if kind == OperatorKind.BITWISE_NOT_OP:
            if not isinstance(value1, int):
                return (error, OperatorSignal.SIG_CHLD1)

        elif not isinstance(value1, int) or not isinstance(value2, int):
            return (error,
                OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

        if kind == OperatorKind.BITWISE_NOT_OP:
            return (~value1, OperatorSignal.SIG_CHLD1)

        elif kind == OperatorKind.BITWISE_OR_OP:
            result = value1 | value2

        elif kind == OperatorKind.BITWISE_AND_OP:
            result = value1 & value2

        elif kind == OperatorKind.BITWISE_XOR_OP:
            result = value1 ^ value2

        elif kind == OperatorKind.LEFT_SHIFT_OP:
            result = value1 << value2

        elif kind == OperatorKind.URIGHT_SHIFT_OP:
            if value1 >= 0:
                # Sign bit is off
                result = value1 >> value2
            else:
                # Sign bit is on
                result = value1 >> 1
                result &= ~sign_mask
                result >>= value2 - 1

        elif kind == OperatorKind.RIGHT_SHIFT_OP:
            if value1 >= 0:
                # Sign bit is off
                result = value1 >> value2
            else:
                # Sign bit is on
                result = value1
                for dummy in xrange(value2):
                    result = (result >> 1) | sign_mask

        else:
            raise Exception("Should not get here")

        return (result, OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

    @staticmethod
    def _doIntegerArithmetic(kind, value1, value2):
        if kind == OperatorKind.ADDITION_OP:
            result = value1 + value2

        elif kind == OperatorKind.SUBTRACTION_OP:
            result = value1 - value2

        elif kind == OperatorKind.MULTIPLICATION_OP:
            result = value1 * value2

        elif kind == OperatorKind.DIVISION_OP:
            if value2 != 0:
                result = value1 / value2
            else:
                result = error

        elif kind == OperatorKind.MODULUS_OP:
            if value2 != 0:
                result = value1 % value2
            else:
                result = error

        else:
            raise Exception("Should not get here")

        return (result, OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

    @staticmethod
    def _doRealArithmetic(kind, value1, value2):
        if kind == OperatorKind.ADDITION_OP:
            result = value1 + value2

        elif kind == OperatorKind.SUBTRACTION_OP:
            result = value1 - value2

        elif kind == OperatorKind.MULTIPLICATION_OP:
            result = value1 * value2

        elif kind == OperatorKind.DIVISION_OP:
            result = value1 / value2

        elif kind == OperatorKind.MODULUS_OP:
            result = error

        else:
            raise Exception("Should not get here")

        return (result, OperatorSignal.SIG_CHLD1 | OperatorSignal.SIG_CHLD2)

    @staticmethod
    def _compareIntegers(kind, value1, value2):
        if kind == OperatorKind.LESS_THAN_OP:
            return value1 < value2

        elif kind == OperatorKind.LESS_OR_EQUAL_OP:
            return value1 <= value2

        elif kind == OperatorKind.EQUAL_OP:
            return value1 == value2

        elif kind == OperatorKind.NOT_EQUAL_OP:
            return value1 != value2

        elif kind == OperatorKind.GREATER_THAN_OP:
            return value1 > value2

        elif kind == OperatorKind.GREATER_OR_EQUAL_OP:
            return value1 >= value2

        else:
            raise Exception("Should not get here")

    @staticmethod
    def _compareStrings(kind, value1, value2, exact):
        if exact:
            value = cmp(value1, value2)
        else:
            value = cmp(value1.lower(), value2.lower())

        if value < 0:
            if kind in (OperatorKind.LESS_THAN_OP,
               OperatorKind.LESS_OR_EQUAL_OP,
               OperatorKind.NOT_EQUAL_OP):
                return True

        elif value == 0:
            if kind in (OperatorKind.LESS_OR_EQUAL_OP,
               OperatorKind.EQUAL_OP,
               OperatorKind.GREATER_OR_EQUAL_OP):
                return True

        else:
            if kind in (OperatorKind.GREATER_THAN_OP,
               OperatorKind.GREATER_OR_EQUAL_OP,
               OperatorKind.NOT_EQUAL_OP):
                return True

        return False

    @staticmethod
    def _coerceToNumber(value1, value2):
        if isinstance(value1, dict) or isinstance(value2, dict):
            return (value1, value2, dict)
        elif isinstance(value1, list) or isinstance(value2, list):
            return (value1, value2, list)
        elif isinstance(value1, basestring) or isinstance(value2, list):
            return (value1, value2, basestring)
        elif isinstance(value1, Undefined) or isinstance(value2, Undefined):
            return (value1, value2, Undefined)
        elif isinstance(value1, Error) or isinstance(value2, Error):
            return (value1, value2, Error)

        # Promote booleans to integers
        if isinstance(value1, bool):
            value1 = 1 if value1 else 0
        if isinstance(value2, bool):
            value2 = 1 if value2 else 0

        # Both value1 and value2 of same numerical type
        if isinstance(value1, int) and isinstance(value2, int):
            return (value1, value2, int)
        if isinstance(value1, float) and isinstance(value2, float):
            return (value1, value2, float)

        # Type promotions required
        if isinstance(value1, int) and isinstance(value2, float):
            value1 = float(value1)
        if isinstance(value1, float) and isinstance(value2, int):
            value2 = float(value2)

        return (value1, value2, float)

    @staticmethod
    def _createFromValue(kind, value, tree):
        if not tree:
            return None

        literal = Literal(value)
        return Operator(kind, literal, tree)

    @staticmethod
    def _createFromTree(kind, tree, value):
        if not tree:
            return None

        literal = Literal(value)
        return Operator(kind, tree, literal)

    @staticmethod
    def checkStringOperator(kind):
        non_strict_kinds = (
            OperatorKind.META_EQUAL_OP,
            OperatorKind.META_NOT_EQUAL_OP,
            OperatorKind.LOGICAL_AND_OP,
            OperatorKind.LOGICAL_OR_OP,
            OperatorKind.MEMBERSHIP_IN,
            OperatorKind.TERNARY_OP)

        if kind in non_strict_kinds:
            return False
        else:
            return True

    @staticmethod
    def getPrecedenceLevel(kind):
        kind_to_precedence = {
            OperatorKind.SUBSCRIPT_OP: 13,
            OperatorKind.LOGICAL_NOT_OP: 12,
            OperatorKind.BITWISE_NOT_OP: 12,
            OperatorKind.UNARY_PLUS_OP: 12,
            OperatorKind.UNARY_MINUS_OP: 12,
            OperatorKind.MULTIPLICATION_OP: 11,
            OperatorKind.DIVISION_OP: 11,
            OperatorKind.MODULUS_OP: 11,
            OperatorKind.ADDITION_OP: 10,
            OperatorKind.SUBTRACTION_OP: 10,
            OperatorKind.LEFT_SHIFT_OP: 9,
            OperatorKind.RIGHT_SHIFT_OP: 9,
            OperatorKind.URIGHT_SHIFT_OP: 9,
            OperatorKind.LESS_THAN_OP: 8,
            OperatorKind.LESS_OR_EQUAL_OP: 8,
            OperatorKind.GREATER_OR_EQUAL_OP: 8,
            OperatorKind.GREATER_THAN_OP: 8,
            OperatorKind.NOT_EQUAL_OP: 7,
            OperatorKind.EQUAL_OP: 7,
            OperatorKind.IS_OP: 7,
            OperatorKind.ISNT_OP: 7,
            OperatorKind.BITWISE_AND_OP: 6,
            OperatorKind.BITWISE_XOR_OP: 5,
            OperatorKind.BITWISE_OR_OP: 4,
            OperatorKind.MEMBERSHIP_IN: 3,
            OperatorKind.LOGICAL_AND_OP: 2,
            OperatorKind.LOGICAL_OR_OP: 1,
            OperatorKind.TERNARY_OP: 0}

        if kind in kind_to_precedence:
            return kind_to_precedence[kind]
        else:
            return -1
