#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Parser",
    "ParserStream",
    ]

from checkbox.lib.enum import enum_value_to_name_string

from checkbox.message.function import Function
from checkbox.message.lexer import Lexer
from checkbox.message.list import List
from checkbox.message.literal import Literal
from checkbox.message.operator import (
    Operator,
    OperatorKind,
    )
from checkbox.message.record import Record
from checkbox.message.reference import Reference
from checkbox.message.token import (
    Token,
    TokenType,
    )
from checkbox.message.types import (
    error,
    undefined,
    )


class ParserError(Exception):

    def __init__(self, token, *expected):
        self.token = token
        self.expected = expected

    def __str__(self):
        token_name = enum_value_to_name_string(TokenType, self.token)
        expected_names = [enum_value_to_name_string(TokenType, expected)
            for expected in self.expected]
        if len(expected_names) == 0:
            expected_name = "nothing"
        elif len(expected_names) == 1:
            expected_name = expected_names[0]
        else:
            expected_name = ", ".join(expected_names[:-1])
            expected_name = " or ".join([expected_name, expected_names[-1]])

        return "Expected %s but got %s" % (expected_name, token_name)


class Parser:

    __slots__ = (
        "_lexer",
        )

    def __init__(self):
        self._lexer = Lexer()

    def peekToken(self):
        if self._lexer.was_initialized:
            return self._lexer.peekToken()
        else:
            return TokenType.TOKEN_ERROR

    def consumeToken(self):
        if self._lexer.was_initialized:
            return self._lexer.consumeToken()
        else:
            return TokenType.TOKEN_ERROR

    def parseRecord(self, file, full=False):
        self._lexer.initialize(file)

        return self._parseRecord(full)

    def parseExpression(self, file, full=False):
        self._lexer.initialize(file)

        return self._parseExpression(full)

    def parseNextExpression(self):
        if not self._lexer.was_initialized:
            return None

        return self._parseExpression(False)

    # expression ::= logical_or_expression
    #   | logical_or_expression '?' Expression ':' Expression
    def _parseExpression(self, full=False):
        tree_l = self._parseLogicalOrExpression()
        if not tree_l:
            return None

        token_type = self._lexer.peekToken()
        if token_type == TokenType.QMARK:
            self._lexer.consumeToken()
            tree_m = self._parseExpression()
            token_type = self._lexer.consumeToken()
            if token_type != TokenType.COLON:
                raise ParserError(token_type, TokenType.COLON)

            tree_r = self._parseExpression()
            if tree_l and tree_m and tree_r:
                return Operator(
                    OperatorKind.TERNARY_OP, tree_l, tree_m, tree_r)

            return None

        if full:
            token_type = self._lexer.consumeToken()
            if token_type != TokenType.END_OF_INPUT:
                raise ParserError(token_type, TokenType.END_OF_INPUT)

        return tree_l

    # logical_or_expression ::= logical_and_expression
    #   | logical_or_expression '||' logical_and_expression
    def _parseLogicalOrExpression(self):
        tree_l = self._parseLogicalAndExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.LOGICAL_OR:
            self._lexer.consumeToken()
            tree_r = self._parseLogicalAndExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.LOGICAL_OR_OP, tree_l, tree_r)
            else:
                return None

        return tree_l

    # logical_and_expression ::= membership_expression
    #   | logical_and_expression '&&' membership_expression
    def _parseLogicalAndExpression(self):
        tree_l = self._parseMembershipExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.LOGICAL_AND:
            self._lexer.consumeToken()
            tree_r = self._parseMembershipExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.LOGICAL_AND_OP, tree_l, tree_r)
            else:
                return None

        return tree_l

    # membership_expression ::= inclusive_or_expression
    #   | membership_expression 'in' inclusive_or_expression
    def _parseMembershipExpression(self):
        tree_l = self._parseInclusiveOrExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.IN:
            self._lexer.consumeToken()
            tree_r = self._parseInclusiveOrExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.MEMBERSHIP_IN, tree_l, tree_r)
            else:
                return None

        return tree_l

    # inclusive_or_expression ::= exclusive_or_expression
    #   | inclusive_or_expression '|' exclusive_or_expression
    def _parseInclusiveOrExpression(self):
        tree_l = self._parseExclusiveOrExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.BITWISE_OR:
            self._lexer.consumeToken()
            tree_r = self._parseExclusiveOrExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.BITWISE_OR_OP, tree_l, tree_r)
            else:
                return None

        return tree_l

    # exclusive_or_expression ::= and_expression
    #   | exclusive_or_expression '^' and_expression
    def _parseExclusiveOrExpression(self):
        tree_l = self._parseAndExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.BITWISE_XOR:
            self._lexer.consumeToken()
            tree_r = self._parseAndExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.BITWISE_XOR_OP, tree_l, tree_r)
            else:
                return None

        return tree_l

    # and_expression ::= equality_expression
    #   | and_expression '&' equality_expression
    def _parseAndExpression(self):
        tree_l = self._parseEqualityExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() == TokenType.BITWISE_AND:
            self._lexer.consumeToken()
            tree_r = self._parseEqualityExpression()
            if tree_l and tree_r:
                tree_l = Operator(OperatorKind.BITWISE_AND_OP, tree_l, tree_r)
            else:
                return None

        return tree_l

    # equality_expression ::= relational_expression
    #   | equality_expression '==' relational_expression
    #   | equality_expression '!=' relational_expression
    #   | equality_expression '=~' relational_expression
    #   | equality_expression '!~' relational_expression
    def _parseEqualityExpression(self):
        token_table = {
            TokenType.EQUAL: OperatorKind.EQUAL_OP,
            TokenType.NOT_EQUAL: OperatorKind.NOT_EQUAL_OP,
            TokenType.META_EQUAL: OperatorKind.META_EQUAL_OP,
            TokenType.META_NOT_EQUAL: OperatorKind.META_NOT_EQUAL_OP}

        tree_l = self._parseRelationalExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            tree_r = self._parseRelationalExpression()
            if tree_l and tree_r:
                tree_l = Operator(token_table[token_type], tree_l, tree_r)
            else:
                return None

        return tree_l

    # relational_expression ::= shift_expression
    #   | relational_expression '<' shift_expression
    #   | relational_expression '>' shift_expression
    #   | relational_expression '<=' shift_expression
    #   | relational_expression '>=' shift_expression
    def _parseRelationalExpression(self):
        token_table = {
            TokenType.LESS_THAN: OperatorKind.LESS_THAN_OP,
            TokenType.GREATER_THAN: OperatorKind.GREATER_THAN_OP,
            TokenType.LESS_OR_EQUAL: OperatorKind.LESS_OR_EQUAL_OP,
            TokenType.GREATER_OR_EQUAL: OperatorKind.GREATER_OR_EQUAL_OP}

        tree_l = self._parseShiftExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            tree_r = self._parseShiftExpression()
            if tree_l and tree_r:
                tree_l = Operator(token_table[token_type], tree_l, tree_r)
            else:
                return None

        return tree_l

    # shift_expression ::= additive_expression
    #   | shift_expression '<<' additive_expression
    #   | shift_expression '>>' additive_expression
    #   | shift_expression '>>>' Aditive_expression
    def _parseShiftExpression(self):
        token_table = {
            TokenType.LEFT_SHIFT: OperatorKind.LEFT_SHIFT_OP,
            TokenType.RIGHT_SHIFT: OperatorKind.RIGHT_SHIFT_OP,
            TokenType.URIGHT_SHIFT: OperatorKind.URIGHT_SHIFT_OP}

        tree_l = self._parseAdditiveExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            tree_r = self._parseAdditiveExpression()
            if tree_l and tree_r:
                tree_l = Operator(token_table[token_type], tree_l, tree_r)
            else:
                return None

        return tree_l

    # additive_expression ::= multiplicative_expression
    #   | additive_expression '+' multiplicative_expression
    #   | additive_expression '-' multiplicative_expression
    def _parseAdditiveExpression(self):
        token_table = {
            TokenType.PLUS: OperatorKind.ADDITION_OP,
            TokenType.MINUS: OperatorKind.SUBTRACTION_OP}

        tree_l = self._parseMultiplicativeExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            tree_r = self._parseMultiplicativeExpression()
            if tree_l and tree_r:
                tree_l = Operator(token_table[token_type], tree_l, tree_r)
            else:
                return None

        return tree_l

    # multiplicative_expression ::= unary_expression
    #   |  multiplicative_expression '*' unary_expression
    #   |  multiplicative_expression '/' unary_expression
    #   |  multiplicative_expression '%' unary_expression
    def _parseMultiplicativeExpression(self):
        token_table = {
            TokenType.MULTIPLY: OperatorKind.MULTIPLICATION_OP,
            TokenType.DIVIDE: OperatorKind.DIVISION_OP,
            TokenType.MODULUS: OperatorKind.MODULUS_OP}

        tree_l = self._parseUnaryExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            tree_r = self._parseUnaryExpression()
            if tree_l and tree_r:
                tree_l = Operator(token_table[token_type], tree_l, tree_r)
            else:
                return None

        return tree_l

    # unary_expression ::= postfix_expression
    #   | unary_operator unary_expression
    # (where unary_operator is one of { -, +, ~, ! })
    def _parseUnaryExpression(self):
        token_table = {
            TokenType.MINUS: OperatorKind.UNARY_MINUS_OP,
            TokenType.PLUS: OperatorKind.UNARY_PLUS_OP,
            TokenType.BITWISE_NOT: OperatorKind.BITWISE_NOT_OP,
            TokenType.LOGICAL_NOT: OperatorKind.LOGICAL_NOT_OP}

        if self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()

            tree_m = self._parseUnaryExpression()
            if tree_m:
                return Operator(token_table[token_type], tree_m)
            else:
                return None
        else:
            return self._parsePostfixExpression()

    # postfix_expression ::= primary_expression
    #   | postfix_expression '.' identifier
    #   | postfix_expression '{' expression '}'
    def _parsePostfixExpression(self):
        token_table = {
            TokenType.OPEN_BRACE: OperatorKind.MULTIPLICATION_OP,
            TokenType.SELECTION: OperatorKind.DIVISION_OP}

        tree_l = self._parsePrimaryExpression()
        if not tree_l:
            return None

        while self._lexer.peekToken() in token_table.keys():
            token_type = self._lexer.consumeToken()
            if token_type == TokenType.OPEN_BRACE:
                tree_r = self._parseExpression()
                if tree_l and tree_r:
                    tree_l = Operator(
                        OperatorKind.SUBSCRIPT_OP, tree_l, tree_r)
                    if self._lexer.consumeToken() == TokenType.CLOSE_BRACE:
                        continue

                return None

            elif token_type == TokenType.SELECTION:
                token = Token()
                token_type = self._lexer.consumeToken(token)
                if token_type != TokenType.IDENTIFIER:
                    raise ParserError(token_type, TokenType.IDENTIFIER)

                tree_l = Reference(token.value, False, tree_l)

        return tree_l

    # primary_expression ::= identifier
    #   | function_call
    #   | '.' identifier
    #   | '(' expression ')'
    #   | literal
    # function_call ::= identifier argument_list
    def _parsePrimaryExpression(self):
        token = Token()

        token_type = self._lexer.peekToken(token)
        if token_type == TokenType.IDENTIFIER:
            self._lexer.consumeToken()
            token_type = self._lexer.peekToken()
            if token_type == TokenType.OPEN_PAREN:
                arguments = self._parseArgumentList()
                return Function(token.value, *arguments)
            else:
                return Reference(token.value, False)

        elif token_type == TokenType.SELECTION:
            self._lexer.consumeToken()
            token_type = self._lexer.consumeToken(token)
            if token_type != TokenType.IDENTIFIER:
                raise ParserError(token_type, TokenType.IDENTIFIER)

            return Reference(token.value, True)

        elif token_type == TokenType.OPEN_PAREN:
            self._lexer.consumeToken()
            tree_l = self._parseExpression()
            if not tree_l:
                return None

            token_type = self._lexer.consumeToken()
            if token_type != TokenType.CLOSE_PAREN:
                raise ParserError(token_type, TokenType.CLOSE_PAREN)

            return Operator(OperatorKind.PARENTHESES_OP, tree_l)

        elif token_type == TokenType.OPEN_BRACE:
            return self._parseRecord()

        elif token_type == TokenType.OPEN_BRACKET:
            return self._parseList()

        elif token_type == TokenType.NULL:
            self._lexer.consumeToken()
            return Literal(None)

        elif token_type == TokenType.UNDEFINED:
            self._lexer.consumeToken()
            return Literal(undefined)

        elif token_type == TokenType.ERROR:
            self._lexer.consumeToken()
            return Literal(error)

        elif token_type == TokenType.BOOLEAN:
            self._lexer.consumeToken()
            return Literal(token.value)

        elif token_type == TokenType.INTEGER:
            self._lexer.consumeToken()
            return Literal(token.value, token.factor)

        elif token_type == TokenType.REAL:
            self._lexer.consumeToken()
            return Literal(token.value, token.factor)

        elif token_type == TokenType.STRING:
            self._lexer.consumeToken()
            return Literal(token.value)

        else:
            return None

    # argument_list ::= '(' list_of_arguments ')'
    # list_of_arguments ::= (epsilon)
    #   | list_of_arguments ',' expression
    def _parseArgumentList(self):
        arguments = []
        token_type = self._lexer.consumeToken()
        if token_type != TokenType.OPEN_PAREN:
            raise ParserError(token_type, TokenType.OPEN_PAREN)

        token_type = self._lexer.peekToken()
        while token_type != TokenType.CLOSE_PAREN:
            tree = self._parseExpression()
            arguments.append(tree)

            # the next token must be a ',' or a ')'
            token_type = self._lexer.peekToken()
            if token_type == TokenType.COMMA:
                self._lexer.consumeToken()
            elif token_type != TokenType.CLOSE_PAREN:
                raise ParserError(token_type, TokenType.CLOSE_PAREN)

        self._lexer.consumeToken()
        return arguments

    # record ::= '{' reference_list '}'
    # reference_list ::= (epsilon)
    #   | reference ';' reference_list
    # reference ::= identifier '=' expression
    def _parseRecord(self, full=False):
        tree_l = Record()

        token_type = self._lexer.consumeToken()
        if token_type != TokenType.OPEN_BRACE:
            return None

        token = Token()
        token_type = self._lexer.peekToken()
        while token_type != TokenType.CLOSE_BRACE:
            token_type = self._lexer.consumeToken(token)
            if token_type == TokenType.SEMICOLON:
                continue

            if token_type != TokenType.IDENTIFIER:
                raise ParserError(token_type, TokenType.IDENTIFIER)

            token_type = self._lexer.consumeToken()
            if token_type != TokenType.BOUND_TO:
                raise ParserError(token_type, TokenType.BOUND_TO)

            tree_r = self._parseExpression()
            if not tree_r:
                return None

            if not tree_l.insert(token.value, tree_r):
                return None

            token_type = self._lexer.peekToken()
            if token_type not in (TokenType.SEMICOLON, TokenType.CLOSE_BRACE):
                raise ParserError(
                    token_type, TokenType.SEMICOLON, TokenType.CLOSE_BRACE)

            # Slurp remaining semicolons
            while token_type == TokenType.SEMICOLON:
                self._lexer.consumeToken()
                token_type = self._lexer.peekToken()

        self._lexer.consumeToken()

        if full:
            token_type = self._lexer.consumeToken()
            if token_type != TokenType.END_OF_INPUT:
                raise ParserError(token_type, TokenType.END_OF_INPUT)

        return tree_l

    # list ::= '[' list_of_expressions ']'
    # list_of_expressions ::= (epsilon)
    #   | expression ',' list_of_expressions
    def _parseList(self, full=False):
        list_of_expressions = []

        token_type = self._lexer.consumeToken()
        if token_type != TokenType.OPEN_BRACKET:
            raise ParserError(token_type, TokenType.OPEN_BRACKET)

        token_type = self._lexer.peekToken()
        while token_type != TokenType.CLOSE_BRACKET:
            tree_r = self._parseExpression()
            if not tree_r:
                return None

            list_of_expressions.append(tree_r)

            token_type = self._lexer.peekToken()
            if token_type == TokenType.COMMA:
                self._lexer.consumeToken()
                token_type = self._lexer.peekToken()

            elif token_type != TokenType.CLOSE_BRACKET:
                raise ParserError(token_type, TokenType.CLOSE_BRACKET)

        self._lexer.consumeToken()

        tree_l = List(list_of_expressions)
        if full:
            token_type = self._lexer.consumeToken()
            if token_type != TokenType.END_OF_INPUT:
                raise ParserError(token_type, TokenType.END_OF_INPUT)

        return tree_l


class ParserStream(Parser):

    def parseRecord(self, file, full=False):
        self._lexer.initialize(file)

        return self._parseStreamRecord(full)

    def parseExpression(self, file, full=False):
        self._lexer.initialize(file)

        return self._parseStreamList(full)

    def _parseStreamRecord(self, full=False):
        tree_l = Record()

        token = Token()
        token_type = self._lexer.peekToken()
        while token_type not in (
            TokenType.NEW_PARAGRAPH, TokenType.END_OF_INPUT):

            token_type = self._lexer.consumeToken(token)
            if token_type != TokenType.IDENTIFIER:
                raise ParserError(token_type, TokenType.IDENTIFIER)

            token_type = self._lexer.consumeToken()
            if token_type != TokenType.BOUND_TO:
                raise ParserError(token_type, TokenType.BOUND_TO)

            tree_r = self._parseExpression()
            if not tree_r:
                return None

            if not tree_l.insert(token.value, tree_r):
                return None

            token_type = self._lexer.peekToken()
            if token_type == TokenType.NEW_LINE:
                self._lexer.consumeToken()
                token_type = self._lexer.peekToken()
            elif token_type not in (
                TokenType.NEW_PARAGRAPH, TokenType.END_OF_INPUT):

                raise ParserError(
                    token_type, TokenType.NEW_PARAGRAPH,
                    TokenType.END_OF_INPUT)

        # Ignore full
        return tree_l

    def _parseStreamList(self, full=False):
        list_of_expressions = []

        token_type = self._lexer.peekToken()
        while token_type != TokenType.END_OF_INPUT:
            tree_r = self._parseStreamRecord()
            if not tree_r:
                return None

            list_of_expressions.append(tree_r)

            token_type = self._lexer.peekToken()
            if token_type == TokenType.NEW_PARAGRAPH:
                self._lexer.consumeToken()

            elif token_type != TokenType.END_OF_INPUT:
                raise ParserError(token_type, TokenType.END_OF_INPUT)

        self._lexer.consumeToken()

        # Ignore full
        return List(list_of_expressions)
