#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Record",
    ]

from copy import copy

from checkbox.message.eval import (
    EvalLookup,
    EvalState,
    )
from checkbox.message.expression import Expression
from checkbox.message.literal import Literal
from checkbox.message.types import (
    error,
    undefined,
    )


class Record(Expression):

    __slots__ = (
        "references",
        "chained_parent",
        )

    def __init__(self, references=dict()):
        super(Record, self).__init__()
        self.references = dict(references)
        self.chained_parent = None

    def __getattr__(self, name):
        return getattr(self.references, name)

    def _setParentScope(self, parent):
        # Already set by base class for this node
        pass

    def copyFrom(self, other):
        super(Record, self).copyFrom(other)
        self.clear()

        self.chained_parent = other.chained_parent
        for key, value in other.references.iteritems():
            tree = copy(value)
            tree.setParentScope(self)
            self.references[key] = tree

    def sameAs(self, other):
        if not super(Record, self).sameAs(other):
            return False

        if len(self.references) != len(other.references):
            return False

        for key, value in self.references.iteritems():
            other_value = other.lookup(key)
            if not other_value:
                return False
            if value != other_value:
                return False

        return True

    def _evaluate(self, state):
        return self.references

    def _evaluateTree(self, state, tree=None):
        value = self._evaluate(state)

        return (value, self)

    def _flatten(self, state, kind):
        new_record = Record()
        old_record = state.current
        state.current = self

        for name, atree in self.references.iteritems():
            evalue, etree = atree.flatten(state)
            if not etree:
                etree = Literal(evalue)

            new_record.insert(name, etree)

        state.current = old_record
        return (undefined, new_record, kind)

    def clear(self):
        self.references = {}
        self.chained_parent = None

    def insert(self, name, tree):
        if not name:
            raise Exception(
                "no reference name when inserting expression in record")

        if not tree:
            raise Exception(
                "no expression when inserting reference '%s' in record"
                % name)

        tree.setParentScope(self)
        self.references[name] = tree

        return True

    def insertNull(self, name):
        tree = Literal(None)
        return self.insert(name, tree)

    def insertBoolean(self, name, value):
        tree = Literal(value)
        return self.insert(name, tree)

    def insertInteger(self, name, value):
        tree = Literal(value)
        return self.insert(name, tree)

    def insertReal(self, name, value):
        tree = Literal(value)
        return self.insert(name, tree)

    def insertString(self, name, value):
        tree = Literal(value)
        return self.insert(name, tree)

    def evaluateName(self, name):
        state = EvalState()
        state.setScope(self)

        lookup, tree = self.lookupInScope(name, state)
        if lookup == EvalLookup.EVAL_OK:
            value = tree.evaluate(state)

        elif lookup == EvalLookup.EVAL_UNDEF:
            value = undefined

        elif lookup == EvalLookup.EVAL_ERROR:
            value = error

        else:
            raise Exception("Should not get here")

        return value

    def evaluateBoolean(self, name, default=None):
        value = self.evaluateName(name)
        if isinstance(value, bool):
            return value

        return default

    def evaluateInteger(self, name, default=None):
        value = self.evaluateName(name)
        if isinstance(value, int):
            return value

        return default

    def evaluateReal(self, name, default=None):
        value = self.evaluateName(name)
        if isinstance(value, float):
            return value

        return default

    def evaluateNumber(self, name, default=None):
        value = self.evaluateName(name)
        if isinstance(value, (int, float)):
            return value

        return default

    def evaluateString(self, name, default=None):
        value = self.evaluateName(name)
        if isinstance(value, basestring):
            return value

        return default

    def lookup(self, name):
        if name in self.references:
            tree = self.references[name]
        elif self.chained_parent:
            tree = self.chained_parent.lookup(name)
        else:
            tree = None

        return tree

    def lookupInScope(self, name, state):
        current = self
        lookup = EvalLookup.EVAL_UNDEF
        tree = None
        while not tree and current:
            state.current = current

            tree = current.lookup(name)
            if tree:
                lookup = EvalLookup.EVAL_OK
                break

            super_scope = current.parent
            if name == "root":
                tree = state.root
                lookup = EvalLookup.EVAL_OK if tree else EvalLookup.EVAL_UNDEF
                break

            elif name == "self":
                tree = state.current
                lookup = EvalLookup.EVAL_OK if tree else EvalLookup.EVAL_UNDEF
                break

            elif name == "parent":
                tree = state.current.parent
                lookup = EvalLookup.EVAL_OK if tree else EvalLookup.EVAL_UNDEF
                break

            else:
                current = super_scope
                if current == self:
                    lookup = EvalLookup.EVAL_UNDEF
                    break

        return (lookup, tree)
