#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Reference",
    ]

from checkbox.message.eval import (
    EvalLookup,
    EvalState,
    )
from checkbox.message.expression import Expression
from checkbox.message.list import List
from checkbox.message.literal import Literal
from checkbox.message.record import Record
from checkbox.message.types import (
    Error,
    Undefined,
    error,
    undefined,
    )


class Reference(Expression):

    __slots__ = (
        "name",
        "absolute",
        "expression",
        )

    def __init__(self, name=None, absolute=False, expression=None):
        super(Reference, self).__init__()
        self.name = name
        self.absolute = absolute
        self.expression = expression

    def copyFrom(self, other):
        super(Reference, self).copyFrom(other)
        self.name = other.name
        self.absolute = other.absolute
        if other.expression:
            self.expression = other.expression.copy()
        else:
            self.expression = None

    def sameAs(self, other):
        if not super(Reference, self).sameAs(other):
            return False

        if (self.absolute != other.absolute or
            self.name != other.name):
            return False

        if ((not self.expression and other.expression) or
            (self.expression and not other.expression)):
            return False

        if not self.expression and not other.expression:
            return True

        return self.expression.sameAs(other.expression)

    def _setParentScope(self, parent):
        if self.expression:
            self.expression._setParentScope(parent)

    def _evaluate(self, state):
        current = state.current

        lookup, tree_l, sig = self._lookupExpression(state)
        if lookup == EvalLookup.EVAL_ERROR:
            value = error

        elif lookup == EvalLookup.EVAL_UNDEF:
            value = undefined

        elif lookup == EvalLookup.EVAL_OK:
            if tree_l in state.cache:
                value = state.cache[tree_l]

            else:
                # Avoid circular reference
                state.cache[tree_l] = undefined

                value = tree_l.evaluate(state)
                state.cache[tree_l] = value

        else:
            raise Exception("Should not get here")

        state.current = current
        return value

    def _evaluateTree(self, state, tree=None):
        current = state.current

        lookup, tree_l, sig = self._lookupExpression(state, True)
        if lookup == EvalLookup.EVAL_ERROR:
            value = error

        elif lookup == EvalLookup.EVAL_UNDEF:
            value = undefined

        elif lookup == EvalLookup.EVAL_OK:
            if tree_l in state.cache:
                state.current = current
                value = state.cache[tree_l]

            else:
                # Avoid circular reference
                state.cache[tree_l] = undefined

                value = tree_l.evaluate(state)
                state.cache[tree_l] = value

        else:
            raise Exception("Should not get here")

        tree = Reference(self.name, self.absolute, sig)

        state.current = current
        return (value, tree)

    def _flatten(self, state, kind):
        current = state.current
        tree = None

        lookup, tree_l, sig = self._lookupExpression(state)
        if lookup == EvalLookup.EVAL_ERROR:
            value = error

        elif lookup == EvalLookup.EVAL_UNDEF:
            tree = self
            value = undefined

        elif lookup == EvalLookup.EVAL_OK:
            if tree_l in state.cache:
                value = state.cache[tree_l]
                if isinstance(value, Undefined):
                    tree = self

            else:
                # Avoid circular reference
                state.cache[tree_l] = undefined

                value, tree = tree_l.flatten(state)
                if tree:
                    if state.flatten_and_inline:
                        return (value, tree, kind)

                    tree = self
                    value = undefined
                    del state.cache[tree_l]
                else:
                    state.cache[tree_l] = value

        else:
            raise Exception("Should not get here")

        state.current = current
        return (value, tree, kind)

    def _lookupExpression(self, state, want_sig=False):
        tree = None
        sig = None

        # reference and .reference
        if not self.expression:
            current = state.root if self.absolute else state.current
            if self.absolute and not current:
                raise Exception("Circular reference detected")
            else:
                value = undefined

        # expression.reference
        else:
            if want_sig:
                value, sig = self.expression._evaluateTree(state)
            else:
                value = self.expression._evaluate(state)

            if isinstance(value, Undefined):
                return (EvalLookup.EVAL_UNDEF, tree, sig)

            elif isinstance(value, Error):
                return (EvalLookup.EVAL_ERROR, tree, sig)

            if not isinstance(value, dict) and not isinstance(value, list):
                return (EvalLookup.EVAL_ERROR, tree, sig)

            current = value

        if isinstance(value, list):
            expressions = []
            for data in value:
                # TODO: really need copy?
                reference = Reference(self.name, False, data.copy())

                state = EvalState()
                state.setScope(state.current)
                if want_sig:
                    value, sig = reference._evaluateTree(state)
                else:
                    value = reference._evaluate(state)

                if isinstance(value, dict):
                    expressions.append(value)

                elif isinstance(value, list):
                    expressions.append(value)

                else:
                    literal = Literal(value)
                    expressions.append(literal)

            tree = List(expressions)

            record = Record()
            tree.setParentScope(record)
            return (EvalLookup.EVAL_OK, tree, sig)

        if current:
            lookup, tree = current.lookupInScope(self.name, state)
        else:
            lookup = EvalLookup.EVAL_UNDEF

        return (lookup, tree, sig)
