#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from unittest import TestCase

from checkbox.message.eval import EvalState
from checkbox.message.function import Function
from checkbox.message.list import List
from checkbox.message.literal import Literal
from checkbox.message.record import Record
from checkbox.message.reference import Reference
from checkbox.message.types import undefined


class TestFunction(TestCase):

    def test_sameAs(self):
        # Same reference
        function1 = Function()
        self.assertNotEquals(function1, None)
        self.assertEquals(function1, function1)

        # Different reference
        function2 = Function()
        self.assertEquals(function1, function2)

    def test_int(self):
        for value in 1, 1.0, "1":
            function = Function("int", Literal(value))
            result = function.evaluate()
            self.assertTrue(isinstance(result, int))
            self.assertEquals(result, 1)

    def test_real(self):
        for value in 1, 1.0, "1.0":
            function = Function("real", Literal(value))
            result = function.evaluate()
            self.assertTrue(result)
            self.assertEquals(result, 1.0)

    def test_string(self):
        for value in 1, "1", u"1":
            function = Function("string", Literal(value))
            result = function.evaluate()
            self.assertTrue(isinstance(result, basestring))
            self.assertEquals(result, "1")

    def test_string_int(self):
            function = Function("string", Literal(1))
            result = function.evaluate()
            self.assertEquals(result, "1")

    def test_string_real(self):
            function = Function("string", Literal(0.0))
            result = function.evaluate()
            self.assertEquals(result, "0.0")

    def test_string_bool(self):
            function = Function("string", Literal(True))
            result = function.evaluate()
            self.assertEquals(result, "true")

    def test_string_list(self):
            function = Function("string", List([Literal(1)]))
            result = function.evaluate()
            self.assertEquals(result, "[ 1, ]")

    def test_string_record(self):
            function = Function("string", Record({"a": Literal(1)}))
            result = function.evaluate()
            self.assertEquals(result, "{ a = 1; }")

    def test_bool(self):
        # False values
        for value in 0, 0.0, "", "false", "FALSE":
            function = Function("bool", Literal(value))
            result = function.evaluate()
            self.assertTrue(isinstance(result, bool))
            self.assertEquals(result, False)

        # True values
        for value in 1, 2, 1.0, "true", "TRUE":
            function = Function("bool", Literal(value))
            result = function.evaluate()
            self.assertTrue(isinstance(result, bool))
            self.assertEquals(result, True)

    def test_evaluate(self):
        function = Function("int", Literal("1"))
        result = function.evaluate()
        self.assertTrue(isinstance(result, int))
        self.assertEquals(result, 1)

    def test_flatten(self):
        function = Function("int", Literal("1"))
        state = EvalState()
        value, tree = function.flatten(state)
        self.assertTrue(isinstance(value, int))
        self.assertEquals(value, 1)
        self.assertEquals(tree, None)

    def test_flatten_expression(self):
        # With non-existent state
        function = Function("int", Reference("a"))
        state = EvalState()
        value, tree = function.flatten(state)
        self.assertEquals(value, undefined)
        self.assertEquals(tree, function)

        # With existent state
        state.setScope(Record({"a": Literal(1)}))
        value, tree = function.flatten(state)
        self.assertTrue(isinstance(value, int))
        self.assertEquals(value, 1)
        self.assertEquals(tree, None)
