#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from unittest import TestCase

from checkbox.message.literal import Literal
from checkbox.message.operator import (
    Operator,
    OperatorKind,
    OperatorSignal,
    )
from checkbox.message.types import Error


class TestOperator(TestCase):

    def test_copyFrom(self):
        literal1 = Literal(1)
        literal2 = Literal(2)

        # Different operators
        operator1 = Operator(OperatorKind.UNARY_PLUS_OP, literal1)
        operator2 = Operator(OperatorKind.UNARY_PLUS_OP, literal2)
        self.assertNotEquals(operator1, operator2)

        # Same operators
        operator1.copyFrom(operator2)
        self.assertEquals(operator1, operator2)

    def test_sameAs(self):
        # Same reference
        literal1 = Literal(1)
        operator1 = Operator(OperatorKind.UNARY_PLUS_OP, literal1)
        self.assertNotEquals(operator1, None)
        self.assertEquals(operator1, operator1)

        # Different reference
        operator2 = Operator(OperatorKind.UNARY_PLUS_OP, literal1)
        self.assertEquals(operator1, operator2)

        # Different value reference, same data
        literal2 = Literal(1)
        operator2 = Operator(OperatorKind.UNARY_PLUS_OP, literal2)
        self.assertEquals(operator1, operator2)

        # Different value reference, different data
        literal2 = Literal(2)
        operator2 = Operator(OperatorKind.UNARY_PLUS_OP, literal2)
        self.assertNotEquals(operator1, operator2)

    def test_evaluate(self):
        # Unary
        literal1 = Literal(1)
        operator = Operator(OperatorKind.UNARY_MINUS_OP, literal1)

        result = operator.evaluate()
        self.assertTrue(isinstance(result, int))
        self.assertEquals(result, -1)

        # Binary
        literal2 = Literal(2)
        operator = Operator(OperatorKind.ADDITION_OP, literal1, literal2)

        result = operator.evaluate()
        self.assertTrue(isinstance(result, int))
        self.assertEquals(result, 3)

        # Ternary
        literal3 = Literal(False)
        operator = Operator(
            OperatorKind.TERNARY_OP, literal3, literal1, literal2)

        result = operator.evaluate()
        self.assertTrue(isinstance(result, int))
        self.assertEquals(result, 2)

    def test_doOperation(self):
        result = Operator._doOperation(OperatorKind.UNARY_PLUS_OP, 1)[0]
        self.assertEquals(result, 1)

        result = Operator._doOperation(
            OperatorKind.UNARY_MINUS_OP, 1)[0]
        self.assertEquals(result, -1)

        result = Operator._doOperation(
            OperatorKind.ADDITION_OP, 1, 2)[0]
        self.assertEquals(result, 3)

        result = Operator._doOperation(
            OperatorKind.TERNARY_OP, True, 2, 3)[0]
        self.assertEquals(result, 2)

    def test_doComparison(self):
        result = Operator._doComparison(
            OperatorKind.EQUAL_OP, "A", "a")[0]
        self.assertEquals(result, False)

        result = Operator._doComparison(
            OperatorKind.META_EQUAL_OP, "A", "a")[0]
        self.assertEquals(result, True)

        result = Operator._doComparison(
            OperatorKind.NOT_EQUAL_OP, "A", "a")[0]
        self.assertEquals(result, True)

        result = Operator._doComparison(
            OperatorKind.META_NOT_EQUAL_OP, "A", "a")[0]
        self.assertEquals(result, False)

        result = Operator._doComparison(
            OperatorKind.EQUAL_OP, "A", 1)[0]
        self.assertTrue(isinstance(result, Error))

    def test_doArithmetic(self):
        result = Operator._doArithmetic(
            OperatorKind.UNARY_MINUS_OP, "a", "b")[0]
        self.assertTrue(isinstance(result, Error))

        result = Operator._doArithmetic(
            OperatorKind.UNARY_MINUS_OP, 1, "b")[0]
        self.assertEquals(result, -1)

        result = Operator._doArithmetic(
            OperatorKind.UNARY_MINUS_OP, 1.0, "b")[0]
        self.assertEquals(result, -1.0)

        result = Operator._doArithmetic(
            OperatorKind.UNARY_MINUS_OP, True, "b")[0]
        self.assertEquals(result, False)

    def test_doLogical(self):
        result, signal = Operator._doLogical(
            OperatorKind.LOGICAL_NOT_OP, 0, 1)
        self.assertTrue(isinstance(result, Error))
        self.assertEquals(signal, OperatorSignal.SIG_CHLD1)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_NOT_OP, True, False)[0]
        self.assertEquals(result, False)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_NOT_OP, False, True)[0]
        self.assertEquals(result, True)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_OR_OP, True, False)[0]
        self.assertEquals(result, True)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_OR_OP, True, True)[0]
        self.assertEquals(result, True)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_OR_OP, False, False)[0]
        self.assertEquals(result, False)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_AND_OP, True, False)[0]
        self.assertEquals(result, False)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_AND_OP, True, True)[0]
        self.assertEquals(result, True)

        result = Operator._doLogical(
            OperatorKind.LOGICAL_AND_OP, False, False)[0]
        self.assertEquals(result, False)

    def test_doMembership(self):
        result, signal = Operator._doMembership(
            OperatorKind.MEMBERSHIP_IN, 0, 1)
        self.assertTrue(isinstance(result, Error))
        self.assertEquals(signal, OperatorSignal.SIG_CHLD2)

        result = Operator._doMembership(
            OperatorKind.MEMBERSHIP_IN, 0, [1])[0]
        self.assertEquals(result, False)

        result = Operator._doMembership(
            OperatorKind.MEMBERSHIP_IN, 0, [0])[0]
        self.assertEquals(result, True)

    def test_doBitwise(self):
        result, signal = Operator._doBitwise(
            OperatorKind.BITWISE_NOT_OP, "a", None)
        self.assertTrue(isinstance(result, Error))
        self.assertEquals(signal, OperatorSignal.SIG_CHLD1)

        result, signal = Operator._doBitwise(
            OperatorKind.BITWISE_NOT_OP, 1, None)
        self.assertEquals(result, -2)
        self.assertEquals(signal, OperatorSignal.SIG_CHLD1)

        result = Operator._doBitwise(
            OperatorKind.BITWISE_OR_OP, 1, 2)[0]
        self.assertEquals(result, 3)

        result = Operator._doBitwise(
            OperatorKind.BITWISE_AND_OP, 1, 2)[0]
        self.assertEquals(result, 0)

        result = Operator._doBitwise(
            OperatorKind.BITWISE_XOR_OP, 1, 2)[0]
        self.assertEquals(result, 3)

        result = Operator._doBitwise(
            OperatorKind.LEFT_SHIFT_OP, 1, 2)[0]
        self.assertEquals(result, 4)

        result = Operator._doBitwise(
            OperatorKind.LEFT_SHIFT_OP, 2, 1)[0]
        self.assertEquals(result, 4)

        result = Operator._doBitwise(
            OperatorKind.URIGHT_SHIFT_OP, 1, 2)[0]
        self.assertEquals(result, 0)

        result = Operator._doBitwise(
            OperatorKind.URIGHT_SHIFT_OP, 2, 1)[0]
        self.assertEquals(result, 1)

        result = Operator._doBitwise(
            OperatorKind.RIGHT_SHIFT_OP, 1, 2)[0]
        self.assertEquals(result, 0)

        result = Operator._doBitwise(
            OperatorKind.RIGHT_SHIFT_OP, 2, 1)[0]
        self.assertEquals(result, 1)

    def test_doIntegerArithmetic(self):
        result = Operator._doIntegerArithmetic(
            OperatorKind.ADDITION_OP, 0, 1)[0]
        self.assertEquals(result, 0 + 1)

        result = Operator._doIntegerArithmetic(
            OperatorKind.SUBTRACTION_OP, 0, 1)[0]
        self.assertEquals(result, 0 - 1)

        result = Operator._doIntegerArithmetic(
            OperatorKind.MULTIPLICATION_OP, 0, 1)[0]
        self.assertEquals(result, 0 * 1)

        result = Operator._doIntegerArithmetic(
            OperatorKind.DIVISION_OP, 0, 1)[0]
        self.assertEquals(result, 0 / 1)

        result = Operator._doIntegerArithmetic(
            OperatorKind.DIVISION_OP, 1, 0)[0]
        self.assertTrue(isinstance(result, Error))

        result = Operator._doIntegerArithmetic(
            OperatorKind.MODULUS_OP, 0, 1)[0]
        self.assertEquals(result, 0 % 1)

        result = Operator._doIntegerArithmetic(
            OperatorKind.MODULUS_OP, 1, 0)[0]
        self.assertTrue(isinstance(result, Error))

    def test_doRealArithmetic(self):
        result = Operator._doRealArithmetic(
            OperatorKind.ADDITION_OP, 0.0, 1.0)[0]
        self.assertEquals(result, 0.0 + 1.0)

        result = Operator._doRealArithmetic(
            OperatorKind.SUBTRACTION_OP, 0.0, 1.0)[0]
        self.assertEquals(result, 0.0 - 1.0)

        result = Operator._doRealArithmetic(
            OperatorKind.MULTIPLICATION_OP, 0.0, 1.0)[0]
        self.assertEquals(result, 0.0 * 1.0)

        result = Operator._doRealArithmetic(
            OperatorKind.DIVISION_OP, 0.0, 1.0)[0]
        self.assertEquals(result, 0.0 / 1.0)

        result = Operator._doRealArithmetic(
            OperatorKind.MODULUS_OP, 0.0, 1.0)[0]
        self.assertTrue(isinstance(result, Error))

    def test_compareIntegers(self):
        for kind in (OperatorKind.LESS_THAN_OP,
                OperatorKind.LESS_OR_EQUAL_OP,
                OperatorKind.NOT_EQUAL_OP):
            result = Operator._compareIntegers(kind, 0, 1)
            self.assertTrue(result)

        for kind in (OperatorKind.GREATER_THAN_OP,
                OperatorKind.GREATER_OR_EQUAL_OP,
                OperatorKind.EQUAL_OP):
            result = Operator._compareIntegers(kind, 0, 1)
            self.assertFalse(result)

    def test_compareStrings(self):
        for kind in (OperatorKind.LESS_THAN_OP,
                OperatorKind.LESS_OR_EQUAL_OP,
                OperatorKind.NOT_EQUAL_OP):
            result = Operator._compareStrings(kind, "A", "a", True)
            self.assertTrue(result)

        for kind in (OperatorKind.GREATER_THAN_OP,
                OperatorKind.GREATER_OR_EQUAL_OP,
                OperatorKind.EQUAL_OP):
            result = Operator._compareStrings(kind, "A", "a", True)
            self.assertFalse(result)

        result = Operator._compareStrings(
            OperatorKind.EQUAL_OP, "A", "a", False)
        self.assertTrue(result)

        result = Operator._compareStrings(
            OperatorKind.NOT_EQUAL_OP, "A", "a", False)
        self.assertFalse(result)

    def test_coerceToNumber(self):
        number_type = Operator._coerceToNumber("a", "a")[2]
        self.assertEquals(number_type, basestring)

        number_type = Operator._coerceToNumber("a", 1)[2]
        self.assertEquals(number_type, basestring)

        value1, value2, number_type = Operator._coerceToNumber(
            True, False)
        self.assertEquals(number_type, int)
        self.assertEquals(value1, 1)
        self.assertEquals(value2, 0)

        value1, value2, number_type = Operator._coerceToNumber(1, 2)
        self.assertEquals(number_type, int)
        self.assertEquals(value1, 1)
        self.assertEquals(value2, 2)

        value1, value2, number_type = Operator._coerceToNumber(1, 2.0)
        self.assertEquals(number_type, float)
        self.assertTrue(isinstance(value1, float))
        self.assertTrue(isinstance(value2, float))

    def test_checkStringOperator(self):
        self.assertTrue(
            Operator.checkStringOperator(OperatorKind.MULTIPLICATION_OP))
        self.assertFalse(
            Operator.checkStringOperator(OperatorKind.LOGICAL_AND_OP))

    def test_getPrecedenceLevel(self):
        multiplication = Operator.getPrecedenceLevel(
            OperatorKind.MULTIPLICATION_OP)
        addition = Operator.getPrecedenceLevel(OperatorKind.ADDITION_OP)
        self.assertTrue(multiplication > addition)
