#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

from unittest import TestCase

from checkbox.message.list import List
from checkbox.message.literal import Literal
from checkbox.message.operator import (
    Operator,
    OperatorKind,
    )
from checkbox.message.record import Record
from checkbox.message.reference import Reference
from checkbox.message.types import (
    error,
    undefined,
    )
from checkbox.message.unparser import Unparser


class TestUnparser(TestCase):

    def test_unparseValue(self):
        # NULL
        string = Unparser.unparseValue(None)
        self.assertEquals(string, "null")

        # ERROR
        string = Unparser.unparseValue(error)
        self.assertEquals(string, "error")

        # UNDEFINED
        string = Unparser.unparseValue(undefined)
        self.assertEquals(string, "undefined")

        # BOOLEAN
        string = Unparser.unparseValue(True)
        self.assertEquals(string, "true")

        # INTEGER
        string = Unparser.unparseValue(1)
        self.assertEquals(string, "1")

        # REAL
        string = Unparser.unparseValue(1.0)
        self.assertEquals(string, "1.000000000000000E+00")

        # STRING
        string = Unparser.unparseValue("a")
        self.assertEquals(string, "\"a\"")

        # RECORD
        string = Unparser.unparseValue({})
        self.assertEquals(string, "{ }")

        # LIST
        string = Unparser.unparseValue([])
        self.assertEquals(string, "[ ]")

    def test_unparseLiteral(self):
        literal = Literal(1)
        string = Unparser.unparseLiteral(literal)
        self.assertEquals(string, "1")

    def test_unparseReference(self):
        reference1 = Reference("a", True)
        string = Unparser.unparseReference(reference1)
        self.assertEquals(string, ".a")

        reference1 = Reference("a", False)
        string = Unparser.unparseReference(reference1)
        self.assertEquals(string, "a")

        reference2 = Reference("b", expression=reference1)
        string = Unparser.unparseReference(reference2)
        self.assertEquals(string, "a.b")

    def test_unparseOperator(self):
        literal1 = Literal(1)
        literal2 = Literal(2)
        literal3 = Literal(True)

        # PARENTHESES_OP
        operator = Operator(OperatorKind.PARENTHESES_OP, literal1)
        string = Unparser.unparseOperator(operator)
        self.assertEquals(string, "(1)")

        # UNARY_PLUS_OP
        operator = Operator(OperatorKind.UNARY_PLUS_OP, literal1)
        string = Unparser.unparseOperator(operator)
        self.assertEquals(string, " +1")

        # TERNARY_OP
        operator = Operator(
            OperatorKind.TERNARY_OP, literal3, literal1, literal2)
        string = Unparser.unparseOperator(operator)
        self.assertEquals(string, "true ? 1 : 2")

        # SUBSCRIPT_OP
        operator = Operator(OperatorKind.SUBSCRIPT_OP, literal1, literal2)
        string = Unparser.unparseOperator(operator)
        self.assertEquals(string, "1[2]")

        # ADDITION_OP
        operator = Operator(OperatorKind.ADDITION_OP, literal1, literal2)
        string = Unparser.unparseOperator(operator)
        self.assertEquals(string, "1 + 2")

    def test_unparseRecord(self):
        record = Record()
        string = Unparser.unparseRecord(record)
        self.assertEquals(string, "{ }")

        literal = Literal(1)
        record.insert("a", literal)
        string = Unparser.unparseRecord(record)
        self.assertEquals(string, "{ a = 1; }")

    def test_unparseList(self):
        list = List()
        string = Unparser.unparseList(list)
        self.assertEquals(string, "[ ]")

        literal = Literal(1)
        list.append(literal)
        string = Unparser.unparseList(list)
        self.assertEquals(string, "[ 1, ]")

    def test_unparseIdentifier(self):
        string = Unparser.unparseIdentifier("a")
        self.assertEquals(string, "a")

        string = Unparser.unparseIdentifier("a b")
        self.assertEquals(string, "'a b'")
