#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Unparser",
    "UnparserStream",
    ]

import re

from checkbox.message.convert import unconvert_escapes
from checkbox.message.function import Function
from checkbox.message.list import List
from checkbox.message.literal import Literal
from checkbox.message.operator import (
    Operator,
    OperatorKind,
    )
from checkbox.message.record import Record
from checkbox.message.reference import Reference
from checkbox.message.types import (
    Error,
    Undefined,
    )


FACTOR_TO_CHAR = (
    "",
    "B",
    "K",
    "M",
    "G",
    "T")

OPERATOR_TO_STRING = (
    "",      # no-op

    " < ",   # comparisons
    " <= ",
    " != ",
    " == ",
    " >= ",
    " > ",
    " is ",
    " isnt ",

    " +",    # arithmetic
    " -",
    " + ",
    " - ",
    " * ",
    " / ",
    " % ",

    " !",    # logical
    " || ",
    " && ",

    " ~",    # bitwise
    " | ",
    " ^ ",
    " & ",
    " << ",
    " >> ",
    " >>> ",

    " in ",  # membership

    " () ",  # misc mixfix --- no "single token" representation
    " [] ",
    " ?: ")

IDENTIFIER_PATTERN = r"[A-Za-z_]\w*"
IDENTIFIER_RE = re.compile(r"^%s$" % IDENTIFIER_PATTERN)


class Unparser:

    string_delimiter = "\""
    name_delimiter = "'"

    @classmethod
    def unparseValue(cls, value, *args):
        buffer = ""
        if value is None:
            buffer += "null"

        elif isinstance(value, basestring):
            string = unconvert_escapes(value, cls.string_delimiter)
            buffer += "%s%s%s" % (
                cls.string_delimiter, string, cls.string_delimiter)

        elif isinstance(value, bool):
            buffer += "true" if value else "false"

        elif isinstance(value, int):
            buffer += "%d" % value

        elif isinstance(value, float):
            # TODO: add support for NaN, INF and -INF
            if value == 0.0:
                buffer += "%.1f" % value

            else:
                buffer += "%.15E" % value

        elif isinstance(value, Undefined):
            buffer += "undefined"

        elif isinstance(value, Error):
            buffer += "error"

        elif isinstance(value, dict):
            buffer += cls.unparseRecord(value, *args)

        elif isinstance(value, list):
            buffer += cls.unparseList(value, *args)

        else:
            raise Exception("Should not get here")

        return buffer

    @classmethod
    def unparseExpression(cls, expression, *args):
        buffer = ""
        if not expression:
            return buffer

        if isinstance(expression, Literal):
            buffer += cls.unparseLiteral(expression, *args)

        elif isinstance(expression, Reference):
            buffer += cls.unparseReference(expression, *args)

        elif isinstance(expression, Operator):
            buffer += cls.unparseOperator(expression, *args)

        elif isinstance(expression, Function):
            buffer += cls.unparseFunction(expression, *args)

        elif isinstance(expression, Record):
            buffer += cls.unparseRecord(expression, *args)

        elif isinstance(expression, List):
            buffer += cls.unparseList(expression, *args)

        else:
            raise Exception("Should not get here")

        return buffer

    @classmethod
    def unparseLiteral(cls, literal, *args):
        buffer = cls.unparseValue(literal.value, *args)
        if isinstance(literal.value, (int, float)):
            buffer += FACTOR_TO_CHAR[literal.factor]

        return buffer

    @classmethod
    def unparseReference(cls, reference, *args):
        if reference.expression:
            buffer = cls.unparseExpression(reference.expression, *args)
            buffer += ".%s" % reference.name
        else:
            buffer = "." if reference.absolute else ""
            buffer += cls.unparseIdentifier(reference.name, *args)

        return buffer

    @classmethod
    def unparseOperator(cls, operator, *args):
        if operator.kind == OperatorKind.PARENTHESES_OP:
            buffer = "(%s)" % cls.unparseExpression(operator._child1, *args)

        elif (operator.kind == OperatorKind.UNARY_PLUS_OP or
            operator.kind == OperatorKind.UNARY_MINUS_OP or
            operator.kind == OperatorKind.LOGICAL_NOT_OP or
            operator.kind == OperatorKind.BITWISE_NOT_OP):

            buffer = "%s%s" % (OPERATOR_TO_STRING[operator.kind],
                cls.unparseExpression(operator._child1, *args))

        elif operator.kind == OperatorKind.TERNARY_OP:
            buffer = "%s ? %s : %s" % (
                cls.unparseExpression(operator._child1, *args),
                cls.unparseExpression(operator._child2, *args),
                cls.unparseExpression(operator._child3, *args))

        elif operator.kind == OperatorKind.SUBSCRIPT_OP:
            buffer = "%s[%s]" % (
                cls.unparseExpression(operator._child1, *args),
                cls.unparseExpression(operator._child2, *args))

        else:
            buffer = "%s%s%s" % (
                cls.unparseExpression(operator._child1, *args),
                OPERATOR_TO_STRING[operator.kind],
                cls.unparseExpression(operator._child2, *args))

        return buffer

    @classmethod
    def unparseFunction(cls, function, *args):
        buffer = function.function_name + "("
        buffer += ",".join([cls.unparseExpression(argument)
            for argument in function.arguments])
        buffer += ")"
        return buffer

    @classmethod
    def unparseRecord(cls, record, *args):
        buffer = "{ "
        for identifier, expression in record.items():
            buffer += "%s = %s; " % (
                cls.unparseIdentifier(identifier, *args),
                cls.unparseExpression(expression, *args))

        buffer += "}"
        return buffer

    @classmethod
    def unparseList(cls, list, *args):
        buffer = "[ "
        for expression in list:
            buffer += "%s, " % cls.unparseExpression(expression, *args)

        buffer += "]"
        return buffer

    @classmethod
    def unparseIdentifier(cls, identifier, *args):
        identifier = unconvert_escapes(identifier, cls.name_delimiter)
        if not IDENTIFIER_RE.match(identifier):
            return "%s%s%s" % (
                cls.name_delimiter, identifier,
                cls.name_delimiter)

        return identifier


class UnparserStream(Unparser):

    @classmethod
    def unparseRecord(cls, record, depth=0):
        buffer = "{ " if depth else ""

        for identifier, expression in record.references.items():
            buffer += "%s = %s" % (
                cls.unparseIdentifier(identifier, depth + 1),
                cls.unparseExpression(expression, depth + 1))
            buffer += "; " if depth else "\n"

        if depth:
            buffer += "}"

        return buffer

    @classmethod
    def unparseList(cls, list, depth=0):
        buffer = "[ " if depth else ""

        for expression in list:
            buffer += cls.unparseExpression(expression, depth + 1)
            buffer += ", " if depth else "\n"

        if depth:
            buffer += "]"

        return buffer
