#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Args",
    ]

import logging

from string import whitespace


class Args(list):

    def __init__(self, *args):
        self.extend(args)

    def appendQuoted(self, quoted):
        raw = self.quotedToRaw(quoted)
        if raw is None:
            return False

        return self.appendRaw(raw)

    def appendRaw(self, raw):
        return self.splitRaw(raw, self)

    def appendRecord(self, record, attribute):
        string = record.evaluateString(attribute)
        if string is not None:
            return self.appendRaw(string)

        else:
            return True

    @classmethod
    def quotedToRaw(cls, quoted):
        if not quoted:
            return True

        index = 0
        while index < len(quoted):
            if quoted[index] in whitespace:
                index += 1
            else:
                break

        assert(quoted[index] == '"')
        index += 1

        raw = ""
        quote_terminated = None
        while index < len(quoted):
            character = quoted[index]
            if character == '"':
                if index + 1 < len(quoted) and quoted[index + 1] == '"':
                    # Repeated double quote treated as an exscape
                    raw += character
                    index += 2
                else:
                    quote_terminated = quoted[index:]
                    index += 1
                    break
            else:
                raw += character
                index += 1

        if not quote_terminated:
            logging.error("Unterminated double quote")
            return None

        while index < len(quoted):
            if quoted[index] in whitespace:
                index += 1
            else:
                break

        if index != len(quoted):
            logging.error("Unexpected character following double quote")
            return None

        return raw

    @classmethod
    def argToRaw(cls, arg, raw=""):
        if raw:
            raw += " "

        if not arg:
            raw += "''"

        for character in arg:
            if character in (" ", "\t", "\n", "\r", "\'"):
                if raw and raw[-1] == "\'":
                    # Combine preceeding quote
                    raw = raw[:-1]

                else:
                    raw += "\'"

                if character == "\'":
                    # Repeat quote to escape it
                    raw += "\'"

                raw += character
                raw += "\'"

            else:
                raw += character

        return raw

    @classmethod
    def joinArgs(cls, args, start=0):
        raw = ""
        for index, arg in enumerate(args):
            if index >= start:
                raw = cls.argToRaw(arg, raw)

        return raw

    @classmethod
    def splitRaw(cls, raw, args):
        if not raw:
            return True

        buffer = ""
        parsed_token = False

        index = 0
        while index < len(raw):
            character = raw[index]
            if character == "'":
                parsed_token = True
                quote = raw[index:]
                index += 1

                while index < len(raw):
                    character = raw[index]
                    if character == quote[0]:
                        if index + 1 < len(raw) and raw[index + 1] == quote[0]:
                            # Repeated quote treated as an exscape
                            buffer += character
                            index += 2
                        else:
                            break
                    else:
                        buffer += character
                        index += 1

                if index == len(raw):
                    logging.error("Unbalanced quote starting here: %s", quote)
                    return False

                # Skip the closing quote
                index += 1

            elif character in (" ", "\t", "\n", "\r"):
                index += 1
                if parsed_token:
                    parsed_token = False
                    args.append(buffer)
                    buffer = ""

            else:
                parsed_token = True
                buffer += character
                index += 1

        if parsed_token:
            args.append(buffer)

        return True
