#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Client",
    "ClientError",
    ]

import logging

from checkbox.io.fifo import Fifo

from checkbox.process.monitor import MonitorStatus
from checkbox.process.operation import OperationCode


class ClientError(Exception):

    pass


class Client:

    __slots__ = (
        "_fifo",
        "_path",
        )

    def __init__(self, path):
        self._path = path
        self._fifo = Fifo()
        self._fifo.setTimeout(None)
        if not self._fifo.connect(path):
            raise ClientError("Error connecting to fifo path: ", path)

    def registerSubfamily(self, pid, watcher_pid, max_snapshot_interval):
        logging.debug("About to register subfamily of %d", pid)
        self._fifo.putInteger(OperationCode.REGISTER_SUBFAMILY)
        self._fifo.putInteger(pid)
        self._fifo.putInteger(watcher_pid)
        self._fifo.putInteger(max_snapshot_interval)

        return self._getEnd()

    def trackEnvironment(self, pid, pid_env_id):
        logging.debug("About to track %d from environment", pid)
        self._fifo.putInteger(OperationCode.TRACK_ENVIRONMENT)
        self._fifo.putInteger(pid)

        for ancestor in pid_env_id.ancestors:
            self._fifo.putBoolean(ancestor.active)
            self._fifo.putString(ancestor.envid)

        return self._getEnd()

    def trackLogin(self, pid, login):
        logging.debug("About to track %d from login %s", pid, login)
        self._fifo.putInteger(OperationCode.TRACK_LOGIN)
        self._fifo.putInteger(pid)
        self._fifo.putString(login)

        return self._getEnd()

    def signalProcess(self, pid, signal):
        logging.debug("About to signal process %d with %d", pid, signal)
        self._fifo.putInteger(OperationCode.SIGNAL_PROCESS)
        self._fifo.putInteger(pid)
        self._fifo.putInteger(signal)

        return self._getEnd()

    def suspendFamily(self, pid):
        logging.debug("About to suspend family %d", pid)

        return self._putPidCode(pid, OperationCode.SUSPEND_FAMILY)

    def resumeFamily(self, pid):
        logging.debug("About to resume family %d", pid)

        return self._putPidCode(pid, OperationCode.RESUME_FAMILY)

    def killFamily(self, pid):
        logging.debug("About to kill family %d", pid)

        return self._putPidCode(pid, OperationCode.KILL_FAMILY)

    def unregisterFamily(self, pid):
        logging.debug("About to unregister family of %d", pid)

        return self._putPidCode(pid, OperationCode.UNREGISTER_FAMILY)

    def _putPidCode(self, pid, code):
        self._fifo.putInteger(code)
        self._fifo.putInteger(pid)

        return self._getEnd()

    def takeSnapshot(self):
        logging.debug("About to take snapshot")

        return self._putCode(OperationCode.TAKE_SNAPSHOT)

    def quit(self):
        logging.debug("About to quit")

        return self._putCode(OperationCode.QUIT)

    def _putCode(self, code):
        self._fifo.putInteger(code)

        return self._getEnd()

    def _getEnd(self):
        if not self._fifo.putEnd():
            raise ClientError(
                "Failed to start connection with process server.")

        status = self._fifo.getInteger()
        if status is None:
            raise ClientError("Failed to read response from process server.")

        return status == MonitorStatus.SUCCESS
