#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Family",
    "FamilyInfo",
    ]

import os
import logging


class FamilyInfo:

    __slots__ = (
        "max_snapshot_interval",
        "login",
        "group",
        )

    def __init__(self):
        self.max_snapshot_interval = -1
        self.login = None
        self.group = None


class FamilyMember:

    __slots__ = (
        "_family",
        "_info",
        "_still_alive",
        )

    def __init__(self, family=None, info=None):
        self._family = family
        self._info = info
        self._still_alive = False

    @property
    def family(self):
        return self._family

    @property
    def info(self):
        """Get process info."""
        return self._info

    def stillAlive(self, info):
        """Update the info object to indicate the process is still alive."""
        self._info = info
        self._still_alive = True

    def moveToSubfamily(self, subfamily):
        """Move a process into a registered family."""
        if self._family:
            self._family._member_list.remove(self)

        subfamily._member_list.append(self)
        self._family = subfamily


class Family:

    __slots__ = (
        "_monitor",
        "_root_pid",
        "_root_birthday",
        "_watcher_pid",
        "_max_snapshot_interval",
        "_exited_user_cpu_time",
        "_exited_sys_cpu_time",
        "_max_image_size",
        "_member_list",
        )

    def __init__(self, monitor, root_pid=0, root_birthday=0, watcher_pid=0,
            max_snapshot_interval=-1):
        self._monitor = monitor
        self._root_pid = root_pid
        self._root_birthday = root_birthday
        self._watcher_pid = watcher_pid
        self._max_snapshot_interval = max_snapshot_interval
        self._exited_user_cpu_time = 0
        self._exited_sys_cpu_time = 0
        self._max_image_size = 0
        self._member_list = []

    @property
    def root_pid(self):
        return self._root_pid

    @property
    def root_birthday(self):
        return self._root_birthday

    @property
    def watcher_pid(self):
        return self._watcher_pid

    @property
    def max_image_size(self):
        return self._max_image_size

    @property
    def max_snapshot_interval(self):
        return self._max_snapshot_interval

    def addMember(self, info):
        member = FamilyMember()
        member._family = self
        member._info = info
        member._still_alive = False
        self._member_list.append(member)
        self._monitor.addMember(member)

    def updateMaxImageSize(self, image_size):
        for member in self._member_list:
            image_size += member._info.image_size

        if image_size > self._max_image_size:
            self._max_image_size = image_size

        return image_size

    def aggregateUsage(self, usage):
        for member in self._member_list:
            # cpu
            usage.user_cpu_time += member.info.user_time
            usage.sys_cpu_time += member.info.sys_time
            usage.percent_cpu += member.info.cpu_usage

            # image size
            usage.total_image_size += member.info.image_size

            # process count
            usage.process_count += 1

        # Factor usage from processes that exited
        usage.user_cpu_time += self._exited_user_cpu_time
        usage.sys_cpu_time += self._exited_sys_cpu_time

    def signalRoot(self, signal):
        try:
            os.kill(self._root_pid, signal)
        except OSError, e:
            logging.info(
                "kill(%d, %d) error: %d", self._root_pid, signal, e.errno)

    def spree(self, signal):
        for member in self._member_list:
            try:
                os.kill(member._info.pid, signal)
            except OSError, e:
                logging.info(
                    "kill(%d, %d) error: %d", self._root_pid, signal, e.errno)

    def removeExitedProcesses(self):
        for member in self._member_list:
            if not member._still_alive:
                if self._root_pid:
                    logging.info(
                        "process %d (of family %d) has exited",
                        member._info.pid, self._root_pid)

                else:
                    logging.info(
                        "process %d (not in monitored family) has exited",
                        member._info.pid)

                # Save cpu usage information
                self._exited_user_cpu_time = member._info.user_time
                self._exited_sys_cpu_time = member._info.sys_time

                # Update monitor table
                self._monitor.removeMember(member)

                self._member_list.remove(member)

            else:
                member._still_alive = False

    def foldIntoParent(self, parent):
        # Fold cpu usage information
        parent._exited_user_cpu_time += self._exited_user_cpu_time
        parent._exited_sys_cpu_time += self._exited_sys_cpu_time

        if not self._member_list:
            return

        for member in self._member_list:
            member._family = parent

        parent._member_list = self._member_list
        self._member_list = []
