#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Handler",
    ]

import os
import logging

from signal import (
    SIGCONT,
    SIGKILL,
    SIGSTOP,
    )

from checkbox.lib.enum import Enum

from checkbox.process.info import Info
from checkbox.process.pidenv import PidEnvId


HandlerDirection = Enum(
    "PATRICIDE",    # parent die first, then kids
    "INFANTICIDE",  # children die first, then parent
    )


class HandlerPid:

    __slots__ = (
        "pid",
        "ppid",
        "birthday",
        "cpu_user_time",
        "cpu_sys_time",
        )

    def __init__(self):
        self.pid = 0
        self.ppid = 0
        self.birthday = 0
        self.cpu_user_time = 0
        self.cpu_sys_time = 0


class Handler:

    __slots__ = (
        "_pid_env_id",
        "_login",
        "_alive_cpu_sys_time",
        "_alive_cpu_user_time",
        "_exited_cpu_sys_time",
        "_exited_cpu_user_time",
        "_max_image_size",
        "_old_pids",
        "_parent_pid",
        )

    def __init__(self, pid):
        self._pid_env_id = PidEnvId()
        self._login = None

        self._parent_pid = pid
        self._old_pids = []
        self._exited_cpu_user_time = 0
        self._exited_cpu_sys_time = 0
        self._max_image_size = 0

    def hardKill(self):
        self.takeSnapshot()
        self._spree(SIGKILL, HandlerDirection.INFANTICIDE)

    def softKill(self, signal):
        self.takeSnapshot()
        self._spree(SIGCONT, HandlerDirection.INFANTICIDE)
        self._spree(signal, HandlerDirection.INFANTICIDE)

    def suspendFamily(self):
        self.takeSnapshot()
        self._spree(SIGSTOP, HandlerDirection.PATRICIDE)

    def resumeFamily(self):
        self._spree(SIGCONT, HandlerDirection.INFANTICIDE)

    def getProcessorUsage(self):
        return (self._exited_cpu_sys_time + self._alive_cpu_sys_time,
            self._exited_cpu_user_time + self._alive_cpu_user_time)

    def getMaxImagesize(self):
        return self._max_image_size

    def setEnvironment(self, pid_env_id):
        self._pid_env_id.copy_from(pid_env_id)

    def setLogin(self, login):
        self._login = login

    def takeSnapshot(self):
        if self._login:
            infos = Info.getLoginInfos(self._login)
        else:
            infos = Info.getFamilyInfos(self._parent_pid)

        infos = list(infos)

        for old_info in self._old_infos:
            current_pid = old_info.pid
            current_pid_exited = True

            for info in infos:
                if info.pid == current_pid:
                    current_pid_exited = False
                    break

            else:
                info = Info.getInfo(current_pid)
                if info:
                    if info.birthday == old_info.birthday:
                        infos.append(current_pid)

                        if not self._login:
                            detached_families = Info.getFamilyInfos(
                                current_pid)
                            if detached_families:
                                for detached_family in detached_families:
                                    if detached_family != current_pid:
                                        infos.append(detached_family)

                                current_pid_exited = False

                            else:
                                current_pid_exited = True

                        else:
                            current_pid_exited = False

            if current_pid_exited:
                self._exited_cpu_user_time += old_info.cpu_user_time
                self._exited_cpu_sys_time += old_info.cpu_sys_time

        self._alive_cpu_sys_time = 0
        self._alive_cpu_user_time = 0
        current_image_size = 0
        new_infos = []
        for info in infos:
            new_infos.append(info)
            self._alive_cpu_sys_time += info.sys_time
            self._alive_cpu_user_time += info.user_time
            current_image_size += info.image_size

        if current_image_size > self._max_image_size:
            self._max_image_size = current_image_size

        self.old_infos = new_infos

        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            family_pids = ", ".join([str(info.pid) for info in infos])
            logging.debug(
                "parent %d family: %s", self._daddy_pid, family_pids)
            logging.debug(
                "alive_cpu_user: %d, exited_cpu_user: %d, max_image: %d",
                self._alive_cpu_user_time, self._exited_cpu_user_time,
                self._max_image_size)

    def currentFamily(self):
        for old_info in self._old_infos:
            yield old_info

    def _safeKill(self, kill_pid, signal):
        pid = kill_pid.pid

        # Check that init is not being killed
        if pid < 2 or self._parent_pid < 2:
            return

        logging.info("About to kill pid %d with signal %d", pid, signal)

        try:
            os.kill(pid, signal)
        except OSError, e:
            logging.info("kill(%d, %d) error: %d", pid, signal, e.errno)

    def _spree(self, signal, direction):
        start = 0
        for end, old_pid in enumerate(self._old_pids):
            if old_pid.ppid == 1 or old_pid.pid == 0:
                if direction == HandlerDirection.PATRICIDE:
                    # Parents go first
                    for index in xrange(start, end):
                        self._safeKill(self._old_pids[index], signal)

                else:
                    # Children go first
                    for index in xrange(end - 1, start - 1, -1):
                        self._safeKill(self._old_pids[index], signal)

                start = end
