#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Info",
    ]

import os
import errno
import logging
import resource

from pwd import getpwnam
from string import digits
from time import time

from checkbox.lib.enum import Enum
from checkbox.lib.stat import Stat

from checkbox.process.pidenv import PidEnvId


HZ = 100
KB = 1024


InfoStatus = Enum(
    "OK",
    "FAMILY_NONE",
    "FAMILY_ALL",
    "FAMILY_SOME",
    "NOPID",
    "PERMISSION",
    "GARBLED",
    "UNSPECIFIED",
    "ALIVE",
    "DEAD",
    "UNCERTAIN")


class InfoRaw:

    __slots__ = (
        "image_size",
        "rss_size",
        "minor_fault",
        "major_fault",
        "pid",
        "ppid",
        "flags",
        "owner",
        "user_time",
        "sys_time",
        "creation_time",
        "sample_time",
        )

    def __init__(self):
        self.image_size = 0L
        self.rss_size = 0L
        self.minor_fault = 0L
        self.major_fault = 0L
        self.pid = 0
        self.ppid = 0
        self.flags = 0L
        self.owner = 0
        self.user_time = 0L
        self.sys_time = 0L
        self.creation_time = 0L
        self.sample_time = 0L


class Info:

    __slots__ = (
        "image_size",
        "rss_size",
        "minor_fault",
        "major_fault",
        "cpu_usage",
        "user_time",
        "sys_time",
        "age",
        "pid",
        "ppid",
        "creation_time",
        "birthday",
        "owner",
        "pid_env_id",
        )

    _pagesize = None

    _boottime = 0
    _boottime_expiration = 0

    def __init__(self):
        self.image_size = 0
        self.rss_size = 0
        self.minor_fault = 0
        self.major_fault = 0
        self.cpu_usage = 0
        self.user_time = 0
        self.sys_time = 0
        self.age = 0
        self.pid = -1
        self.ppid = -1
        self.creation_time = 0
        self.birthday = 0
        self.owner = 0
        self.pid_env_id = PidEnvId()

    @classmethod
    def getInfo(cls, pid):
        info = Info()
        info_raw = cls._getInfoRaw(pid)
        if not info_raw:
            return None

        if not cls._pagesize:
            cls._pagesize = resource.getpagesize() / KB

        # Convert sizes from bytes to kilobytes
        if info_raw.flags & 64 and info_raw.ppid != 1:
            info.image_size = 0
            info.rss_size = 0

        else:
            info.image_size = info_raw.image_size / KB
            info.rss_size = info_raw.rss_size * cls._pagesize

        # Convert time from jiffies to seconds
        info.user_time = info_raw.user_time / HZ
        info.sys_time = info_raw.sys_time / HZ

        # Keep the birthday in jiffies for process comparison
        info.birthday = info_raw.creation_time

        # Convert the creation time from jiffies
        if not cls._checkBoottime(info_raw.sample_time):
            logging.info("Problem getting boottime")
            return None

        # Convert from jiffies to seconds
        birthday_seconds = info_raw.creation_time / HZ

        # Convert from seconds since boottime to seconds since epoch
        info.creation_time = birthday_seconds + cls._boottime

        # Calculate age
        info.age = info_raw.sample_time - info.creation_time
        if info.age < 0:
            info.age = 0

        info.owner = info_raw.owner
        info.pid = info_raw.pid
        info.ppid = info_raw.ppid

        # Get environment information
        path = os.path.join("/proc", str(info.pid), "environ")
        try:
            with open(path) as file:
                env_buffer = file.read()
        except IOError:
            # Unlike other things, this is optional
            pass
        else:
            # Convert NUL delimited strings to the usual format
            env_entries = env_buffer.split("\0")
            env = dict((entry.split("=", 1)) for entry in env_entries if entry)

            # Filter and insert into pid environment table
            info.pid_env_id.filter_and_insert(env)

        return info

    @classmethod
    def getInfos(cls):
        for pid in cls._getPIDs():
            info = cls.getInfo(pid)
            if info:
                yield info

    @classmethod
    def getLoginPIDs(cls, login):
        pwd = getpwnam(login)
        owner = pwd.pw_uid

        for info in cls.getInfos():
            if info.owner == owner:
                logging.debug(
                    "Found pid %d owned by %s (uid=%d)",
                    info.pid, login, owner)
                yield info.pid

    @classmethod
    def _isInFamily(cls, pids, pid_env_id, child):
        for pid in pids:
            if child.ppid == pid:
                logging.debug(
                    "PID %d is in family of %d",
                    child.pid, pid)
                return True

            if pid_env_id == child.pid_env_id:
                logging.debug(
                    "PID %d is predicted to be in family of %d",
                    child.pid, pid)
                return True

        return False

    @classmethod
    def getFamilyInfos(cls, pid, pid_env_id):
        """
        Get an iterator of the processes sharing the ancestor in the
        given arguments.

        :param pid: Identifier of the daddy.
        :param pid_env_id: Environment identifer of the daddy.
        """
        infos = list(cls.getInfos())

        # Find the daddy in all processes.
        for del_index, info in enumerate(infos):
            if info.pid == pid:
                logging.debug("Found family pid %d on the system", pid)
                break
        else:
            # Find the conceptual daddy regardless of actual parentage.
            for del_index, info in enumerate(infos):
                if info.pid_env_id == pid_env_id:
                    logging.debug(
                        "Parent pid %d is gone. Found descendant %d",
                        pid, info.pid)
                    break
            else:
                # Otherwise truly give up.
                logging.debug("Parent pid %d not found on the system", pid)
                return

        # Take whatever daddy was found and use that as the ancestor.
        del infos[del_index]
        pids = [info.pid]
        yield info

        # Find all the children that are in the same family.
        while True:
            del_index = 0
            for info in infos:
                if cls._isInFamily(pids, pid_env_id, info):
                    pids.append(info.pid)
                    yield info
                else:
                    infos[del_index] = info
                    del_index += 1

            if del_index != len(infos):
                del infos[del_index:]
            else:
                break

    @classmethod
    def _getPIDs(cls):
        try:
            for name in os.listdir("/proc"):
                if name[0] in digits:
                    yield int(name)
        except OSError:
            # Ignore error reading /proc
            pass

    @classmethod
    def _getInfoRaw(cls, pid):
        path = os.path.join("/proc", str(pid), "stat")

        # Try to read the entry a few times
        attempts = 0
        max_attempts = 5
        while attempts < max_attempts:
            attempts += 1

            status = InfoStatus.OK
            info_raw = None

            try:
                with open(path) as file:
                    line = file.readline()
            except IOError, error:
                if error.errno == errno.ENOENT:
                    status = InfoStatus.NOPID
                    logging.debug("Pid %d does not exist", pid)

                elif error.errno == errno.EACCESS:
                    status = InfoStatus.PERMISSION
                    logging.debug("No permission to open %s", path)

                else:
                    status = InfoStatus.UNSPECIFIED
                    logging.info(
                        "Error opening %s, errno %d", path, error.errno)

                # Immediate failure, try again.
                continue

            elements = line.split()
            if len(elements) < 35:
                status = InfoStatus.UNSPECIFIED
                logging.info("Unexpected short can on %s", path)
                continue

            # Fill the raw structure
            info_raw = InfoRaw()
            info_raw.pid = int(elements[0])
            info_raw.ppid = int(elements[3])
            info_raw.flags = long(elements[8])
            info_raw.minor_fault = long(elements[9])
            info_raw.major_fault = long(elements[11])
            info_raw.user_time = long(elements[13])
            info_raw.sys_time = long(elements[14])
            info_raw.creation_time = long(elements[21])
            info_raw.image_size = long(elements[22])
            info_raw.rss_size = long(elements[23])
            info_raw.sample_time = long(time())

            # Check pid
            if pid == info_raw.pid:
                break

            status = InfoStatus.GARBLED

        if status != InfoStatus.OK:
            if status == InfoStatus.GARBLED:
                logging.info(
                    "Found garbage in %s after %d attempts", path, attempts)

        else:
            stat = Stat.create_from_path(path)
            result = stat.get_result()
            if result:
                info_raw.owner = result.st_uid

            else:
                info_raw.owner = 0

        return info_raw

    @classmethod
    def _checkBoottime(cls, now):
        if cls._boottime_expiration < now:
            # Get uptime_boottime
            uptime_boottime = 0
            try:
                with open("/proc/uptime") as file:
                    line = file.readline()
            except IOError:
                # Check later
                pass
            else:
                elements = line.split()
                if len(elements) >= 1:
                    uptime = float(elements[0])
                    uptime_boottime = long(now - uptime + 0.5)

            # Get stat_boottime
            stat_boottime = 0
            try:
                with open("/proc/stat") as file:
                    for line in file.xreadlines():
                        if line.startswith("btime"):
                            break
            except IOError:
                # Check later
                pass
            else:
                elements = line.split()
                if len(elements) >= 2:
                    stat_boottime = long(elements[1])

            if (stat_boottime == 0
                and uptime_boottime == 0
                and cls._boottime == 0):
                logging.info(
                    "Problem opening /proc/stat"
                    " and /proc/uptime for boottime")
                return False

            if stat_boottime != 0 or uptime_boottime != 0:
                old_boottime = cls._boottime

                # Use older of the two boottime estimates
                if stat_boottime == 0:
                    cls._boottime = uptime_boottime

                elif uptime_boottime == 0:
                    cls._boottime = stat_boottime

                else:
                    cls._boottime = min(stat_boottime, uptime_boottime)

                # Update once every minute
                cls._boottime_expiration = now + 60

                logging.info(
                    "new boottime = %lu; old_boottime = %lu; "
                    "/proc/stat boottime = %lu; /proc/uptime boottime = %lu",
                    cls._boottime, old_boottime,
                    stat_boottime, uptime_boottime)

        return True
