#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "ManagerDirect",
    ]

import logging

from checkbox.lib.decorators import check_key_error

from checkbox.daemon.server import Server

from checkbox.process.handler import Handler
from checkbox.process.manager import (
    Manager,
    ManagerError,
    )


class ManagerDirect(Manager):

    __slots__ = (
        "_table",
        )

    is_direct = True

    def __init__(self):
        self._table = dict()

    def registerSubfamily(self, pid, watcher_pid, snapshot_interval):
        handler = Handler(pid)

        # Register a timer with the daemon core
        timer_id = Server.registerTimer(
            handler.takeSnapshot, 2, snapshot_interval)

        # Insert the handler and timer_id in the table
        if pid in self._table:
            logging.info(
                "PID %d already registered in direct family", pid)
            Server.unregisterTimer(timer_id)
            return False

        self._table[pid] = (handler, timer_id,)
        return True

    @check_key_error(False)
    def trackEnvironment(self, pid, pid_env_id):
        handler, timer_id = self._table[pid]
        handler.setEnvironment(pid_env_id)
        return True

    @check_key_error(False)
    def trackLogin(self, pid, login):
        handler, timer_id = self._table[pid]
        handler.setLogin(login)
        return True

    def signalProcess(self, pid, signal):
        raise ManagerError("signalProcess should never be called")

    @check_key_error(False)
    def suspendFamily(self, pid):
        handler, timer_id = self._table[pid]
        handler.suspendFamily()
        return True

    @check_key_error(False)
    def resumeFamily(self, pid):
        handler, timer_id = self._table[pid]
        handler.resumeFamily()
        return True

    @check_key_error(False)
    def killFamily(self, pid):
        handler, timer_id = self._table[pid]
        handler.hardKill()
        return True

    def unregisterFamily(self, pid):
        try:
            # Remove the pid from the table
            handler, timer_id = self._table.pop(pid)
        except KeyError:
            logging.info(
                "PID %d not registered in direct family", pid)
            return False

        # Cancel the timer in the daemon core
        Server.unregisterTimer(timer_id)
        return True
