#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "ManagerIndirect",
    ]

import os
import logging

from time import sleep

from checkbox.lib import param
from checkbox.lib.singleton import Singleton

from checkbox.daemon.server import server

from checkbox.process.args import Args
from checkbox.process.client import (
    Client,
    ClientError,
    )
from checkbox.process.config import get_path
from checkbox.process.manager import (
    Manager,
    ManagerError,
    )


class ManagerIndirect(Manager, Singleton):

    __slots__ = (
        "_client",
        "_log",
        "_path",
        "_pid",
        "_reaper_id",
        )

    _initialized = False

    is_direct = False

    def __init__(self, suffix=None):
        """Instantiate a indirect manager singleton."""
        if ManagerIndirect._initialized:
            return

        ManagerIndirect._initialized = True
        self._pid = -1
        self._reaper_id = False

        self._path = path_base = get_path()
        if suffix:
            self._path += ".%s" % suffix

        self._log = param.get("PROCESS_LOG")
        if self._log and suffix:
            self._log += ".%s" % suffix

        # Determine if the server needs to be started
        path_base_from_env = param.get("PROCESS_PATH_BASE")
        if not path_base_from_env \
           or path_base != path_base_from_env:
            if not self._startServer():
                raise ManagerError("Unable to start process server")

            os.environ["PROCESS_PATH_BASE"] = path_base
            os.environ["PROCESS_PATH"] = self._path
        else:
            path_from_env = param.get("PROCESS_PATH")
            if not path_from_env:
                raise ManagerError(
                    "PROCESS_PATH_BASE in environment but not PROCESS_PATH")

            self._path = path_from_env

        # Create the process client
        try:
            self._client = Client(self._path)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()

    def __del__(self):
        if self._pid != -1:
            self._stopServer()
            del os.environ["PROCESS_PATH_BASE"]
            del os.environ["PROCESS_PATH"]

        ManagerIndirect._initialized = False

    def registerSubfamily(self, pid, watcher_pid, max_snapshot_interval):
        # This is only called from forked children, so don't restart
        # the server because the parent would still not see the changes.
        try:
            return self._client.registerSubfamily(
                pid, watcher_pid, max_snapshot_interval)
        except ClientError, error:
            logging.info(error.message)
            return False

    def trackEnvironment(self, pid, pid_env_id):
        # See registerSubfamily
        try:
            return self._client.trackEnvironment(pid, pid_env_id)
        except ClientError, error:
            logging.info(error.message)
            return False

    def trackLogin(self, pid, login):
        # See registerSubfamily
        try:
            return self._client.trackLogin(pid, login)
        except ClientError, error:
            logging.info(error.message)
            return False

    def signalProcess(self, pid, signal):
        try:
            return self._client.signalProcess(pid, signal)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()
            return False

    def suspendFamily(self, pid):
        try:
            return self._client.suspendFamily(pid)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()
            return False

    def resumeFamily(self, pid):
        try:
            return self._client.resumeFamily(pid)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()
            return False

    def killFamily(self, pid):
        try:
            return self._client.killFamily(pid)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()
            return False

    def unregisterFamily(self, pid):
        try:
            return self._client.unregisterFamily(pid)
        except ClientError, error:
            logging.info(error.message)
            self._restartServer()
            return False

    def _startServer(self):
        assert(self._pid == -1)

        # Build executable
        path = param.get("PROCESS_SERVER")
        if not path:
            bin_path = param.get("BIN")
            if not bin_path:
                logging.info("PROCESS_SERVER not defined in configuration")
                return False

            path = os.path.join(path, "checkbox-process-server")

        executable = path

        # Build args
        args = Args()
        args.append(os.path.basename(executable))

        if self._log:
            args.append("--log=%s" % self._log)
            args.append("--log-level=debug")

        max_snapshot_interval = param.get("PROCESS_MAX_SNAPSHOT_INTERVAL")
        if max_snapshot_interval is not None:
            args.append("--interval=%s" % max_snapshot_interval)

        # TODO: get principal

        args.append(self._path)

        # Build reaper
        if not self._reaper_id:
            self._reaper_id = server.registerReaper(
                function=self._reapServer)

            if not self._reaper_id:
                logging.info(
                    "Unable to register a reaper for the process server")
                return False

        # Start the process server with a pipe coming back on stderr.
        pipe_ends = [-1, -1]
        if not server.createPipe(pipe_ends):
            logging.info("Error creating pipe ends for the process server")
            return False

        std = [-1, -1, pipe_ends[1]]

        self._pid = server.createProcess(
            executable, args, self._reaper_id, False, std=std)
        if self._pid < 0:
            logging.info("Unable to execute the process server")
            server.closePipe(pipe_ends[0])
            server.closePipe(pipe_ends[1])
            return False

        # Close the pipe ends handed to the child and then block on the pipe
        if not server.closePipe(pipe_ends[1]):
            logging.info("Error closing process server pipe end")
            server.killProcess(self._pid)
            server.closePipe(pipe_ends[0])
            self._pid = -1
            return False

        error = server.readPipe(pipe_ends[0], 80)
        if error != "":
            server.killProcess(self._pid)
            server.closePipe(pipe_ends[0])
            self._pid = -1
            if error is None:
                logging.info("Error reading pipe from process server")
                return False

            logging.info("Error received from process server: %s", error)
            return False

        if not server.closePipe(pipe_ends[0]):
            logging.info("Error closing pipe to process server")
            server.killProcess(self._pid)
            self._pid = -1
            return False

        return True

    def _stopServer(self):
        if not self._client.quit():
            logging.info("Error telling process server to quit")

        self._pid = -1

    def _restartServer(self):
        if not param.getBoolean("PROCESS_RESTART_ON_ERROR", False):
            raise ManagerError("process server has failed")

        self._client = None
        while not self._client:
            if self._pid != -1:
                logging.info("Attempting to restart process server")
                self._pid = -1
                if not self._startServer():
                    raise ManagerError("Unable to start process server")
            else:
                logging.info("Waiting a second to restart process server")
                sleep(1)

            try:
                self._client = Client(self._path)
            except ClientError, error:
                logging.info(error.message)

    def _reapServer(self, pid, status):
        if self._pid == -1 or self._pid != pid:
            logging.info(
                "process server %d exited with status %d", pid, status)
        else:
            logging.info(
                "process server %d exited unexpectedly with status %d",
                pid, status)
            self._restartServer()

        return 0
