#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Monitor",
    "MonitorStatus",
    ]

import logging

from checkbox.lib.enum import Enum

from checkbox.process.info import Info
from checkbox.process.family import Family
from checkbox.process.tracker import (
    EnvironmentTracker,
    LoginTracker,
    ParentTracker,
    PIDTracker,
    )
from checkbox.process.tree import Tree
from checkbox.process.usage import Usage


MonitorStatus = Enum(
    "SUCCESS",
    "BAD_ROOT_PID",
    "BAD_WATCHER_PID",
    "BAD_SNAPSHOT_INTERVAL",
    "ALREADY_REGISTERED",
    "FAMILY_NOT_FOUND",
    "PROCESS_NOT_FOUND",
    "PROCESS_NOT_FAMILY",
    "UNREGISTER_ROOT",
    "BAD_ENVIRONMENT_INFO",
    "BAD_LOGIN_INFO",
    "NO_GROUP_ID_AVAILABLE",
    )


class Monitor:

    __slots__ = (
        "_tree_root",
        "_tree_table",
        "_member_table",
        "_pid_tracker",
        "_login_tracker",
        "_environment_tracker",
        "_parent_tracker",
        "_everybody_else",
        )

    def __init__(self, pid, birthday, snapshot_interval):
        assert(snapshot_interval >= -1)

        self._tree_table = {}
        self._member_table = {}

        # Create trackers
        self._pid_tracker = PIDTracker(self)
        self._login_tracker = LoginTracker(self)
        self._environment_tracker = EnvironmentTracker(self)
        self._parent_tracker = ParentTracker(self)

        # Create container for all processes not being tracked
        self._everybody_else = Family(self)

        # Create root family
        family = Family(self, pid, birthday, pid, snapshot_interval)
        self._pid_tracker.addMapping(family, pid, birthday)

        # Make this family the root of the tree
        self._tree_root = Tree(family)
        self._tree_table[pid] = self._tree_root

        self.takeSnapshot()

    @property
    def snapshot_interval(self):
        """Get the snapshot interval for this family monitor."""
        return self._getSnapshotInterval(self._tree_root)

    def registerSubfamily(self, pid, watcher_pid, max_snapshot_interval):
        """Create a subfamily which can then be signalled and accounted for.

        :param pid: Root pid, which must be positive.
        :param watcher_pid: Watcher pid, zero indicates the pid is unwatched.
        :param max_snapshot_interval: Maximum snapshot interval,
            -1 for infinite.
        """
        # Root pid must be positive
        if pid <= 0:
            logging.info("Bad root pid: %d", pid)
            return MonitorStatus.BAD_ROOT_PID

        # Watcher pid must be non-negative
        if watcher_pid < 0:
            logging.info("Bad watcher pid: %d", watcher_pid)
            return MonitorStatus.BAD_WATCHER_PID

        # Maximum snapshot interval must either be non-negative
        # or -1 for infinite
        if max_snapshot_interval < -1:
            logging.info("Bad snapshot interval: %d", max_snapshot_interval)
            return MonitorStatus.BAD_SNAPSHOT_INTERVAL

        self.takeSnapshot()

        member = self._member_table.get(pid)
        if not member or member.family == self._everybody_else:
            logging.info("PID %d not in process tree", pid)
            return MonitorStatus.PROCESS_NOT_FAMILY

        if pid in self._tree_table:
            logging.info("PID %d already registered", pid)
            return MonitorStatus.ALREADY_REGISTERED

        # Register a new family
        family = Family(
            self, pid, member.info.birthday, watcher_pid,
            max_snapshot_interval)

        self._pid_tracker.addMapping(family, pid, member.info.birthday)

        # Find the family what will be the new subfamily's parent
        parent_root_pid = member.family.root_pid
        parent_tree = self._tree_table[parent_root_pid]
        child_tree = parent_tree.addChild(family)
        assert(child_tree)

        # Move the family's root process into the correct family
        member.moveToSubfamily(family)
        self._tree_table[pid] = child_tree

        return MonitorStatus.SUCCESS

    def unregisterFamily(self, pid):
        """Remove a family.

        :param pid: Pid of the subfamily.
        """
        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return MonitorStatus.FAMILY_NOT_FOUND

        if not tree.getParent():
            logging.info("Cannot unregister root family")
            return MonitorStatus.UNREGISTER_ROOT

        self._unregisterFamily(tree)
        return MonitorStatus.SUCCESS

    def trackEnvironment(self, pid, pid_env_id):
        """Start tracking a family using envionment variables.

        :param pid: Root pid to track.
        :param pid_env_id: Environment variables.
        """
        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return MonitorStatus.FAMILY_NOT_FOUND

        self._environment_tracker.addMapping(tree.getData(), pid_env_id)
        return MonitorStatus.SUCCESS

    def trackLogin(self, pid, login):
        """Start tracking a family using login.

        :param pid: Root pid to track.
        :param login: Login user name.
        """
        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return MonitorStatus.FAMILY_NOT_FOUND

        self._login_tracker.addMapping(tree.getData(), login)
        return MonitorStatus.SUCCESS

    def signalProcess(self, pid, signal):
        """Send a signal to a single process.

        :param pid: Pid to signal.
        :param signal: Signal to send.
        """
        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return MonitorStatus.FAMILY_NOT_FOUND

        tree.getData().signalRoot(signal)
        return MonitorStatus.SUCCESS

    def signalFamily(self, pid, signal):
        """Send a signal to all processes in the family.

        :param pid: Root pid to signal.
        :param signal: Signal to send.
        """
        # Get as up-to-date as possible
        self.takeSnapshot()

        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return MonitorStatus.FAMILY_NOT_FOUND

        self._signalFamily(tree, signal)
        return MonitorStatus.SUCCESS

    def addMember(self, member):
        """Add to the member table.

        :param member: Family member to add.
        """
        self._member_table[member.info.pid] = member

    def removeMember(self, member):
        """Remove from the member table.

        :param member: Family member to remove.
        """
        del self._member_table[member.info.pid]

    def lookupMember(self, pid):
        """Lookup from the member table by pid.

        :param pid: Member pid to lookup.
        """
        return self._member_table.get(pid)

    def addFamilyMember(self, family, info, string):
        """Used by trackers to make associations between
           processes and families being monitored.

        :param family: Family where to add member.
        :param info: Process information.
        :param string: String for logging purposes.
        """
        member = self._member_table.get(info.pid)
        if not member:
            # Member is not already associated with a family
            logging.info(
                "%s: Found family %d for process %d",
                string, family.root_pid, info.pid)
            family.addMember(info)
            return True

        elif family != member.family:
            # Member is already associated with a family
            node = self._tree_table[family.root_pid]
            while node.getParent():
                node = node.getParent()
                if node.getData() == member.family:
                    logging.info(
                        "%s: Found family %d for process %d"
                        " (more specific than current family)",
                        string, family.root_pid, info.pid)

                    member.moveToSubfamily(family)
                    return True

            logging.info(
                "%s: Found family %d for process %d"
                " (less specific - ignoring)",
                string, family.root_pid, info.pid)
            return False

        else:
            # Process already in family, nothing to do
            logging.info(
                "%s: Found family %d for process %d"
                " (already determined)",
                string, family.root_pid, info.pid)
            return False

    def getUsage(self, pid):
        try:
            tree = self._tree_table[pid]
        except KeyError:
            logging.info("Family with root %d not found", pid)
            return None

        usage = Usage()
        self._getTreeUsage(tree, usage)

        # Max image size is handled from the top level family.
        usage.max_image_size = tree.getData().max_image_size

        return usage

    def _getTreeUsage(self, tree, usage):
        tree.getData().aggregateUsage(usage)

        for child in tree.getChildren():
            self._getTreeUsage(child, usage)

    def takeSnapshot(self):
        """Take a snapshot of all processes on the system to update the
           families being tracked."""
        # Scan through the latest snapshot
        new_list = []
        for info in Info.getInfos():
            if info.pid == 0:
                continue

            member = self._member_table.get(info.pid)
            if member and member.info.birthday == info.birthday:
                # Update process with latest info
                member.stillAlive(info)

            else:
                # Unknown process, add to new list
                new_list.append(info)

        # Purge exited processes
        self._removeExitedProcesses(self._tree_root)
        self._everybody_else.removeExitedProcesses()

        # Handle new processes
        self._pid_tracker.findProcesses(new_list)
        self._login_tracker.findProcesses(new_list)
        self._environment_tracker.findProcesses(new_list)
        self._parent_tracker.findProcesses(new_list)

        # Add remaining processes to everybody else
        for info in new_list:
            if info.pid not in self._member_table:
                logging.info(
                    "Process %d is not monitored by any family", info.pid)
                self._everybody_else.addMember(info)

        # Cleanup unwatched families
        self._removeUnwatchedFamilies(self._tree_root)

        # Update maximum image size
        self._updateMaxImageSizes(self._tree_root)

    def _signalFamily(self, tree, signal):
        # Signal the current tree
        tree.getData().spree(signal)

        # Recurse on children
        for child in tree.getChildren():
            self._signalFamily(child, signal)

    def _unregisterFamily(self, tree):
        # Remove tree from trackers
        self._pid_tracker.removeMapping(tree.getData())
        self._login_tracker.removeMapping(tree.getData())
        self._environment_tracker.removeMapping(tree.getData())

        # Remove from tree table
        assert(tree.getData().root_pid in self._tree_table)
        del self._tree_table[tree.getData().root_pid]

        # Move all family members to the current family
        parent = tree.getParent()
        assert(parent)
        tree.getData().foldIntoParent(parent.getData())

        # Remove the current node
        tree.remove()

    def _getSnapshotInterval(self, tree):
        tree_value = tree.getData().max_snapshot_interval

        # Recurse on children
        for child in tree.getChildren():
            child_value = self._getSnapshotInterval(child)
            if tree_value == -1:
                tree_value = child_value

            elif child_value < tree_value:
                tree_value = child_value

        return tree_value

    def _updateMaxImageSizes(self, tree):
        # Recurse on children
        sum = 0
        for child in tree.getChildren():
            sum += self._updateMaxImageSizes(child)

        return tree.getData().updateMaxImageSize(sum)

    def _removeExitedProcesses(self, tree):
        family = tree.getData()
        family.removeExitedProcesses()

        # Recurse on children
        for child in tree.getChildren():
            self._removeExitedProcesses(child)

    def _removeUnwatchedFamilies(self, tree):
        # Recurse on children
        for child in tree.getChildren():
            self._removeUnwatchedFamilies(child)

        family = tree.getData()
        watcher_pid = family.watcher_pid
        if not watcher_pid:
            return

        member = self._member_table.get(watcher_pid)
        if member:
            if member.info.birthday <= family.root_birthday:
                # This family's watcher is still around
                return

            logging.info("Watcher %d found with later birthdate", watcher_pid)

        if not tree.getParent():
            raise Exception("Server parent has exited")

        self._unregisterFamily(tree)
