#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "PidEnvId",
    "PidEnvIdStatus",
    ]

import re

from checkbox.lib.enum import Enum


PidEnvIdStatus = Enum(
    "OK",
    "NO_SPACE",
    "OVERSIZED",
    "BAD_FORMAT",
    )


class PidEnvIdEntry:

    __slots__ = (
        "active",
        "envid",
        )

    prefix = "_PROCESS_ANCESTOR_"

    def __init__(self):
        self.active = False
        self.envid = ""


class PidEnvId:

    __slots__ = (
        "ancestors",
        )

    _max_ancestors = 32
    _format_pattern = re.compile(r"%s(?P<ppid>\d+)=(?P<pid>\d+)"
        % PidEnvIdEntry.prefix)

    def __init__(self):
        self.ancestors = [PidEnvIdEntry() for i in xrange(self._max_ancestors)]

    def __cmp__(self, other):
        if not self:
            if not other:
                return 0
            else:
                return 1

        elif not other:
            return -1

        else:
            searched = found = 0
            for ancestor in self.ancestors:
                if not ancestor.active:
                    break

                for other_ancestor in other.ancestors:
                    if not other_ancestor.active:
                        break

                    if ancestor.envid == other_ancestor.envid:
                        found += 1

                searched += 1

            # No match if there is nothing to match
            if not searched:
                return -1

            return cmp(searched, found)

    def copy(self):
        other = PidEnvId()
        other.copy_from(self)
        return other

    def copy_from(self, other):
        for ancestor, other_ancestor in zip(self.ancestors, other.ancestors):
            ancestor.active = other_ancestor.active
            ancestor.envid = other_ancestor.envid

    def filter_and_insert(self, env):
        i = 0
        for key, value in env.items():
            if key.startswith(PidEnvIdEntry.prefix):
                if i == self._max_ancestors:
                    return PidEnvIdStatus.NO_SPACE

                self.ancestors[i].envid = "%s=%s" % (key, value)
                self.ancestors[i].active = True
                i += 1

        return PidEnvIdStatus.OK

    def append(self, line):
        """Add line into the pidenvid structure

        :param line: Line created by format.
        """
        for ancestor in self.ancestors:
            if not ancestor.active:
                ancestor.envid = line
                ancestor.active = True
                return PidEnvIdStatus.OK

        return PidEnvIdStatus.NO_SPACE

    def append_direct(self, ppid, pid):
        line = self.format_to_envid(ppid, pid)
        return self.append(line)

    @classmethod
    def format_to_envid(cls, ppid, pid):
        return "%s%d=%d" % (PidEnvIdEntry.prefix, ppid, pid)

    @classmethod
    def format_from_envid(cls, line):
        match = cls._format_pattern.match(line)
        if not match:
            return PidEnvIdStatus.BAD_FORMAT

        ppid = int(match.group("ppid"))
        pid = int(match.group("pid"))

        return (ppid, pid)
