#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Server",
    ]

import sys
import logging

from signal import (
    SIGCONT,
    SIGSTOP,
    SIGKILL,
    )
from time import time

from checkbox.lib.enum import enum_value_to_name_string

from checkbox.io.fifo import Fifo

from checkbox.process.monitor import MonitorStatus
from checkbox.process.operation import OperationCode
from checkbox.process.pidenv import PidEnvId


class Server:

    __slots__ = (
        "_monitor",
        "_fifo",
        "_quit_received",
        )

    def __init__(self, monitor, path):
        self._monitor = monitor
        self._fifo = Fifo()
        if not self._fifo.listen(path):
            raise Exception("Could not listen on fifo path: %s" % path)

        self._quit_received = False

    def setPrincipal(self, principal):
        """Set the client principal which will be able to connect to
           this server.

        :param principal: Principal name.
        """
        if not self._fifo.setPrincipal(principal):
            raise Exception("Error setting client principal.")

    def run(self):
        """The server wait loop."""
        snapshot_countdown = sys.maxint
        operation_to_function = {
            OperationCode.REGISTER_SUBFAMILY: self._registerSubfamily,
            OperationCode.TRACK_ENVIRONMENT: self._trackEnvironment,
            OperationCode.TRACK_LOGIN: self._trackLogin,
            OperationCode.SIGNAL_PROCESS: self._signalProcess,
            OperationCode.SUSPEND_FAMILY: self._suspendFamily,
            OperationCode.RESUME_FAMILY: self._resumeFamily,
            OperationCode.KILL_FAMILY: self._killFamily,
            OperationCode.UNREGISTER_FAMILY: self._unregisterFamily,
            OperationCode.TAKE_SNAPSHOT: self._takeSnapshot,
            OperationCode.QUIT: self._quit,
            }

        while not self._quit_received:
            snapshot_interval = self._monitor.snapshot_interval
            if snapshot_interval == -1:
                snapshot_countdown -= 1

            elif snapshot_countdown > snapshot_interval:
                snapshot_countdown = snapshot_interval

            # Check to run the timer handler
            if snapshot_countdown == 0:
                self._monitor.takeSnapshot()

                snapshot_countdown = snapshot_interval

            time_before = time()

            self._fifo.setTimeout(snapshot_countdown)
            is_ready = self._fifo.accept()
            if is_ready < 0:
                raise Exception("Failed trying to accept client.")

            if not is_ready:
                # Set the countdown to 0 to execute the timer handler
                snapshot_countdown = 0
                continue

            # Adjust the countdown
            snapshot_countdown -= time() - time_before
            snapshot_countdown = max(0, snapshot_countdown)

            # Read the code from the client
            code = self._fifo.getInteger()
            logging.debug(
                "Got operation code: %s",
                enum_value_to_name_string(OperationCode, code))

            try:
                function = operation_to_function[code]
            except KeyError:
                raise Exception("Unknown operation code: %d", code)

            status = function()
            self._fifo.getEnd()

            self._fifo.putInteger(status)
            self._fifo.putEnd()

    def _registerSubfamily(self):
        pid = self._fifo.getInteger()
        watcher_pid = self._fifo.getInteger()
        max_snapshot_interval = self._fifo.getInteger()

        return self._monitor.registerSubfamily(
            pid, watcher_pid, max_snapshot_interval)

    def _trackEnvironment(self):
        fifo = self._fifo
        pid = fifo.getInteger()
        pid_env_id = PidEnvId()
        for ancestor in pid_env_id.ancestors:
            ancestor.active = fifo.getBoolean()
            ancestor.envid = fifo.getString()

        return self._monitor.trackEnvironment(pid, pid_env_id)

    def _trackLogin(self):
        pid = self._fifo.getInteger()
        login = self._fifo.getString()

        return self._monitor.trackLogin(pid, login)

    def _signalProcess(self):
        pid = self._fifo.getInteger()
        signal = self._fifo.getInteger()

        return self._monitor.signalProcess(pid, signal)

    def _suspendFamily(self):
        pid = self._fifo.getInteger()

        return self._monitor.signalProcess(pid, SIGSTOP)

    def _resumeFamily(self):
        pid = self._fifo.getInteger()

        return self._monitor.signalProcess(pid, SIGCONT)

    def _killFamily(self):
        pid = self._fifo.getInteger()

        return self._monitor.signalProcess(pid, SIGKILL)

    def _unregisterFamily(self):
        pid = self._fifo.getInteger()

        return self._monitor.unregisterFamily(pid)

    def _takeSnapshot(self):
        self._monitor.takeSnapshot()

        return MonitorStatus.SUCCESS

    def _quit(self):
        self._quit_received = True

        return MonitorStatus.SUCCESS
