#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "DroneMixin",
    ]

import os
import socket


try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except:
    MAXFD = 256


class Drone:

    __slots__ = (
        "_sock",
        )

    def __init__(self, port):
        # Connect to the controller port
        s = socket.socket()
        s.connect(("127.0.0.1", port))
        self._sock = s.makefile()
        s.close()

    def run(self):
        sock = self._sock
        pids = set()

        while True:
            argv = sock.readline().split()

            if not argv:
                return 0

            elif argv[0] == "SPAWN":
                if len(argv) != 2:
                    return "usage: SPAWN <port>"
                port = int(argv[1])
                pid = fork_drone(port)
                pids.add(pid)
                sock.write("%d\n" % pid)
                sock.flush()

            elif argv[0] == "REAP":
                if len(argv) != 2:
                    return "usage: REAP <pid>"
                pid = int(argv[1])
                pids.remove(pid)
                os.waitpid(pid, 0)

            else:
                return "Invalid operation: " + argv[0]


class DroneMixin:

    def setUp(self):
        super(DroneMixin, self).setUp()

        self._sock = sock = socket.socket()
        sock.bind(("127.0.0.1", 0))
        sock.listen(5)

        # Start the first drone.
        self._port = sock.getsockname()[1]
        self._init_pid = fork_drone(self._port)

        # Store drone info.
        init_sock = self.accept()
        self._drone_info = {
            self._init_pid: (init_sock, os.getpid()),
            }

    def tearDown(self):
        super(DroneMixin, self).tearDown()

        for pid in self._drone_info.keys():
            self.kill(pid)

    @property
    def init_pid(self):
        return self._init_pid

    def accept(self):
        s = self._sock.accept()[0]
        sock = s.makefile()
        s.close()

        return sock

    def spawn(self, parent_pid=None):
        if parent_pid is None:
            parent_pid = self._init_pid

        # Send the spawn operation to the parent drone.
        parent_sock = self._drone_info[parent_pid][0]
        parent_sock.write("SPAWN %d\n" % self._port)
        parent_sock.flush()

        # Read the new process's PID from it's parent.
        child_pid = int(parent_sock.readline())
        child_sock = self.accept()
        self._drone_info[child_pid] = (child_sock, parent_pid)

        return child_pid

    def reap(self, child_pid, parent_pid=None):
        if parent_pid is None:
            parent_pid = self._init_pid

        # Send the reap operation to the parent drone.
        parent_sock = self._drone_info[parent_pid][0]
        parent_sock.write("REAP %d\n" % child_pid)

        # Ingore errors in case the child was killed.
        try:
            parent_sock.flush()
        except socket.error:
            pass

    def kill(self, pid):
        # Close the drone connection.
        (sock, parent_pid) = self._drone_info.pop(pid)
        sock.close()

        # Reap it ourselves if it's the initial drone.
        if parent_pid == os.getpid():
            os.waitpid(self._init_pid, 0)
        elif parent_pid in self._drone_info:
            self.reap(pid, parent_pid)


def fork_drone(port):
    # Fork drone process.
    pid = os.fork()
    if pid == 0:
        # Close file descriptors.
        if hasattr(os, "closerange"):
            os.closerange(3, MAXFD)
        else:
            for i in xrange(3, MAXFD):
                try:
                    os.close(i)
                except:
                    pass

        drone = Drone(port)
        drone.run()
        os._exit(0)

    return pid
