#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = []

import os
import signal
import time

from unittest import TestCase

from checkbox.process.tests.drone import DroneMixin

from checkbox.process.info import Info
from checkbox.process.family import FamilyMember
from checkbox.process.monitor import (
    Monitor,
    MonitorStatus,
    )
from checkbox.process.pidenv import PidEnvId


class TestMonitor(DroneMixin, TestCase):

    def setUp(self):
        super(TestMonitor, self).setUp()

        self.birthday = time.time()
        self.snapshot_interval = -1

    def makeMonitor(self, pid=None, birthday=None, snapshot_interval=None):
        if pid is None:
            pid = self.init_pid
        if birthday is None:
            birthday = self.birthday
        if snapshot_interval is None:
            snapshot_interval = self.snapshot_interval

        return Monitor(pid, birthday, snapshot_interval)

    def test_registerSubfamilyBadRootPid(self):
        monitor = self.makeMonitor()
        self.assertEquals(
            monitor.registerSubfamily(0, 0, -1),
            MonitorStatus.BAD_ROOT_PID)

    def test_registerSubfamilyBadWatcherPid(self):
        monitor = self.makeMonitor()
        self.assertEquals(
            monitor.registerSubfamily(1, -1, -1),
            MonitorStatus.BAD_WATCHER_PID)

    def test_registerSubfamilyBadSnapshotInterval(self):
        monitor = self.makeMonitor()
        self.assertEquals(
            monitor.registerSubfamily(1, 0, -2),
            MonitorStatus.BAD_SNAPSHOT_INTERVAL)

    def test_registerSubfamilyProcessNotFamily(self):
        monitor = self.makeMonitor()
        pid = self.spawn()
        self.assertEquals(
            monitor.registerSubfamily(pid, 0, -1),
            MonitorStatus.PROCESS_NOT_FAMILY)

    def test_registerSubfamilyAlreadyRegistered(self):
        monitor = self.makeMonitor()
        monitor.trackLogin(self.init_pid, os.getlogin())

        pid = self.spawn()
        monitor.registerSubfamily(pid, 0, -1)
        self.assertEquals(
            monitor.registerSubfamily(pid, 0, -1),
            MonitorStatus.ALREADY_REGISTERED)

    def test_registerSubfamilySuccess(self):
        monitor = self.makeMonitor()
        monitor.trackLogin(self.init_pid, os.getlogin())

        pid = self.spawn()
        self.assertEquals(
            monitor.registerSubfamily(pid, 0, -1),
            MonitorStatus.SUCCESS)

    def test_unregisterFamilyFamilyNotFound(self):
        monitor = self.makeMonitor()

        pid = self.spawn()
        self.assertEquals(
            monitor.unregisterFamily(pid),
            MonitorStatus.FAMILY_NOT_FOUND)

    def test_unregisterFamilyUnregisterRoot(self):
        monitor = self.makeMonitor()

        self.assertEquals(
            monitor.unregisterFamily(self.init_pid),
            MonitorStatus.UNREGISTER_ROOT)

    def test_unregisterFamilySuccess(self):
        monitor = self.makeMonitor()
        monitor.trackLogin(self.init_pid, os.getlogin())

        pid = self.spawn()
        monitor.registerSubfamily(pid, 0, -1)
        self.assertEquals(
            monitor.unregisterFamily(pid),
            MonitorStatus.SUCCESS)

    def test_trackLoginFamilyNotFound(self):
        monitor = self.makeMonitor()

        pid = self.spawn()
        self.assertEquals(
            monitor.trackLogin(pid, os.getlogin()),
            MonitorStatus.FAMILY_NOT_FOUND)

    def test_trackLogin(self):
        pid1 = self.spawn()
        monitor = self.makeMonitor(pid1)
        monitor.trackLogin(pid1, os.getlogin())

        pid2 = self.spawn(pid1)
        self.assertEquals(monitor.lookupMember(pid2), None)

        monitor.takeSnapshot()
        self.assertNotEquals(monitor.lookupMember(pid2), None)

    def test_trackEnvironmentFamilyNotFound(self):
        monitor = self.makeMonitor()

        pid = self.spawn()
        self.assertEquals(
            monitor.trackEnvironment(pid, os.getlogin()),
            MonitorStatus.FAMILY_NOT_FOUND)

    def test_trackEnvironment(self):
        pid1 = self.spawn()
        pid_env_id = PidEnvId()
        monitor = self.makeMonitor(pid1)
        monitor.trackEnvironment(pid1, pid_env_id)

        pid2 = self.spawn()
        self.assertEquals(monitor.lookupMember(pid2), None)

        monitor.takeSnapshot()
        self.assertNotEquals(monitor.lookupMember(pid2), None)

    def test_signalProcessFamilyNotFound(self):
        monitor = self.makeMonitor()

        pid = self.spawn()
        self.assertEquals(
            monitor.signalProcess(pid, signal.SIGTERM),
            MonitorStatus.FAMILY_NOT_FOUND)

    def test_signalProcess(self):
        pid = self.spawn()
        monitor = self.makeMonitor(pid)

        self.assertEquals(
            monitor.signalProcess(pid, signal.SIGTERM),
            MonitorStatus.SUCCESS)
        self.reap(pid)

    def test_signalFamilyFamilyNotFound(self):
        monitor = self.makeMonitor()

        pid = self.spawn()
        self.assertEquals(
            monitor.signalFamily(pid, signal.SIGTERM),
            MonitorStatus.FAMILY_NOT_FOUND)

    def test_signalFamily(self):
        monitor = self.makeMonitor()
        monitor.trackLogin(self.init_pid, os.getlogin())

        pid1 = self.spawn()
        monitor.registerSubfamily(pid1, 0, -1)

        pid2 = self.spawn(pid1)
        self.assertEquals(
            monitor.signalFamily(pid1, signal.SIGTERM),
            MonitorStatus.SUCCESS)
        self.reap(pid2, pid1)
        self.reap(pid1)

    def test_addMember(self):
        monitor = self.makeMonitor()
        pid = self.spawn()
        self.assertEquals(monitor.lookupMember(pid), None)

        member = FamilyMember(info=Info.getInfo(pid))
        monitor.addMember(member)
        self.assertNotEquals(monitor.lookupMember(pid), None)

    def test_removeMember(self):
        pid = self.spawn()
        monitor = self.makeMonitor(pid)
        self.assertNotEquals(monitor.lookupMember(pid), None)

        member = FamilyMember(info=Info.getInfo(pid))
        monitor.removeMember(member)
        self.assertEquals(monitor.lookupMember(pid), None)

    def test_takeSnapshot(self):
        monitor = self.makeMonitor()
        monitor.takeSnapshot()

        pid = self.spawn()
        self.assertEquals(monitor.lookupMember(pid), None)

        monitor.takeSnapshot()
        self.assertNotEquals(monitor.lookupMember(pid), None)
