#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Timeslice",
    ]

import math
import time


class Timeslice:

    __slots__ = (
        "_timeslice",
        "_min_interval",
        "_max_interval",
        "_default_interval",
        "_initial_interval",
        "_last_duration",
        "_start_time",
        "_next_start_time",
        "_never_ran_before",
        "_expedite_next_run",
        )

    def __init__(self):
        self._timeslice = 0.0
        self._min_interval = 0.0
        self._max_interval = 0.0
        self._default_interval = 0.0
        self._initial_interval = -1.0
        self._last_duration = 0.0
        self._start_time = 0.0
        self._next_start_time = 0.0
        self._never_ran_before = True
        self._expedite_next_run = True

    def copy(self):
        timeslice = Timeslice()
        timeslice.copy_from(self)
        return timeslice

    def copy_from(self, other):
        self._timeslice = other._timeslice
        self._min_interval = other._min_interval
        self._max_interval = other._max_interval
        self._default_interval = other._default_interval
        self._initial_interval = other._initial_interval
        self._last_duration = other._last_duration
        self._start_time = other._start_time
        self._next_start_time = other._next_start_time
        self._never_ran_before = other._never_ran_before
        self._expedite_next_run = other._expedite_next_run

    def get_last_duration(self):
        """Get the timeslice last duration."""
        return self._last_duration

    def get_timeslice(self):
        """Get the timeslice."""
        return self._timeslice

    def set_timeslice(self, timeslice):
        """Set the timeslice.

        :param timeslice: Timeslice to set.
        """
        self._timeslice = timeslice
        self.update_next_start_time()

    def get_min_interval(self):
        """Get the timeslice minimum interval."""
        return self._min_interval

    def set_min_interval(self, min_interval):
        """Set the timeslice minimum interval.

        :param min_interval: Minimum interval to set.
        """
        self._min_interval = min_interval
        self.update_next_start_time()

    def get_max_interval(self):
        """Get the timeslice maximum interval."""
        return self._max_interval

    def set_max_interval(self, max_interval):
        """Set the timeslice maximum interval.

        :param max_interval: Maximum interval to set.
        """
        self._max_interval = max_interval
        self.update_next_start_time()

    def get_default_interval(self):
        """Get the timeslice default interval."""
        return self._default_interval

    def set_default_interval(self, default_interval):
        """Set the timeslice default interval.

        :param default_interval: Default interval to set.
        """
        self._default_interval = default_interval
        self.update_next_start_time()

    def get_initial_interval(self):
        """Get the timeslice initial interval."""
        return self._initial_interval

    def set_initial_interval(self, initial_interval):
        """Set the timeslice initial interval.

        :param initial_interval: Initial interval to set.
        """
        self._initial_interval = initial_interval
        self.update_next_start_time()

    def get_start_time(self):
        return self._start_time

    def set_start_time_now(self):
        """Set the start time to now."""
        self._start_time = time.time()

    def set_finish_time_now(self):
        finish_time = time.time()
        self._last_duration = finish_time - self._start_time
        self._never_ran_before = False
        self._expedite_next_run = False
        self.update_next_start_time()

    def reset(self):
        self._last_duration = 0.0
        self._start_time = time.time()
        self._never_ran_before = True
        self._expedite_next_run = False
        self.update_next_start_time()

    def update_next_start_time(self):
        delay = self._default_interval

        if self._expedite_next_run:
            delay = 0.0

        if not self._start_time:
            self.set_start_time_now()

        elif self._timeslice > 0:
            slice_delay = self._last_duration / self._timeslice
            if delay < slice_delay:
                delay = slice_delay

        if self._max_interval > 0 and delay > self._max_interval:
            delay = self._max_interval

        if delay < self._min_interval:
            delay = self._min_interval

        if self._never_ran_before and self._initial_interval >= 0:
            delay = self._initial_interval

        self._next_start_time = math.floor(delay + self._start_time + 0.5)

    def get_next_start_time(self):
        return self._next_start_time

    def get_time_to_next_run(self):
        delta = int(self._next_start_time - time.time())
        if delta < 0:
            return 0

        return delta

    def is_time_to_run(self):
        return self.get_time_to_next_run() == 0

    def expedite_next_run(self):
        self._expedite_next_run = True
        self.update_next_start_time()
