#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "EnvironmentTracker",
    "LoginTracker",
    "ParentTracker",
    "PIDTracker",
    ]

import pwd
import logging


class Tracker:

    __slots__ = (
        "_monitor",
        )

    def __init__(self, monitor):
        self._monitor = monitor

    def findProcesses(self, info_list):
        for info in info_list:
            self._checkProcess(info)

    def _checkProcess(self, info):
        raise NotImplementedError(
            "this function must be overridden by subclasses")


class TrackerTag:

    def test(self, info):
        raise NotImplementedError(
            "this function must be overridden by subclasses")


class TrackerTable:

    __slots__ = (
        "_table",
        )

    def __init__(self):
        self._table = {}

    def addMapping(self, family, tag):
        self._table[family] = tag

    def removeMapping(self, family):
        if family in self._table:
            del self._table[family]

    def findFamily(self, info):
        for family, tag in self._table.items():
            if tag.test(info):
                return family

        return None


class EnvironmentTag(TrackerTag):

    __slots__ = (
        "_pid_env_id",
        )

    def __init__(self, pid_env_id=None):
        self._pid_env_id = pid_env_id

    def get_pid_env_id(self):
        return self._pid_env_id

    def test(self, info):
        return self._pid_env_id == info.pid_env_id


class EnvironmentTracker(Tracker):

    __slots__ = (
        "_table",
        )

    def __init__(self, monitor):
        super(EnvironmentTracker, self).__init__(monitor)
        self._table = TrackerTable()

    def addMapping(self, family, pid_env_id):
        self._table.addMapping(family, EnvironmentTag(pid_env_id))

    def removeMapping(self, family):
        self._table.removeMapping(family)

    def _checkProcess(self, info):
        family = self._table.findFamily(info)
        if family is not None:
            self._monitor.addFamilyMember(family, info, "ENV")
            return True

        return False


class LoginTag(TrackerTag):

    __slots__ = (
        "_login",
        "_uid",
        )

    def __init__(self, login=None):
        self._login = login

        if login:
            try:
                passwd = pwd.getpwnam(login)

            except KeyError:
                logging.info("Login account not found: %s", login)
                self._uid = -1

            else:
                self._uid = passwd.pw_uid

    def get_login(self):
        return self._login

    def test(self, info):
        return info.owner == self._uid


class LoginTracker(Tracker):

    __slots__ = (
        "_table",
        )

    def __init__(self, monitor):
        super(LoginTracker, self).__init__(monitor)
        self._table = TrackerTable()

    def addMapping(self, family, login):
        self._table.addMapping(family, LoginTag(login))

    def removeMapping(self, family):
        self._table.removeMapping(family)

    def _checkProcess(self, info):
        family = self._table.findFamily(info)
        if family is not None:
            self._monitor.addFamilyMember(family, info, "LOGIN")
            return True

        return False


class ParentTracker(Tracker):

    __slots__ = (
        "_keep_checking",
        )

    def __init__(self, monitor):
        super(ParentTracker, self).__init__(monitor)
        self._keep_checking = False

    def findProcesses(self, info):
        self._keep_checking = True
        while self._keep_checking:
            self._keep_checking = False
            super(ParentTracker, self).findProcesses(info)

    def _checkProcess(self, info):
        member = self._monitor.lookupMember(info.ppid)
        if member is None:
            return False

        if member.family.root_pid == 0:
            return False

        if member.info.birthday > info.birthday:
            return False

        if self._monitor.addFamilyMember(member.family, info, "PARENT"):
            self._keep_checking = True
            return True

        return False


class PIDTag(TrackerTag):

    __slots__ = (
        "_pid",
        "_birthday",
        )

    def __init__(self, pid=0, birthday=0):
        self._pid = pid
        self._birthday = birthday

    def test(self, info):
        return info.pid == self._pid and info.birthday == self._birthday


class PIDTracker(Tracker):

    __slots__ = (
        "_table",
        )

    def __init__(self, monitor):
        super(PIDTracker, self).__init__(monitor)
        self._table = TrackerTable()

    def addMapping(self, family, pid, birthday):
        self._table.addMapping(family, PIDTag(pid, birthday))

    def removeMapping(self, family):
        self._table.removeMapping(family)

    def _checkProcess(self, info):
        family = self._table.findFamily(info)
        if family is not None:
            self._monitor.addFamilyMember(family, info, "PID")
            return True

        return False
