#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "Tree",
    ]


class Tree:

    __slots__ = (
        "_data",
        "_parent",
        "_sibling",
        "_child",
        )

    def __init__(self, data, parent=None, sibling=None):
        self._data = data
        self._parent = parent
        self._sibling = sibling
        self._child = None

    def getData(self):
        """Get this node's data."""
        return self._data

    def getParent(self):
        """Get this node's parent."""
        return self._parent

    def getChild(self):
        """Get the youngest child which can be used to get the rest of
           the children from its sibling."""
        return self._child

    def getChildren(self):
        """Get an iterator of the children from the youngest to its
           siblings."""
        child = self._child
        while child:
            yield child
            child = child._sibling

    def getSibling(self):
        """Get the next sibling from a list of siblings."""
        return self._sibling

    def addChild(self, data):
        """Add a child node to this tree's root node.

        :param data: Data to add.
        """
        self._child = Tree(data, self, self._child)
        return self._child

    def remove(self):
        """Detach the node from the tree, with all children inherited
           by the parent. This does not work if if called on the root,
           since there is nowhere to reparent the children)"""

        assert(self._parent)

        if self._child:
            # Children need to be reparented
            child = self._child
            while child._sibling:
                child._parent = self._parent
                child = child._sibling

            child._parent = self._parent

            # Merge the children into siblings
            child._sibling = self._sibling
            self._sibling = self._child

        # Sibling need to be relinked
        if self._parent._child == self:
            self._parent._child = self._sibling

        else:
            previous = self._parent._child
            while previous._sibling != self:
                previous = previous._sibling

            previous._sibling = self._sibling

        # Sever all ties
        self._child = self._sibling = self._parent = None
