#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "ConfigMixin",
    ]

import os
import errno
import logging

from ConfigParser import (
    ConfigParser,
    Error as ConfigError,
    )
from optparse import (
    OptionGroup,
    OptionValueError,
    )


class ConfigMixin:

    # Config defaults
    user_config = "~/.checkbox.conf"
    system_config = "/etc/checkbox.conf"

    def addOptions(self, parser):
        """See Application."""
        super(ConfigMixin, self).addOptions(parser)

        group = OptionGroup(parser, "Config options")
        group.add_option("--no-config",
            action="store_true",
            help="Do not read any configuration files")
        group.add_option("--config",
            metavar="FILE",
            help="Read from the configuration file instead of "
                " the system file %s and the user file %s"
                % (self.system_config, self.user_config))

        parser.add_option_group(group)

    def parseOptions(self, options, args):
        """See Application."""
        super(ConfigMixin, self).parseOptions(options, args)

        if options.no_config:
            return

        # Ignore the system config if the user specifies a config
        # on the command line.
        if options.config:
            if not read_config(options.config):
                raise OptionValueError(
                    "No configuration file: %s" % options.config)
        else:
            # Read the user config and then the system config.
            read_config(self.user_config)
            read_config(self.system_config)


def read_config(filename):
    """Parse configuration files and set the environment."""
    try:
        fp = open(filename)
    except IOError, error:
        if error.errno == errno.ENOENT:
            return False
        else:
            raise

    logging.debug("Reading configuration file: %s", filename)

    parser = ConfigParser()
    try:
        parser.readfp(fp, filename)
    except ConfigError, error:
        raise OptionValueError(error.message)

    # Set DEFAULT section as default values in the environment.
    defaults = {}
    environ = os.environ
    for key, value in parser.defaults().iteritems():
        key = key.upper()
        value = value % defaults
        defaults[key] = value
        environ.setdefault(key, value)

    # Set other sections using the name as the prefix.
    for section in parser.sections():
        for key, value in parser.items(section):
            key = key.upper()
            if key not in defaults:
                key = "%s_%s" % (section, key)
                value = value % defaults
                environ.setdefault(key, value)

    return True
