#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "LoggerMixin",
    ]

import sys
import time
import logging

from optparse import OptionGroup


class Formatter(logging.Formatter):
    """Formatter encoding our preferred output format."""

    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = '%(asctime)s %(levelname)-7s %(message)s'
        if datefmt is None:
            datefmt = "%Y-%m-%d %H:%M:%S"
        logging.Formatter.__init__(self, fmt, datefmt)
        self.converter = time.gmtime  # Output should be UTC


class LoggerMixin:

    # Logger defaults
    default_log_level = "warning"

    def addOptions(self, parser):
        """See Application."""
        super(LoggerMixin, self).addOptions(parser)

        group = OptionGroup(parser, "Logger options")
        group.add_option("--log",
            metavar="FILE",
            help="The file to write the log to.")
        group.add_option("--log-level",
            default=self.default_log_level,
            help="One of debug, info, warning, error or critical.")

        parser.add_option_group(group)

    def parseOptions(self, options, args):
        """See Application."""
        super(LoggerMixin, self).parseOptions(options, args)

        log = options.log
        if log:
            if log == "-":
                log_stream, log_file = sys.stdout, None
            else:
                log_stream, log_file = None, log
        else:
            log_stream, log_file = sys.stderr, None

        get_logger(options.log_level, log_stream, log_file)


def reset_logger(logger):
    """Reset the state of the given Logger instance.

    :param logger: Logger instance to reset.
    """
    for handler in logger.handlers[:]:
        handler.flush()
        try:
            handler.close()
        except KeyError:
            pass
        logger.removeHandler(handler)


def get_logger(level, log_stream=None, log_file=None, name=None):
    """Get a Logger instance with a handler to the given parameters.

    :param level: Log level for the handler.
    :param log_stream: Stream to use for logging.
    :param log_file: Or, the file to use for logging.
    :param name: Optional name of the Logger, defaults to root.
    """
    root_logger = logging.getLogger()
    reset_logger(root_logger)

    # Log to the given file or the default stream
    if log_file:
        handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler(log_stream)

    # Set the level on the handler rather than the logger
    level = logging.getLevelName(level.upper())
    handler.setLevel(level)
    root_logger.setLevel(logging.NOTSET)

    # Do the same for the formatter
    formatter = Formatter()
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    return logging.getLogger(name)
