#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "run",
    ]

import sys

from optparse import (
    OptionGroup,
    OptionValueError,
    )

from checkbox.lib.enum import (
    EnumError,
    enum_name_to_value,
    enum_value_to_name_string,
    )

from checkbox.process.client import (
    Client,
    ClientError,
    )
from checkbox.process.operation import OperationCode

from checkbox.scripts.application import (
    Application,
    ApplicationError,
    )
from checkbox.scripts.mixins.logger import LoggerMixin


class ProcessClientApplication(LoggerMixin, Application):

    # Application defaults
    usage = """\
Usage: %prog [OPTIONS] PATH


Operations:
  REGISTER_SUBFAMILY <PID> <WATCHER_PID> <MAX_SNAPSHOT_INTERVAL>
  SIGNAL_PROCESS <SIGNAL> [<PID>]
  SUSPEND_FAMILY [<PID>]
  RESUME_FAMILY [<PID>]
  KILL_FAMILY [<PID>]
  UNREGISTER_FAMILY <PID>
  TAKE_SNAPSHOT
  QUIT
"""

    def addOptions(self, parser):
        """See Application."""
        super(ProcessClientApplication, self).addOptions(parser)

        group = OptionGroup(parser, "Client options")
        group.add_option("-i", "--input",
            help="Input file, defaults to stdin.")
        group.add_option("-o", "--output",
            help="Output file, defaults to stdout.")

        parser.add_option_group(group)

    def parseOptions(self, options, args):
        """See Application."""
        super(ProcessClientApplication, self).parseOptions(options, args)

        # Check arguments
        if len(args) != 1:
            raise OptionValueError("Must specify a single path")

        # Input file
        if options.input is not None:
            try:
                self.input = open(options.input)
            except IOError, error:
                raise OptionValueError(
                    "Invalid input file: %s", error.strerror)
        else:
            self.input = sys.stdin

        # Output file
        if options.output is not None:
            try:
                self.output = open(options.output, "w")
            except IOError, error:
                raise OptionValueError(
                    "Invalid output file: %s", error.strerror)
        else:
            self.output = sys.stdout

        path = args[0]
        try:
            self.client = Client(path)
        except ClientError, error:
            raise OptionValueError(error.message)

    def process(self):
        """See Application."""
        operation_to_function = {
            OperationCode.REGISTER_SUBFAMILY: self.registerSubfamily,
            OperationCode.SIGNAL_PROCESS: self.signalProcess,
            OperationCode.SUSPEND_FAMILY: self.suspendFamily,
            OperationCode.RESUME_FAMILY: self.resumeFamily,
            OperationCode.KILL_FAMILY: self.killFamily,
            OperationCode.UNREGISTER_FAMILY: self.unregisterFamily,
            OperationCode.TAKE_SNAPSHOT: self.takeSnapshot,
            OperationCode.QUIT: self.quit,
            }

        ok = True
        while ok:
            line = self.input.readline()
            tokens = line.split()
            if not tokens:
                break

            name = tokens.pop(0)
            try:
                code = enum_name_to_value(OperationCode, name)
            except EnumError:
                raise ApplicationError("Invalid operation name: %s" % name)

            try:
                function = operation_to_function[code]
            except KeyError:
                raise ApplicationError(
                    "Invalid operation name: %d"
                    % enum_value_to_name_string(OperationCode, code))

            try:
                ok = function(*tokens)
            except TypeError:
                raise ApplicationError(
                    "Invalid argument count: %s" % line)
            except ValueError:
                raise ApplicationError(
                    "Invalid argument types: %s" % line)

    def registerSubfamily(self, pid, watcher_pid, max_snapshot_interval):
        pid = int(pid)
        watcher_pid = int(watcher_pid)
        max_snapshot_interval = int(max_snapshot_interval)

        return self.client.registerSubfamily(
            pid, watcher_pid, max_snapshot_interval)

    def signalProcess(self, signal, pid=0):
        signal = int(signal)
        if pid != 0:
            pid = int(pid)

        return self.client.signalProcess(pid, signal)

    def suspendFamily(self, pid=0):
        if pid != 0:
            pid = int(pid)

        return self.client.suspendFamily(pid)

    def resumeFamily(self, pid=0):
        if pid != 0:
            pid = int(pid)

        return self.client.resumeFamily(pid)

    def killFamily(self, pid=0):
        if pid != 0:
            pid = int(pid)

        return self.client.killFamily(pid)

    def unregisterFamily(self, pid):
        pid = int(pid)

        return self.client.unregisterFamily(pid)

    def takeSnapshot(self):
        return self.client.takeSnapshot()

    def quit(self):
        return self.client.quit()


def run():
    application = ProcessClientApplication()
    application.run()
