#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "run",
    ]

import os
import sys

from optparse import (
    OptionGroup,
    OptionValueError,
    )

from checkbox.lib.file import fd_close

from checkbox.process.info import Info
from checkbox.process.monitor import Monitor
from checkbox.process.server import Server

from checkbox.scripts.application import (
    Application,
    ApplicationError,
    )
from checkbox.scripts.mixins.logger import LoggerMixin


class ProcessServerApplication(LoggerMixin, Application):

    # Application defaults
    usage = "Usage: %prog [OPTIONS] PATH"

    # Process server defaults
    default_interval = 60

    def addOptions(self, parser):
        """See Application."""
        super(ProcessServerApplication, self).addOptions(parser)

        group = OptionGroup(parser, "Server options")
        group.add_option("-p", "--principal",
            help="Local client principal.")
        group.add_option("-i", "--interval",
            type="int",
            default=self.default_interval,
            help="Maximum snapshot interval.")

        parser.add_option_group(group)

    def parseOptions(self, options, args):
        """See Application."""
        # Close stdin and stdout since we don't use them.
        fd_close(sys.stdin.fileno())
        fd_close(sys.stdout.fileno())

        super(ProcessServerApplication, self).parseOptions(options, args)

        # Check options
        if options.interval < -1:
            raise OptionValueError("Must be non-negative or -1")

        # Check arguments
        if len(args) != 1:
            raise OptionValueError("Must specify a single path")

        self.path = args[0]
        self.interval = options.interval
        self.principal = options.principal

    def process(self):
        # Get the PID and birthday of the parent.
        parent_info = self.getParentInfo()

        # Initialize the engine for monitoring process families.
        monitor = Monitor(
            parent_info.pid, parent_info.birthday, self.interval)

        # Initialize the server for accepting requests from clients.
        server = Server(monitor, self.path)

        # Set the client to accept connections from.
        server.setPrincipal(self.principal)

        # Close stderr now that the server has been initialized.
        fd_close(sys.stderr.fileno())

        # Finally, run the server.
        server.run()

    def getParentInfo(self):
        own_info = Info.getInfo(os.getpid())
        if own_info is None:
            raise ApplicationError("Failed on own PID")

        parent_info = Info.getInfo(own_info.ppid)
        if parent_info is None:
            raise ApplicationError("Failed on parent PID")

        if parent_info.birthday > own_info.birthday:
            raise ApplicationError(
                "Parent PID birthday (%d) is later than own PID (%d)"
                % (parent_info.birthday, own_info.birthday))

        return parent_info


def run():
    application = ProcessServerApplication()
    application.run()
