#
# Copyright (c) 2012 Canonical
#
# This file is part of Checkbox.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__metaclass__ = type

__all__ = [
    "UniqueMixin",
    ]

import os
import sys

from itertools import count


class UniqueMixin:

    # This allocates process-wide unique integers.  We count on Python doing
    # only cooperative threading to make this safe across threads.
    _unique_int_counter = count(100000)

    def getUniqueInteger(self, bits=None):
        """Return an integer unique to this factory instance.

        For each thread, this will be a series of increasing numbers,
        but the starting point will be unique per thread.

        :param bits: Optional number of bits for the integer.
        """
        integer = UniqueMixin._unique_int_counter.next()
        if bits is not None:
            integer = integer % (2 ** bits)

        return integer

    def getUniqueIPAddress(self):
        """Return an IP address unique to this factory instance.

        Invalid addresses like 0.0.0.0 and broadcast addresses like
        255.255.255.255 might be returned.
        """
        return ".".join([
            str(self.getUniqueInteger(8)) for octet in xrange(4)])

    def getUniqueString(self, prefix=None):
        """Return a string unique to this factory instance.

        :param prefix: Used as a prefix for the unique string. If
            unspecified, defaults to "generic-string".
        """
        if prefix is None:
            frame = sys._getframe(2)
            source_filename = frame.f_code.co_filename
            # Dots and dashes cause trouble with some consumers of these
            # names.
            source = (
                os.path.basename(source_filename)
                .replace("_", "-")
                .replace(".", "-"))
            if source.startswith(
                    "<doctest "):
                # Like "-<doctest xx-build-summary-txt[10]>".
                source = (source
                    .replace("<doctest ", "")
                    .replace("[", "")
                    .replace("]>", ""))
            prefix = "unique-from-%s-line%d" % (
                source, frame.f_lineno)
        string = "%s-%s" % (prefix, self.getUniqueInteger())
        return string

    def getUniqueUnicode(self):
        return self.getUniqueString().decode("latin-1")
